#include <time.h>

#include <Y2/Y.h>
#include <Y2/Ylib.h>

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include "guiutils.h"
#include "dspview.h"
#include "menubutton.h"
#include "conbar.h"
#include "config.h"

#include "images/icon_y_menu_20x20.xpm"
#include "images/icon_yserver_location_20x20.xpm"
#include "images/icon_yconsole_20x20.xpm"
#include "images/icon_yaudio_20x20.xpm"
#include "images/icon_ymixer_20x20.xpm"
#include "images/icon_yplayer_20x20.xpm"
#include "images/icon_options2_20x20.xpm"
#include "images/icon_sound_bm_10x10.xpm"
#include "images/icon_cycle_load_bm_10x10.xpm"
#include "images/icon_close_20x20.xpm"
#include "images/icon_reload_20x20.xpm"
#include "images/icon_shellout_20x20.xpm"
#include "images/icon_shutdown_20x20.xpm"


static void ConBarRealizeCB(GtkWidget *widget, gpointer data);
static void ConBarLocationCB(GtkWidget *widget, gpointer data);
static void ConBarButtonClickedCB(GtkWidget *widget, gpointer data);
static void ConBarShelloutToggledCB(GtkWidget *widget, gpointer data);
static gint ConBarFocusChangeEventCB(
	GtkWidget *widget, GdkEventFocus *focus, gpointer data
);
static gint ConBarConfigureNotifyCB(
	GtkWidget *widget, GdkEventConfigure *configure, gpointer data
);
static gint ConBarKeyEventCB(
	GtkWidget *widget, GdkEventKey *key, gpointer data
);
static gint ConBarButtonEventCB(
	GtkWidget *widget, GdkEventButton *button, gpointer data
);

gchar *ConBarGetLocation(connection_bar_struct *cb);
void ConBarSetLocation(
	connection_bar_struct *cb, const gchar *loc_str,
	gboolean emit_signal, gboolean record_history
);

void ConBarSetConnectState(connection_bar_struct *cb, gboolean is_connected);

dspview_struct *ConBarGetDSPView(connection_bar_struct *cb);

void ConBarSetAudioColors(
	connection_bar_struct *cb,
	GdkColor *color_bg, GdkColor *color_fg,
	GdkColor *color_fg_dim
);
void ConBarSetAudioFonts(
	connection_bar_struct *cb,
	const gchar *font_name_large, const gchar *font_name_medium,
	const gchar *font_name_small
);

static void ConBarDrawAudioPage3(
	connection_bar_struct *cb,
	gint state, GtkStyle *style,
	GdkDrawable *drawable, gint width, gint height
);
static void ConBarDrawAudioPage2(
	connection_bar_struct *cb,
	gint state, GtkStyle *style,
	GdkDrawable *drawable, gint width, gint height
);
static void ConBarDrawAudioPage1(
	connection_bar_struct *cb,
	gint state, GtkStyle *style,
	GdkDrawable *drawable, gint width, gint height
);
static void ConBarDrawAudioNotConnectedPage(
	connection_bar_struct *cb,
	gint state, GtkStyle *style,
	GdkDrawable *drawable, gint width, gint height
);
void ConBarSetAudio(
	connection_bar_struct *cb,
	gulong start_time, const gchar *audio_mode_name,
	gint sample_rate, gint sample_size, gint channels,
	glong cycle_us, gfloat cycle_load,
	gint num_fragments, gint fragment_size,
	gboolean flip_stereo,  gboolean shellout,
	const guint8 *buf, gint buf_len,
	gfloat *channel_volume_avg,
	gint version_major, gint version_minor,
	const gchar *server_name, const gchar *sound_name
);
void ConBarSetShellOut(connection_bar_struct *cb, gboolean shellout);

connection_bar_struct *ConBarNew(
	guint display_flags, gint dspview_width, gint dspview_height,
	gpointer client_data,
	void (*connect_cb)(GtkWidget *, const gchar *, gpointer),
	void (*server_list_cb)(GtkWidget *, gpointer),
	void (*close_cb)(GtkWidget *, gpointer),
	void (*new_console_cb)(GtkWidget *, gpointer),
	void (*new_audio_cb)(GtkWidget *, gpointer),
	void (*new_mixer_cb)(GtkWidget *, gpointer),
	void (*new_player_cb)(GtkWidget *, gpointer),
	void (*options_cb)(GtkWidget *, gpointer),
	void (*refresh_cb)(GtkWidget *, gpointer),
	gboolean (*shellout_cb)(GtkWidget *, gboolean, gpointer),
	void (*shutdown_cb)(GtkWidget *, gpointer),
	GtkWidget *parent, GtkAccelGroup *accelgrp
);
void ConBarDraw(
	connection_bar_struct *cb,
	const guint8 *buf, gint len
);
void ConBarMap(connection_bar_struct *cb);
void ConBarUnmap(connection_bar_struct *cb);
void ConBarDelete(connection_bar_struct *cb);


#define CONBAR_AUDIO_FONT_NAME_LARGE	\
 "-adobe-helvetica-bold-r-*-*-16-*-*-*-*-*-*-*"
#define CONBAR_AUDIO_FONT_NAME_MEDIUM	\
 "-adobe-helvetica-bold-r-*-*-14-*-*-*-*-*-*-*"
#define CONBAR_AUDIO_FONT_NAME_SMALL	\
 "-adobe-helvetica-bold-r-*-*-10-*-*-*-*-*-*-*"

#define CONBAR_AUDIO_DA_WIDTH	320
#define CONBAR_AUDIO_DA_HEIGHT	50

#define CONBAR_DO_UPDATE_AUDIO_PM(p)	\
{ if((p) != NULL) ConBarSetAudio( \
   (p), (p)->start_time, (p)->audio_mode_name, \
   (p)->sample_rate, (p)->sample_size, (p)->channels, \
   (p)->cycle_us, (p)->cycle_load, \
   (p)->num_fragments, (p)->fragment_size, \
   (p)->flip_stereo, (p)->shellout, \
   NULL, 0, (p)->channel_volume_avg, \
   (p)->version_major, (p)->version_minor, \
   (p)->server_name, (p)->sound_name \
  ); }



#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define ABSOLUTE(x)	(((x) < 0) ? ((x) * -1) : (x))


/*
 *	Audio GtkDrawingArea "realize" signal callback.
 */
static void ConBarRealizeCB(GtkWidget *widget, gpointer data)
{
	gchar **xpm_data;
	GdkBitmap **mask;
	GdkWindow *window;
	GtkWidget *w;
	GtkStyle *style;
	connection_bar_struct *cb = CONBAR(data);
	if(cb == NULL)
	    return;

	w = cb->audio_da;
	if(w == NULL)
	    return;

	window = w->window;
	style = gtk_widget_get_style(w);
	if((window == NULL) || (style == NULL))
	    return;

#define DO_LOAD_MASK	\
{ \
 if(*mask == NULL) \
 { \
  GdkPixmap *pixmap = gdk_pixmap_create_from_xpm_d( \
   window, mask, &style->bg[GTK_STATE_NORMAL], xpm_data \
  ); \
  gdk_pixmap_unref(pixmap); \
 } \
}
	/* Create bitmaps as needed */
	mask = &cb->audio_shellout_bm;
	xpm_data = icon_sound_bm_10x10_xpm;
	DO_LOAD_MASK

	mask = &cb->audio_cycle_load_bm;
	xpm_data = icon_cycle_load_bm_10x10_xpm;
	DO_LOAD_MASK

#undef DO_LOAD_MASK
}

/*
 *	Location GtkCombo "activate" signal callback.
 */
static void ConBarLocationCB(GtkWidget *widget, gpointer data)
{
	connection_bar_struct *cb = CONBAR(data);
	if(cb == NULL)
	    return;

	if(cb->connect_cb != NULL)
	{
	    const gchar *loc_str = ConBarGetLocation(cb);
	    if(loc_str != NULL)
		cb->connect_cb(widget, loc_str, cb->client_data);
	}
}

/*
 *	Menu GtkButton "clicked" signal callback.
 */
static void ConBarButtonClickedCB(GtkWidget *widget, gpointer data)
{
	connection_bar_struct *cb = (connection_bar_struct *)data;
	if(cb == NULL)
	    return;

}

/*
 *	Shellout GtkToggleButton "toggled" signal callback.
 */
static void ConBarShelloutToggledCB(GtkWidget *widget, gpointer data)
{
	gboolean shellout;
	connection_bar_struct *cb = CONBAR(data);
	if(cb == NULL)
	    return;

	shellout = (widget != NULL) ?
	    GTK_TOGGLE_BUTTON(widget)->active : FALSE;
	if(cb->shellout_cb != NULL)
	{
	    gboolean actual_shellout = cb->shellout_cb(
		widget, shellout, cb->client_data
	    );
	    if(actual_shellout != shellout)
	    {
		GTK_TOGGLE_BUTTON(widget)->active = actual_shellout;
		gtk_widget_queue_draw(widget);
	    }
	}
}

/*
 *	Audio GtkDrawingArea "focus_in_event" signal callback.
 */
static gint ConBarFocusChangeEventCB(
	GtkWidget *widget, GdkEventFocus *focus, gpointer data
)
{
	GtkWidget *w;
	connection_bar_struct *cb = CONBAR(data);
	if((focus == NULL) || (cb == NULL))
	    return(FALSE);

	w = cb->audio_da;
	if(w == NULL)
	    return(FALSE);

	if(focus->in)
	    GTK_WIDGET_SET_FLAGS(w, GTK_HAS_FOCUS);
	else
	    GTK_WIDGET_UNSET_FLAGS(w, GTK_HAS_FOCUS);

	CONBAR_DO_UPDATE_AUDIO_PM(cb);
/*	gdk_window_clear(w->window); */

	return(TRUE);
}

/*
 *	Audio GtkDrawingArea "configure_event" signal callback.
 */
static gint ConBarConfigureNotifyCB(
	GtkWidget *widget, GdkEventConfigure *configure, gpointer data
)
{
	gint status = FALSE;
	connection_bar_struct *cb = CONBAR(data);
	if((configure == NULL) || (cb == NULL))
	    return(status);

	GDK_PIXMAP_UNREF(cb->audio_pm)
	cb->audio_pm = GDK_PIXMAP_NEW(
	    configure->width, configure->height
	);
	if(cb->audio_pm != NULL)
	    CONBAR_DO_UPDATE_AUDIO_PM(cb);

	status = TRUE;

	return(status);
}

/*
 *      Audio GtkDrawingArea "key_press_event" signal callback.
 */
static gint ConBarKeyEventCB(
	GtkWidget *widget, GdkEventKey *key, gpointer data
)
{
	gint status = FALSE;
	gboolean keystate;
	connection_bar_struct *cb = CONBAR(data);
	if((key == NULL) || (cb == NULL))
	    return(status);

	keystate = (key->type == GDK_KEY_PRESS) ? TRUE : FALSE;

#define DO_STOP_KEY_SIGNAL	\
{ if(widget != NULL) \
 gtk_signal_emit_stop_by_name( \
  GTK_OBJECT(widget), \
  keystate ? "key_press_event" : "key_release_event" \
 ); \
}

	switch(key->keyval)
	{
	  case GDK_Up:
	  case GDK_KP_Up:
	  case GDK_Left:
	  case GDK_KP_Left:
	  case GDK_Page_Up:
	  case GDK_KP_Page_Up:
	  case 'b': case 'p':
	    if(keystate)
	    {
		cb->audio_page_num--;
		if(cb->audio_page_num < 0)
		    cb->audio_page_num = CONBAR_TOTAL_AUDIO_PAGES - 1;
		CONBAR_DO_UPDATE_AUDIO_PM(cb);
	    }
	    DO_STOP_KEY_SIGNAL
	    status = TRUE;
	    break;

	  case GDK_Down:
	  case GDK_KP_Down:
	  case GDK_Right:
	  case GDK_KP_Right:
	  case GDK_Page_Down:
	  case GDK_KP_Page_Down:
	  case ' ':
	  case GDK_Return:
	  case 'n':
	    if(keystate)
	    {
		cb->audio_page_num++;
		if(cb->audio_page_num >= CONBAR_TOTAL_AUDIO_PAGES)
		    cb->audio_page_num = 0;
		CONBAR_DO_UPDATE_AUDIO_PM(cb);
	    }
	    DO_STOP_KEY_SIGNAL
	    status = TRUE;
	    break;

	  case '1':
	    if(keystate)
	    {
		cb->audio_page_num = 0;
		CONBAR_DO_UPDATE_AUDIO_PM(cb);
	    }
	    status = TRUE;
	    break;

	  case '2':
	    if(keystate)
	    {
		cb->audio_page_num = 1;
		CONBAR_DO_UPDATE_AUDIO_PM(cb);
	    }
	    status = TRUE;
	    break;

	  case '3':
	    if(keystate)
	    {
		cb->audio_page_num = 2;
		CONBAR_DO_UPDATE_AUDIO_PM(cb);
	    }
	    status = TRUE;
	    break;

	  case GDK_Home:
	    if(keystate)
	    {
		cb->audio_page_num = 0;
		CONBAR_DO_UPDATE_AUDIO_PM(cb);
	    }
	    status = TRUE;
	    break;

	  case GDK_End:
	    if(keystate)
	    {
		cb->audio_page_num = CONBAR_TOTAL_AUDIO_PAGES - 1;
		CONBAR_DO_UPDATE_AUDIO_PM(cb);
	    }
	    status = TRUE;
	    break;

	}




#undef DO_STOP_KEY_SIGNAL

	return(status);
}

/*
 *      Audio GtkDrawingArea "button_press_event" signal callback.
 */
static gint ConBarButtonEventCB(
	GtkWidget *widget, GdkEventButton *button, gpointer data
)
{
	gint status = FALSE;
	connection_bar_struct *cb = CONBAR(data);
	if((button == NULL) || (cb == NULL))
	    return(status);

	if(((widget != NULL) ? !GTK_WIDGET_HAS_FOCUS(widget) : FALSE) &&
	   (button->type == GDK_BUTTON_PRESS)
	)
	    gtk_widget_grab_focus(widget);

	switch(button->button)
	{
	  case 1:
	    switch((gint)button->type)
	    {
	      case GDK_BUTTON_PRESS:
		cb->audio_page_num++;
		if(cb->audio_page_num >= CONBAR_TOTAL_AUDIO_PAGES)
		    cb->audio_page_num = 0;
		CONBAR_DO_UPDATE_AUDIO_PM(cb);
		status = TRUE;
		break;

	      case GDK_BUTTON_RELEASE:
		status = TRUE;
		break;
	    }
	    break;
	}

	return(status);
}



/*
 *	Returns the value of the location bar or NULL on error.
 */
gchar *ConBarGetLocation(connection_bar_struct *cb)
{
	GtkCombo *combo;

	if(cb == NULL)
	    return(NULL);

	combo = (GtkCombo *)cb->location_combo;
	if(combo == NULL)
	    return(NULL);

	return(gtk_entry_get_text(GTK_ENTRY(combo->entry)));
}

/*
 *	Sets the value of the location bar to loc_str.
 *
 *	If emit_signal is TRUE then the connect_cb will be called.
 *
 *	If record_history is TRUE then the value of loc_str will be
 *	added to the location combo's history list.
 */
void ConBarSetLocation(
	connection_bar_struct *cb, const gchar *loc_str,
	gboolean emit_signal, gboolean record_history
)
{
	GtkCombo *combo;

	if(cb == NULL)
	    return;

	combo = (GtkCombo *)cb->location_combo;
	if(combo == NULL)
	    return;

	if(loc_str == NULL)
	{
	    gtk_entry_set_text(GTK_ENTRY(combo->entry), "");
	    return;
	}

	gtk_entry_set_text(GTK_ENTRY(combo->entry), loc_str);

	if(record_history)
	{
	    GUIComboAddItem(GTK_WIDGET(combo), loc_str);
	}

	if(emit_signal && (cb->connect_cb != NULL))
	{
	    cb->connect_cb(combo->entry, loc_str, cb->client_data);
	}
}


/*
 *	Updates all widgets on the connection bar depending on the
 *	given connection state.
 */
void ConBarSetConnectState(connection_bar_struct *cb, gboolean is_connected)
{
	gboolean sensitive;
	GtkWidget *w;
	dspview_struct *dv;


#define DO_SET_SENSITIVE	\
{ if(w != NULL) gtk_widget_set_sensitive(w, sensitive); } \

	if(cb == NULL)
	    return;

	dv = cb->dspview;

	/* No change in connected state? */
	if(cb->is_connected == is_connected)
	    return;

	/* Update connected state */
	cb->is_connected = is_connected;


	sensitive = is_connected;

	w = (dv != NULL) ? dv->view_da : NULL;
	DO_SET_SENSITIVE
	w = cb->audio_da;
	DO_SET_SENSITIVE
	w = cb->shellout_toggle;
	DO_SET_SENSITIVE	
	w = cb->shutdown_btn;
	DO_SET_SENSITIVE
	

#undef DO_SET_SENSITIVE
}


/*
 *	Returns the pointer to the DSP view or NULL on error.
 */
dspview_struct *ConBarGetDSPView(connection_bar_struct *cb)
{
	return(CONBAR_DSPVIEW(cb));
}


/*
 *	Sets the colors for the Audio drawing area.
 *
 *	The given colors need not be allocated.
 */
void ConBarSetAudioColors(
	connection_bar_struct *cb,
	GdkColor *color_bg, GdkColor *color_fg,
	GdkColor *color_fg_dim
)
{
	gint state;
	GtkRcStyle *rcstyle;
	GtkWidget *w;


	if(cb == NULL)
	    return;

	w = cb->audio_da;
	if(w == NULL)
	    return;

	rcstyle = gtk_rc_style_new();

	state = GTK_STATE_NORMAL;
	rcstyle->color_flags[state] = (
	    ((color_fg != NULL) ? GTK_RC_FG | GTK_RC_TEXT : 0) |
	    ((color_bg != NULL) ? GTK_RC_BASE : 0)
	);
	if(color_fg != NULL)
	{
	    memcpy(&rcstyle->fg[state], color_fg, sizeof(GdkColor));
	    memcpy(&rcstyle->text[state], color_fg, sizeof(GdkColor));
	}
	if(color_bg != NULL)
	{
	    memcpy(&rcstyle->base[state], color_bg, sizeof(GdkColor));
	}

	state = GTK_STATE_INSENSITIVE;
	rcstyle->color_flags[state] = (
	    ((color_fg_dim != NULL) ? GTK_RC_FG | GTK_RC_TEXT : 0) |
	    ((color_bg != NULL) ? GTK_RC_BASE : 0)
	);
	if(color_fg_dim != NULL)
	{
	    memcpy(&rcstyle->fg[state], color_fg_dim, sizeof(GdkColor));
	    memcpy(&rcstyle->text[state], color_fg_dim, sizeof(GdkColor));
	}
	if(color_bg != NULL)
	{
	    memcpy(&rcstyle->base[state], color_bg, sizeof(GdkColor));
	}

	gtk_widget_modify_style(w, rcstyle);

	gtk_rc_style_unref(rcstyle);
}

/*
 *      Sets the fonts for the Audio drawing area.
 */
void ConBarSetAudioFonts(
	connection_bar_struct *cb,
	const gchar *font_name_large, const gchar *font_name_medium,
	const gchar *font_name_small
)
{
	const gchar *font_name;
	GdkFont **font;

	if(cb == NULL)
	    return;

#define DO_RELOAD_FONT	\
{ if(font_name != NULL) { \
   if(*font != NULL) gdk_font_unref(*font); \
   *font = gdk_font_load(font_name); \
}}

	font_name = font_name_large;
	font = &cb->audio_font_large;
	DO_RELOAD_FONT

	font_name = font_name_medium;
	font = &cb->audio_font_medium;
	DO_RELOAD_FONT

	font_name = font_name_small;
	font = &cb->audio_font_small;
	DO_RELOAD_FONT

#undef DO_RELOAD_FONT
}

/*
 *	Called by ConBarSetAudio() to draw page 3.
 *
 *	All inputs assumed valid.  The background is expected to be
 *	already drawn.  The is_connected member is expected to be TRUE.
 */
static void ConBarDrawAudioPage3(
	connection_bar_struct *cb,
	gint state, GtkStyle *style,
	GdkDrawable *drawable, gint width, gint height
)
{
	GdkFont *font;
	gint font_extent_x = 0, font_next_y = 2;

	/* Draw uptime and cycle load */
	font = cb->audio_font_medium;
	if(font != NULL)
	{
	    gint x, y;
	    gint lbearing, rbearing, swidth, ascent, descent;
	    gulong up_time = (cb->start_time > 0) ?
		time(NULL) - cb->start_time : 0;
	    gchar *s;

	    if(up_time > 0)
		s = g_strdup_printf(
		    "UpTime: %.2i:%.2i:%.2i  Cycle Load: %.1f%%",
		    (gint)(up_time / 3600),
		    (gint)((up_time / 60) % 60),
		    (gint)(up_time % 60),
		    cb->cycle_load * 100.0f
		);
	    else
		s = g_strdup_printf(
		    "UpTime: *Unknown*  Cycle Load: %.1f%%",
		    cb->cycle_load * 100.0f
		);
	    gdk_string_extents(
		font, s,
		&lbearing, &rbearing, &swidth, &ascent, &descent
	    );
	    x = 2;
	    y = font_next_y;
	    gdk_draw_string(
		drawable, font, style->text_gc[state],
		x + lbearing, y + ascent,
		s
	    );
	    g_free(s);
	    font_extent_x = MAX(font_extent_x, x + swidth);
	    font_next_y += ascent + descent + 5;
	}

	/* Draw other statistics */
	font = cb->audio_font_small;
	if(font != NULL)
	{
	    gint x, y;
	    gint lbearing, rbearing, swidth, ascent, descent;
	    gchar *s = g_strdup_printf(
		"Maximum Channels: %i",
		YMixerValues
	    );
	    gdk_string_extents(
		font, s,
		&lbearing, &rbearing, &swidth, &ascent, &descent
	    );
	    x = 4;
	    y = font_next_y;
	    gdk_draw_string(
		drawable, font, style->text_gc[state],
		x + lbearing, y + ascent,
		s
	    );
	    g_free(s);
	    font_extent_x = MAX(font_extent_x, x + swidth);
	    font_next_y += ascent + descent + 5;
	}
}

/*
 *	Called by ConBarSetAudio() to draw page 2.
 *
 *	All inputs assumed valid.  The background is expected to be
 *	already drawn.  The is_connected member is expected to be TRUE.
 */
static void ConBarDrawAudioPage2(
	connection_bar_struct *cb,
	gint state, GtkStyle *style,
	GdkDrawable *drawable, gint width, gint height
)
{
	GdkFont *font;
	gint font_extent_x = 0, font_next_y = 2;

	/* Draw server name and vendor name */
	font = cb->audio_font_medium;
	if(font != NULL)
	{
	    gint x, y;
	    gint lbearing, rbearing, swidth, ascent, descent;
	    gchar *s = g_strdup_printf(
		"Server: \"%s\"  Vendor: \"%s\"",
		(cb->server_name != NULL) ?
		    cb->server_name : "*Untitled*",
		""
	    );
	    gdk_string_extents(
		font, s,
		&lbearing, &rbearing, &swidth, &ascent, &descent
	    );
	    x = 2;
	    y = font_next_y;
	    gdk_draw_string(
		drawable, font, style->text_gc[state],
		x + lbearing, y + ascent,
		s
	    );
	    g_free(s);
	    font_extent_x = MAX(font_extent_x, x + swidth);
	    font_next_y += ascent + descent + 5;
	}

	/* Draw sound name */
	font = cb->audio_font_medium;
	if(font != NULL)
	{
	    gint x, y;
	    gint lbearing, rbearing, swidth, ascent, descent;
	    gchar *s = g_strdup_printf(
		"Sound Name: \"%s\"",
		(cb->sound_name != NULL) ?
		    cb->sound_name : "*Untitled*"
	    );
	    gdk_string_extents(
		font, s,
		&lbearing, &rbearing, &swidth, &ascent, &descent
	    );
	    x = 2;
	    y = font_next_y;
	    gdk_draw_string(
		drawable, font, style->text_gc[state],
		x + lbearing, y + ascent,
		s
	    );
	    g_free(s);
	    font_extent_x = MAX(font_extent_x, x + swidth);
	    font_next_y += ascent + descent + 5;
	}

	/* Draw protocol version and info */
	font = cb->audio_font_small;
	if(font != NULL)
	{
	    gint x, y;
	    gint lbearing, rbearing, swidth, ascent, descent;
	    gchar *s = g_strdup_printf(
"Protocol Ver: %i.%i  Protocol Max Len: %i bytes",
		cb->version_major, cb->version_minor,
		YNetRecvBufLen
	    );
	    gdk_string_extents(
		font, s,
		&lbearing, &rbearing, &swidth, &ascent, &descent
	    );
	    x = 4;
	    y = font_next_y;
	    gdk_draw_string(
		drawable, font, style->text_gc[state],
		x + lbearing, y + ascent,
		s
	    );
	    g_free(s);
	    font_extent_x = MAX(font_extent_x, x + swidth);
	    font_next_y += ascent + descent + 5;
	}
}

/*
 *	Called by ConBarSetAudio() to draw page 1.
 *
 *	All inputs assumed valid.  The background is expected to be
 *	already drawn.  The is_connected member is expected to be TRUE.
 */
static void ConBarDrawAudioPage1(
	connection_bar_struct *cb,
	gint state, GtkStyle *style,
	GdkDrawable *drawable, gint width, gint height
)
{
	GdkFont *font;
	gint font_extent_x = 0, font_next_y = 2;

	/* Draw audio mode name */
	font = cb->audio_font_medium;
	if(font != NULL)
	{
	    gint        x = 5,
			y = font_next_y;
	    gint lbearing, rbearing, swidth, ascent, descent;
	    gchar *s = g_strdup_printf(
		"Audio Mode: \"%s\"",
		(cb->audio_mode_name != NULL) ?
		    cb->audio_mode_name : "*Unknown*"
	    );
	    gdk_string_extents(
		font, s,
		&lbearing, &rbearing, &swidth, &ascent, &descent
	    );
	    gdk_draw_string(
		drawable, font, style->text_gc[state],
		x + lbearing, y + ascent,
		s
	    );
	    g_free(s);
	    font_extent_x = MAX(font_extent_x, x + swidth);
	    font_next_y += ascent + descent + 5;
	}

	/* Draw main stats */
	font = cb->audio_font_large;
	if(font != NULL)
	{
	    gint	x = 5,
			y = font_next_y;
	    gint lbearing, rbearing, swidth, ascent, descent;
	    gchar *s = g_strdup_printf(
		"%i Hz  %i Bits  %s%s",
		cb->sample_rate, cb->sample_size,
		(cb->channels == 4) ? "Quad" :
		    ((cb->channels == 2) ? "Stereo" : "Mono"),
		cb->flip_stereo ? " (Flip)" : ""
	    );
	    gdk_string_extents(
		font, s,
		&lbearing, &rbearing, &swidth, &ascent, &descent
	    );
	    gdk_draw_string(
		drawable, font, style->text_gc[state],
		x + lbearing, y + ascent,
		s
	    );
	    g_free(s);
	    font_extent_x = MAX(font_extent_x, x + swidth);
	    font_next_y += ascent + descent + 5;
	}

	/* Draw minor stats */
	font = cb->audio_font_small;
	if(font != NULL)
	{
	    gint        x = 5,
			y = font_next_y;
	    gint lbearing, rbearing, swidth, ascent, descent;
	    gchar *s = g_strdup_printf(
		"Cycle: %ld us  Buffer: %i*%i bytes",
		cb->cycle_us, cb->num_fragments, cb->fragment_size
	    );
	    gdk_string_extents(
		font, s,
		&lbearing, &rbearing, &swidth, &ascent, &descent
	    );
	    gdk_draw_string(
		drawable, font, style->text_gc[state],
		x + lbearing, y + ascent,
		s
	    );
	    g_free(s);
	    font_extent_x = MAX(font_extent_x, x + swidth);
	    font_next_y += ascent + descent + 5;
	}

	/* Draw shellout icon and server load */
	if(cb->channel_volume_avg != NULL)
	{
	    gfloat *cv = cb->channel_volume_avg;
	    GdkBitmap *mask;


	    /* Shell out icon and volume */
	    mask = cb->audio_shellout_bm;
	    if(mask != NULL)
	    {
		gboolean color_set;
		gint    bm_width = 10, bm_height = 10;
		gint	bar_width = 10, bar_height = 3,
			total_bars = 10;
		gint    x = font_extent_x + 10,
			y = 2 + (total_bars * bar_height / 2) -
			    (bm_height / 2);
		gint i, ci, xr, yr;
		GdkGC *gc = style->fg_gc[
		    (cb->shellout) ? GTK_STATE_INSENSITIVE : state
		];


		/* Draw icon */
		gdk_gc_set_clip_mask(gc, mask);
		gdk_gc_set_clip_origin(gc, x, y);
		gdk_draw_rectangle(
		    drawable, gc, TRUE, x, y, bm_width, bm_height
		);
		gdk_gc_set_clip_mask(gc, NULL);

		/* Draw volume ticks for channel 0 */
		ci = 0;
		xr = x + bm_width + 4 + (0 * bar_width);
		yr = 2 + total_bars * bar_height;
		color_set = FALSE;
		gc = style->fg_gc[state];
		for(i = 0; i < total_bars; i++)
		{
		    if(!color_set &&
		       (((gfloat)i / (gfloat)total_bars) >= cv[ci])
		    )
		    {
			gc = style->fg_gc[GTK_STATE_INSENSITIVE];
			color_set = TRUE;
		    }
		    gdk_draw_rectangle(
			drawable, gc, TRUE, xr, yr,
			bar_width - 2, bar_height - 1
		    );
		    yr -= bar_height;
		}

		/* Draw volume ticks for channel 1 */
		ci = (cb->channels >= 2) ? 1 : 0;
		xr = x + bm_width + 4 + (1 * bar_width);
		yr = 2 + total_bars * bar_height;
		color_set = FALSE;
		gc = style->fg_gc[state];
		for(i = 0; i < total_bars; i++)
		{
		    if(!color_set &&
		       (((gfloat)i / (gfloat)total_bars) >= cv[ci])
		    )
		    {
			gc = style->fg_gc[GTK_STATE_INSENSITIVE];
			color_set = TRUE;
		    }
		    gdk_draw_rectangle(
			drawable, gc, TRUE, xr, yr,
			bar_width - 2, bar_height - 1
		    );
		    yr -= bar_height;
		}

		/* Draw volume ticks for channel 2 */
		ci = (cb->channels >= 4) ? 2 : 0;
		xr = x + bm_width + 4 + (2 * bar_width);
		yr = 2 + total_bars * bar_height;
		color_set = FALSE;
		gc = style->fg_gc[state];
		for(i = 0; i < total_bars; i++)
		{
		    if(!color_set &&
		       (((gfloat)i / (gfloat)total_bars) >= cv[ci])
		    )
		    {
			gc = style->fg_gc[GTK_STATE_INSENSITIVE];
			color_set = TRUE;
		    }
		    gdk_draw_rectangle(
			drawable, gc, TRUE, xr, yr,
			bar_width - 2, bar_height - 1
		    );
		    yr -= bar_height;
		}

		/* Draw volume ticks for channel 3 */
		ci = (cb->channels >= 4) ? 3 : ((cb->channels == 2) ? 1 : 0);
		xr = x + bm_width + 4 + (3 * bar_width);
		yr = 2 + total_bars * bar_height;
		color_set = FALSE;
		gc = style->fg_gc[state];
		for(i = 0; i < total_bars; i++)
		{
		    if(!color_set &&
		       (((gfloat)i / (gfloat)total_bars) >= cv[ci])
		    )
		    {
			gc = style->fg_gc[GTK_STATE_INSENSITIVE];
			color_set = TRUE;
		    }
		    gdk_draw_rectangle(
			drawable, gc, TRUE, xr, yr,
			bar_width - 2, bar_height - 1
		    );
		    yr -= bar_height;
		}

	    }

	    /* Cycle load icon and bars */
	    mask = cb->audio_cycle_load_bm;
	    if(mask != NULL)
	    {
		gint    bm_width = 10, bm_height = 10;
		gint i, xp, total_bars = 10;
		gint	bar_width = 3, bar_height = 6, bar_spacing = 1;
		gint    x = font_extent_x + 10,
			y = height - bm_height - 3;
		gboolean color_set = FALSE;
		GdkGC *gc = style->fg_gc[state];


		/* Draw icon */
		gdk_gc_set_clip_mask(gc, mask);
		gdk_gc_set_clip_origin(gc, x, y);
		gdk_draw_rectangle(
		    drawable, gc, TRUE, x, y, bm_width, bm_height
		);
		gdk_gc_set_clip_mask(gc, NULL);

		/* Draw bars */
		x += bm_width + 4;
		y += (bm_height / 2) - (bar_height / 2);
		gdk_gc_set_line_attributes(
		    gc,
		    bar_width,
		    GDK_LINE_SOLID, GDK_CAP_ROUND, GDK_JOIN_MITER
		);
		gdk_gc_set_line_attributes(
		    style->fg_gc[GTK_STATE_INSENSITIVE],
		    bar_width,
		    GDK_LINE_SOLID, GDK_CAP_ROUND, GDK_JOIN_MITER
		);
		for(i = 0, xp = x; i < total_bars; i++, xp += (bar_width + bar_spacing))
		{
		    if(!color_set &&
		       (((gfloat)i / (gfloat)total_bars) >= cb->cycle_load)
		    )
		    {
			gc = style->fg_gc[GTK_STATE_INSENSITIVE];
			color_set = TRUE;
		    }
		    gdk_draw_line(
			drawable, gc,
			xp, y, xp, y + bar_height
		    );
		}
		gdk_gc_set_line_attributes(
		    style->fg_gc[state],
		    1, GDK_LINE_SOLID, GDK_CAP_NOT_LAST, GDK_JOIN_MITER
		);
		gdk_gc_set_line_attributes(
		    style->fg_gc[GTK_STATE_INSENSITIVE],
		    1, GDK_LINE_SOLID, GDK_CAP_NOT_LAST, GDK_JOIN_MITER
		);
	    }
	}
}

/*
 *	Called by ConBarSetAudio() to draw the "Not Connected" page.
 *
 *	All inputs assumed valid.  The background is expected to be
 *	already drawn.  The is_connected member is expected to be FALSE.
 */
static void ConBarDrawAudioNotConnectedPage(
	connection_bar_struct *cb,
	gint state, GtkStyle *style,
	GdkDrawable *drawable, gint width, gint height
)
{
	GdkFont *font;
	gint font_extent_x = 0, font_next_y = 2;

	/* Draw not connected label? */
	font = cb->audio_font_large;
	if(font != NULL)
	{
	    gint x, y;
	    gint lbearing, rbearing, swidth, ascent, descent;
	    const gchar *s = "Not Connected";
	    gdk_string_extents(
		font, s,
		&lbearing, &rbearing, &swidth, &ascent, &descent
	    );
	    x = (width / 2) - (swidth / 2);
	    y = (height / 2) - ((ascent + descent) / 2);
	    gdk_draw_string(
		drawable, font, style->text_gc[state],
		x + lbearing, y + ascent,
		s
	    );
	    font_extent_x = MAX(font_extent_x, x + swidth);
	    font_next_y += ascent + descent + 5;
	}
}

/*
 *	Draws the Audio pixmap buffer and sets it as the background
 *	for the Audio GtkDrawingArea widget's GdkWindow.
 *
 *	This will also update the tooltips for the DSPView if the
 *	given values change from the previous call.
 */
void ConBarSetAudio(
	connection_bar_struct *cb,
	gulong start_time, const gchar *audio_mode_name,
	gint sample_rate, gint sample_size, gint channels,
	glong cycle_us, gfloat cycle_load,
	gint num_fragments, gint fragment_size,
	gboolean flip_stereo, gboolean shellout,
	const guint8 *buf, gint buf_len,
	gfloat *channel_volume_avg,
	gint version_major, gint version_minor,
	const gchar *server_name, const gchar *sound_name
)
{
	gboolean sample_rate_changed;
	gint state, width, height;
	GdkPixmap *pixmap;
	GdkWindow *window;
	GtkStyle *style;
	GtkWidget *w;


	if(cb == NULL)
	    return;

	/* Check if sample rate changed */
	sample_rate_changed = (cb->sample_rate != sample_rate) ?
	    TRUE : FALSE;

	/* Update DSPView tooltip */
	if((CONBAR_DSPVIEW(cb) != NULL) && sample_rate_changed)
	{
	    dspview_struct *dv = CONBAR_DSPVIEW(cb);
	    GtkWidget *w = DSPVIEW_VIEW(dv);
	    gchar *s = NULL;

	    if(sample_rate > 0)
		s = g_strdup_printf(
		    "%i Hz  %i bits  %s",
		    sample_rate, sample_size,
		    (channels == 4) ? "Quad" :
			((channels == 2) ? "Stereo" : "Mono")
		);
	    GUISetWidgetTip(w, s);
	    g_free(s);
	}


	/* Record audio values */
	cb->start_time = start_time;
	if(audio_mode_name != cb->audio_mode_name)
	{
	    g_free(cb->audio_mode_name);
	    cb->audio_mode_name = (audio_mode_name != NULL) ?
		g_strdup(audio_mode_name) : NULL;
	}
	cb->sample_rate = sample_rate;
	cb->sample_size = sample_size;
	cb->channels = CLIP(channels, 0, 4);
	cb->cycle_us = cycle_us;
	cb->cycle_load = cycle_load;
	cb->num_fragments = num_fragments;
	cb->fragment_size = fragment_size;
	cb->flip_stereo = flip_stereo;
	cb->shellout = shellout;
	cb->version_major = version_major;
	cb->version_minor = version_minor;
	if(server_name != cb->server_name)
	{
	    g_free(cb->server_name);
	    cb->server_name = (server_name != NULL) ?
		g_strdup(server_name) : NULL;
	}
	if(sound_name != cb->sound_name)
	{
	    g_free(cb->sound_name);
	    cb->sound_name = (sound_name != NULL) ?
		g_strdup(sound_name) : NULL;
	}

	if(channel_volume_avg != cb->channel_volume_avg)
	{
	    if((channel_volume_avg != NULL) &&
	       (cb->channels > 0)
	    )
		memcpy(
		    cb->channel_volume_avg, channel_volume_avg,
		    cb->channels * sizeof(gfloat)
		);
	}

	/* Get Audio drawing area */
	w = cb->audio_da;
	window = (w != NULL) ? w->window : NULL;
	pixmap = cb->audio_pm;
	if((window == NULL) || (pixmap == NULL))
	    return;

	/* Get widget state, GC, and style */
	state = GTK_WIDGET_STATE(w);
	style = gtk_widget_get_style(w);
	gdk_window_get_size(pixmap, &width, &height);
	if((style == NULL) || (width <= 0) || (height <= 0))
	    return;


	/* Draw background */
	gdk_draw_rectangle(
	    pixmap, style->base_gc[state], TRUE,
	    0, 0, width, height
	);

	/* Begin drawing page specificly.  First check if we are
	 * connected.  If so then draw the page specified by
	 * audio_page_num.  Otherwise draw the "Not Connected" page.
	 */
	if(cb->is_connected)
	{
	    switch(cb->audio_page_num)
	    {
	      case 2:	/* Page 3 */
		ConBarDrawAudioPage3(
		    cb, state, style, pixmap, width, height
		);
		break;
	      case 1:	/* Page 2 */
		ConBarDrawAudioPage2(
		    cb, state, style, pixmap, width, height
		);
		break;
	      default:	/* Page 1 */
		ConBarDrawAudioPage1(
		    cb, state, style, pixmap, width, height
		);
		break;
	    }
	}
	else
	{
	    ConBarDrawAudioNotConnectedPage(
		cb, state, style, pixmap, width, height
	    );
	}

	/* Draw focus? */
	if(GTK_WIDGET_HAS_FOCUS(w) && (state != GTK_STATE_INSENSITIVE))
	{
	    gdk_draw_rectangle(
		pixmap, style->fg_gc[state], FALSE,
		0, 0, width - 1, height - 1
	    );
	}

	/* Set newly drawn background pixmap to the Audio window */
	gdk_window_set_back_pixmap(window, pixmap, FALSE);
}


/*
 *	Updates the shellout button state.
 */
void ConBarSetShellOut(connection_bar_struct *cb, gboolean shellout)
{
	GtkWidget *w;


	if(cb == NULL)
	    return;

	w = cb->shellout_toggle;
	if(w != NULL)
	    gtk_toggle_button_set_active(
		GTK_TOGGLE_BUTTON(w), shellout
	    );
}


/*
 *	Creates a new connection bar and parents it to the given parent.
 */
connection_bar_struct *ConBarNew(
	guint display_flags, gint dspview_width, gint dspview_height,
	gpointer client_data,
	void (*connect_cb)(GtkWidget *, const gchar *, gpointer),
	void (*server_list_cb)(GtkWidget *, gpointer),
	void (*close_cb)(GtkWidget *, gpointer),
	void (*new_console_cb)(GtkWidget *, gpointer),
	void (*new_audio_cb)(GtkWidget *, gpointer),
	void (*new_mixer_cb)(GtkWidget *, gpointer),
	void (*new_player_cb)(GtkWidget *, gpointer),
	void (*options_cb)(GtkWidget *, gpointer),
	void (*refresh_cb)(GtkWidget *, gpointer),
	gboolean (*shellout_cb)(GtkWidget *, gboolean, gpointer),
	void (*shutdown_cb)(GtkWidget *, gpointer),
	GtkWidget *parent, GtkAccelGroup *accelgrp
)
{
	gint border_minor = 2, border_major = 5;
	GtkWidget *w, *parent2, *parent3, *parent4, *main_box;
	connection_bar_struct *cb = (connection_bar_struct *)g_malloc0(
	    sizeof(connection_bar_struct)
	);
	if(cb == NULL)
	    return(NULL);

	/* Reset values */
	cb->initialized = TRUE;
	cb->map_state = FALSE;
	cb->display_flags = display_flags;
	cb->visibility = GDK_VISIBILITY_FULLY_OBSCURED;
	cb->accelgrp = accelgrp;

	cb->is_connected = TRUE;	/* Since widgets sensitive upon creation */
	cb->audio_page_num = 0;

	cb->start_time = 0;
	cb->audio_mode_name = NULL;
	cb->sample_rate = cb->sample_size = cb->channels = 0;
	cb->cycle_us = 0;
	cb->cycle_load = 0.0f;
	cb->num_fragments = cb->fragment_size = 0;
	cb->flip_stereo = FALSE;
	cb->shellout = FALSE;
	cb->version_major = cb->version_minor = 0;
	cb->server_name = cb->sound_name = NULL;

	cb->client_data = client_data;
	cb->connect_cb = connect_cb;
	cb->server_list_cb = server_list_cb;
	cb->close_cb = close_cb;
	cb->new_console_cb = new_console_cb;
	cb->new_audio_cb = new_audio_cb;
	cb->new_mixer_cb = new_mixer_cb;
	cb->new_player_cb = new_player_cb;
	cb->refresh_cb = refresh_cb;
	cb->options_cb = options_cb;
	cb->shellout_cb = shellout_cb;
	cb->shutdown_cb = shutdown_cb;

	cb->audio_font_large = gdk_font_load(
	    CONBAR_AUDIO_FONT_NAME_LARGE
	);
	cb->audio_font_medium = gdk_font_load(
	    CONBAR_AUDIO_FONT_NAME_MEDIUM
	);
	cb->audio_font_small = gdk_font_load(
	    CONBAR_AUDIO_FONT_NAME_SMALL
	);

	cb->audio_shellout_bm = NULL;
	cb->audio_cycle_load_bm = NULL;


	/* Need to hold one refcount of the accel group for ourselves */
	if(accelgrp != NULL)
	    gtk_accel_group_ref(accelgrp);

	/* Create toplevel */
	cb->toplevel = main_box = parent2 = w = gtk_hbox_new(FALSE, 0);
	if(parent != NULL)
	{
	    if(GTK_IS_BOX(parent))
		gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	    else if(GTK_IS_CONTAINER(parent))
		gtk_container_add(GTK_CONTAINER(parent), w);
	}
	gtk_container_border_width(GTK_CONTAINER(w), border_minor);
	parent = w;

	/* Create left pane vbox */
	cb->left_pane_vbox = w = gtk_vbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent2 = w;

	/* Any need to create widgets on first row? */
	if((display_flags & CONBAR_DISPLAY_FLAG_WINDOW_MENU) ||
	   (display_flags & CONBAR_DISPLAY_FLAG_LOCATION)
	)
	{
	    w = gtk_hbox_new(FALSE, border_major);
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent3 = w;

	    /* Create menu button? */
	    if(display_flags & CONBAR_DISPLAY_FLAG_WINDOW_MENU)
	    {
		GtkWidget *menu;

		/* Create menu button */
		cb->window_menu_btn = w = MenuButtonNewH(
		    "Windows", (guchar **)icon_y_menu_20x20_xpm, &menu
		);
		cb->window_menu = menu;
/*		gtk_widget_set_usize(w, 90, -1); */
		gtk_button_set_relief(GTK_BUTTON(w), GTK_RELIEF_NONE);
		MenuButtonSetMapTrigger(w, MENU_BUTTON_MAP_TYPE_PRESSED);
		gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
		gtk_signal_connect(
		    GTK_OBJECT(w), "clicked",
		    GTK_SIGNAL_FUNC(ConBarButtonClickedCB), cb
		);
		GUISetWidgetTip(
		    w,
		    "Click for Windows menu"
		);
		gtk_widget_show(w);

		/* Create menu */
		if(menu != NULL)
		{
		    guchar **icon;
		    const gchar *label;
		    gint accel_key = 0;
		    guint accel_mods = 0;
		    gpointer mclient_data = cb->client_data;
		    void (*func_cb)(GtkWidget *w, gpointer) = NULL;
		    gpointer fw;

#define DO_ADD_MENU_ITEM_LABEL  \
{ \
 w = GUIMenuItemCreate( \
  menu, GUI_MENU_ITEM_TYPE_LABEL, accelgrp, \
  icon, label, accel_key, accel_mods, (void **)&fw, \
  mclient_data, func_cb \
 ); \
}
#define DO_ADD_MENU_SEP         \
{ \
 w = GUIMenuItemCreate( \
  menu, GUI_MENU_ITEM_TYPE_SEPARATOR, NULL, \
  NULL, NULL, 0, 0, NULL, \
  NULL, NULL \
 ); \
}
		    if(new_console_cb != NULL)
		    {
		        icon = (guchar **)icon_yconsole_20x20_xpm;
		        label = "New Console";
		        accel_key = 'n';
		        accel_mods = GDK_CONTROL_MASK;
		        func_cb = new_console_cb;
		        DO_ADD_MENU_ITEM_LABEL
		    }

		    if(new_audio_cb != NULL)
		    {
			icon = (guchar **)icon_yaudio_20x20_xpm;
			label = "New Audio";
			accel_key = 'a';
			accel_mods = GDK_CONTROL_MASK;
			func_cb = new_audio_cb;
			DO_ADD_MENU_ITEM_LABEL
		    }

		    if(new_mixer_cb != NULL)
		    {
			icon = (guchar **)icon_ymixer_20x20_xpm;
			label = "New Mixer";
			accel_key = 'm';
			accel_mods = GDK_CONTROL_MASK;
			func_cb = new_mixer_cb;
			DO_ADD_MENU_ITEM_LABEL
		    }

		    if(new_player_cb != NULL)
		    {
			icon = (guchar **)icon_yplayer_20x20_xpm;
			label = "New Player";
			accel_key = 'p';
			accel_mods = GDK_CONTROL_MASK;
			func_cb = new_player_cb;
			DO_ADD_MENU_ITEM_LABEL
		    }

		    DO_ADD_MENU_SEP

		    icon = (guchar **)icon_options2_20x20_xpm;
		    label = "Options...";
		    accel_key = 'o';
		    accel_mods = GDK_CONTROL_MASK;
		    func_cb = options_cb;
		    DO_ADD_MENU_ITEM_LABEL

		    DO_ADD_MENU_SEP

		    icon = (guchar **)icon_close_20x20_xpm;
		    label = "Close";
		    accel_key = 'w';
		    accel_mods = GDK_CONTROL_MASK;
		    func_cb = close_cb;
		    DO_ADD_MENU_ITEM_LABEL

		    if(!GTK_WIDGET_REALIZED(menu))
			gtk_widget_realize(menu);

#undef DO_ADD_MENU_SEP
#undef DO_ADD_MENU_ITEM_LABEL
		}	/* Create menu */
	    }

	    /* Create location combo? */
	    if(display_flags & CONBAR_DISPLAY_FLAG_LOCATION)
	    {
		gpointer combo_rtn;
		GtkCombo *combo;

		w = gtk_hbox_new(FALSE, 0);
		gtk_box_pack_start(GTK_BOX(parent3), w, TRUE, TRUE, 0);
		gtk_widget_show(w);
		parent4 = w;

		cb->location_btn = w = GUIButtonPixmapLabelH(
		    (guint8 **)icon_yserver_location_20x20_xpm,
#ifdef PROG_LANGUAGE_ENGLISH
		    "Location:",
#endif
#ifdef PROG_LANGUAGE_SPANISH
		    "La Ubicacin:",
#endif
#ifdef PROG_LANGUAGE_FRENCH
		    "Emplacement:",
#endif
		    NULL
		);
/*		gtk_widget_set_usize(w, 100, -1); */
		gtk_button_set_relief(GTK_BUTTON(w), GTK_RELIEF_NONE);
		gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
		gtk_signal_connect(
		    GTK_OBJECT(w), "clicked",
		    GTK_SIGNAL_FUNC(server_list_cb), cb->client_data
		);
		GUISetWidgetTip(
		    w,
		    "Click for a list of Y servers"
		);
		if(accelgrp != NULL)
		    gtk_accel_group_add(
			accelgrp, 'l', GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE,
			GTK_OBJECT(w), "clicked"
		    );
		gtk_widget_show(w);

		w = (GtkWidget *)GUIComboCreate(
		    NULL,	/* Label */
		    "",		/* Initial value */
		    NULL,	/* Initial list */
		    10,		/* Maximum items */
		    &combo_rtn,
		    cb, ConBarLocationCB, NULL
		);
		combo = (GtkCombo *)combo_rtn;
		cb->location_combo = GTK_WIDGET(combo);
		gtk_box_pack_start(GTK_BOX(parent4), w, TRUE, TRUE, 0);
		GUISetWidgetTip(
		    combo->entry,
"Enter the address of the Y server that you want to connect to (format\
 of address should be \"<address>:<port>\") or click on \"Location\"\
 for a list of Y servers"
		);
		gtk_widget_show(w);
	    }
	}

	/* Any need to create widgets on second row? */
	if(display_flags & CONBAR_DISPLAY_FLAG_AUDIO)
	{
	    gint	width = CONBAR_AUDIO_DA_WIDTH,
			height = CONBAR_AUDIO_DA_HEIGHT;

	    /* Hbox for Audio widgets */
	    w = gtk_hbox_new(FALSE, border_major);
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_container_border_width(GTK_CONTAINER(w), border_minor);
	    gtk_widget_show(w);
	    parent3 = w;

	    w = gtk_frame_new(NULL);
	    gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_IN);
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent4 = w;

	    /* Audio drawing area */
	    cb->audio_da = w = gtk_drawing_area_new();
	    gtk_drawing_area_size(GTK_DRAWING_AREA(w), width, height);
	    gtk_widget_add_events(
		w,
		GDK_STRUCTURE_MASK | GDK_EXPOSURE_MASK |
		GDK_KEY_PRESS_MASK | GDK_KEY_RELEASE_MASK |
		GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
		GDK_ENTER_NOTIFY_MASK | GDK_LEAVE_NOTIFY_MASK |
		GDK_FOCUS_CHANGE_MASK
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "configure_event",
		GTK_SIGNAL_FUNC(ConBarConfigureNotifyCB), cb
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "focus_in_event",
		GTK_SIGNAL_FUNC(ConBarFocusChangeEventCB), cb
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "focus_out_event",
		GTK_SIGNAL_FUNC(ConBarFocusChangeEventCB), cb
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "key_press_event",
		GTK_SIGNAL_FUNC(ConBarKeyEventCB), cb
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "key_release_event",
		GTK_SIGNAL_FUNC(ConBarKeyEventCB), cb
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "button_press_event",
		GTK_SIGNAL_FUNC(ConBarButtonEventCB), cb
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "button_release_event",
		GTK_SIGNAL_FUNC(ConBarButtonEventCB), cb
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "realize",
		GTK_SIGNAL_FUNC(ConBarRealizeCB), cb
	    );
	    gtk_container_add(GTK_CONTAINER(parent4), w);
	    GTK_WIDGET_SET_FLAGS(w, GTK_CAN_FOCUS | GTK_CAN_DEFAULT);
	    gtk_widget_show(w);

	    cb->audio_pm = NULL;
	}


	/* Create DSPView, refresh, shellout, or shutdown? */
	if((display_flags & CONBAR_DISPLAY_FLAG_DSPVIEW) ||
	   (display_flags & CONBAR_DISPLAY_FLAG_REFRESH) ||
	   (display_flags & CONBAR_DISPLAY_FLAG_SHELLOUT) ||
	   (display_flags & CONBAR_DISPLAY_FLAG_SHUTDOWN)
	)
	{
	    /* Right pane vbox */
	    cb->right_pane_vbox = w = gtk_vbox_new(
		(display_flags & CONBAR_DISPLAY_FLAG_AUDIO) ? FALSE :
		    ((display_flags & CONBAR_DISPLAY_FLAG_REFRESH) ?
			FALSE : TRUE),
		border_major
	    );
	    gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent2 = w;

	    /* Create DSPView? */
	    if(display_flags & CONBAR_DISPLAY_FLAG_DSPVIEW)
	    {
		dspview_struct *dv;

		cb->dspview = dv = DSPViewNew(
		    parent2,
		    (dspview_width > 0) ? dspview_width : 80 + 2,
		    (dspview_height > 0) ? dspview_height : 32 + 2
		);
		w = DSPVIEW_VIEW(dv);
		gtk_widget_add_events(
		    w,
		    GDK_ENTER_NOTIFY_MASK | GDK_LEAVE_NOTIFY_MASK
		);
		DSPViewMap(dv);
	    }

	    parent3 = NULL;

	    /* Create refresh button? */
	    if(display_flags & CONBAR_DISPLAY_FLAG_REFRESH)
	    {
		if(parent3 == NULL)
		{
		    w = gtk_hbox_new(FALSE, border_minor);
		    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
		    gtk_widget_show(w);
		    parent3 = w;
		}

		cb->refresh_btn = w = GUIButtonPixmap(
		    (guint8 **)icon_reload_20x20_xpm
		);
		gtk_widget_set_usize(w, 20 + 2, 20 + 2);
		gtk_button_set_relief(GTK_BUTTON(w), GTK_RELIEF_NONE);
		gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
		gtk_signal_connect(
		    GTK_OBJECT(w), "clicked",
		    GTK_SIGNAL_FUNC(refresh_cb), cb->client_data
		);
		GUISetWidgetTip(
		    w,
		    "Refresh values from Y server"
		);
		if(accelgrp != NULL)
		    gtk_accel_group_add(
			accelgrp, GDK_F5, 0, GTK_ACCEL_VISIBLE,
			GTK_OBJECT(w), "clicked"
		    );
		gtk_widget_show(w);
	    }

	    /* Create shellout toggle? */
	    if(display_flags & CONBAR_DISPLAY_FLAG_SHELLOUT)
	    {
		if(parent3 == NULL)
		{
		    w = gtk_hbox_new(FALSE, border_minor);
		    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
		    gtk_widget_show(w);
		    parent3 = w;
		}

		cb->shellout_toggle = w = GUIToggleButtonPixmap(
		    (guint8 **)icon_shellout_20x20_xpm
		);
		gtk_widget_set_usize(w, 20 + 2, 20 + 2);
		gtk_button_set_relief(GTK_BUTTON(w), GTK_RELIEF_NONE);
		gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
		gtk_signal_connect(
		    GTK_OBJECT(w), "toggled",
		    GTK_SIGNAL_FUNC(ConBarShelloutToggledCB), cb
		);
		GUISetWidgetTip(
		    w,
		    "Toggle shellout/unshell of Y server"
		);
		gtk_widget_show(w);
	    }

	    /* Create shutdown button? */
	    if(display_flags & CONBAR_DISPLAY_FLAG_SHUTDOWN)
	    {
		if(parent3 == NULL)
		{
		    w = gtk_hbox_new(FALSE, border_minor);
		    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
		    gtk_widget_show(w);
		    parent3 = w;
		}

		cb->shutdown_btn = w = GUIButtonPixmap(
		    (guint8 **)icon_shutdown_20x20_xpm
		);
		gtk_widget_set_usize(w, 20 + 2, 20 + 2);
		gtk_button_set_relief(GTK_BUTTON(w), GTK_RELIEF_NONE);
		gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
		gtk_signal_connect(
		    GTK_OBJECT(w), "clicked",
		    GTK_SIGNAL_FUNC(shutdown_cb), cb->client_data
		);
		GUISetWidgetTip(
		    w,
		    "Shutdown Y server"
		);
		gtk_widget_show(w);
	    }

	}

	return(cb);
}

/*
 *	Redraws the DSPView and the Audio display.
 */
void ConBarDraw(
	connection_bar_struct *cb,
	const guint8 *buf, gint len
)
{
	GdkWindow *window;
	GtkWidget *w;


	if(cb == NULL)
	    return;

	DSPViewDraw(cb->dspview, buf, len);

	w = cb->audio_da;
	window = (w != NULL) ? w->window : NULL;
	if(window != NULL)
	    gdk_window_clear(window);
}

/*
 *	Maps the connection bar.
 */
void ConBarMap(connection_bar_struct *cb)
{
	if((cb == NULL) ? TRUE : !cb->initialized)
	    return;

	if(!cb->map_state)
	{
	    GtkWidget *w = cb->toplevel;
	    if(w != NULL)
		gtk_widget_show(w);
	    cb->map_state = TRUE;
	}
}

/*
 *	Unmaps the connection bar.
 */
void ConBarUnmap(connection_bar_struct *cb)
{
	if((cb == NULL) ? TRUE : !cb->initialized)
	    return;

	if(cb->map_state)
	{
	    GtkWidget *w = cb->toplevel;
	    if(w != NULL)
		gtk_widget_hide(w);
	    cb->map_state = FALSE;
	}
}

/*
 *	Destroys the connection bar.
 */
void ConBarDelete(connection_bar_struct *cb)
{
	gchar **s;
	GdkBitmap **bm;
	GdkPixmap **pm;
	GdkFont **font;
	GtkWidget **w;


	if(cb == NULL)
	    return;

	if(cb->initialized)
	{
#define DO_FREE_S	\
{ if(*s != NULL) { g_free(s); *s = NULL; } }
#define DO_UNREF_FONT	\
{ if(*font != NULL) { gdk_font_unref(*font); *font = NULL; } }
#define DO_UNREF_BITMAP		\
{ if(*bm != NULL) { gdk_bitmap_unref(*bm); *bm = NULL; } }
#define DO_UNREF_PIXMAP		\
{ if(*pm != NULL) { gdk_pixmap_unref(*pm); *pm = NULL; } }
#define DO_DESTROY_WIDGET	\
{ if(*w != NULL) { gtk_widget_destroy(*w); *w = NULL; } }

	    /* Audio drawing area bitmaps */
	    bm = &cb->audio_shellout_bm;
	    DO_UNREF_BITMAP
	    bm = &cb->audio_cycle_load_bm;
	    DO_UNREF_BITMAP

	    /* DSPView */
	    DSPViewDelete(cb->dspview);
	    cb->dspview = NULL;

	    w = &cb->shutdown_btn;
	    DO_DESTROY_WIDGET
	    w = &cb->shellout_toggle;
	    DO_DESTROY_WIDGET
	    w = &cb->refresh_btn;
	    DO_DESTROY_WIDGET
	    w = &cb->audio_da;
	    DO_DESTROY_WIDGET
	    w = &cb->location_combo;
	    DO_DESTROY_WIDGET
	    w = &cb->location_btn;
	    DO_DESTROY_WIDGET
	    w = &cb->window_menu_btn;
	    DO_DESTROY_WIDGET
	    w = &cb->right_pane_vbox;
	    DO_DESTROY_WIDGET
	    w = &cb->left_pane_vbox;
	    DO_DESTROY_WIDGET
	    w = &cb->toplevel;
	    DO_DESTROY_WIDGET

	    font = &cb->audio_font_large;
	    DO_UNREF_FONT
	    font = &cb->audio_font_medium;
	    DO_UNREF_FONT
	    font = &cb->audio_font_small;
	    DO_UNREF_FONT

	    pm = &cb->audio_pm;
	    DO_UNREF_PIXMAP

	    if(cb->accelgrp != NULL)
	    {
		gtk_accel_group_unref(cb->accelgrp);
		cb->accelgrp = NULL;
	    }

	    s = &cb->audio_mode_name;
	    DO_FREE_S
	    s = &cb->server_name;
	    DO_FREE_S
	    s = &cb->sound_name;
	    DO_FREE_S

#undef DO_DESTROY_WIDGET
#undef DO_UNREF_PIXMAP
#undef DO_UNREF_BITMAP
#undef DO_UNREF_FONT
#undef DO_FREE_S
	}

	g_free(cb);
}
