/*
				   Connection Bar
 */
#ifndef CONBAR_H
#define CONBAR_H

#include <gtk/gtk.h>
#include "dspview.h"

/*
 *	Display flags:
 */
#define CONBAR_DISPLAY_FLAG_WINDOW_MENU		(1 << 0)
#define CONBAR_DISPLAY_FLAG_LOCATION		(1 << 1)
#define CONBAR_DISPLAY_FLAG_DSPVIEW		(1 << 2)
#define CONBAR_DISPLAY_FLAG_AUDIO		(1 << 3)
#define CONBAR_DISPLAY_FLAG_REFRESH		(1 << 4)
#define CONBAR_DISPLAY_FLAG_SHELLOUT		(1 << 5)
#define CONBAR_DISPLAY_FLAG_SHUTDOWN		(1 << 6)

/*
 *	Total audio pages:
 */
#define CONBAR_TOTAL_AUDIO_PAGES		3

/*
 *	Connection Bar:
 */
typedef struct {

	gboolean initialized;
	gboolean map_state;
	guint display_flags;		/* Any of CONBAR_DISPLAY_FLAG_* */
	GdkVisibilityState visibility;
	GtkAccelGroup *accelgrp;

	GdkFont		*audio_font_large,
			*audio_font_medium,
			*audio_font_small;

	GdkBitmap	*audio_shellout_bm,
			*audio_cycle_load_bm;

	GtkWidget	*toplevel,
			*left_pane_vbox, *right_pane_vbox,
			*window_menu_btn,
			*window_menu,
			*location_btn,	/* Server list button */
			*location_combo,
			*audio_da,	/* Audio stats */
			*refresh_btn,
			*shellout_toggle,
			*shutdown_btn;
	GdkPixmap	*audio_pm;	/* Back buffer for audio_da */

	dspview_struct *dspview;

	gboolean is_connected;
	gint audio_page_num;

	gpointer client_data;
	void (*connect_cb)(GtkWidget *, const gchar *, gpointer);
	void (*server_list_cb)(GtkWidget *, gpointer);
	void (*close_cb)(GtkWidget *, gpointer);
	void (*new_console_cb)(GtkWidget *, gpointer);
	void (*new_audio_cb)(GtkWidget *, gpointer);
	void (*new_mixer_cb)(GtkWidget *, gpointer);
	void (*new_player_cb)(GtkWidget *, gpointer);
	void (*options_cb)(GtkWidget *, gpointer);
	void (*refresh_cb)(GtkWidget *, gpointer);
	gboolean (*shellout_cb)(GtkWidget *, gboolean, gpointer);
	void (*shutdown_cb)(GtkWidget *, gpointer);

	/* Recorded Audio values */
	gulong start_time;
	gchar *audio_mode_name;
	gint sample_rate, sample_size, channels;
	glong cycle_us;
	gfloat cycle_load;
	gint num_fragments, fragment_size;
	gboolean flip_stereo, shellout;
	gint version_major, version_minor;
	gchar *server_name, *sound_name;

	gfloat channel_volume_avg[4];

} connection_bar_struct;


#define CONBAR(p)	((connection_bar_struct *)(p))
#define CONBAR_TOPLEVEL(p)		(((p) != NULL) ? \
 (CONBAR(p)->toplevel) : NULL)
#define CONBAR_DSPVIEW(p)		(((p) != NULL) ? \
 (CONBAR(p)->dspview) : NULL)
#define CONBAR_MENU_BUTTON(p)		(((p) != NULL) ? \
 (CONBAR(p)->menu_btn) : NULL)
#define CONBAR_MENU(p)			(((p) != NULL) ? \
 (CONBAR(p)->menu) : NULL)
#define CONBAR_LOCATION_BUTTON(p)	(((p) != NULL) ? \
 (CONBAR(p)->location_btn) : NULL)
#define CONBAR_LOCATION_COMBO(p)	(((p) != NULL) ? \
 (CONBAR(p)->location_combo) : NULL)
#define CONBAR_AUDIO(p)			(((p) != NULL) ? \
 (CONBAR(p)->audio_da) : NULL)
#define CONBAR_REFRESH_BUTTON(p)	(((p) != NULL) ? \
 (CONBAR(p)->refresh_btn) : NULL)
#define CONBAR_SHELLOUT_TOGGLE(p)	(((p) != NULL) ? \
 (CONBAR(p)->shellout_toggle) : NULL)
#define CONBAR_SHUTDOWN_BUTTON(p)	(((p) != NULL) ? \
 (CONBAR(p)->shutdown_btn) : NULL)

#define CONBAR_IS_CONNECTED	(((p) != NULL) ? \
 (CONBAR(p)->is_connected) : FALSE)



extern gchar *ConBarGetLocation(connection_bar_struct *cb);
extern void ConBarSetLocation(
	connection_bar_struct *cb, const gchar *loc_str,
	gboolean emit_signal, gboolean record_history
);

extern void ConBarSetConnectState(connection_bar_struct *cb, gboolean is_connected);

extern dspview_struct *ConBarGetDSPView(connection_bar_struct *cb);

extern void ConBarSetAudioColors(
	connection_bar_struct *cb,
	GdkColor *color_bg, GdkColor *color_fg,
	GdkColor *color_fg_dim
);
extern void ConBarSetAudioFonts(
	connection_bar_struct *cb,
	const gchar *font_name_large, const gchar *font_name_medium,
	const gchar *font_name_small
);
extern void ConBarSetAudio(
	connection_bar_struct *cb, 
	gulong start_time, const gchar *audio_mode_name,
	gint sample_rate, gint sample_size, gint channels,
	glong cycle_us, gfloat cycle_load,
	gint num_fragments, gint fragment_size,
	gboolean flip_stereo, gboolean shellout,
	const guint8 *buf, gint buf_len,
	gfloat *channel_volume_avg,
	gint version_major, gint version_minor,
	const gchar *server_name, const gchar *sound_name
);
extern void ConBarSetShellOut(connection_bar_struct *cb, gboolean shellout);

extern connection_bar_struct *ConBarNew(
	guint display_flags, gint dspview_width, gint dspview_height,
	gpointer client_data,
	void (*connect_cb)(GtkWidget *, const gchar *, gpointer),
	void (*server_list_cb)(GtkWidget *, gpointer),
	void (*close_cb)(GtkWidget *, gpointer),
	void (*new_console_cb)(GtkWidget *, gpointer),
	void (*new_audio_cb)(GtkWidget *, gpointer),
	void (*new_mixer_cb)(GtkWidget *, gpointer),
	void (*new_player_cb)(GtkWidget *, gpointer),
	void (*options_cb)(GtkWidget *, gpointer),
	void (*refresh_cb)(GtkWidget *, gpointer),
	gboolean (*shellout_cb)(GtkWidget *, gboolean, gpointer),
	void (*shutdown_cb)(GtkWidget *, gpointer),
	GtkWidget *parent, GtkAccelGroup *accelgrp
);
extern void ConBarDraw(
	connection_bar_struct *cb,
	const guint8 *buf, gint len
);
extern void ConBarMap(connection_bar_struct *cb);
extern void ConBarUnmap(connection_bar_struct *cb);
extern void ConBarDelete(connection_bar_struct *cb);




#endif	/* CONBAR_H */
