/*
				DSP View Widget
 */
#ifndef DSPVIEW_H
#define DSPVIEW_H

#include <gtk/gtk.h>


/*
 *	Display Styles:
 */
typedef enum {
	DSPVIEW_DISPLAY_STYLE_WAVE,		/* Solid Volume waves */
	DSPVIEW_DISPLAY_STYLE_SGRAPH,		/* Sine (plotted) graph */
	DSPVIEW_DISPLAY_STYLE_ANALYZER,		/* Horizontal blocks */
	DSPVIEW_DISPLAY_STYLE_LIGHTENING,	/* Lightening wave */
	DSPVIEW_DISPLAY_STYLE_PLASMA,		/* Blobs of plasma */
	DSPVIEW_DISPLAY_STYLE_NOVA		/* Exploding star */
} dspview_display_style;

/*
 *	Number of analyzer blocks to display:
 */
#define DSPVIEW_TOTAL_ANALYZER_BLOCKS	20

/*
 *	Merge Modes:
 *
 *	Note that certain merge modes do not have affect on all
 *	dspview_display_styles since some of them impose their own
 *	merge modes
 */
typedef enum {
	DSPVIEW_MERGE_MODE_SET,
	DSPVIEW_MERGE_MODE_SET_ALPHA,		/* Transparent by alpha */
	DSPVIEW_MERGE_MODE_ADD,
	DSPVIEW_MERGE_MODE_SUBTRACT,
	DSPVIEW_MERGE_MODE_BLEND,
	DSPVIEW_MERGE_MODE_MULTIPLY
} dspview_merge_mode;

/*
 *	Default DSP color (0xAABBGGRR):
 */
#define DSPVIEW_DEF_DSP_COLOR	0xffffa0c4

/*
 *	Default fade speed coefficient:
 */
#define DSPVIEW_DEF_FADE_COEFF	0.3

/*
 *	DSPView:
 */
typedef struct _dspview_struct		dspview_struct;
struct _dspview_struct {

	gboolean	map_state;
	GtkWidget	*toplevel,
			*view_da;

	gint		channels;	/* 1, 2, or 4 */

	dspview_display_style	display_style;
	dspview_merge_mode	merge_mode;

	guint32		dsp_color;	/* 0xAABBGGRR */

	gfloat		fade_coeff;	/* Fade speed coefficient (0.0 to 1.0],
					 * 1.0 for strongest and fastest fade
					 *
					 * The fade speed coefficient has
					 * different affects depending on
					 * the display_style and has no
					 * affect on certain display
					 * styles */

	/* view_da "back" buffer 32 bit 0xAABBGGRR format */
	guint8	*buf;
	gint	width, height, bpl, bpp;

	/* DSP "history" buffer (not used with some display styles), 
	 * allocation size and format must match the view_da "back"
	 * buffer buf
	 */
	guint8  *dsp_buf;

	/* Background image buffer (shared) */
	const guint8	*bg_buf;
	gint		bg_width, bg_height,
			bg_bpl, bg_bpp;

	/* Analyzer history block positions, used only if display_style
	 * is set to DSPVIEW_DISPLAY_STYLE_ANALYZER
	 */
	guint8 analyzer_history_buf[DSPVIEW_TOTAL_ANALYZER_BLOCKS];

};
#define DSPVIEW(p)	((dspview_struct *)(p))
#define DSPVIEW_TOPLEVEL(p)	(((p) != NULL) ? \
 (DSPVIEW(p)->toplevel) : NULL)
#define DSPVIEW_VIEW(p)		(((p) != NULL) ? \
 (DSPVIEW(p)->view_da) : NULL)


extern dspview_struct *DSPViewNew(
	GtkWidget *parent, gint width, gint height
);
extern void DSPViewSetBG(
	dspview_struct *dv,
	const guint8 *buf,
	gint width, gint height, gint bpl, gint bpp
);
extern void DSPViewSetChannels(
	dspview_struct *dv, gint channels
);
extern void DSPViewSetDisplayStyle(
	dspview_struct *dv, dspview_display_style style
);
extern void DSPViewSetMergeMode(
	dspview_struct *dv, dspview_merge_mode mode
);
extern void DSPViewSetDSPColor(
	dspview_struct *dv, guint32 c
);
extern void DSPViewSetFadeCoeff(
	dspview_struct *dv, gfloat v
);
extern void DSPViewDraw(
	dspview_struct *dv,
	const guint8 *buf, gint len
);
extern void DSPViewMap(dspview_struct *dv);
extern void DSPViewUnmap(dspview_struct *dv);
extern void DSPViewDelete(dspview_struct *dv);


#endif	/* DSPVIEW_H */
