#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <time.h>
#include <signal.h>
#include <unistd.h>

#include <Imlib.h>

#include <Y2/Y.h>
#include <Y2/Ylib.h>
#include <Y2/Ymixercodes.h>

#include <gtk/gtk.h>
#include <gdk/gdkrgb.h>
#include <gdk/gdkx.h>

#include "../include/string.h"
#include "../include/disk.h"
#include "../include/imlibio.h"

#include "guiutils.h"
#include "cdialog.h"
#include "fprompt.h"
#include "csd.h"
#include "fsd.h"
#include "fb.h"
#include "pdialog.h"

#include "cfg.h"
#include "cfg_fio.h"
#include "optwin.h"

#include "yconsolewin.h"
#include "yconsolecb.h"
#include "yaudio.h"
#include "yaudiocb.h"
#include "ymixer.h"
#include "ymixercb.h"
#include "yplayer.h"
#include "yplayercb.h"
#include "optwin.h"
#include "ycoptwin.h"
#include "yserverlist.h"
#include "startupdialog.h"
#include "yc.h"
#include "ycop.h"
#include "ycimg.h"
#include "yccfglist.h"
#include "config.h"


static void YCSignalCB(int s);
static gint YCTimeoutCB(gpointer data);
static gint YCTimeout2CB(gpointer data);
static gint YCTimeout3CB(gpointer data);

static void YCSetDefaultConfiguration(
	yc_core_struct *core_ptr, cfg_item_struct *cfg_list
);
static yc_core_struct *YCNew(
	gint argc, gchar *argv[], gpointer imlib_handle,
	gint *err_rtn
);
static void YCDelete(yc_core_struct *core_ptr, gboolean save_config);


static yc_core_struct *yc_core;


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)

#define GET_CFG_I(p)		YC_GET_CFG_I(core_ptr, (p))
#define GET_CFG_L(p)		YC_GET_CFG_L(core_ptr, (p))
#define GET_CFG_UL(p)		YC_GET_CFG_UL(core_ptr, (p))
#define GET_CFG_F(p)		YC_GET_CFG_F(core_ptr, (p))
#define GET_CFG_D(p)		YC_GET_CFG_D(core_ptr, (p))
#define GET_CFG_S(p)		YC_GET_CFG_S(core_ptr, (p))
#define GET_CFG_COLOR(p)	YC_GET_CFG_COLOR(core_ptr, (p))
#define GET_CFG_INTLIST(p)	YC_GET_CFG_INTLIST(core_ptr, (p))

#define SET_CFG_I(p,x)		YC_SET_CFG_I(core_ptr, (p), (x))
#define SET_CFG_L(p,x)		YC_SET_CFG_L(core_ptr, (p), (x))
#define SET_CFG_UL(p,x)		YC_SET_CFG_UL(core_ptr, (p), (x))
#define SET_CFG_F(p,x)		YC_SET_CFG_F(core_ptr, (p), (x))
#define SET_CFG_D(p,x)		YC_SET_CFG_D(core_ptr, (p), (x))
#define SET_CFG_S(p,x)		YC_SET_CFG_S(core_ptr, (p), (x))
#define SET_CFG_COLOR(p,x)	YC_SET_CFG_COLOR(core_ptr, (p), (x))
#define SET_CFG_INTLIST(p,x)	YC_SET_CFG_INTLIST(core_ptr, (p), (x))


/*
 *	UNIX signal callback.
 */
static void YCSignalCB(int s)
{
	static int segfault_count = 0;
	yc_core_struct *core_ptr = yc_core;
	if(core_ptr == NULL)
	    return;

	switch(s)
	{
	  case SIGINT:
	  case SIGTERM:
	    core_ptr->close_all_windows = TRUE;
	    break;
	  case SIGSEGV:
	    segfault_count++;   /* Count this segfault */
	    fprintf(
		stderr,
 "%s triggered a segmentation fault (%i times)\n",
		PROG_NAME,
		segfault_count
	    );
	    /* 3 or more segfaults? */
	    if(segfault_count >= 3)
	    {
		fprintf(
		    stderr,
 "%s attempting immediate process exit().\n",
		    PROG_NAME
		);
		exit(1);        /* Explicit exit */
	    }
	    else
	    {
		core_ptr->close_all_windows = TRUE;
	    }
	    break;
#ifdef SIGHUP
	  case SIGHUP:
	    break;
#endif
#ifdef SIGPIPE
	  case SIGPIPE:
	    break;
#endif
	}
}

/*
 *	Timeout callback.
 */
static gint YCTimeoutCB(gpointer data)
{
	gint i;
	yconsole_struct *yc;
	yaudio_struct *ya;
	ymixer_struct *ym;
	yplayer_struct *yp;
	optwin_struct *ow;
	yc_core_struct *core_ptr = YC_CORE(data);
	if(core_ptr == NULL)
	    return(FALSE);

	/* Iterate through all YConsole windows */
	for(i = 0; i < core_ptr->total_yconsoles; i++)
	{
	    yc = core_ptr->yconsole[i];
	    if(yc == NULL)
		continue;

	    if(!yc->map_state)
		continue;

	    YConsoleManageCB(yc);
	}

	/* Iterate through all YAudio windows */
	for(i = 0; i < core_ptr->total_yaudios; i++)
	{
	    ya = core_ptr->yaudio[i];
	    if(ya == NULL)
		continue;

	    if(!ya->map_state)
		continue;

	    YAudioManageCB(ya);
	}

	/* Iterate through all YMixer windows */
	for(i = 0; i < core_ptr->total_ymixers; i++)
	{
	    ym = core_ptr->ymixer[i];
	    if(ym == NULL)
		continue;

	    if(!ym->map_state)
		continue;

	    YMixerManageCB(ym);
	}

	/* Iterate through all YPlayer windows */
	for(i = 0; i < core_ptr->total_yplayers; i++)
	{
	    yp = core_ptr->yplayer[i];
	    if(yp == NULL)
		continue;

	    if(!yp->map_state)
		continue;

	    YPlayerManageCB(yp);
	}

	/* Options window */
	ow = core_ptr->optwin;
	if(ow != NULL)
	{
	    if(ow->map_state)
		YCOptWinManageCB(ow, core_ptr);
	}


	return(TRUE);
}

/*
 *	Secondary timeout callback.
 *
 *	Called in 500 ms intervals.
 */
static gint YCTimeout2CB(gpointer data)
{
	gboolean all_windows_closed = TRUE;
	gint i;
	yconsole_struct *yc;
	yaudio_struct *ya;
	ymixer_struct *ym;
	yplayer_struct *yp;
	startup_dialog_struct *sd;
	yc_core_struct *core_ptr = YC_CORE(data);
	if(core_ptr == NULL)
	    return(FALSE);


	/* Iterate through all YConsole windows */
	for(i = 0; i < core_ptr->total_yconsoles; i++)
	{
	    yc = core_ptr->yconsole[i];
	    if(yc == NULL)
		continue;

	    if(!yc->map_state)
	    {
		YConsoleDelete(yc);
		core_ptr->yconsole[i] = yc = NULL;
		continue;
	    }

	    if(all_windows_closed)
		all_windows_closed = FALSE;

	    YConsoleManage2CB(yc);
	}

	/* Iterate through all YAudio windows */
	for(i = 0; i < core_ptr->total_yaudios; i++)
	{
	    ya = core_ptr->yaudio[i];
	    if(ya == NULL)
		continue;

	    if(!ya->map_state)
	    {
		YAudioDelete(ya);
		core_ptr->yaudio[i] = ya = NULL;
		continue;
	    }

	    if(all_windows_closed)
		all_windows_closed = FALSE;

	    YAudioManage2CB(ya);
	}

	/* Iterate through all YMixer windows */
	for(i = 0; i < core_ptr->total_ymixers; i++)
	{
	    ym = core_ptr->ymixer[i];
	    if(ym == NULL)
		continue;

	    if(!ym->map_state)
	    {
		YMixerDelete(ym);
		core_ptr->ymixer[i] = ym = NULL;
		continue;
	    }

	    if(all_windows_closed)
		all_windows_closed = FALSE;

	    YMixerManage2CB(ym);
	}

	/* Iterate through all YPlayer windows */
	for(i = 0; i < core_ptr->total_yplayers; i++)
	{
	    yp = core_ptr->yplayer[i];
	    if(yp == NULL)
		continue;

	    if(!yp->map_state)
	    {
		YPlayerDelete(yp);
		core_ptr->yplayer[i] = yp = NULL;
		continue;
	    }

	    if(all_windows_closed)
		all_windows_closed = FALSE;

	    YPlayerManage2CB(yp);
	}

	/* YServer list window */
	if(core_ptr->yserver_list != NULL)
	{
	    yserver_list_struct *sl = core_ptr->yserver_list;
	    if(sl->map_state)
	    {
		if(all_windows_closed)
		    all_windows_closed = FALSE;
	    }
	}

	/* Options window */
	if(core_ptr->optwin != NULL)
	{
	    optwin_struct *ow = core_ptr->optwin;
	    if(ow->map_state)
	    {
		if(all_windows_closed)
		    all_windows_closed = FALSE;
	    }
	}

	/* Iterate through all Startup Dialogs */
	for(i = 0; i < core_ptr->total_startup_dialogs; i++)
	{
	    sd = core_ptr->startup_dialog[i];
	    if(sd == NULL)
		continue;

	    if(!sd->map_state)
	    {
		StartupDialogDelete(sd);
		core_ptr->startup_dialog[i] = sd = NULL;
		continue;
	    }

	    if(all_windows_closed)
		all_windows_closed = FALSE;
	}


	/* When all windows are closed, do not call this timeout anymore
	 * and break out of the main GTK+ loop.
	 */
	if(all_windows_closed || core_ptr->close_all_windows)
	{
	    core_ptr->toid2 = (guint)-1;
	    gtk_main_quit();
	    return(FALSE);
	}
	else
	{
	    return(TRUE);
	}
}


/*
 *	Third timeout callback.
 *
 *      Called in 60000 ms intervals.
 */
static gint YCTimeout3CB(gpointer data)
{
	yc_core_struct *core_ptr = YC_CORE(data);
	if(core_ptr == NULL)
	    return(FALSE);

	/* Check for unused memory and deallocate them */
	YCReclaimMemory(core_ptr);

	return(TRUE);
}


/*
 *	Sets default values to the specified cfg list.
 */
static void YCSetDefaultConfiguration(
	yc_core_struct *core_ptr, cfg_item_struct *cfg_list
)
{
	gchar *s, *cwd, *home;
	GList *glist;
	cfg_color_struct color;
	cfg_intlist_struct *intlist;

	if(cfg_list == NULL)
	    return;

	cwd = STRDUP(g_get_current_dir());
	home = STRDUP(g_getenv("HOME"));

	SET_CFG_I(YC_CFG_PARM_VERSION_MAJOR, PROG_VERSION_MAJOR);
	SET_CFG_I(YC_CFG_PARM_VERSION_MINOR, PROG_VERSION_MINOR);
	SET_CFG_I(YC_CFG_PARM_VERSION_RELEASE, PROG_VERSION_RELEASE);

	SET_CFG_S(YC_CFG_PARM_DIR_START_UP, cwd);
	s = g_strdup_printf("%s%c%s", home, DIR_DELIMINATOR, YC_DEF_DATA_DIR_LOCAL);
	SET_CFG_S(YC_CFG_PARM_DIR_LOCAL, s);
	g_free(s);
	SET_CFG_S(YC_CFG_PARM_DIR_GLOBAL, YC_DEF_DATA_DIR_GLOBAL);
	s = g_getenv("TMPDIR");
	if(s == NULL)
#ifdef P_tmpdir
	    s = P_tmpdir;
#else
	    s = "/tmp";
#endif
	SET_CFG_S(YC_CFG_PARM_DIR_TMP, s);

	s = g_strdup_printf("%s%c%s%c%s",
	    home, DIR_DELIMINATOR, YC_DEF_DATA_DIR_LOCAL,
	    DIR_DELIMINATOR, YC_DEF_SERVERS_LIST_FILE
	);
	SET_CFG_S(YC_CFG_PARM_FILE_SERVERS_LIST, s);
	g_free(s);

	SET_CFG_I(YC_CFG_PARM_SHOW_TOOLTIPS, TRUE);

	SET_CFG_I(YC_CFG_PARM_STARTUP_CONSOLE, TRUE);
	SET_CFG_I(YC_CFG_PARM_STARTUP_AUDIO, FALSE);
	SET_CFG_I(YC_CFG_PARM_STARTUP_MIXER, FALSE);
	SET_CFG_I(YC_CFG_PARM_STARTUP_PLAYER, FALSE);
	SET_CFG_S(YC_CFG_PARM_STARTUP_LOCATION, "127.0.0.1:9433");

	SET_CFG_I(YC_CFG_PARM_DSPVIEW_DISPLAY_STYLE, DSPVIEW_DISPLAY_STYLE_NOVA);
	SET_CFG_I(YC_CFG_PARM_DSPVIEW_MERGE_MODE, DSPVIEW_MERGE_MODE_SET_ALPHA);
	if(TRUE)
	{
	    guint32 v = DSPVIEW_DEF_DSP_COLOR;
	    color.r = (gfloat)((v & 0x000000ff) >> 0) / (gfloat)(0xff);
	    color.g = (gfloat)((v & 0x0000ff00) >> 8) / (gfloat)(0xff);
	    color.b = (gfloat)((v & 0x00ff0000) >> 16) / (gfloat)(0xff);
	    color.a = (gfloat)((v & 0xff000000) >> 24) / (gfloat)(0xff);
	    SET_CFG_COLOR(YC_CFG_PARM_DSPVIEW_COLOR, &color);
	}
	SET_CFG_F(YC_CFG_PARM_DSPVIEW_TRANSPARENCY, 1.0f);
	SET_CFG_F(YC_CFG_PARM_DSPVIEW_FADE_COEFF, DSPVIEW_DEF_FADE_COEFF);

	SET_CFG_S(YC_CFG_PARM_DSPVIEW_FILE_IMAGE_BG_LARGE, "");
	SET_CFG_I(YC_CFG_PARM_DSPVIEW_IMAGE_BG_LARGE_FIT_METHOD, YC_IMG_FIT_METHOD_TILE);
	SET_CFG_S(YC_CFG_PARM_DSPVIEW_FILE_IMAGE_BG_SMALL, "");
	SET_CFG_I(YC_CFG_PARM_DSPVIEW_IMAGE_BG_SMALL_FIT_METHOD, YC_IMG_FIT_METHOD_TILE);

	color.r = 0.0;
	color.g = 0.0;
	color.b = 0.0;
	color.a = 1.0;
	SET_CFG_COLOR(YC_CFG_PARM_AUDIO_DISPLAY_COLOR_BG, &color);
	color.r = 0.0;
	color.g = 1.0;
	color.b = 0.0;
	color.a = 1.0;
	SET_CFG_COLOR(YC_CFG_PARM_AUDIO_DISPLAY_COLOR_FG, &color);
	color.r = 0.0;
	color.g = 0.5;
	color.b = 0.0;
	color.a = 1.0;
	SET_CFG_COLOR(YC_CFG_PARM_AUDIO_DISPLAY_COLOR_FG_DIM, &color);
	SET_CFG_S(YC_CFG_PARM_AUDIO_DISPLAY_FONT_LARGE,
	    "-adobe-helvetica-bold-r-*-*-16-*-*-*-*-*-*-*"
	);
	SET_CFG_S(YC_CFG_PARM_AUDIO_DISPLAY_FONT_MEDIUM,
	    "-adobe-helvetica-bold-r-*-*-14-*-*-*-*-*-*-*"
	);
	SET_CFG_S(YC_CFG_PARM_AUDIO_DISPLAY_FONT_SMALL,
	    "-adobe-helvetica-bold-r-*-*-10-*-*-*-*-*-*-*"
	);

	SET_CFG_I(YC_CFG_PARM_MIXER_COLOR_USE_DEFAULT, FALSE);
	color.r = 0.9;
	color.g = 0.9;
	color.b = 0.9;
	color.a = 1.0;
	SET_CFG_COLOR(YC_CFG_PARM_MIXER_COLOR_FG, &color);
	color.r = 0.3;
	color.g = 0.3;
	color.b = 0.3;
	color.a = 1.0;
	SET_CFG_COLOR(YC_CFG_PARM_MIXER_COLOR_BG, &color);
	color.r = 0.5;
	color.g = 0.5;
	color.b = 0.5;
	color.a = 1.0;
	SET_CFG_COLOR(YC_CFG_PARM_MIXER_COLOR_BG_PRELIGHT, &color);
	color.r = 0.9;
	color.g = 0.9;
	color.b = 0.9;
	color.a = 1.0;
	SET_CFG_COLOR(YC_CFG_PARM_MIXER_COLOR_TEXT, &color);
	color.r = 0.1;
	color.g = 0.1;
	color.b = 0.1;
	color.a = 1.0;
	SET_CFG_COLOR(YC_CFG_PARM_MIXER_COLOR_BASE, &color);

	SET_CFG_I(YC_CFG_PARM_CONSOLE_SAVE_POSITIONS, TRUE);
	SET_CFG_I(YC_CFG_PARM_CONSOLE_X, 0);
	SET_CFG_I(YC_CFG_PARM_CONSOLE_Y, 0);
	SET_CFG_I(YC_CFG_PARM_CONSOLE_WIDTH, 400);
	SET_CFG_I(YC_CFG_PARM_CONSOLE_HEIGHT, 100);

	SET_CFG_I(YC_CFG_PARM_AUDIO_SAVE_POSITIONS, TRUE);
	SET_CFG_I(YC_CFG_PARM_AUDIO_X, 0);
	SET_CFG_I(YC_CFG_PARM_AUDIO_Y, 0);
	SET_CFG_I(YC_CFG_PARM_AUDIO_WIDTH, 400);
	SET_CFG_I(YC_CFG_PARM_AUDIO_HEIGHT, 250);

	glist = NULL;
	glist = g_list_append(glist, (gpointer)160);
	glist = g_list_append(glist, (gpointer)75);
	glist = g_list_append(glist, (gpointer)30);
	glist = g_list_append(glist, (gpointer)60);
	intlist = CFGIntListNew(glist);
	g_list_free(glist);
	SET_CFG_INTLIST(YC_CFG_PARM_AUDIO_MODES_LIST_COLUMN_WIDTHS, intlist);
	CFGIntListDelete(intlist);

	SET_CFG_I(YC_CFG_PARM_AUDIO_PAGE, 0);

	SET_CFG_I(YC_CFG_PARM_MIXER_SAVE_POSITIONS, TRUE);
	SET_CFG_I(YC_CFG_PARM_MIXER_X, 0);
	SET_CFG_I(YC_CFG_PARM_MIXER_Y, 0);
	SET_CFG_I(YC_CFG_PARM_MIXER_WIDTH, 400);
	SET_CFG_I(YC_CFG_PARM_MIXER_HEIGHT, 200);
	SET_CFG_I(YC_CFG_PARM_MIXER_MIXER_SCALE_LABEL_TYPE,
	    YMIXER_MIXER_SCALE_LABEL_TYPE_ICON
	);

	glist = NULL;
	glist = g_list_append(glist, (gpointer)YMixerCodeVolume);
	glist = g_list_append(glist, (gpointer)YMixerCodeBass);
	glist = g_list_append(glist, (gpointer)YMixerCodeTreble);
	glist = g_list_append(glist, (gpointer)YMixerCodeGainIn);
	glist = g_list_append(glist, (gpointer)YMixerCodeGainOut);
	glist = g_list_append(glist, (gpointer)YMixerCodePCM);
	glist = g_list_append(glist, (gpointer)YMixerCodeSynth);
	glist = g_list_append(glist, (gpointer)YMixerCodeCD);
	glist = g_list_append(glist, (gpointer)YMixerCodeSpeaker);
	glist = g_list_append(glist, (gpointer)YMixerCodeMic);
	glist = g_list_append(glist, (gpointer)YMixerCodeRec);
	intlist = CFGIntListNew(glist);
	g_list_free(glist);
	SET_CFG_INTLIST(YC_CFG_PARM_MIXER_MIXER_SCALES, intlist);
	CFGIntListDelete(intlist);

	SET_CFG_I(YC_CFG_PARM_MIXER_SHOW_MIXER_VALUE, FALSE);

	SET_CFG_I(YC_CFG_PARM_PLAYER_SAVE_POSITIONS, TRUE);
	SET_CFG_I(YC_CFG_PARM_PLAYER_X, 0);
	SET_CFG_I(YC_CFG_PARM_PLAYER_Y, 0);
	SET_CFG_I(YC_CFG_PARM_PLAYER_WIDTH, 550);
	SET_CFG_I(YC_CFG_PARM_PLAYER_HEIGHT, 200);
	SET_CFG_I(YC_CFG_PARM_PLAYER_SNDOBJ_PANED_POSITION, 350);

	glist = NULL;
	glist = g_list_append(glist, (gpointer)120);
	glist = g_list_append(glist, (gpointer)70);
	glist = g_list_append(glist, (gpointer)70);
	glist = g_list_append(glist, (gpointer)70);
	glist = g_list_append(glist, (gpointer)80);
	intlist = CFGIntListNew(glist);
	g_list_free(glist);
	SET_CFG_INTLIST(
	    YC_CFG_PARM_PLAYER_SNDOBJ_LIST_COLUMN_WIDTHS,
	    intlist
	);
	CFGIntListDelete(intlist);

	s = g_strdup_printf("%s%c%s%c%s",
	    home, DIR_DELIMINATOR, YC_DEF_DATA_DIR_LOCAL,
	    DIR_DELIMINATOR, YC_DEF_PLAYLIST_FILE
	);
	SET_CFG_S(YC_CFG_PARM_PLAYER_SNDOBJ_DEFAULT_PLAYLIST_FILE, s);
	g_free(s);
	SET_CFG_I(YC_CFG_PARM_PLAYER_SNDOBJ_CHANGE_MODE, TRUE);
	SET_CFG_I(YC_CFG_PARM_PLAYER_SNDOBJ_REPEAT, FALSE);
	SET_CFG_F(YC_CFG_PARM_PLAYER_SNDOBJADV_VOLUME_X, 0.0f);
	SET_CFG_F(YC_CFG_PARM_PLAYER_SNDOBJADV_VOLUME_Y, 0.0f);
	SET_CFG_F(YC_CFG_PARM_PLAYER_SNDOBJADV_VOLUME_A, 1.0f);
	SET_CFG_I(YC_CFG_PARM_PLAYER_SNDOBJADV_REPEAT, FALSE);

	glist = NULL;
	glist = g_list_append(glist, (gpointer)35);
	glist = g_list_append(glist, (gpointer)60);
	glist = g_list_append(glist, (gpointer)50);
	glist = g_list_append(glist, (gpointer)5);
	intlist = CFGIntListNew(glist);
	g_list_free(glist);
	SET_CFG_INTLIST(
	    YC_CFG_PARM_PLAYER_AUDIOCD_TRACKS_LIST_COLUMN_WIDTHS,
	    intlist
	);
	CFGIntListDelete(intlist);

	SET_CFG_I(YC_CFG_PARM_PLAYER_PAGE, 0);

	SET_CFG_I(YC_CFG_PARM_SERVER_LIST_SAVE_POSITIONS, TRUE);
	SET_CFG_I(YC_CFG_PARM_SERVER_LIST_X, 0);
	SET_CFG_I(YC_CFG_PARM_SERVER_LIST_Y, 0);
	SET_CFG_I(YC_CFG_PARM_SERVER_LIST_WIDTH, 300);
	SET_CFG_I(YC_CFG_PARM_SERVER_LIST_HEIGHT, 350);

	glist = NULL;
	glist = g_list_append(glist, (gpointer)120);
	glist = g_list_append(glist, (gpointer)150);
	glist = g_list_append(glist, (gpointer)30);
	intlist = CFGIntListNew(glist);
	g_list_free(glist);
	SET_CFG_INTLIST(YC_CFG_PARM_SERVER_LIST_COLUMN_WIDTHS, intlist);
	CFGIntListDelete(intlist);

	SET_CFG_I(YC_CFG_PARM_STARTUP_DIALOG_SAVE_POSITIONS, TRUE);
	SET_CFG_I(YC_CFG_PARM_STARTUP_DIALOG_X, 0);
	SET_CFG_I(YC_CFG_PARM_STARTUP_DIALOG_Y, 0);
	SET_CFG_I(YC_CFG_PARM_STARTUP_DIALOG_WIDTH, 600);
	SET_CFG_I(YC_CFG_PARM_STARTUP_DIALOG_HEIGHT, 300);
	SET_CFG_I(YC_CFG_PARM_STARTUP_DIALOG_AUTO_CLOSE, FALSE);

	g_free(cwd);
	g_free(home);
}

/*
 *	Creates a new YConsole core structure.
 */
static yc_core_struct *YCNew(
	gint argc, gchar *argv[], gpointer imlib_handle,
	gint *err_rtn
)
{
	gboolean	startup_yconsole = FALSE,
			startup_yaudio = FALSE,
			startup_ymixer = FALSE,
			startup_yplayer = FALSE,
			startup_yserver_list = FALSE;
	gboolean startup_window_command_line_override = FALSE;
	gboolean created_window = FALSE;
	gint yplayer_audiocd_track_number = 0;
	gchar *yplayer_sndobj_path = NULL;
	gint i;
	const gchar *cstrptr, *arg_ptr;
	gchar *startup_address = NULL;
	yc_core_struct *core_ptr;
	cfg_item_struct src_cfg_list[] = YC_CONFIGURATION_LIST;
	cfg_item_struct *cfg_list;

	*err_rtn = 0;


	/* Handle command line arguments */
	for(i = 1; i < argc; i++)
	{
	    arg_ptr = argv[i];
	    if(arg_ptr == NULL)
		continue;

	    /* Startup with YConsole? */
	    if(!strcasecmp(arg_ptr, "--yconsole") ||
	       !strcasecmp(arg_ptr, "-yconsole")
	    )
	    {
		startup_yconsole = TRUE;
		startup_window_command_line_override = TRUE;
	    }
	    /* Startup with YAudio? */
	    else if(!strcasecmp(arg_ptr, "--yaudio") ||
		    !strcasecmp(arg_ptr, "-yaudio")
	    )
	    {
		startup_yaudio = TRUE;
		startup_window_command_line_override = TRUE;
	    }
	    /* Startup with YMixer? */
	    else if(!strcasecmp(arg_ptr, "--ymixer") ||
		    !strcasecmp(arg_ptr, "-ymixer")
	    )
	    {
		startup_ymixer = TRUE;
		startup_window_command_line_override = TRUE;
	    }
	    /* Startup with YPlayer? */
	    else if(!strcasecmp(arg_ptr, "--yplayer") ||
		    !strcasecmp(arg_ptr, "-yplayer")
	    )
	    {
		startup_yplayer = TRUE;
		startup_window_command_line_override = TRUE;
	    }
	    /* Startup with YPlayer playing a sound object? */
	    else if(!strcasecmp(arg_ptr, "--yplayer_sound_object") ||
		    !strcasecmp(arg_ptr, "-yplayer_sound_object") ||
		    !strcasecmp(arg_ptr, "--yplayer_sndobj") ||
		    !strcasecmp(arg_ptr, "-yplayer_sndobj")
	    )
	    {
		startup_yplayer = TRUE;
		startup_window_command_line_override = TRUE;

		i++;
		arg_ptr = (i < argc) ? argv[i] : NULL;
		if(arg_ptr != NULL)
		{
		    g_free(yplayer_sndobj_path);
		    yplayer_sndobj_path = g_strdup(arg_ptr);
		}
		else
		{
		    fprintf(
			stderr,
			"%s: Requires argument.\n",
			argv[i - 1]
		    );
		}
	    }
	    /* Startup with YPlayer playing an audio CD track? */
	    else if(!strcasecmp(arg_ptr, "--yplayer_audiocd") ||
		    !strcasecmp(arg_ptr, "-yplayer_audiocd")
	    )
	    {
		startup_yplayer = TRUE;
		startup_window_command_line_override = TRUE;

		i++;
		arg_ptr = (i < argc) ? argv[i] : NULL;
		if(arg_ptr != NULL)
		{
		    yplayer_audiocd_track_number = atoi(arg_ptr);
		}
		else
		{
		    fprintf(
			stderr,
			"%s: Requires argument.\n",
			argv[i - 1]
		    );
		}
	    }
	    /* Startup with YServer List? */
	    else if(!strcasecmp(arg_ptr, "--yserver_list") ||
		    !strcasecmp(arg_ptr, "-yserver_list") ||
		    !strcasecmp(arg_ptr, "--yserver-list") ||
		    !strcasecmp(arg_ptr, "-yserver-list")
	    )
	    {
		startup_yserver_list = TRUE;
		startup_window_command_line_override = TRUE;
	    }

	    /* Recorder? */
	    else if(!strcasecmp(arg_ptr, "--recorder") ||
		    !strcasecmp(arg_ptr, "-recorder")
	    )
	    {
		i++;
		arg_ptr = (i < argc) ? argv[i] : NULL;
		if(arg_ptr != NULL)
		{
		    g_free(startup_address);
		    startup_address = g_strdup(arg_ptr);
		}
		else
		{
		    fprintf(
			stderr,
			"%s: Requires argument.\n",
			argv[i - 1]
		    );
		}
	    }
	    /* All else assume address? */
	    else if((*arg_ptr != '+') && (*arg_ptr != '-'))
	    {
		g_free(startup_address);
		startup_address = g_strdup(arg_ptr);
	    }


	}


	/* Allocate the core */
	core_ptr = (yc_core_struct *)g_malloc0(sizeof(yc_core_struct));
	if(core_ptr == NULL)
	{
	    *err_rtn = 3;
	    return(NULL);
	}

	/* Reset core values */

	/* Configuration file location and configuration list */
	cstrptr = g_getenv("HOME");
	core_ptr->cfg_file = g_strdup_printf(
	    "%s%c%s%c%s",
	    (cstrptr != NULL) ? cstrptr : "/", DIR_DELIMINATOR,
	    YC_DEF_DATA_DIR_LOCAL, DIR_DELIMINATOR,
	    YC_DEF_CONFIG_FILE
	);
	core_ptr->cfg_list = cfg_list = CFGItemListCopyList(src_cfg_list);
	/* Set default configuration */
	YCSetDefaultConfiguration(core_ptr, core_ptr->cfg_list);
	/* Load configuration from file (if file exists) */
	CFGFileOpen(core_ptr->cfg_file, core_ptr->cfg_list);

	core_ptr->imlib_handle = imlib_handle;	/* Shared */

	core_ptr->close_all_windows = FALSE;

	core_ptr->yconsole = NULL;
	core_ptr->total_yconsoles = 0;

	core_ptr->ymixer = NULL;
	core_ptr->total_ymixers = 0;

	core_ptr->yplayer = NULL;
	core_ptr->total_yplayers = 0;

	core_ptr->yserver_list = NULL;
	core_ptr->optwin = NULL;
	core_ptr->optwin_dspview = NULL;

	core_ptr->startup_dialog = NULL;
	core_ptr->total_startup_dialogs = 0;

	/* Load cursors */
	core_ptr->cursor_busy = gdk_cursor_new(GDK_WATCH);
	core_ptr->cursor_text = gdk_cursor_new(GDK_XTERM);

	/* Load default DSPView background images and allocate image
	 * buffers on the core structure.
	 */
	core_ptr->dsp_bg_img = NULL;
	core_ptr->dsp_bg_img_small = NULL;
	YCImgLoadDefault(core_ptr);


	/* Need to reset timeout callback ids */
	core_ptr->toid = (guint)-1;
	core_ptr->toid2 = (guint)-1;
	core_ptr->toid3 = (guint)-1;


	/* Begin updating globals that depend on the configuration
	 * (which we now have loaded).
	 */

	/* If no startup address was specified from the command line
	 * then use the address set in the configuration
	 */
	if(startup_address == NULL)
	{
	    cstrptr = GET_CFG_S(YC_CFG_PARM_STARTUP_LOCATION);
	    startup_address = g_strdup(
		(cstrptr != NULL) ? cstrptr : "127.0.0.1:9433"
	    );
	}

	/* Global tooltips state */
	GUISetGlobalTipsState(
	    GET_CFG_I(YC_CFG_PARM_SHOW_TOOLTIPS) ? TRUE : FALSE
	);

	/* Load configuration specified DSPView images.  Note that the
	 * pointers to the DSPView images will not be changed by this
	 * call.  Windows need to be created after this call so that
	 * they can reference these images without expecting the
	 * DSPView image pointers to change.
	 */
	YCImgLoad(
	    core_ptr,
	    GET_CFG_S(YC_CFG_PARM_DSPVIEW_FILE_IMAGE_BG_LARGE),
	    GET_CFG_I(YC_CFG_PARM_DSPVIEW_IMAGE_BG_LARGE_FIT_METHOD),
	    GET_CFG_S(YC_CFG_PARM_DSPVIEW_FILE_IMAGE_BG_SMALL),
	    GET_CFG_I(YC_CFG_PARM_DSPVIEW_IMAGE_BG_SMALL_FIT_METHOD)
	);



	/* Begin getting values for initial window creation */

	/* Get startup window from configuration if one was not
	 * specified from the command line.
	 */
	if(!startup_window_command_line_override)
	{
	    startup_yconsole = GET_CFG_I(YC_CFG_PARM_STARTUP_CONSOLE);
	    startup_yaudio = GET_CFG_I(YC_CFG_PARM_STARTUP_AUDIO);
	    startup_ymixer = GET_CFG_I(YC_CFG_PARM_STARTUP_MIXER);
	    startup_yplayer = GET_CFG_I(YC_CFG_PARM_STARTUP_PLAYER);
	}

	/* Create first window */
	if(startup_yconsole)
	{
	    if(YCNewYConsole(core_ptr, startup_address) >= 0)
		created_window = TRUE;
	}
	if(startup_yaudio)
	{
	    if(YCNewYAudio(core_ptr, startup_address) >= 0)
		created_window = TRUE;
	}
	if(startup_ymixer)
	{
	    if(YCNewYMixer(core_ptr, startup_address) >= 0)
		created_window = TRUE;
	}
	if(startup_yplayer)
	{
	    if(yplayer_sndobj_path != NULL)
	    {
		if(YCNewYPlayerSoundObject(
		    core_ptr, startup_address, yplayer_sndobj_path
		) >= 0)
		    created_window = TRUE;
	    }
	    else if(yplayer_audiocd_track_number > 0)
	    {
		if(YCNewYPlayerAudioCD(
		    core_ptr, startup_address, yplayer_audiocd_track_number
		) >= 0)
		    created_window = TRUE;
	    }
	    else
	    {
		if(YCNewYPlayer(core_ptr, startup_address) >= 0)
		    created_window = TRUE;
	    }
	}
	if(startup_yserver_list)
	{
	    if(YCNewYServerList(core_ptr, NULL) >= 0)
		created_window = TRUE;
	}
	/* If no windows were created then create a YConsole */
	if(!created_window)
	{
	    if(YCNewYConsole(core_ptr, startup_address) >= 0)
		created_window = TRUE;
	}



	/* Set timeout callbacks */
	core_ptr->toid = gtk_timeout_add(
	    100, (GtkFunction)YCTimeoutCB, core_ptr
	);
	core_ptr->toid2 = gtk_timeout_add(
	    500, (GtkFunction)YCTimeout2CB, core_ptr
	);
	core_ptr->toid3 = gtk_timeout_add(
	    60000, (GtkFunction)YCTimeout3CB, core_ptr
	);


	/* Deallocate locals */
	g_free(startup_address);
	g_free(yplayer_sndobj_path);

	return(core_ptr);
}

/*
 *	Deletes the given YConsole core structure.
 */
static void YCDelete(yc_core_struct *core_ptr, gboolean save_config)
{
	gint i;
	GdkCursor **cursor;


	if(core_ptr == NULL)
	    return;

	/* Save configuration? */
	if(save_config)
	{
	    /* YServers list */
	    YServerListFileSave(
		core_ptr->yserver_list,
		GET_CFG_S(YC_CFG_PARM_FILE_SERVERS_LIST)
	    );

	    /* Configuration items */
	    CFGFileSave(core_ptr->cfg_file, core_ptr->cfg_list);
	}

	/* Remove timeout callbacks */
	if(core_ptr->toid != (guint)-1)
	{
	    gtk_timeout_remove(core_ptr->toid);
	    core_ptr->toid = (guint)-1;
	}
	if(core_ptr->toid2 != (guint)-1)
	{
	    gtk_timeout_remove(core_ptr->toid2);
	    core_ptr->toid2 = (guint)-1;
	}
	if(core_ptr->toid3 != (guint)-1)
	{
	    gtk_timeout_remove(core_ptr->toid3);
	    core_ptr->toid3 = (guint)-1;
	}


	/* Delete all Startup Dialogs */
	for(i = 0; i < core_ptr->total_startup_dialogs; i++)
	    StartupDialogDelete(core_ptr->startup_dialog[i]);
	g_free(core_ptr->startup_dialog);
	core_ptr->startup_dialog = NULL;
	core_ptr->total_startup_dialogs = 0;

	/* Delete DSPView on the options window */
	DSPViewDelete(core_ptr->optwin_dspview);
	core_ptr->optwin_dspview = NULL;

	/* Delete options window */
	OptWinDelete(core_ptr->optwin);
	core_ptr->optwin = NULL;

	/* Delete YServer list window */
	YServerListDelete(core_ptr->yserver_list);
	core_ptr->yserver_list = NULL;

	/* Delete all YConsole windows */
	for(i = 0; i < core_ptr->total_yconsoles; i++)
	    YConsoleDelete(core_ptr->yconsole[i]);
	g_free(core_ptr->yconsole);
	core_ptr->yconsole = NULL;
	core_ptr->total_yconsoles = 0;

	/* Delete all YAudio windows */
	for(i = 0; i < core_ptr->total_yaudios; i++)
	    YAudioDelete(core_ptr->yaudio[i]);
	g_free(core_ptr->yaudio);
	core_ptr->yaudio = NULL;
	core_ptr->total_yaudios = 0;

	/* Delete all YMixer windows */
	for(i = 0; i < core_ptr->total_ymixers; i++)
	    YMixerDelete(core_ptr->ymixer[i]);
	g_free(core_ptr->ymixer);
	core_ptr->ymixer = NULL;
	core_ptr->total_ymixers = 0;

	/* Delete all YPlayer windows */
	for(i = 0; i < core_ptr->total_yplayers; i++)
	    YPlayerDelete(core_ptr->yplayer[i]);
	g_free(core_ptr->yplayer);
	core_ptr->yplayer = NULL;
	core_ptr->total_yplayers = 0;


	/* Delete DSPView background images.  Since other windows
	 * will have pointers to these images, these images need to
	 * be deleted after the windows are deleted.
	 */
	g_free(core_ptr->dsp_bg_img);
	core_ptr->dsp_bg_img = NULL;
	g_free(core_ptr->dsp_bg_img_small);
	core_ptr->dsp_bg_img_small = NULL;


	/* Destroy cursors */
#define DO_DESTROY_CURSOR	\
{ if(*cursor != NULL) { gdk_cursor_destroy(*cursor); *cursor = NULL; } }

	cursor = &core_ptr->cursor_busy;
	DO_DESTROY_CURSOR
	cursor = &core_ptr->cursor_text;
	DO_DESTROY_CURSOR

#undef DO_DESTROY_CURSOR


	/* Imlib handle is shared, do not delete */
	core_ptr->imlib_handle = NULL;

	/* Delete configuration list */
	CFGItemListDeleteList(core_ptr->cfg_list);
	core_ptr->cfg_list = NULL;
	g_free(core_ptr->cfg_file);
	core_ptr->cfg_file = NULL;

	/* Deallocate structure itself */
	g_free(core_ptr);
}


int main(int argc, char *argv[])
{
	gboolean initialized_gtk = FALSE;
	gint i, err_val;
	const gchar *arg_ptr;
	gpointer imlib_handle;
	yc_core_struct *core_ptr;


	yc_core = NULL;


	/* Handle basic command line arguments */
	for(i = 1; i < argc; i++)
	{
	    arg_ptr = argv[i];
	    if(arg_ptr == NULL)
		continue;

	    /* Help? */
	    if(!strcasecmp(arg_ptr, "--help") ||
	       !strcasecmp(arg_ptr, "-help") ||
	       !strcasecmp(arg_ptr, "--h") ||
	       !strcasecmp(arg_ptr, "-h") ||
	       !strcasecmp(arg_ptr, "-?")
	    )
	    {
		printf(PROG_HELP_MESG);
		return(0);
	    }
	    /* Version? */
	    else if(!strcasecmp(arg_ptr, "--version") ||
		    !strcasecmp(arg_ptr, "-version")
	    )
	    {
		printf(
		    "%s Version %s\n%s",
		    PROG_NAME, PROG_VERSION,
		    PROG_COPYRIGHT
		);
		return(0);
	    }
	}


	/* Set up UNIX signal callbacks */
	signal(SIGINT, YCSignalCB);
	signal(SIGTERM, YCSignalCB);
	signal(SIGSEGV, YCSignalCB);
#ifdef SIGHUP
	signal(SIGHUP, YCSignalCB);
#endif
#ifdef SIGPIPE
	signal(SIGPIPE, YCSignalCB);
#endif

	/* Reset timezone */
	tzset();


	/* Initialize GTK+ as needed */
	if(!initialized_gtk)
	{
	    if(!gtk_init_check(&argc, &argv))
	    {
		fprintf(
		    stderr,
 "This program requires X.\n"
		);
		return(1);
	    }
	    initialized_gtk = TRUE;
	}

	gdk_rgb_init();

	imlib_handle = Imlib_init(GDK_DISPLAY());

	CDialogInit();
	FPromptInit();
	CSDInit();
	FSDInit();
	FileBrowserInit();
	PDialogInit();

	/* Create core structure */
	yc_core = core_ptr = YCNew(
	    argc, argv, imlib_handle, &err_val
	);

	/* If core is successfully created then enter main loop */
	if(core_ptr != NULL)
	    gtk_main();

	/* Delete core structure */
	yc_core = NULL;
	YCDelete(core_ptr, TRUE);
	core_ptr = NULL;


	PDialogShutdown();
	FileBrowserShutdown();
	FSDShutdown();
	CSDShutdown();
	FPromptShutdown();
	CDialogShutdown();

	return(err_val);
}
