#include <stdlib.h>
#include <string.h>

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include "guiutils.h"

#include "optwin.h"
#include "optwinop.h"
#include "config.h"

#include "images/icon_ok_20x20.xpm"
#include "images/icon_select_20x20.xpm"
#include "images/icon_save_20x20.xpm"
#include "images/icon_cancel_20x20.xpm"


static gint OptWinDeleteEventCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
);
static void OptWinSwitchPageCB(
        GtkNotebook *notebook, GtkNotebookPage *page, guint page_num,
        gpointer data
);
static void OptWinOKCB(GtkWidget *widget, gpointer data);
static void OptWinApplyCB(GtkWidget *widget, gpointer data);
static void OptWinSaveCB(GtkWidget *widget, gpointer data);
static void OptWinCancelCB(GtkWidget *widget, gpointer data);

optwin_cfgref_struct *OptWinCfgRefNew(
        gint id,                /* Unique ID to identify this cfgref. */
        GtkWidget *w,           /* Widget that holds the value. */
        const gchar *cfg_parm   /* Configuration parameter name. */
);
void OptWinCfgRefDelete(optwin_cfgref_struct *cfgref);
optwin_cfgref_struct *OptWinCfgRefAppend(
        optwin_struct *optwin,
        gint id,                /* Unique ID to identify this cfgref. */
        GtkWidget *w,           /* Widget that holds the value. */
        const gchar *cfg_parm   /* Configuration parameter name. */
);
void OptWinCfgRefSetColor(
        optwin_cfgref_struct *cfgref, GdkColor *fg_c, GdkColor *bg_c
);

optwin_cfgref_struct *OptWinCfgRefMatchID(
        optwin_struct *optwin, gint id, gint *n
);

optwin_struct *OptWinNew(
        gpointer core_ptr, const gchar *title, guint8 **icon_data
);
GtkWidget *OptWinPageAppend(
	optwin_struct *optwin, GtkWidget *w
);
void OptWinMap(optwin_struct *optwin);
void OptWinUnmap(optwin_struct *optwin);
void OptWinDelete(optwin_struct *optwin);



#define OPTWIN_WIDTH		640
#define OPTWIN_HEIGHT		480

#define OPTWIN_BTN_WIDTH	(100 + (2 * 3))
#define OPTWIN_BTN_HEIGHT	(30 + (2 * 3))



/*
 *	"delete_event" signal callback.
 */
static gint OptWinDeleteEventCB(
        GtkWidget *widget, GdkEvent *event, gpointer data
)
{
	optwin_struct *optwin = OPTWIN(data);
	if(optwin == NULL)
	    return(TRUE);

	OptWinCancelCB(NULL, optwin);

	return(TRUE);
}

/*
 *	GtkNotebook "switch_page" signal callback.
 */
static void OptWinSwitchPageCB(
        GtkNotebook *notebook, GtkNotebookPage *page, guint page_num,
        gpointer data
)
{
	static gboolean reenterent = FALSE;
        optwin_struct *optwin = OPTWIN(data);
        if(optwin == NULL)
            return;

	if(reenterent)
	    return;
	else
	    reenterent = TRUE;




	reenterent = FALSE;
}

/*
 *	OK button callback.
 */
static void OptWinOKCB(GtkWidget *widget, gpointer data)
{
        static gboolean reenterent = FALSE;
        optwin_struct *optwin = OPTWIN(data);
        if(optwin == NULL)
            return;

        if(reenterent)
            return;
        else
            reenterent = TRUE;

	OptWinDoApply(optwin);
	OptWinUnmap(optwin);

        reenterent = FALSE;
}

/*
 *	Apply button callback.
 */
static void OptWinApplyCB(GtkWidget *widget, gpointer data)
{
        static gboolean reenterent = FALSE;
        optwin_struct *optwin = OPTWIN(data);
        if(optwin == NULL)
            return;

        if(reenterent)
            return;
        else
            reenterent = TRUE;

        OptWinDoApply(optwin);

        reenterent = FALSE;
}

/*
 *	Save button callback.
 */
static void OptWinSaveCB(GtkWidget *widget, gpointer data)
{
        static gboolean reenterent = FALSE;
        optwin_struct *optwin = OPTWIN(data);
        if(optwin == NULL)
            return;

        if(reenterent)
            return;
        else
            reenterent = TRUE;

/* Need to work on this. */

        reenterent = FALSE;
}

/*
 *	Cancel button callback.
 */
static void OptWinCancelCB(GtkWidget *widget, gpointer data)
{
        static gboolean reenterent = FALSE;
        optwin_struct *optwin = OPTWIN(data);
        if(optwin == NULL)
            return;

        if(reenterent)
            return;
        else
            reenterent = TRUE;

	OptWinUnmap(optwin);

        reenterent = FALSE;
}


/*
 *	Creates a new cfgref.
 */
optwin_cfgref_struct *OptWinCfgRefNew(
        gint id,                /* Unique ID to identify this cfgref. */
        GtkWidget *w,           /* Widget that holds the value. */
        const gchar *cfg_parm   /* Configuration parameter name. */
)
{
	optwin_cfgref_struct *cfgref = (optwin_cfgref_struct *)g_malloc0(
	    sizeof(optwin_cfgref_struct)
	);
	if(cfgref == NULL)
	    return(cfgref);

	cfgref->id = id;
	cfgref->optwin = NULL;
	cfgref->w = w;
	cfgref->browse = NULL;
	cfgref->cfg_parm = (cfg_parm != NULL) ? g_strdup(cfg_parm) : NULL;

	/* Reset special values. */
	cfgref->radio_value = 0;
	cfgref->colormap = NULL;
	memset(&cfgref->fg_color, 0x00, sizeof(GdkColor));
        memset(&cfgref->bg_color, 0x00, sizeof(GdkColor));
	cfgref->slist = NULL;

	cfgref->client_data = NULL;
	cfgref->destroy_cb = NULL;

	return(cfgref);
}

/*
 *	Deallocates all resources referenced on the given cfgref and
 *	the structure itself.
 *
 *	If destroy_cb is not NULL then it will be called first.
 */
void OptWinCfgRefDelete(optwin_cfgref_struct *cfgref)
{
	GdkColormap *colormap;
	GtkWidget **w;


	if(cfgref == NULL)
	    return;

	/* Call destroy callback first. */
	if(cfgref->destroy_cb != NULL)
	    cfgref->destroy_cb(cfgref, cfgref->client_data);

#define DO_DESTROY_WIDGET       \
{ if(*w != NULL) { gtk_widget_destroy(*w); *w = NULL; } }

	/* Destroy widgets. */
	w = &cfgref->w;
	DO_DESTROY_WIDGET
	w = &cfgref->browse;
	DO_DESTROY_WIDGET
	w = &cfgref->parent;
	DO_DESTROY_WIDGET


	/* Deallocate other resources referenced under special
	 * situations.
	 */

	/* Colormap and colors. */
	colormap = cfgref->colormap;
	if(colormap != NULL)
	{
	    gdk_colormap_free_colors(colormap, &cfgref->fg_color, 1);
            gdk_colormap_free_colors(colormap, &cfgref->bg_color, 1);
	    gdk_colormap_unref(colormap);
	    cfgref->colormap = colormap = NULL;
	}

	/* Stack list. */
	StackListDelete(cfgref->slist);
	cfgref->slist = NULL;

#undef DO_DESTROY_WIDGET

	g_free(cfgref->cfg_parm);
	cfgref->cfg_parm = NULL;

	g_free(cfgref);
}

/*
 *	Appends a new cfgref structure to the optwin, returns the pointer
 *	to the new cfgref structure.
 */
optwin_cfgref_struct *OptWinCfgRefAppend(
        optwin_struct *optwin,
        gint id,                /* Unique ID to identify this cfgref. */
        GtkWidget *w,           /* Widget that holds the value. */
        const gchar *cfg_parm   /* Configuration parameter name. */
)
{
	gint i;
	optwin_cfgref_struct *cfgref;


	if(optwin == NULL)
	    return(NULL);

	/* Sanitize total. */
	if(optwin->total_cfgrefs < 0)
	    optwin->total_cfgrefs = 0;

	/* Allocate more pointers. */
	i = optwin->total_cfgrefs;
	optwin->total_cfgrefs = i + 1;

	optwin->cfgref = (optwin_cfgref_struct **)g_realloc(
	    optwin->cfgref,
	    optwin->total_cfgrefs * sizeof(optwin_cfgref_struct *)
	);
	if(optwin->cfgref == NULL)
	{
	    optwin->total_cfgrefs = 0;
	    return(NULL);
	}

	/* Allocate a new structure. */
	optwin->cfgref[i] = cfgref = OptWinCfgRefNew(
	    id, w, cfg_parm
	);
	if(cfgref != NULL)
	{
	    cfgref->optwin = optwin;
	}

	return(cfgref);
}

/*
 *	Sets the color for the cfgref, allocating it (and the colormap
 *	as needed).  The widget for this cfgref must be defined and be
 *	valid.
 *
 *	The given GdkColor need not be allocated.
 */
void OptWinCfgRefSetColor(
	optwin_cfgref_struct *cfgref, GdkColor *fg_c, GdkColor *bg_c
)
{
	GtkWidget *w;
	GdkColormap *colormap;
	GdkWindow *window;


	if(cfgref == NULL)
	    return;

	/* Get widget for this cfgref, it must not be NULL.  We will
	 * realize it as needed and it must have a GdkWindow.
	 */
        w = OPTWIN_CFGREF_WIDGET(cfgref);
	if(w == NULL)
	    return;

	if(!GTK_WIDGET_REALIZED(w))
	    gtk_widget_realize(w);

	if(GTK_WIDGET_NO_WINDOW(w))
	    return;

	window = w->window;
	if(window == NULL)
	    return;

	/* Get colormap, check if it is allocated. */
        colormap = cfgref->colormap;
	if(colormap == NULL)
	{
	    /* Colormap was not allocated, so we need to get the
	     * colormap from the widget's GdkWindow and then allocate
	     * the color.
	     */
	    cfgref->colormap = colormap = gdk_window_get_colormap(window);
	    if(colormap != NULL)
	    {
		/* Need to have cfgref hold on to it. */
		gdk_colormap_ref(colormap);

		/* Set and allocate new colors. */
		if(fg_c != NULL)
		    memcpy(&cfgref->fg_color, fg_c, sizeof(GdkColor));
                if(bg_c != NULL)
                    memcpy(&cfgref->bg_color, bg_c, sizeof(GdkColor));

		gdk_colormap_alloc_color(
		    colormap, &cfgref->fg_color, TRUE, TRUE
		);
                gdk_colormap_alloc_color(
                    colormap, &cfgref->bg_color, TRUE, TRUE
                );
	    }
	}
	else
	{
	    /* Colormap was already allocated, so we free the existing
	     * color and set a new one.  The new one will be allocated.
	     */
	    gdk_colormap_free_colors(
		colormap, &cfgref->fg_color, 1
	    );
            gdk_colormap_free_colors(
                colormap, &cfgref->bg_color, 1
            );

	    if(fg_c != NULL)
		memcpy(&cfgref->fg_color, fg_c, sizeof(GdkColor));
            if(bg_c != NULL)
                memcpy(&cfgref->bg_color, bg_c, sizeof(GdkColor));

	    gdk_colormap_alloc_color(
		colormap, &cfgref->fg_color, TRUE, TRUE
	    );
            gdk_colormap_alloc_color(
                colormap, &cfgref->bg_color, TRUE, TRUE
            );
	}
}


/*
 *	Returns the pointer to the first cfgref found on the options
 *	window that matches the given id.
 */
optwin_cfgref_struct *OptWinCfgRefMatchID(
	optwin_struct *optwin, gint id, gint *n
)
{
	gint i;
	optwin_cfgref_struct *cfgref;


	if(n != NULL)
	    *n = -1;

	if(optwin == NULL)
	    return(NULL);

	for(i = 0; i < optwin->total_cfgrefs; i++)
	{
	    cfgref = optwin->cfgref[i];
	    if(cfgref == NULL)
		continue;

	    if(OPTWIN_CFGREF_ID(cfgref) == id)
	    {
		if(n != NULL)
		    *n = i;
		return(cfgref);
	    }
	}

	return(NULL);
}


/*
 *	Creates a new optwin_struct.
 */
optwin_struct *OptWinNew(
	gpointer core_ptr, const gchar *title, guint8 **icon_data
)
{
	gint /*border_minor = 2,*/ border_major = 5;
	GdkWindow *window;
	GtkAccelGroup *accelgrp;
	GtkWidget *w, *parent, *parent2;
	optwin_struct *optwin = (optwin_struct *)g_malloc0(
	    sizeof(optwin_struct)
	);
	if(optwin == NULL)
	    return(optwin);


	/* Reset values. */
	optwin->initialized = TRUE;
	optwin->map_state = FALSE;
	optwin->processing = FALSE;
	optwin->core_ptr = core_ptr;


	/* Keyboard accelerator group. */
	optwin->accelgrp = accelgrp = gtk_accel_group_new();


        /* Toplevel. */
	optwin->toplevel = w = gtk_window_new(GTK_WINDOW_DIALOG);
	gtk_widget_set_usize(w, OPTWIN_WIDTH, OPTWIN_HEIGHT);
        gtk_widget_realize(w);
        gtk_window_set_title(GTK_WINDOW(w), title);
        window = w->window;
        if(window != NULL)
        {
            gdk_window_set_decorations(
                window,
		GDK_DECOR_BORDER | GDK_DECOR_TITLE | GDK_DECOR_MENU |
		GDK_DECOR_MINIMIZE
            );
            gdk_window_set_functions(
                window,
                GDK_FUNC_MOVE | GDK_FUNC_MINIMIZE | GDK_FUNC_CLOSE
            );
	    GUISetWMIcon(window, (u_int8_t **)icon_data);
        }
        gtk_signal_connect(
            GTK_OBJECT(w), "delete_event",
            GTK_SIGNAL_FUNC(OptWinDeleteEventCB), optwin
        );
        gtk_container_border_width(GTK_CONTAINER(w), 0);
        gtk_accel_group_attach(accelgrp, GTK_OBJECT(w));
        parent = w;


	/* Main vbox. */
	w = gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(parent), w);
	gtk_widget_show(w);
	parent = w;


	/* Vbox for main notebook. */
        w = gtk_vbox_new(FALSE, 0);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_box_pack_start(GTK_BOX(parent), w, TRUE, TRUE, 0);
        gtk_widget_show(w);
        parent2 = w;

	/* Main notebook. */
	optwin->notebook = w = gtk_notebook_new();
	gtk_notebook_set_tab_pos(GTK_NOTEBOOK(w), GTK_POS_TOP);
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, TRUE, 0);
	gtk_notebook_set_scrollable(GTK_NOTEBOOK(w), TRUE);
	gtk_notebook_set_show_tabs(GTK_NOTEBOOK(w), TRUE);
	gtk_notebook_set_show_border(GTK_NOTEBOOK(w), TRUE);
/*	gtk_notebook_set_page(GTK_NOTEBOOK(w), 0); */
	gtk_signal_connect(
	    GTK_OBJECT(w), "switch_page",
	    GTK_SIGNAL_FUNC(OptWinSwitchPageCB), optwin
	);
	gtk_widget_show(w);
	parent2 = w;



        /* Separator. */
        w = gtk_hseparator_new();
        gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
        gtk_widget_show(w);


        /* Buttons hbox. */
        w = gtk_hbox_new(TRUE, 0);
        gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, border_major);
        gtk_widget_show(w);
        parent2 = w;

        /* OK button. */
        optwin->ok_btn = w = GUIButtonPixmapLabelH(
            (u_int8_t **)icon_ok_20x20_xpm, "OK", NULL
        );
        gtk_widget_set_usize(w, OPTWIN_BTN_WIDTH, OPTWIN_BTN_HEIGHT);
        GTK_WIDGET_SET_FLAGS(w, GTK_CAN_DEFAULT);
        gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(w), "clicked",
            GTK_SIGNAL_FUNC(OptWinOKCB), optwin
        );
        gtk_accel_group_add(
            accelgrp, GDK_Return, 0, GTK_ACCEL_VISIBLE,
            GTK_OBJECT(w), "clicked"
        );
        gtk_accel_group_add(
            accelgrp, GDK_Return, 0, GTK_ACCEL_VISIBLE,
            GTK_OBJECT(w), "clicked"
        );
        gtk_accel_group_add(
            accelgrp, GDK_3270_Enter, 0, GTK_ACCEL_VISIBLE,
            GTK_OBJECT(w), "clicked"
        );
        gtk_accel_group_add(
            accelgrp, GDK_KP_Enter, 0, GTK_ACCEL_VISIBLE,
            GTK_OBJECT(w), "clicked"
        );
        gtk_accel_group_add(
            accelgrp, GDK_ISO_Enter, 0, GTK_ACCEL_VISIBLE,
            GTK_OBJECT(w), "clicked"
        );
        gtk_widget_show(w);

        /* Apply button. */
        optwin->apply_btn = w = GUIButtonPixmapLabelH(
            (u_int8_t **)icon_select_20x20_xpm,
#ifdef PROG_LANGUAGE_ENGLISH
            "Apply",
#endif
#ifdef PROG_LANGUAGE_SPANISH
            "Aplique",
#endif
#ifdef PROG_LANGUAGE_FRENCH
            "S'appliquer",
#endif
	    NULL
        );
        gtk_widget_set_usize(w, OPTWIN_BTN_WIDTH, OPTWIN_BTN_HEIGHT);
        GTK_WIDGET_SET_FLAGS(w, GTK_CAN_DEFAULT);
        gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(w), "clicked",
            GTK_SIGNAL_FUNC(OptWinApplyCB), optwin
        );
        gtk_widget_show(w);

        /* Save button. */
        optwin->save_btn = w = GUIButtonPixmapLabelH(
            (u_int8_t **)icon_save_20x20_xpm,
#ifdef PROG_LANGUAGE_ENGLISH
            "Save",
#endif
#ifdef PROG_LANGUAGE_SPANISH
            "Salve",
#endif
#ifdef PROG_LANGUAGE_FRENCH
            "Epargner",
#endif
	    NULL
        );
        gtk_widget_set_usize(w, OPTWIN_BTN_WIDTH, OPTWIN_BTN_HEIGHT);
        GTK_WIDGET_SET_FLAGS(w, GTK_CAN_DEFAULT);
        gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(w), "clicked",
            GTK_SIGNAL_FUNC(OptWinSaveCB), optwin
        );
/*	gtk_widget_show(w); */

        /* Cancel button. */
        optwin->cancel_btn = w = GUIButtonPixmapLabelH(
            (u_int8_t **)icon_cancel_20x20_xpm,
#ifdef PROG_LANGUAGE_ENGLISH
            "Cancel",
#endif
#ifdef PROG_LANGUAGE_SPANISH
            "Cancele",
#endif
#ifdef PROG_LANGUAGE_FRENCH
            "Annuler",
#endif
	    NULL
        );
        gtk_widget_set_usize(w, OPTWIN_BTN_WIDTH, OPTWIN_BTN_HEIGHT);
        GTK_WIDGET_SET_FLAGS(w, GTK_CAN_DEFAULT);
        gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(w), "clicked",
            GTK_SIGNAL_FUNC(OptWinCancelCB), optwin
        );
        gtk_accel_group_add(
            accelgrp, GDK_Escape, 0, GTK_ACCEL_VISIBLE,
            GTK_OBJECT(w), "clicked"
        );
        gtk_widget_show(w);



	return(optwin);
}

/*
 *	Appends a new page to the given optwin, using the given widget
 *	as the tab widget.
 *
 *	The given widget should not be referenced again after this
 *	call.
 */
GtkWidget *OptWinPageAppend(
        optwin_struct *optwin, GtkWidget *w
)
{
	gint i;
	GtkWidget *parent, *parent2;


	if(w == NULL)
	    return(NULL);

	if(optwin == NULL)
	{
	    gtk_widget_destroy(w);
	    return(NULL);
	}

	parent = optwin->notebook;

	/* Create a new vbox as parent2 and append it to the notebook
	 * as a page using the given widget w as the tab widget.
	 */
	parent2 = gtk_vbox_new(FALSE, 0);
	gtk_notebook_append_page(
	    GTK_NOTEBOOK(parent), parent2, w
	);
	gtk_widget_show(parent2);

	/* Widget w should not be referenced again after this point. */
	w = NULL;


	/* Record new toplevel vbox of the appended page. */
	if(optwin->total_pages < 0)
	    optwin->total_pages = 0;

	i = optwin->total_pages;
	optwin->total_pages = i + 1;

	optwin->page = (GtkWidget **)g_realloc(
	    optwin->page,
	    optwin->total_pages * sizeof(GtkWidget *)
	);
	if(optwin->page == NULL)
	{
	    optwin->total_pages = 0;
	}
	else
	{
	    optwin->page[i] = parent2;
	}

	/* Return the newly created vbox of the appended page. */
	return(parent2);
}

/*
 *	Maps the given optwin.
 */
void OptWinMap(optwin_struct *optwin)
{
	GtkWidget *w;

        if(optwin == NULL)
            return;

        w = optwin->toplevel;
	gtk_widget_show_raise(w);
	optwin->map_state = TRUE;

	w = optwin->ok_btn;
	if(w != NULL)
	{
	    gtk_widget_grab_focus(w);
	    gtk_widget_grab_default(w);
	}
}

/*
 *	Unmaps the given optwin.
 */
void OptWinUnmap(optwin_struct *optwin)
{
        GtkWidget *w;

        if(optwin == NULL)
            return;

	w = optwin->toplevel;
	if(w == NULL)
	    return;

	if(optwin->map_state)
	{
	    gtk_widget_hide(w);
	    optwin->map_state = FALSE;
	}
}

/*
 *	Deallocates all resources on the given optwin and deallocates the
 *	structure itself.
 */
void OptWinDelete(optwin_struct *optwin)
{
	gint i;
	GtkWidget **w;


	if(optwin == NULL)
	    return;

	if(optwin->initialized)
	{
#define DO_DESTROY_WIDGET       \
{ \
 if((*w) != NULL) \
 { \
  GtkWidget *tmp_w = *w; \
  (*w) = NULL; \
  gtk_widget_destroy(tmp_w); \
 } \
}

	    /* Deallocate all cfgrefs, this will destroy all widgets 
	     * that we referenced as instructed by the calling function.
	     */
	    for(i = 0; i < optwin->total_cfgrefs; i++)
		OptWinCfgRefDelete(optwin->cfgref[i]);
	    g_free(optwin->cfgref);
	    optwin->cfgref = NULL;
	    optwin->total_cfgrefs = 0;

	    /* Deallocate toplevel widgets on all notebook pages. */
	    for(i = 0; i < optwin->total_pages; i++)
	    {
		w = &optwin->page[i];
		DO_DESTROY_WIDGET
	    }
            g_free(optwin->page);
            optwin->page = NULL;
            optwin->total_pages = 0;


            w = &optwin->notebook;
            DO_DESTROY_WIDGET

	    w = &optwin->ok_btn;
	    DO_DESTROY_WIDGET
	    w = &optwin->apply_btn;
	    DO_DESTROY_WIDGET
            w = &optwin->save_btn;
	    DO_DESTROY_WIDGET
            w = &optwin->cancel_btn;
	    DO_DESTROY_WIDGET

            w = &optwin->toplevel;
            DO_DESTROY_WIDGET


            if(optwin->accelgrp != NULL)
            {
                gtk_accel_group_unref(optwin->accelgrp);
                optwin->accelgrp = NULL;
            }

#undef DO_DESTROY_WIDGET
	}

	/* Deallocate structure itself. */
	g_free(optwin);
}
