/*
                               Options Window
 */

#ifndef OPTWIN_H
#define OPTWIN_H

#include <gtk/gtk.h>
#include "stacklist.h"


/*
 *	Configuration reference structure:
 *
 *	References a configuration parameter with a widget on the 
 *	options window.
 */
typedef struct {

	gint id;		/* Unique ID to identify this cfgref. */
	gpointer optwin;	/* Pointer back to options window. */
	GtkWidget *w;		/* Widget that holds the value. */
	GtkWidget *browse;	/* Browse widget (optional). */
	GtkWidget *parent;	/* Parent of referenced widgets (optional). */
	gchar *cfg_parm;	/* Configuration parameter name. */

	/* Following are members use in special or extended cases. */
	gint radio_value;	/* Used for active GtkRadioButton. */
	GdkColormap *colormap;	/* Used for the GdkColor. */
	GdkColor fg_color, bg_color;

	/* Stack list, used instead of the GtkWidget w if not NULL. */
	stack_list_struct *slist;

	/* Destroy callback
	 * (optwin_cfgref_struct *cfgref, gpointer client_data)
	 */
	gpointer client_data;
	void (*destroy_cb)(gpointer, gpointer);

} optwin_cfgref_struct;

#define OPTWIN_CFGREF(p)	((optwin_cfgref_struct *)(p))
#define OPTWIN_CFGREF_ID(p)		(((p) != NULL) ? \
 (OPTWIN_CFGREF(p)->id) : 0)
#define OPTWIN_CFGREF_WIDGET(p)		(((p) != NULL) ? \
 (OPTWIN_CFGREF(p)->w) : NULL)
#define OPTWIN_CFGREF_PARM(p)		(((p) != NULL) ? \
 (OPTWIN_CFGREF(p)->cfg_parm) : NULL)
#define OPTWIN_CFGREF_PARENT(p)		(((p) != NULL) ? \
 (OPTWIN_CFGREF(p)->parent) : NULL)


/*
 *	Options window structure:
 */
typedef struct {

	gboolean  initialized;
	gboolean  map_state;
	gboolean  processing;
	gpointer core_ptr;

	GtkAccelGroup *accelgrp;

	GtkWidget *toplevel;

	GtkWidget *notebook;

	GtkWidget **page;
	gint total_pages;

	GtkWidget *ok_btn, *apply_btn, *save_btn, *cancel_btn;

	optwin_cfgref_struct **cfgref;
	gint total_cfgrefs;

} optwin_struct;

#define OPTWIN(p)	((optwin_struct *)(p))
#define OPTWIN_TOPLEVEL(p)	(((p) != NULL) ? \
 (OPTWIN(p)->toplevel) : NULL)
#define OPTWIN_NOTEBOOK(p)	(((p) != NULL) ? \
 (OPTWIN(p)->notebook) : NULL)
#define OPTWIN_PAGES(p)		(((p) != NULL) ? \
 (OPTWIN(p)->total_pages) : 0)
#define OPTWIN_CFGREFS(p)	(((p) != NULL) ? \
 (OPTWIN(p)->total_cfgrefs) : 0)


extern optwin_cfgref_struct *OptWinCfgRefNew(
	gint id,		/* Unique ID to identify this cfgref. */
	GtkWidget *w,		/* Widget that holds the value. */
	const gchar *cfg_parm	/* Configuration parameter name. */
);
extern void OptWinCfgRefDelete(optwin_cfgref_struct *cfgref);
extern optwin_cfgref_struct *OptWinCfgRefAppend(
	optwin_struct *optwin,
	gint id,                /* Unique ID to identify this cfgref. */
	GtkWidget *w,           /* Widget that holds the value. */
	const gchar *cfg_parm	/* Configuration parameter name. */
);
extern void OptWinCfgRefSetColor(
        optwin_cfgref_struct *cfgref, GdkColor *fg_c, GdkColor *bg_c
);

extern optwin_cfgref_struct *OptWinCfgRefMatchID(
        optwin_struct *optwin, gint id, gint *n
);

extern optwin_struct *OptWinNew(
        gpointer core_ptr, const gchar *title, guint8 **icon_data
);
extern GtkWidget *OptWinPageAppend(
	optwin_struct *optwin, GtkWidget *w
);
extern void OptWinMap(optwin_struct *optwin);
extern void OptWinUnmap(optwin_struct *optwin);
extern void OptWinDelete(optwin_struct *optwin);






#endif	/* OPTWIN_H */
