#include <stdlib.h>
#include <string.h>

#include <Y2/Y.h>
#include <Y2/Ylib.h>

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include "guiutils.h"

#include "cfg.h"
#include "conbar.h"
#include "yaudio.h"
#include "yaudiocb.h"
#include "yc.h"
#include "ycop.h"
#include "yccfglist.h"
#include "config.h"

#include "images/icon_yaudio_48x48.xpm"
#include "images/icon_audio_mode_mono_20x20.xpm"
#include "images/icon_audio_mode_stereo_20x20.xpm"
#include "images/icon_select_20x20.xpm"
#include "images/icon_goto_20x20.xpm"
#include "images/icon_reload_20x20.xpm"
#include "images/icon_calculator_20x20.xpm"

GtkRcStyle *YAudioAudioDisplayRCStyleNew(yaudio_struct *ya);
void YAudioSetTitle(yaudio_struct *ya, const gchar *s);
gchar *YAudioGetLocation(yaudio_struct *ya);
gint YAudioDoConnect(
	yaudio_struct *ya,
	const gchar *address, const gchar *start_command
);
void YAudioDoDisconnect(yaudio_struct *ya, gboolean force_shutdown);

void YAudioSelectAudioModeByName(yaudio_struct *ya, const gchar *name);

void YAudioUpdateAudioPresetModes(yaudio_struct *ya);
void YAudioUpdateAudioValues(yaudio_struct *ya);
void YAudioSetAudioSensitive(yaudio_struct *ya, gboolean sensitive);

void YAudioSavePosition(yaudio_struct *ya);

yaudio_struct *YAudioNew(gpointer core_ptr);
void YAudioUpdateMenus(yaudio_struct *ya);
void YAudioSetBusy(yaudio_struct *ya, gboolean is_busy);
void YAudioMap(yaudio_struct *ya);
void YAudioUnmap(yaudio_struct *ya);
void YAudioDelete(yaudio_struct *ya);


/*
 *      Returns a rcstyle specifying the style for the audio displays.
 */
GtkRcStyle *YAudioAudioDisplayRCStyleNew(yaudio_struct *ya)
{
	gint state;
	GdkColor *ct;
	GtkRcStyle *rcstyle;
	const cfg_color_struct *cs;
	yc_core_struct *core_ptr;


	if(ya == NULL)
	    return(NULL);

	core_ptr = YC_CORE(ya->core_ptr);
	if(core_ptr == NULL)
	    return(NULL);

	rcstyle = gtk_rc_style_new();

#define GET_CFG_COLOR(p)        (YC_GET_CFG_COLOR(core_ptr, (p)))
#define DO_CONVERT_COLOR        \
{ if((cs != NULL) && (ct != NULL)) { \
 ct->red        = (gushort)-1 * cs->r; \
 ct->green      = (gushort)-1 * cs->g; \
 ct->blue       = (gushort)-1 * cs->b; \
} }
	state = GTK_STATE_NORMAL;
	rcstyle->color_flags[state] = GTK_RC_FG |
	    GTK_RC_TEXT | GTK_RC_BASE;
	cs = GET_CFG_COLOR(YC_CFG_PARM_AUDIO_DISPLAY_COLOR_FG);
	ct = &rcstyle->fg[state];
	DO_CONVERT_COLOR
	ct = &rcstyle->text[state];
	DO_CONVERT_COLOR
	cs = GET_CFG_COLOR(YC_CFG_PARM_AUDIO_DISPLAY_COLOR_BG);
	ct = &rcstyle->base[state];
	DO_CONVERT_COLOR

	state = GTK_STATE_PRELIGHT;
	rcstyle->color_flags[state] = GTK_RC_FG |
	    GTK_RC_TEXT | GTK_RC_BASE;
	cs = GET_CFG_COLOR(YC_CFG_PARM_AUDIO_DISPLAY_COLOR_FG);
	ct = &rcstyle->fg[state];
	DO_CONVERT_COLOR
	ct = &rcstyle->text[state];
	DO_CONVERT_COLOR
	cs = GET_CFG_COLOR(YC_CFG_PARM_AUDIO_DISPLAY_COLOR_BG);
	ct = &rcstyle->base[state];
	DO_CONVERT_COLOR

	state = GTK_STATE_ACTIVE;
	rcstyle->color_flags[state] = GTK_RC_FG |
	    GTK_RC_TEXT | GTK_RC_BASE;
	cs = GET_CFG_COLOR(YC_CFG_PARM_AUDIO_DISPLAY_COLOR_FG);
	ct = &rcstyle->fg[state];
	DO_CONVERT_COLOR
	ct = &rcstyle->text[state];
	DO_CONVERT_COLOR
	cs = GET_CFG_COLOR(YC_CFG_PARM_AUDIO_DISPLAY_COLOR_FG_DIM);
	ct = &rcstyle->base[state];
	DO_CONVERT_COLOR

	state = GTK_STATE_INSENSITIVE;
	rcstyle->color_flags[state] = GTK_RC_FG |
	    GTK_RC_TEXT | GTK_RC_BASE;
	cs = GET_CFG_COLOR(YC_CFG_PARM_AUDIO_DISPLAY_COLOR_FG_DIM);
	ct = &rcstyle->fg[state];
	DO_CONVERT_COLOR
	ct = &rcstyle->text[state];
	DO_CONVERT_COLOR
	cs = GET_CFG_COLOR(YC_CFG_PARM_AUDIO_DISPLAY_COLOR_BG);
	ct = &rcstyle->base[state];
	DO_CONVERT_COLOR

	state = GTK_STATE_SELECTED;
	rcstyle->color_flags[state] = GTK_RC_FG | GTK_RC_BG |
	    GTK_RC_TEXT | GTK_RC_BASE;
	cs = GET_CFG_COLOR(YC_CFG_PARM_AUDIO_DISPLAY_COLOR_BG);
	ct = &rcstyle->fg[state];
	DO_CONVERT_COLOR
	ct = &rcstyle->text[state];
	DO_CONVERT_COLOR
	cs = GET_CFG_COLOR(YC_CFG_PARM_AUDIO_DISPLAY_COLOR_FG);
	ct = &rcstyle->bg[state];
	DO_CONVERT_COLOR
	ct = &rcstyle->base[state];
	DO_CONVERT_COLOR

	return(rcstyle);
}

/* 
 *	Sets the title string to s, if s is NULL then values from the
 *	yconsole will be used to form the title.
 */
void YAudioSetTitle(yaudio_struct *ya, const gchar *s)
{
	gchar *title;
	GtkWidget *w;


	if(ya == NULL)
	    return;

	w = YAUDIO_TOPLEVEL(ya);
	if(w == NULL)
	    return;

	/* Title string given? */
	if(s != NULL)
	{
	    gtk_window_set_title(GTK_WINDOW(w), s);
	    return;
	}

	/* Format title string */
	if(YAUDIO_IS_CONNECTED(ya))
	{
	    s = ConBarGetLocation(YAUDIO_CONBAR(ya));
	    title = g_strdup_printf("YAudio: %s", s);
	}
	else
	{
	    title = g_strdup("YAudio");
	}
	gtk_window_set_title(GTK_WINDOW(w), title);

	g_free(title);
}


/*
 *	Returns the current location.
 */
gchar *YAudioGetLocation(yaudio_struct *ya)
{
	return((ya != NULL) ? ConBarGetLocation(YAUDIO_CONBAR(ya)) : NULL);
}

/*
 *	Connects the yconsole to the Y server at the addres s.  If the
 *	yconsole is already connected then it will be disconneted first.
 *
 *	Returns non-zero on error and 0 on success.
 */
gint YAudioDoConnect(
	yaudio_struct *ya,
	const gchar *address, const gchar *start_command
)
{
	YConnection *con;


	if(ya == NULL)
	    return(-1);

	/* Disconnect from Y server as needed */
	YAudioDoDisconnect(ya, FALSE);

	/* If no address is given then only perform disconnect */
	if(address == NULL)
	    return(0);

	/* Connect to Y server */
	ya->con = con = YOpenConnection(NULL, address);
	/* If not able to connect on first try and a server start
	 * command is given, then execute that command and try
	 * connecting again.
	 */
	if((con == NULL) && (start_command != NULL))
	{
	    /* Start server (with dialog) and try connecting again */
	    YCNewStartupDialog(
		YC_CORE(ya->core_ptr), YAUDIO_TOPLEVEL(ya),
		start_command, TRUE
	    );
	    ya->con = con = YOpenConnection(NULL, address);
	}
	if(con != NULL)
	{
	    /* Connected to Y server, now get some initial values */
	    YEventAudioStats stat_buf;
	    dspview_struct *dv = ConBarGetDSPView(YAUDIO_CONBAR(ya));

	    /* Update connection bar connected state */
	    ConBarSetConnectState(YAUDIO_CONBAR(ya), TRUE);

	    /* YSHM sound buffer pointer */
	    ya->yshm_buf = (guint8 *)YSHMOpenSound(
		con, &ya->yshm_buf_len, &ya->yshm_id
	    );

	    /* Audio stats */
	    if(!YGetAudioStats(con, &stat_buf))
	    {
		g_free(ya->audio_mode_name);
		ya->audio_mode_name = g_strdup(stat_buf.mode_name);
		ya->sample_rate = stat_buf.sample_rate;
		ya->sample_size = stat_buf.sample_size;
		ya->channels = stat_buf.channels;
		ya->cycle_us = stat_buf.cycle_us;
		ya->cycle_load = 0.0;
		ya->num_fragments = stat_buf.num_fragments;
		ya->fragment_size = stat_buf.fragment_size;
		ya->shellout = FALSE;

		/* Update audio values on connection bar */
		YAUDIO_SET_CONBAR_AUDIO(ya);

		ConBarSetShellOut(YAUDIO_CONBAR(ya), ya->shellout);
		DSPViewSetChannels(dv, ya->channels);
	    }

	    /* Update Audio preset modes list and Audio value widgets */
	    YAudioUpdateAudioPresetModes(ya);
	    YAudioSelectAudioModeByName(ya, ya->audio_mode_name);
	    YAudioUpdateAudioValues(ya);
	    YAudioSetAudioSensitive(ya, TRUE);

	    /* Record location history on connection bar */
	    ConBarSetLocation(YAUDIO_CONBAR(ya), address, FALSE, TRUE);

	    /* Update title */
	    YAudioSetTitle(ya, NULL);
	}

	return(0);
}

/*
 *	Disconnects the yconsole from the Y server.
 */
void YAudioDoDisconnect(yaudio_struct *ya, gboolean force_shutdown)
{
	dspview_struct *dv;


	if(ya == NULL)
	    return;


	/* Update connection bar connected state */
	ConBarSetConnectState(YAUDIO_CONBAR(ya), FALSE);


	/* Close YSHM sound buffer */
	YSHMCloseSound(YAUDIO_CONNECTION(ya), ya->yshm_buf);
	ya->yshm_buf = NULL;
	ya->yshm_buf_len = 0;
	ya->yshm_id = -1;

	/* Close connection or shutdown? */
	if(force_shutdown)
	    YShutdownServer(YAUDIO_CONNECTION(ya));
	else
	    YCloseConnection(YAUDIO_CONNECTION(ya), False);
	ya->con = NULL;


	/* Reset Y audio values */
	memset(ya->channel_volume_avg, 0x00, 4 * sizeof(gfloat));
	ya->start_time = 0;
	g_free(ya->audio_mode_name);
	ya->audio_mode_name = NULL;
	ya->sample_rate = 0;
	ya->sample_size = 0;
	ya->channels = 0;
	ya->cycle_us = 0;
	ya->cycle_load = 0.0;
	ya->num_fragments = 0;
	ya->fragment_size = 0;
	ya->flip_stereo = FALSE;
	ya->shellout = FALSE;
	ya->version_major = ya->version_minor = 0;
	g_free(ya->server_name);
	ya->server_name = NULL;
	g_free(ya->sound_name);
	ya->sound_name = NULL;

	/* Update audio values on connection bar */
	YAUDIO_SET_CONBAR_AUDIO(ya);

	/* Update Audio preset modes list and Audio value widgets */
	YAudioUpdateAudioPresetModes(ya);
	YAudioUpdateAudioValues(ya);
	YAudioSetAudioSensitive(ya, FALSE);

	dv = ConBarGetDSPView(YAUDIO_CONBAR(ya));
	DSPViewSetChannels(dv, ya->channels);

	/* Update title */
	YAudioSetTitle(ya, NULL);
}

/*
 *	Selects the audio mode on the Audio preset modes list who's
 *	name matches the given name.
 */
void YAudioSelectAudioModeByName(yaudio_struct *ya, const gchar *name)
{
	gint i;
	GtkCList *clist;
	YAudioModeValuesStruct *m;

	if((ya == NULL) || (name == NULL))
	    return;

	clist = (GtkCList *)YAUDIO_MODES_CLIST(ya);
	if(clist == NULL)
	    return;

	/* Iterate through all rows, looking for an Audio preset mode
	 * who's name matches the given name.
	 */
	for(i = 0; i < clist->rows; i++)
	{
	    m = (YAudioModeValuesStruct *)gtk_clist_get_row_data(
		clist, i
	    );
	    if(m == NULL)
		continue;

	    /* This Audio preset mode's name matches the given name? */
	    if(!strcasecmp(m->name, name))
	    {
		/* Select this row */
		gtk_clist_freeze(clist);
		gtk_clist_unselect_all(clist);
		clist->focus_row = i;
		gtk_clist_select_row(clist, i, 0);
		gtk_clist_thaw(clist);
	    }
	}
}

/*
 *	Regets the Audio preset modes and updates the modes clist.
 *
 *	If not connected then the list is cleared.
 */
void YAudioUpdateAudioPresetModes(yaudio_struct *ya)
{
	YConnection *con;
	GtkCList *clist;


	clist = (GtkCList *)YAUDIO_MODES_CLIST(ya);
	if(clist == NULL)
	    return;


	/* Delete existing items in the clist */
	gtk_clist_freeze(clist);
	gtk_clist_clear(clist);
	gtk_clist_thaw(clist);


	/* If connected, then get Audio preset modes from Y server */
	con = YAUDIO_CONNECTION(ya);
	if(con != NULL)
	{
	    gint total_modes;
	    YAudioModeValuesStruct **mode = YGetAudioModes(
		con, &total_modes
	    );
	    if(mode != NULL)
	    {
		gint i, row;
		gchar *v[4];
		YAudioModeValuesStruct *m;

		/* Add all Audio preset modes to the clist */
		gtk_clist_freeze(clist);
		for(i = 0; i < total_modes; i++)
		{
		    m = mode[i];
		    if(m == NULL)
			continue;

		    v[0] = g_strdup(m->name);
		    v[1] = g_strdup_printf("%i Hz", m->sample_rate);
		    v[2] = g_strdup_printf("%i", m->sample_size);
		    v[3] = g_strdup(
			(m->channels == 4) ? "Quad" :
			    ((m->channels == 2) ? "Stereo" : "Mono")
		    );
		    row = gtk_clist_append(clist, v);
		    gtk_clist_set_row_data_full(
			clist, row,
			m, YAudioAudioPresetModeDestroyCB
		    );
		    gtk_clist_set_pixtext(
			clist, row, 0,
			v[0], YC_LIST_PIXMAP_TEXT_SPACING,
			(m->channels == 1) ?
			    ya->pixmap_audio_mode_mono :
			    ya->pixmap_audio_mode_stereo,
			(m->channels == 1) ?
			    ya->mask_audio_mode_mono :
			    ya->mask_audio_mode_stereo
		    );
		    g_free(v[0]);
		    g_free(v[1]);
		    g_free(v[2]);
		    g_free(v[3]);

		    /* Structure was transfered to the clist item data
		     * so do not reference it again from the pointer
		     * array.
		     */
		    mode[i] = m = NULL;
		}
		gtk_clist_thaw(clist);

		/* Only free the pointer array, each structure was
	         * transfered to the clist.
		 */
		g_free(mode);
	    }
	}
}

/*
 *      Regets the Audio values and updates the Audio value widgets.
 */
void YAudioUpdateAudioValues(yaudio_struct *ya)
{
	GtkAdjustment *adj;
	GtkCombo *combo;
/*	YConnection *con; */
/*	YEventAudioStats a; */
	gchar numstr[80];


	if(ya == NULL)
	    return;
/*
	con = YAUDIO_CONNECTION(ya);
	if(con == NULL)
	    return;
 */
	/* Get current Audio values from Y server */
/*
	if(YGetAudioStats(con, &a))
	    return;
 */

	/* Sample rate */
	sprintf(numstr, "%i", ya->sample_rate);
	combo = (GtkCombo *)ya->sample_rate_combo;
	if(combo != NULL)
	    gtk_entry_set_text(GTK_ENTRY(combo->entry), numstr);

	/* Sample size */
	sprintf(numstr, "%i", ya->sample_size);
	combo = (GtkCombo *)ya->sample_size_combo;
	if(combo != NULL)
	    gtk_entry_set_text(GTK_ENTRY(combo->entry), numstr);

	/* Channels */
	strcpy(
	    numstr,
	    (ya->channels == 4) ? "Quad" :
		((ya->channels == 2) ? "Stereo" : "Mono")
	);
	combo = (GtkCombo *)ya->channels_combo;
	if(combo != NULL)
	    gtk_entry_set_text(GTK_ENTRY(combo->entry), numstr);

	/* Buffer fragment size */
	sprintf(numstr, "%i", ya->fragment_size);
	combo = (GtkCombo *)ya->fragment_size_combo;
	if(combo != NULL)
	    gtk_entry_set_text(GTK_ENTRY(combo->entry), numstr);

	/* Cycle in microseconds */
	adj = ya->cycle_adj;
	if(adj != NULL)
	{
	    adj->value = (gfloat)ya->cycle_us;
	    gtk_signal_emit_by_name(GTK_OBJECT(adj), "value_changed");
	}
}

/*
 *	Sets all the Audio widgets on the window sensitive or
 *	insensitive.
 */
void YAudioSetAudioSensitive(yaudio_struct *ya, gboolean sensitive)
{
	GtkWidget *w;

	if(ya == NULL)
	    return;

#define DO_SET_SENSITIVE	\
{ if(w != NULL) gtk_widget_set_sensitive(w, sensitive); }

	w = ya->current_mode_btn;
	DO_SET_SENSITIVE
	w = ya->select_mode_btn;
	DO_SET_SENSITIVE

	w = ya->sample_rate_combo;
	DO_SET_SENSITIVE
	w = ya->sample_size_combo;
	DO_SET_SENSITIVE
	w = ya->channels_combo;
	DO_SET_SENSITIVE
	w = ya->fragment_size_combo;
	DO_SET_SENSITIVE
	w = ya->cycle_spin;
	DO_SET_SENSITIVE
	w = ya->calculate_cycle_btn;
	DO_SET_SENSITIVE
	w = ya->set_values_btn;
	DO_SET_SENSITIVE
	w = ya->current_values_btn;
	DO_SET_SENSITIVE

#undef DO_SET_SENSITIVE
}

/*
 *	Save current position value of window to configuration list on
 *	core structure.
 */
void YAudioSavePosition(yaudio_struct *ya)
{
	GtkWidget *w;
	yc_core_struct *core_ptr;


	if(ya == NULL)
	    return;

	core_ptr = YC_CORE(ya->core_ptr);
	if(core_ptr == NULL)
	    return;

	/* Do not save positions? */
	if(!YC_GET_CFG_I(core_ptr, YC_CFG_PARM_AUDIO_SAVE_POSITIONS))
	    return;

	/* Save toplevel position */
	w = YAUDIO_TOPLEVEL(ya);
	if(w != NULL)
	{
	    GdkWindow *window = w->window;
	    gint x = 0, y = 0, width, height;


	    if(window != NULL)
		gdk_window_get_root_origin(window, &x, &y);
	    YC_SET_CFG_I(core_ptr, YC_CFG_PARM_AUDIO_X, x);
	    YC_SET_CFG_I(core_ptr, YC_CFG_PARM_AUDIO_Y, y);

	    width = w->allocation.width;
	    height = w->allocation.height;
	    YC_SET_CFG_I(core_ptr, YC_CFG_PARM_AUDIO_WIDTH, width);
	    YC_SET_CFG_I(core_ptr, YC_CFG_PARM_AUDIO_HEIGHT, height);
	}

	/* Save notebook page */
	w = ya->notebook;
	if(w != NULL)
	{
	    GtkNotebook *notebook = GTK_NOTEBOOK(w);
	    YC_SET_CFG_I(
		core_ptr, YC_CFG_PARM_AUDIO_PAGE,
		gtk_notebook_get_current_page(notebook)
	    );
	}

	/* Save Audio preset modes list column widths */
	w = YAUDIO_MODES_CLIST(ya);
	if(w != NULL)
	{
	    gint i;
	    GList *glist = NULL;
	    GtkCList *clist = GTK_CLIST(w);
	    const gint columns = clist->columns;
	    GtkCListColumn *column;
	    cfg_intlist_struct *intlist;

	    for(i = 0; i < columns; i++)
	    {
		column = &clist->column[i];
		glist = g_list_append(glist, (gpointer)column->width);
	    }
	    intlist = CFGIntListNew(glist);
	    g_list_free(glist);
	    YC_SET_CFG_INTLIST(
		core_ptr, YC_CFG_PARM_AUDIO_MODES_LIST_COLUMN_WIDTHS,
		intlist
	    );
	    CFGIntListDelete(intlist);
	}
}


/*
 *	Create a new YAudio.
 */
yaudio_struct *YAudioNew(gpointer core_ptr)
{
	gint i, border_major = 5, border_minor = 2;
	gchar *heading[4];
	gint list_column_width[4];
	GList *glist;
	GdkWindow *window;
	GtkAdjustment *adj;
	GtkAccelGroup *accelgrp;
	GtkStyle *style;
	GtkWidget *w, *parent, *parent2, *parent3, *parent4,
			*toplevel, *menu;
	GtkCombo *combo;
	GtkNotebook *notebook;
	GtkCList *clist;
	gpointer combo_rtn;
	connection_bar_struct *cb;
	dspview_struct *dv;
	yc_core_struct *cp = YC_CORE(core_ptr);
	yaudio_struct *ya = (yaudio_struct *)g_malloc0(
	    sizeof(yaudio_struct)
	);
	if(ya == NULL)
	    return(NULL);


	/* Reset values */
	ya->initialized = TRUE;
	ya->map_state = FALSE;
	ya->busy_count = 0;
	ya->accelgrp = accelgrp = gtk_accel_group_new();
	ya->core_ptr = core_ptr;

	ya->con = NULL;
	ya->yshm_buf = NULL;
	ya->yshm_buf_len = 0;
	ya->yshm_id = -1;

	ya->start_time = 0;
	ya->audio_mode_name = NULL;
	ya->sample_rate = 0;
	ya->sample_size = 0;
	ya->channels = 0;
	ya->cycle_us = 0;
	ya->cycle_load = 0.0f;
	ya->num_fragments = 0;
	ya->fragment_size = 0;
	ya->flip_stereo = FALSE;
	ya->shellout = FALSE;
	ya->version_major = ya->version_minor = 0;
	ya->server_name = ya->sound_name = NULL;


	/* Create toplevel */
	ya->toplevel = toplevel = w = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_widget_set_name(w, "yaudio window");
	if(cp != NULL)
	{
	    if(YC_GET_CFG_I(cp, YC_CFG_PARM_AUDIO_SAVE_POSITIONS))
	    {
		const cfg_intlist_struct *intlist;

		gtk_widget_set_uposition(
		    w,
		    YC_GET_CFG_I(cp, YC_CFG_PARM_AUDIO_X),
		    YC_GET_CFG_I(cp, YC_CFG_PARM_AUDIO_Y)
		);
		gtk_widget_set_usize(
		    w,
		    YC_GET_CFG_I(cp, YC_CFG_PARM_AUDIO_WIDTH),
		    YC_GET_CFG_I(cp, YC_CFG_PARM_AUDIO_HEIGHT)
		);

		intlist = YC_GET_CFG_INTLIST(
		    cp, YC_CFG_PARM_AUDIO_MODES_LIST_COLUMN_WIDTHS
		);
		if(intlist != NULL)
		{
		    GList *glist = intlist->list;
		    for(i = 0; i < 4; i++)
		    {
			if(glist == NULL)
			    break;

			list_column_width[i] = (gint)glist->data;
			glist = g_list_next(glist);
		    }
		}
	    }
	}
	gtk_window_set_policy(GTK_WINDOW(w), TRUE, TRUE, TRUE);
	gtk_widget_realize(w);
	window = w->window;
	if(window != NULL)
	{
	    gdk_window_set_decorations(
		window,
		GDK_DECOR_BORDER | GDK_DECOR_RESIZEH |
		GDK_DECOR_TITLE | GDK_DECOR_MENU |
		GDK_DECOR_MINIMIZE | GDK_DECOR_MAXIMIZE
	    );
	    gdk_window_set_functions(
		window,
		GDK_FUNC_MOVE | GDK_FUNC_RESIZE | GDK_FUNC_MINIMIZE |
		GDK_FUNC_MAXIMIZE | GDK_FUNC_CLOSE
	    );
	    GUISetWMIcon(window, (guint8 **)icon_yaudio_48x48_xpm);
	}
	gtk_accel_group_attach(accelgrp, GTK_OBJECT(w));
	gtk_signal_connect(
	    GTK_OBJECT(w), "delete_event",
	    GTK_SIGNAL_FUNC(YAudioDeleteEventCB), ya
	);
	parent = w;

	/* Main vbox */
	w = gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(parent), w);
	gtk_widget_show(w);
	parent = w;

	/* Create connection bar */
	ya->conbar = cb = ConBarNew(
	    CONBAR_DISPLAY_FLAG_WINDOW_MENU | CONBAR_DISPLAY_FLAG_LOCATION |
	    CONBAR_DISPLAY_FLAG_DSPVIEW,
	    cp->dsp_bg_img_small_width + 2, cp->dsp_bg_img_small_height + 2,
	    ya,
	    YAudioConnectCB, YAudioYServerListCB, YAudioCloseCB,
	    YAudioNewYConsoleCB, YAudioNewYAudioCB, YAudioNewYMixerCB,
	    YAudioNewYPlayerCB, YAudioUptionsCB,
	    YAudioRefreshCB, YAudioShelloutCB, YAudioShutdownCB,
	    parent, accelgrp
	);
	w = CONBAR_AUDIO(cb);
	dv = ConBarGetDSPView(cb);
	ConBarMap(cb);

	if(cp != NULL)
	    DSPViewSetBG(
		dv, cp->dsp_bg_img_small,
		cp->dsp_bg_img_small_width, cp->dsp_bg_img_small_height,
		cp->dsp_bg_img_small_bpl, 4
	    );
	DSPViewSetDisplayStyle(dv, DSPVIEW_DISPLAY_STYLE_NOVA);
	DSPViewSetMergeMode(dv, DSPVIEW_MERGE_MODE_SET);
	DSPViewSetDSPColor(dv, DSPVIEW_DEF_DSP_COLOR);  /* 0xAABBGGRR */
	w = DSPVIEW_VIEW(dv);
	gtk_widget_add_events(
	    w, GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "button_press_event",
	    GTK_SIGNAL_FUNC(YAudioDSPViewButtonCB), ya
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "button_release_event",
	    GTK_SIGNAL_FUNC(YAudioDSPViewButtonCB), ya
	);


	/* Frame for notebook */
	w = gtk_frame_new(NULL);
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_IN);
	gtk_box_pack_start(GTK_BOX(parent), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent2 = w;

	/* Main notebook */
	ya->notebook = w = gtk_notebook_new();
	notebook = GTK_NOTEBOOK(w);
	gtk_notebook_set_tab_pos(notebook, GTK_POS_TOP);
	gtk_notebook_set_scrollable(notebook, TRUE);
	gtk_notebook_set_show_tabs(notebook, TRUE);
	gtk_notebook_set_show_border(notebook, TRUE);
	gtk_container_add(GTK_CONTAINER(parent2), w);
/*	gtk_signal_connect(
	    GTK_OBJECT(w), "switch_page",
	    GTK_SIGNAL_FUNC(YAudioSwitchPageCB), ya
	);
 */
	gtk_widget_show(w);


	/* Create Audio preset modes list page */
	w = gtk_label_new("Preset Modes");
	gtk_widget_show(w);
	parent2 = gtk_vbox_new(FALSE, 0);
	gtk_container_border_width(GTK_CONTAINER(parent2), border_major);
	gtk_notebook_append_page(notebook, parent2, w);
	gtk_widget_show(parent2);

	/* Scrolled window for clist */
	w = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(
	    GTK_SCROLLED_WINDOW(w),
	    GTK_POLICY_AUTOMATIC,
	    GTK_POLICY_AUTOMATIC
	);
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent3 = w;

	/* Audio preset modes clist */
	heading[0] = "Name";
	heading[1] = "Sample Rate";
	heading[2] = "Bits";
	heading[3] = "Channels";
	ya->modes_clist = w = gtk_clist_new_with_titles(4, heading);
	clist = GTK_CLIST(w);
	gtk_clist_column_titles_passive(clist);
	for(i = 0; i < 4; i++)
	    gtk_clist_set_column_width(clist, i, list_column_width[i]);
	gtk_clist_set_row_height(clist, YC_LIST_ROW_SPACING);
	gtk_widget_add_events(
	    w,
	    GDK_KEY_PRESS_MASK | GDK_KEY_RELEASE_MASK |
	    GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "key_press_event",
	    GTK_SIGNAL_FUNC(YAudioModesListKeyEventCB), ya
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "key_release_event",
	    GTK_SIGNAL_FUNC(YAudioModesListKeyEventCB), ya
	);
	gtk_signal_connect_after(
	    GTK_OBJECT(w), "button_press_event",
	    GTK_SIGNAL_FUNC(YAudioModesListButtonEventCB), ya
	);
	gtk_signal_connect_after(
	    GTK_OBJECT(w), "button_release_event",
	    GTK_SIGNAL_FUNC(YAudioModesListButtonEventCB), ya
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "select_row",
	    GTK_SIGNAL_FUNC(YAudioModesListSelectRowCB), ya
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "unselect_row",
	    GTK_SIGNAL_FUNC(YAudioModesListUnselectRowCB), ya
	);
	gtk_container_add(GTK_CONTAINER(parent3), w);
	gtk_widget_show(w);

	/* Load pixmap and mask pairs for mode list items */
	style = gtk_widget_get_style(w);
	ya->pixmap_audio_mode_mono = gdk_pixmap_create_from_xpm_d(
	    window, &ya->mask_audio_mode_mono,
	    &style->bg[GTK_STATE_NORMAL],
	    (gchar **)icon_audio_mode_mono_20x20_xpm
	);
	ya->pixmap_audio_mode_stereo = gdk_pixmap_create_from_xpm_d(
	    window, &ya->mask_audio_mode_stereo,
	    &style->bg[GTK_STATE_NORMAL],
	    (gchar **)icon_audio_mode_stereo_20x20_xpm
	);


	/* Right click menu */
	ya->modes_menu = menu = GUIMenuCreate();
	if(menu != NULL)
	{
	    guchar **icon;
	    const gchar *label;
	    gint accel_key;
	    guint accel_mods;
	    gpointer mclient_data = ya;
	    void (*func_cb)(GtkWidget *w, gpointer);
	    gpointer fw;

#define DO_ADD_MENU_ITEM_LABEL	\
{ \
 w = GUIMenuItemCreate( \
  menu, GUI_MENU_ITEM_TYPE_LABEL, accelgrp, \
  icon, label, accel_key, accel_mods, (void **)&fw, \
  mclient_data, func_cb \
 ); \
}
#define DO_ADD_MENU_SEP		\
{ \
 w = GUIMenuItemCreate( \
  menu, GUI_MENU_ITEM_TYPE_SEPARATOR, NULL, \
  NULL, NULL, 0, 0, NULL, \
  NULL, NULL \
 ); \
}
	    icon = (guchar **)icon_select_20x20_xpm;
	    label = "Select";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = YAudioAudioModesListSelectModeCB;
	    DO_ADD_MENU_ITEM_LABEL
	    ya->modes_select_mi = w;

	    icon = (guchar **)icon_goto_20x20_xpm;
	    label = "Current";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = YAudioAudioModesListCurrentModeCB;
	    DO_ADD_MENU_ITEM_LABEL
	    ya->modes_current_mi = w;

#undef DO_ADD_MENU_SEP
#undef DO_ADD_MENU_ITEM_LABEL
	}

	/* HBox for buttons */
	w = gtk_hbox_new(TRUE, 0);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;

	/* Select mode button */
	ya->select_mode_btn = w = GUIButtonPixmapLabelH(
	    (guint8 **)icon_select_20x20_xpm, "Select", NULL
	);
	GTK_WIDGET_SET_FLAGS(w, GTK_CAN_DEFAULT);
	gtk_widget_set_usize(
	    w,
	    GUI_BUTTON_HLABEL_WIDTH_DEF, GUI_BUTTON_HLABEL_HEIGHT_DEF
	);
	gtk_box_pack_start(GTK_BOX(parent3), w, TRUE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(YAudioAudioModesListSelectModeCB), ya
	);
	GUISetWidgetTip(w, "Click to select the Audio preset mode");
	gtk_widget_show(w);

	/* Current mode button */
	ya->current_mode_btn = w = GUIButtonPixmapLabelH(
	    (guint8 **)icon_goto_20x20_xpm, "Current", NULL
	);
	GTK_WIDGET_SET_FLAGS(w, GTK_CAN_DEFAULT);
	gtk_widget_set_usize(
	    w,
	    GUI_BUTTON_HLABEL_WIDTH_DEF, GUI_BUTTON_HLABEL_HEIGHT_DEF
	);
	gtk_box_pack_start(GTK_BOX(parent3), w, TRUE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(YAudioAudioModesListCurrentModeCB), ya
	);
	GUISetWidgetTip(
	    w,
	    "Click to go to the current Audio preset mode (if known)"
	);
	gtk_widget_show(w);


	/* Create Audio values page */
	w = gtk_label_new("Values");
	gtk_widget_show(w);
	parent2 = gtk_vbox_new(FALSE, border_major);
	gtk_container_border_width(GTK_CONTAINER(parent2), border_major);
	gtk_notebook_append_page(notebook, parent2, w);
	gtk_widget_show(parent2);

	/* Hbox for sample rate and sample size combos */
	w = gtk_hbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;

	/* Sample rate combo */
	glist = NULL;
	glist = g_list_append(glist, "8000");
	glist = g_list_append(glist, "11025");
	glist = g_list_append(glist, "22050");
	glist = g_list_append(glist, "44100");
	glist = g_list_append(glist, "11025");
	w = GUIComboCreate(
	    "Sample Rate:",
	    (const gchar *)((glist != NULL) ? glist->data : NULL),
	    glist,
	    g_list_length(glist),
	    &combo_rtn,
	    NULL, NULL, NULL
	);
	g_list_free(glist);
	combo = (GtkCombo *)combo_rtn;
	ya->sample_rate_combo = GTK_WIDGET(combo);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	gtk_entry_set_editable(GTK_ENTRY(combo->entry), FALSE);
	gtk_widget_set_usize(combo->entry, 60, -1);
	GUISetWidgetTip(
	    combo->entry,
	    "Select the sample rate (in Hz)"
	);

	/* Sample size combo */
	glist = NULL;
	glist = g_list_append(glist, "8");
	glist = g_list_append(glist, "16");
	w = GUIComboCreate(
	    "Sample Size:",
	    (const gchar *)((glist != NULL) ? glist->data : NULL),
	    glist,
	    g_list_length(glist),
	    &combo_rtn,
	    NULL, NULL, NULL
	);
	g_list_free(glist);
	combo = (GtkCombo *)combo_rtn;
	ya->sample_size_combo = GTK_WIDGET(combo);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	gtk_entry_set_editable(GTK_ENTRY(combo->entry), FALSE);
	gtk_widget_set_usize(combo->entry, 35, -1);
	GUISetWidgetTip(
	    combo->entry,
	    "Select the sample size (in bits)"
	);


	/* Hbox for channels and buffer fragment size combos */
	w = gtk_hbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;

	/* Channels combo */
	glist = NULL;
	glist = g_list_append(glist, "Mono");
	glist = g_list_append(glist, "Stereo");
	glist = g_list_append(glist, "Quad");
	w = GUIComboCreate(
	    "Channels:",
	    (const gchar *)((glist != NULL) ? glist->data : NULL),
	    glist,
	    g_list_length(glist),
	    &combo_rtn,
	    NULL, NULL, NULL
	);
	g_list_free(glist);
	combo = (GtkCombo *)combo_rtn;
	ya->channels_combo = GTK_WIDGET(combo);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	gtk_entry_set_editable(GTK_ENTRY(combo->entry), FALSE);
	gtk_widget_set_usize(combo->entry, 60, -1);
	GUISetWidgetTip(
	    combo->entry,
	    "Select channels"
	);

	/* Buffer fragment size */
	glist = NULL;
	glist = g_list_append(glist, "256");
	glist = g_list_append(glist, "512");
	glist = g_list_append(glist, "1024");
	glist = g_list_append(glist, "2048");
	glist = g_list_append(glist, "4096");
	glist = g_list_append(glist, "8192");
	glist = g_list_append(glist, "16384");
	w = GUIComboCreate(
	    "Buffer Fragment Size:",
	    (const gchar *)((glist != NULL) ? glist->data : NULL),
	    glist,
	    g_list_length(glist),
	    &combo_rtn,
	    NULL, NULL, NULL
	);
	g_list_free(glist);
	combo = (GtkCombo *)combo_rtn;
	ya->fragment_size_combo = GTK_WIDGET(combo);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	gtk_entry_set_editable(GTK_ENTRY(combo->entry), FALSE);
	gtk_widget_set_usize(combo->entry, 60, -1);
	GUISetWidgetTip(
	    combo->entry,
	    "Select the buffer fragment size (in bytes)"
	);


	/* Hbox for cycle spin and calculate cucle button */
	w = gtk_hbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;

	/* Hbox for cycle spin */
	w = gtk_hbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent4 = w;

	w = gtk_label_new("Cycle:");
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	ya->cycle_adj = adj = (GtkAdjustment *)gtk_adjustment_new(
	    0.0f, 1000.0f, 1000000.0f,
	    1000.0f, 10000.0f, 0.0f
	);
	gtk_object_ref(GTK_OBJECT(adj));
	ya->cycle_spin = w = gtk_spin_button_new(adj, 1.0, 0);
	gtk_widget_set_usize(w, 70, -1);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	GUISetWidgetTip(
	    w,
	    "Set the cycle interval (in microseconds), or click on\
 \"Calculate\" to automatically calculate the cycle based on the\
 other Audio values"
	);
	gtk_widget_show(w);

	/* Calculate cycle button */
	ya->calculate_cycle_btn = w = GUIButtonPixmapLabelH(
	    (guint8 **)icon_calculator_20x20_xpm, "Calculate", NULL
	);
	gtk_widget_set_usize(
	    w,
	    GUI_BUTTON_HLABEL_WIDTH, GUI_BUTTON_HLABEL_HEIGHT
	);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(YAudioAudioValuesCalculateCycleCB), ya
	);
	GUISetWidgetTip(
	    w,
	    "Click to automatically calculate the \"theoretical\"\
 cycle interval"
	);
	gtk_widget_show(w);



	/* Hbox to fill the gap */
	w = gtk_hbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, TRUE, 0);
	gtk_widget_show(w);




	w = gtk_hseparator_new();
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);


	/* Hbox for buttons */
	w = gtk_hbox_new(TRUE, 0);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;

	/* Set Audio values button */
	ya->set_values_btn = w = GUIButtonPixmapLabelH(
	    (guint8 **)icon_select_20x20_xpm, "Set", NULL
	);
	GTK_WIDGET_SET_FLAGS(w, GTK_CAN_DEFAULT);
	gtk_widget_set_usize(
	    w,
	    GUI_BUTTON_HLABEL_WIDTH_DEF, GUI_BUTTON_HLABEL_HEIGHT_DEF
	);
	gtk_box_pack_start(GTK_BOX(parent3), w, TRUE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(YAudioAudioValuesSetCB), ya
	);
	GUISetWidgetTip(w, "Click to set Audio values");
	gtk_widget_show(w);

	/* Current Audio values button */
	ya->current_values_btn = w = GUIButtonPixmapLabelH(
	    (guint8 **)icon_reload_20x20_xpm, "Current", NULL
	);
	GTK_WIDGET_SET_FLAGS(w, GTK_CAN_DEFAULT);
	gtk_widget_set_usize(
	    w,
	    GUI_BUTTON_HLABEL_WIDTH_DEF, GUI_BUTTON_HLABEL_HEIGHT_DEF
	);
	gtk_box_pack_start(GTK_BOX(parent3), w, TRUE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(YAudioAudioValuesCurrentCB), ya
	);
	GUISetWidgetTip(w, "Click to get current Audio values");
	gtk_widget_show(w);


	/* Select notebook page */
	gtk_notebook_set_page(
	    notebook, YC_GET_CFG_I(cp, YC_CFG_PARM_AUDIO_PAGE)
	);

	YAudioSetTitle(ya, NULL);
	YAudioUpdateMenus(ya);

	return(ya);
}

/*
 *	Update menus on yconsole.
 */
void YAudioUpdateMenus(yaudio_struct *ya)
{
	gboolean sensitive;
	gint row;
	GList *glist;
	GtkWidget *w;
	GtkCList *clist;
	YConnection *con;


	if(ya == NULL)
	    return;

	con = YAUDIO_CONNECTION(ya);

	clist = (GtkCList *)YAUDIO_MODES_CLIST(ya);
	if(clist == NULL)
	    return;

	/* Get last selected row or -1 if none selected */
	glist = clist->selection_end;
	row = (glist != NULL) ? (gint)glist->data : -1;


#define DO_SET_SENSITIVE	\
 { if(w != NULL) gtk_widget_set_sensitive(w, sensitive); }

	/* Connected and row selected? */
	sensitive = ((con != NULL) && (row > -1)) ? TRUE : FALSE;
	w = ya->modes_select_mi;
	DO_SET_SENSITIVE
	w = ya->select_mode_btn;
	DO_SET_SENSITIVE

	/* Connected? */
	sensitive = (con != NULL) ? TRUE : FALSE;
	w = ya->modes_current_mi;
	DO_SET_SENSITIVE
	w = ya->current_mode_btn;
	DO_SET_SENSITIVE

	/* Do not modify Audio values widgets */

#undef DO_SET_SENSITIVE
}

/*
 *	Mark yconsole as busy or not.
 */
void YAudioSetBusy(yaudio_struct *ya, gboolean is_busy)
{
	GdkCursor *cursor;
	GtkWidget *w;
	yc_core_struct *core_ptr;


	if(ya == NULL)
	    return;

	core_ptr = YC_CORE(ya->core_ptr);
	if(core_ptr == NULL)
	    return;

	w = YAUDIO_TOPLEVEL(ya);
	if(w != NULL)
	{
	    if(is_busy)
	    {
		/* Increase busy count */
		ya->busy_count++;

		/* If already busy then don't change anything */
		if(ya->busy_count > 1)
		    return;

		cursor = core_ptr->cursor_busy;
	    }
	    else
	    {
		/* Reduce busy count */
		ya->busy_count--;
		if(ya->busy_count < 0)
		    ya->busy_count = 0;

		/* If still busy do not change anything */
		if(ya->busy_count > 0)
		    return;

		cursor = NULL;  /* Use default cursor */
	    }

	    /* Update toplevel window's cursor */
	    if(w->window != NULL)
	    {
		gdk_window_set_cursor(w->window, cursor);
		gdk_flush();
	    }
	}
}

/*
 *	Map yconsole.
 */
void YAudioMap(yaudio_struct *ya)
{
	if(ya == NULL)
	    return;

	if(TRUE)
	{
	    GtkWidget *w = YAUDIO_TOPLEVEL(ya);
	    if(w != NULL)
		gtk_widget_show_raise(w);
	    ya->map_state = TRUE;
	}
}

/*
 *	Unmap yconsole.
 */
void YAudioUnmap(yaudio_struct *ya)
{
	if(ya == NULL)
	    return;

	if(TRUE)
	{
	    GtkWidget *w = YAUDIO_TOPLEVEL(ya);
	    if(w != NULL)
		gtk_widget_hide(w);
	    ya->map_state = FALSE;
	}
}

/*
 *	Delete the yconsole.
 */
void YAudioDelete(yaudio_struct *ya)
{
	GdkBitmap **mask;
	GdkPixmap **pixmap;
	GtkAdjustment **adj;
	GtkWidget **w;


	if(ya == NULL)
	    return;

	if(ya->initialized)
	{
#define DO_UNREF_ADJ		\
{ if(*adj != NULL) { gtk_object_unref(GTK_OBJECT(*adj)); *adj = NULL; } }

#define DO_UNREF_PIXMAP_MASK    \
{ if(*pixmap != NULL) { gdk_pixmap_unref(*pixmap); *pixmap = NULL; } \
  if(*mask != NULL) { gdk_bitmap_unref(*mask); *mask = NULL; } }

#define DO_DESTROY_WIDGET       \
{ if(*w != NULL) { gtk_widget_destroy(*w); *w = NULL; } }

	    /* Disconnect from Y server as needed, this will reset
	     * all connection values.
	     */
	    YAudioDoDisconnect(ya, FALSE);

	    /* Delete connection bar */
	    ConBarDelete(YAUDIO_CONBAR(ya));
	    ya->conbar = NULL;

	    /* Destroy widgets */
	    w = &ya->sample_rate_combo;
	    DO_DESTROY_WIDGET
	    w = &ya->sample_size_combo;
	    DO_DESTROY_WIDGET
	    w = &ya->channels_combo;
	    DO_DESTROY_WIDGET
	    w = &ya->fragment_size_combo;
	    DO_DESTROY_WIDGET
	    w = &ya->cycle_spin;
	    DO_DESTROY_WIDGET
	    w = &ya->calculate_cycle_btn;
	    DO_DESTROY_WIDGET
	    w = &ya->set_values_btn;
	    DO_DESTROY_WIDGET
	    w = &ya->current_values_btn;
	    DO_DESTROY_WIDGET
	    adj = &ya->cycle_adj;
	    DO_UNREF_ADJ

	    w = &ya->select_mode_btn;
	    DO_DESTROY_WIDGET
	    w = &ya->current_mode_btn;
	    DO_DESTROY_WIDGET
	    w = &ya->modes_menu;
	    ya->modes_select_mi = NULL;
	    ya->modes_current_mi = NULL;
	    DO_DESTROY_WIDGET
	    w = &ya->modes_clist;
	    DO_DESTROY_WIDGET

	    w = &ya->notebook;
	    DO_DESTROY_WIDGET
	    w = &ya->toplevel;
	    DO_DESTROY_WIDGET

	    if(ya->accelgrp != NULL)
	    {
		gtk_accel_group_unref(ya->accelgrp);
		ya->accelgrp = NULL;
	    }

	    pixmap = &ya->pixmap_audio_mode_mono;
	    mask = &ya->mask_audio_mode_mono;
	    DO_UNREF_PIXMAP_MASK
	    pixmap = &ya->pixmap_audio_mode_stereo;
	    mask = &ya->mask_audio_mode_stereo;
	    DO_UNREF_PIXMAP_MASK

#undef DO_DESTROY_WIDGET
#undef DO_UNREF_ADJ
#undef DO_UNREF_PIXMAP_MASK
	}

	/* Deallocate structure itself */
	g_free(ya);
}
