/*
			     YAudio Window
 */

#ifndef YAUDIO_H
#define YAUDIO_H

#include <Y2/Y.h>
#include <Y2/Ylib.h>
#include <gtk/gtk.h>

#include "conbar.h"

/*
 *	YAudio window structure:
 */
typedef struct {

	gboolean	initialized;
	gboolean	map_state;
	gint		busy_count;
	GtkAccelGroup	*accelgrp;
	gpointer	core_ptr;

	GtkWidget	*toplevel;
	connection_bar_struct *conbar;
	GtkWidget	*notebook;

	/* Widgets on Audio preset modes page. */
	GtkWidget	*modes_clist,
			*modes_menu,
			*modes_select_mi,
			*modes_current_mi,
			*current_mode_btn,
			*select_mode_btn;
	GdkPixmap	*pixmap_audio_mode_mono;
	GdkBitmap	*mask_audio_mode_mono;
	GdkPixmap       *pixmap_audio_mode_stereo;
	GdkBitmap       *mask_audio_mode_stereo;

	/* Widgets on Audio values page. */
	GtkWidget       *sample_rate_combo,
			*sample_size_combo,
			*channels_combo,
			*fragment_size_combo,	/* Buffer fragment size in bytes. */
			*cycle_spin,
			*calculate_cycle_btn,
			*set_values_btn,
			*current_values_btn;
	GtkAdjustment	*cycle_adj;

	/* Y server values. */
	YConnection *con;

	guint8 *yshm_buf;
	gint yshm_buf_len, yshm_id;

	gfloat channel_volume_avg[4];

	gulong start_time;
	gchar *audio_mode_name;
	gint sample_rate, sample_size, channels;
	glong cycle_us;
	gfloat cycle_load;
	gint num_fragments, fragment_size;
	gboolean flip_stereo, shellout;
	gint version_major, version_minor;
	gchar *server_name, *sound_name;

} yaudio_struct;

#define YAUDIO(p)	((yaudio_struct *)(p))
#define YAUDIO_TOPLEVEL(p)	(((p) != NULL) ? \
 (YAUDIO(p)->toplevel) : NULL)
#define YAUDIO_NOTEBOOK(p)	(((p) != NULL) ? \
 (YAUDIO(p)->notebook) : NULL)
#define YAUDIO_MODES_CLIST(p)	(((p) != NULL) ? \
 (YAUDIO(p)->modes_clist) : NULL)
#define YAUDIO_CONBAR(p)	(((p) != NULL) ? \
 (YAUDIO(p)->conbar) : NULL)

#define YAUDIO_IS_CONNECTED(p)	(((p) != NULL) ? \
 ((YAUDIO(p)->con) != NULL) : FALSE)
#define YAUDIO_CONNECTION(p)	 (((p) != NULL) ? \
 (YAUDIO(p)->con) : NULL)

#define YAUDIO_SET_CONBAR_AUDIO(p)	\
{ \
 if((p) != NULL) \
  ConBarSetAudio( \
   YAUDIO_CONBAR(p), \
   (p)->start_time, (p)->audio_mode_name, \
   (p)->sample_rate, (p)->sample_size, (p)->channels, \
   (p)->cycle_us, (p)->cycle_load, \
   (p)->num_fragments, (p)->fragment_size, \
   (p)->flip_stereo, (p)->shellout, \
   (p)->yshm_buf, (p)->yshm_buf_len, \
   (p)->channel_volume_avg, \
   (p)->version_major, (p)->version_minor, \
   (p)->server_name, (p)->sound_name \
  ); \
}

extern GtkRcStyle *YAudioAudioDisplayRCStyleNew(yaudio_struct *ya);
extern void YAudioSetTitle(yaudio_struct *ya, const gchar *s);
extern gchar *YAudioGetLocation(yaudio_struct *ya);
extern gint YAudioDoConnect(
	yaudio_struct *ya,
	const gchar *address, const gchar *start_command
);
extern void YAudioDoDisconnect(yaudio_struct *ya, gboolean force_shutdown);

extern void YAudioSelectAudioModeByName(yaudio_struct *ya, const gchar *name);

extern void YAudioUpdateAudioPresetModes(yaudio_struct *ya);
extern void YAudioUpdateAudioValues(yaudio_struct *ya);
extern void YAudioSetAudioSensitive(yaudio_struct *ya, gboolean sensitive);

extern void YAudioSavePosition(yaudio_struct *ya);

extern yaudio_struct *YAudioNew(gpointer core_ptr);
extern void YAudioUpdateMenus(yaudio_struct *ya);
extern void YAudioSetBusy(yaudio_struct *ya, gboolean is_busy);
extern void YAudioMap(yaudio_struct *ya);
extern void YAudioUnmap(yaudio_struct *ya);
extern void YAudioDelete(yaudio_struct *ya);



#endif	/* YAUDIO_H */
