#include <stdlib.h>
#include <string.h>

#include <Y2/Y.h>
#include <Y2/Ylib.h>

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include "guiutils.h"

#include "cfg.h"
#include "conbar.h"
#include "yaudio.h"
#include "yaudiocb.h"
#include "yc.h"
#include "ycop.h"
#include "ycanalyze.h"
#include "yccfglist.h"
#include "config.h"


void YAudioAudioPresetModeDestroyCB(gpointer data);
gint YAudioDeleteEventCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
);
void YAudioDestroyCB(GtkObject *object, gpointer data);

gint YAudioDSPViewButtonCB(
	GtkWidget *widget, GdkEventButton *button, gpointer data
);

gint YAudioModesListKeyEventCB(
	GtkWidget *widget, GdkEventKey *key, gpointer data
);
gint YAudioModesListButtonEventCB(
	GtkWidget *widget, GdkEventButton *button, gpointer data
);
void YAudioModesListSelectRowCB(
	GtkCList *clist, gint row, gint column, GdkEvent *event,
	gpointer data
);
void YAudioModesListUnselectRowCB(
	GtkCList *clist, gint row, gint column, GdkEvent *event,
	gpointer data
);
void YAudioAudioModesListSelectModeCB(GtkWidget *widget, gpointer data);
void YAudioAudioModesListCurrentModeCB(GtkWidget *widget, gpointer data);

void YAudioAudioValuesCalculateCycleCB(
	GtkWidget *widget, gpointer data
);
void YAudioAudioValuesSetCB(GtkWidget *widget, gpointer data);
void YAudioAudioValuesCurrentCB(GtkWidget *widget, gpointer data);

void YAudioConnectCB(
	GtkWidget *widget, const gchar *s, gpointer data
);
void YAudioYServerListCB(GtkWidget *widget, gpointer data);
void YAudioCloseCB(GtkWidget *widget, gpointer data);
void YAudioNewYConsoleCB(GtkWidget *widget, gpointer data);
void YAudioNewYMixerCB(GtkWidget *widget, gpointer data);
void YAudioNewYAudioCB(GtkWidget *widget, gpointer data);
void YAudioNewYPlayerCB(GtkWidget *widget, gpointer data);
void YAudioUptionsCB(GtkWidget *widget, gpointer data);
void YAudioRefreshCB(GtkWidget *widget, gpointer data);
gboolean YAudioShelloutCB(
	GtkWidget *widget, gboolean shellout, gpointer data
);
void YAudioShutdownCB(GtkWidget *widget, gpointer data);
void YAudioManageCB(yaudio_struct *ya);
void YAudioManage2CB(yaudio_struct *ya);

void YAudioReconfiguredCB(
	yaudio_struct *ya, gpointer core_ptr
);


#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define ABSOLUTE(x)	(((x) < 0) ? ((x) * -1) : (x))


/*
 *	Audio modes GtkCList item destroy signal callback.
 */
void YAudioAudioPresetModeDestroyCB(gpointer data)
{
	YAudioModeValuesStruct *m = (YAudioModeValuesStruct *)data;
	if(m == NULL)
	    return;
/* printf("Destroyed 0x%.8x\n", (guint)m); */
	g_free(m);
}

/*
 *	Toplevel "delete_event" signal callback.
 */
gint YAudioDeleteEventCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
)
{
	YAudioCloseCB(widget, data);
	return(TRUE);
}

/*
 *	Toplevel "destroy" signal callback.
 */
void YAudioDestroyCB(GtkObject *object, gpointer data)
{

}


/*
 *	DSPView "button_press_event" signal callback.
 */
gint YAudioDSPViewButtonCB(
	GtkWidget *widget, GdkEventButton *button, gpointer data
)
{
	gint status = FALSE;
	yaudio_struct *ya = YAUDIO(data);
	if(ya == NULL)
	    return(status);

	switch(button->button)
	{
	  case 3:
	    switch((gint)button->type)
	    {
	      case GDK_BUTTON_PRESS:
		YCNewOptWinSelectPage(
		    YC_CORE(ya->core_ptr), YAUDIO_TOPLEVEL(ya),
		    1
		);
		status = TRUE;
		break;

	      case GDK_BUTTON_RELEASE:
		status = TRUE;
		break;
	    }
	}

	return(status);
}

/*
 *      Audio preset modes list "button_press_event" signal callback.
 */
gint YAudioModesListKeyEventCB(
	GtkWidget *widget, GdkEventKey *key, gpointer data
)
{
	gint status = FALSE;
	gboolean key_press;
	GtkCList *clist;
	yaudio_struct *ya = YAUDIO(data);
	if((key == NULL) || (ya == NULL))
	    return(status);

	clist = (GtkCList *)ya->modes_clist;
	if(clist == NULL)
	    return(status);

	key_press = (key->type == GDK_KEY_PRESS) ? TRUE : FALSE;

	switch(key->keyval)
	{
	  case GDK_Return:
	  case GDK_KP_Enter:
	  case GDK_ISO_Enter:
	  case GDK_3270_Enter:
	    if(key_press)
	    {
		YAudioAudioModesListSelectModeCB(widget, data);
		status = TRUE;
	    }
	    break;

	  case GDK_Escape:
	    if(key_press)
	    {
		status = TRUE;
	    }
	    break;

	  case GDK_Insert:
	    if(key_press)
	    {
		status = TRUE;
	    }
	    break;

	  case GDK_Delete:
	    if(key_press)
	    {
		status = TRUE;
	    }
	    break;

	}

	return(status);
}

/*
 *	Audio preset modes list "button_press_event" signal callback.
 */
gint YAudioModesListButtonEventCB(
	GtkWidget *widget, GdkEventButton *button, gpointer data
)
{
	gint status = FALSE;
	gint row, column;
	GtkWidget *w;
	GtkCList *clist;
	yaudio_struct *ya = YAUDIO(data);
	if((button == NULL) || (ya == NULL))
	    return(status);

	clist = (GtkCList *)ya->modes_clist;
	if(clist == NULL)
	    return(status);

	if(button->window != clist->clist_window)
	    return(status);

	switch(button->button)
	{
	  case 3:
	    switch((gint)button->type)
	    {
	      case GDK_BUTTON_PRESS:
		gtk_clist_get_selection_info(
		    clist, button->x, button->y, &row, &column
		);
		gtk_clist_freeze(clist);
		gtk_clist_unselect_all(clist);
		gtk_clist_select_row(clist, row, column);
		gtk_clist_thaw(clist);

		w = ya->modes_menu;
		if(w != NULL)
		    gtk_menu_popup(
			GTK_MENU(w), NULL, NULL,
			NULL, NULL,
			button->button, button->time
		    );

		status = TRUE;
		break;
	    }
	    break;

	  case 1:
	    switch((gint)button->type)
	    {
	      case GDK_2BUTTON_PRESS:
		YAudioAudioModesListSelectModeCB(widget, data);
		status = TRUE;
		break;
	    }
	    break;
	}

	return(status);
}

/*
 *	Audio preset modes list "select_row" signal callback.
 */
void YAudioModesListSelectRowCB(
	GtkCList *clist, gint row, gint column, GdkEvent *event,
	gpointer data
)
{
	yaudio_struct *ya = YAUDIO(data);
	if((clist == NULL) || (ya == NULL))
	    return;

	if(gtk_clist_row_is_visible(clist, row) !=
	    GTK_VISIBILITY_FULL
	)
	    gtk_clist_moveto(
		clist,
		row, -1,	/* Row, column. */
		0.5, 0.0	/* Row, column. */
	    );

	YAudioUpdateMenus(ya);
}

/*
 *      Audio preset modes list "unselect_row" signal callback.
 */
void YAudioModesListUnselectRowCB(
	GtkCList *clist, gint row, gint column, GdkEvent *event,
	gpointer data
)
{
	yaudio_struct *ya = YAUDIO(data);
	if(ya == NULL)
	    return;

	YAudioUpdateMenus(ya);
}

/*
 *	Audio preset modes list select mode callback.
 */
void YAudioAudioModesListSelectModeCB(GtkWidget *widget, gpointer data)
{
	gint row;
	const gchar *name;
	GList *glist;
	GtkCList *clist;
	YConnection *con;
	YAudioModeValuesStruct *m;
	yaudio_struct *ya = YAUDIO(data);
	if(ya == NULL)
	    return;

	clist = (GtkCList *)ya->modes_clist;
	if(clist == NULL)
	    return;

	/* Get last selected row. */
	glist = clist->selection_end;
	row = (glist != NULL) ? (gint)glist->data : -1;
	if((row < 0) || (row >= clist->rows))
	    return;

	/* Get row data. */
	m = (YAudioModeValuesStruct *)gtk_clist_get_row_data(
	    clist, row
	);
	if(m == NULL)
	    return;

	con = YAUDIO_CONNECTION(ya);
	if(con == NULL)
	    return;

	/* Change to new Audio preset mode. */
	name = m->name;
	if((name != NULL) ? (*name != '\0') : FALSE)
	{
	    /* Change Audio preset mode. */
	    if(!YChangeAudioModePreset(con, name))
	    {
		/* Audio preset mode change successful. */
		YEventAudioStats stat_buf;
		dspview_struct *dv = CONBAR_DSPVIEW(YAUDIO_CONBAR(ya));

		/* Reopen YSHM sound buffer. */
/*
		YSHMCloseSound(con, ya->yshm_buf);
		ya->yshm_buf = (guint8 *)YSHMOpenSound(
		    con, &ya->yshm_buf_len, &ya->yshm_id
		);
 */
		/* Reget Audio values to obtain accurate values after
		 * change.
		 */
		if(!YGetAudioStats(con, &stat_buf))
		{
		    g_free(ya->audio_mode_name);
		    ya->audio_mode_name = g_strdup(stat_buf.mode_name);
		    ya->sample_rate = stat_buf.sample_rate;
		    ya->sample_size = stat_buf.sample_size;
		    ya->channels = stat_buf.channels;
		    ya->cycle_us = stat_buf.cycle_us;
		    ya->num_fragments = stat_buf.num_fragments;
		    ya->fragment_size = stat_buf.fragment_size;
		    ya->shellout = FALSE;

		    /* Update audio values on connection bar. */
		    YAUDIO_SET_CONBAR_AUDIO(ya);
		    ConBarSetShellOut(YAUDIO_CONBAR(ya), ya->shellout);
		    DSPViewSetChannels(dv, ya->channels);
		}
	    }
	}
}

/*
 *	Audio preset modes list current mode callback.
 */
void YAudioAudioModesListCurrentModeCB(GtkWidget *widget, gpointer data)
{
	yaudio_struct *ya = YAUDIO(data);
	if(ya == NULL)
	    return;
	YAudioSelectAudioModeByName(
	    ya, ya->audio_mode_name
	);
}


/*
 *	Audio values calculate cycle callback.
 */
void YAudioAudioValuesCalculateCycleCB(
	GtkWidget *widget, gpointer data
)
{
	gint	sample_rate = 0, sample_size = 0, channels = 0,
		fragment_size = 0;
	GtkAdjustment *adj;
	GtkCombo *combo;
	yaudio_struct *ya = YAUDIO(data);
	if(ya == NULL)
	    return;

	adj = ya->cycle_adj;
	if(adj == NULL)
	    return;

	/* Begin getting values for calculation. */
	/* Sample rate. */
	combo = (GtkCombo *)ya->sample_rate_combo;
	if(combo != NULL)
	    sample_rate = atoi(gtk_entry_get_text(
		GTK_ENTRY(combo->entry)
	    ));

	/* Sample size. */
	combo = (GtkCombo *)ya->sample_size_combo;
	if(combo != NULL)
	    sample_size = atoi(gtk_entry_get_text(
		GTK_ENTRY(combo->entry)
	    ));

	/* Channels. */
	combo = (GtkCombo *)ya->channels_combo;
	if(combo != NULL)
	{
	    const gchar *s = gtk_entry_get_text(
		GTK_ENTRY(combo->entry)
	    );
	    if(!strcasecmp(s, "Quad"))
		channels = 4;
	    else if(!strcasecmp(s, "Stereo"))
		channels = 2;
	    else
		channels = 1;
	}

	/* Buffer fragment size (in bytes). */
	combo = (GtkCombo *)ya->fragment_size_combo;
	if(combo != NULL)
	    fragment_size = atoi(gtk_entry_get_text(
		GTK_ENTRY(combo->entry)
	    ));


	/* Calculate cycle. */
	adj->value = (gfloat)YCalculateCycle(
	    YAUDIO_CONNECTION(ya),
	    sample_rate, channels, sample_size, fragment_size
	);
	gtk_signal_emit_by_name(GTK_OBJECT(adj), "value_changed");
}

/*
 *	Audio values set callback.
 */
void YAudioAudioValuesSetCB(GtkWidget *widget, gpointer data)
{
	gint    sample_rate = 0, sample_size = 0, channels = 0,
		fragment_size = 0;
	glong cycle_us = 30000;
	GtkAdjustment *adj;
	GtkCombo *combo;
	YConnection *con;
	yaudio_struct *ya = YAUDIO(data);
	if(ya == NULL)
	    return;

	/* Begin getting values for calculation. */
	/* Sample rate. */
	combo = (GtkCombo *)ya->sample_rate_combo;
	if(combo != NULL)
	    sample_rate = atoi(gtk_entry_get_text(
		GTK_ENTRY(combo->entry)
	    ));

	/* Sample size. */
	combo = (GtkCombo *)ya->sample_size_combo;
	if(combo != NULL)
	    sample_size = atoi(gtk_entry_get_text(
		GTK_ENTRY(combo->entry)
	    ));

	/* Channels. */
	combo = (GtkCombo *)ya->channels_combo;
	if(combo != NULL)
	{
	    const gchar *s = gtk_entry_get_text(
		GTK_ENTRY(combo->entry)
	    );
	    if(!strcasecmp(s, "Quad"))
		channels = 4;
	    else if(!strcasecmp(s, "Stereo"))
		channels = 2;
	    else
		channels = 1;
	}

	/* Buffer fragment size (in bytes). */
	combo = (GtkCombo *)ya->fragment_size_combo;
	if(combo != NULL)
	    fragment_size = atoi(gtk_entry_get_text(
		GTK_ENTRY(combo->entry)
	    ));

	/* Cycle. */
	adj = ya->cycle_adj;
	if(adj != NULL)
	    cycle_us = (glong)adj->value;


	con = YAUDIO_CONNECTION(ya);
	if(con == NULL)
	    return;

	/* Set Audio values. */
	if(!YSetAudioModeValues(
	    con,
	    sample_size, channels, sample_rate,
	    0,		/* Direction (0 = play). */
	    True,	/* Allow fragmenting. */
	    2,		/* Number of buffer fragments. */
	    fragment_size	/* Buffer fragment size in bytes. */
	))
	{
	    /* Audio values set successfully. */
	    YEventAudioStats stat_buf;
	    dspview_struct *dv = CONBAR_DSPVIEW(YAUDIO_CONBAR(ya));

	    /* Need to set cycle. */
	    YSetCycle(con, cycle_us);

	    /* Reopen YSHM sound buffer. */
/*
	    YSHMCloseSound(con, ya->yshm_buf);
	    ya->yshm_buf = (guint8 *)YSHMOpenSound(
		con, &ya->yshm_buf_len, &ya->yshm_id
	    );
 */
	    /* Reget Audio values to obtain accurate values after
	     * change.
	     */
	    if(!YGetAudioStats(con, &stat_buf))
	    {
		g_free(ya->audio_mode_name);
		ya->audio_mode_name = g_strdup(stat_buf.mode_name);
		ya->sample_rate = stat_buf.sample_rate;
		ya->sample_size = stat_buf.sample_size;
		ya->channels = stat_buf.channels;
		ya->cycle_us = stat_buf.cycle_us;
		ya->num_fragments = stat_buf.num_fragments;
		ya->fragment_size = stat_buf.fragment_size;
		ya->shellout = FALSE;

		/* Update audio values on connection bar. */
		YAUDIO_SET_CONBAR_AUDIO(ya);
		ConBarSetShellOut(YAUDIO_CONBAR(ya), ya->shellout);
		DSPViewSetChannels(dv, ya->channels);
	    }
	}
}

/*
 *	Audio values get current values callback.
 */
void YAudioAudioValuesCurrentCB(GtkWidget *widget, gpointer data)
{
	yaudio_struct *ya = YAUDIO(data);
	if(ya == NULL)
	    return;

	YAudioUpdateAudioValues(ya);
}


/*
 *	Connect callback.
 */
void YAudioConnectCB(
	GtkWidget *widget, const gchar *s, gpointer data
)
{
	gchar *strptr;
	gchar *s2 = (s != NULL) ? g_strdup(s) : NULL;
	yaudio_struct *ya = YAUDIO(data);
	if(ya == NULL)
	    return;

	YAudioSetBusy(ya, TRUE);

	/* Empty string? Then use default location. */
	if(*s2 == '\0')
	{
	    g_free(s2);
	    s2 = g_strdup("127.0.0.1:9433");
	}

	/* No port number? Then append it. */
	strptr = strchr(s2, ':');
	if(strptr == NULL)
	{
	    strptr = g_strdup_printf("%s:9433", s2);
	    g_free(s2);
	    s2 = strptr;
	}

	/* Do connect. */
	YAudioDoConnect(ya, s2, NULL);

	/* Deallocate coppied location. */
	g_free(s2);

	YAudioSetBusy(ya, FALSE);

}

/*
 *	Y server list callback.
 */
void YAudioYServerListCB(GtkWidget *widget, gpointer data)
{
	yaudio_struct *ya = YAUDIO(data);
	if(ya == NULL)
	    return;

	YCNewYServerList(YC_CORE(ya->core_ptr), YAUDIO_TOPLEVEL(ya));
}

/*
 *	Close callback.
 */
void YAudioCloseCB(GtkWidget *widget, gpointer data)
{
	yaudio_struct *ya = YAUDIO(data);
	YAudioSavePosition(ya);
	YAudioUnmap(ya);
}

/*
 *	New YConsole window callback.
 */
void YAudioNewYConsoleCB(GtkWidget *widget, gpointer data)
{
	yaudio_struct *ya = YAUDIO(data);
	if(ya != NULL)
	    YCNewYConsole(
		YC_CORE(ya->core_ptr),
		YAudioGetLocation(ya)
	    );
}

/*
 *	New YMixer window callback.
 */
void YAudioNewYMixerCB(GtkWidget *widget, gpointer data)
{
	yaudio_struct *ya = YAUDIO(data);
	if(ya != NULL)
	    YCNewYMixer(
		YC_CORE(ya->core_ptr),
		YAudioGetLocation(ya)
	    );
}

/*
 *	New YAudio window callback.
 */
void YAudioNewYAudioCB(GtkWidget *widget, gpointer data)
{
	yaudio_struct *ya = YAUDIO(data);
	if(ya != NULL)
	    YCNewYAudio(
		YC_CORE(ya->core_ptr),
		YAudioGetLocation(ya)
	    );
}

/*
 *	New YPlayer window callback.
 */
void YAudioNewYPlayerCB(GtkWidget *widget, gpointer data)
{
	yaudio_struct *ya = YAUDIO(data);
	if(ya != NULL)
	    YCNewYPlayer(
		YC_CORE(ya->core_ptr),
		YAudioGetLocation(ya)
	    );
}

/*
 *	Map options window callback.
 */
void YAudioUptionsCB(GtkWidget *widget, gpointer data)
{
	yaudio_struct *ya = YAUDIO(data);
	if(ya != NULL)
	    YCNewOptWin(
		YC_CORE(ya->core_ptr),
		YAUDIO_TOPLEVEL(ya)
	    );
}

/*
 *	Refresh callback.
 */
void YAudioRefreshCB(GtkWidget *widget, gpointer data)
{
	gchar *s;
	yaudio_struct *ya = YAUDIO(data);
	if(ya == NULL)
	    return;

	YAudioDoDisconnect(ya, FALSE);
	s = ConBarGetLocation(YAUDIO_CONBAR(ya));
	s = (s != NULL) ? g_strdup(s) : NULL;
	YAudioDoConnect(ya, s, NULL);
	g_free(s);

}

/*
 *	Shellout callback.
 *
 *	Returns actual shellout state.
 */
gboolean YAudioShelloutCB(
	GtkWidget *widget, gboolean shellout, gpointer data
)
{
	YConnection *con;
	yaudio_struct *ya = YAUDIO(data);
	if(ya == NULL)
	    return(ya->shellout);

	con = YAUDIO_CONNECTION(ya);
	if(con == NULL)
	    return(ya->shellout);

	if(shellout == ya->shellout)
	    return(ya->shellout);

	if(shellout)
	{
	    if(YSetAudioModeValues(
		con,
		8,      /* Sample size in bits. */
		1,      /* Number of channels. */
		0,      /* Sample rate set to 0 implies shellout. */
		0,      /* Direction, 0 = play, 1 = record. */
		True,
		2,
		1024
	    ))
		ya->shellout = FALSE;
	    else
		ya->shellout = TRUE;
	}
	else
	{
	    if(YChangeAudioModePreset(
		con,
		(ya->audio_mode_name != NULL) ?
		    ya->audio_mode_name : "Default"
	    ))
		ya->shellout = TRUE;
	    else
		ya->shellout = FALSE;
	}

	YAUDIO_SET_CONBAR_AUDIO(ya);

	return(ya->shellout);
}

/*
 *	Shutdown Y server callback.
 */
void YAudioShutdownCB(GtkWidget *widget, gpointer data)
{
	yaudio_struct *ya = YAUDIO(data);
	if(ya == NULL)
	    return;

	YAudioDoDisconnect(ya, TRUE);
}

/*
 *	YAudio manage callback.
 */
void YAudioManageCB(yaudio_struct *ya)
{
	YConnection *con;
	YEvent yevent;
	dspview_struct *dv;
	connection_bar_struct *cb;


	if(ya == NULL)
	    return;

	cb = YAUDIO_CONBAR(ya);
	dv = ConBarGetDSPView(cb);
	if(dv == NULL)
	    return;

	/* Manage Y events. */
	con = YAUDIO_CONNECTION(ya);
	while((con != NULL) && (YGetNextEvent(con, &yevent, False) > 0))
	{
	    switch(yevent.type)
	    {
	      case YAudioChange:
		/* Update Audio values on conbar and DSPView. */
		if(yevent.audio.preset)
		{
		    YEventAudioStats stat_buf;

		    g_free(ya->audio_mode_name);
		    ya->audio_mode_name = g_strdup(yevent.audio.mode_name);

		    if(!YGetAudioStats(con, &stat_buf))
		    {
			ya->sample_rate = stat_buf.sample_rate;
			ya->sample_size = stat_buf.sample_size;
			ya->channels = stat_buf.channels;
			ya->cycle_us = stat_buf.cycle_us;
			ya->num_fragments = stat_buf.num_fragments;
			ya->fragment_size = stat_buf.fragment_size;
			ya->flip_stereo = stat_buf.flip_stereo;
			ya->shellout = FALSE;
		    }
		}
		else
		{
		    YEventAudioChange *audio = &yevent.audio;
		    ya->sample_rate = audio->sample_rate;
		    ya->sample_size = audio->sample_size;
		    ya->channels = audio->channels;
		    ya->num_fragments = audio->num_fragments;
		    ya->fragment_size = audio->fragment_size_bytes;
		    ya->flip_stereo = FALSE;
		    ya->shellout = FALSE;
		}
		YAUDIO_SET_CONBAR_AUDIO(ya);
		ConBarSetShellOut(cb, ya->shellout);
		DSPViewSetChannels(dv, yevent.audio.channels);
		YAudioSelectAudioModeByName(ya, ya->audio_mode_name);
		break;

	      case YCycleChange:
		/* Update audio values. */
		ya->cycle_us = yevent.cycle.cycle_us;
		YAUDIO_SET_CONBAR_AUDIO(ya);
		break;

	      case YDisconnect: case YShutdown:
		YAudioDoDisconnect(ya, FALSE);
		con = NULL;	/* So we break out of this loop. */
		break;

	      case YSHMSound:
		if(yevent.yshmsound.minor_op_code == YSHMSoundClose)
		{
		    /* Need to reopen YSHM sound buffer when Audio
		     * mode changes.
		     */
		    YSHMCloseSound(con, ya->yshm_buf);
		    ya->yshm_buf = (guint8 *)YSHMOpenSound(
			con, &ya->yshm_buf_len, &ya->yshm_id
		    );
		}
		break;
	    }
	}

	/* Redraw DSP view. */
	ConBarDraw(cb, ya->yshm_buf, ya->yshm_buf_len);
}

/*
 *      YAudio manage 2 callback.
 *
 *	This function is called in longer successions.
 */
void YAudioManage2CB(yaudio_struct *ya)
{
	YConnection *con;


	if(ya == NULL)
	    return;

	con = ya->con;
	if(con != NULL)
	{
	    YEventServerStats stat_buf;

/* Don't need this since average volume not displayed.
	    YCAnalyzeAverageChannelVolume(
		ya->yshm_buf, ya->yshm_buf_len,
		ya->channel_volume_avg,
		CLIP(ya->channels, 1, 4)
	    );
*/

	    if(!YGetServerStats(con, &stat_buf))
	    {
		ya->start_time = stat_buf.start_time;
		ya->cycle_load = stat_buf.cycle_load;
		ya->version_major = stat_buf.protocol_version_major;
		ya->version_minor = stat_buf.protocol_version_minor;
		g_free(ya->server_name);
		ya->server_name = g_strdup(stat_buf.server_name);
		g_free(ya->sound_name);
		ya->sound_name = g_strdup(stat_buf.sound_name);
		YAUDIO_SET_CONBAR_AUDIO(ya);
	    }
	}
}


/*
 *	YAudio reconfigured signal callback.
 */
void YAudioReconfiguredCB(
	yaudio_struct *ya, gpointer core_ptr
)
{
	GtkRcStyle *rcstyle;
	GtkWidget *w;
	dspview_struct *dv;
	yc_core_struct *cp = YC_CORE(core_ptr);
	if((ya == NULL) || (cp == NULL))
	    return;

	/* Update connection bar's audio display. */
	if(TRUE)
	{
	    GdkColor *c, cs[3];
	    const cfg_color_struct *color;

#define DO_CONVERT_COLOR	\
{ if((c != NULL) && (color != NULL)) { \
 c->red		= (gushort)-1 * color->r; \
 c->green	= (gushort)-1 * color->g; \
 c->blue	= (gushort)-1 * color->b; \
} }
	    color = YC_GET_CFG_COLOR(
		cp, YC_CFG_PARM_AUDIO_DISPLAY_COLOR_BG
	    );
	    c = &cs[0];
	    DO_CONVERT_COLOR

	    color = YC_GET_CFG_COLOR(
		cp, YC_CFG_PARM_AUDIO_DISPLAY_COLOR_FG
	    );
	    c = &cs[1];
	    DO_CONVERT_COLOR

	    color = YC_GET_CFG_COLOR(
		cp, YC_CFG_PARM_AUDIO_DISPLAY_COLOR_FG_DIM
	    );
	    c = &cs[2];
	    DO_CONVERT_COLOR

	    ConBarSetAudioColors(
		YAUDIO_CONBAR(ya), &cs[0], &cs[1], &cs[2]
	    );
#undef DO_CONVERT_COLOR

	    ConBarSetAudioFonts(
		YAUDIO_CONBAR(ya),
		YC_GET_CFG_S(cp, YC_CFG_PARM_AUDIO_DISPLAY_FONT_LARGE),
		YC_GET_CFG_S(cp, YC_CFG_PARM_AUDIO_DISPLAY_FONT_MEDIUM),
		YC_GET_CFG_S(cp, YC_CFG_PARM_AUDIO_DISPLAY_FONT_SMALL)
	    );
	}

	/* Update connection bar's DSPView. */
	dv = ConBarGetDSPView(YAUDIO_CONBAR(ya));
	if(dv != NULL)
	{
	    const cfg_color_struct *color;
	    gfloat alpha_coeff;

	    DSPViewSetDisplayStyle(
		dv,
		YC_GET_CFG_I(cp, YC_CFG_PARM_DSPVIEW_DISPLAY_STYLE)
	    );
	    DSPViewSetMergeMode(
		dv,
		YC_GET_CFG_I(cp, YC_CFG_PARM_DSPVIEW_MERGE_MODE)
	    );
	    color = YC_GET_CFG_COLOR(cp, YC_CFG_PARM_DSPVIEW_COLOR);
	    alpha_coeff = YC_GET_CFG_F(cp, YC_CFG_PARM_DSPVIEW_TRANSPARENCY);
	    if(color != NULL)
		DSPViewSetDSPColor(
		    dv,
		    (guint32)(
			((guint32)(color->r * 0xff) << 0) |
			((guint32)(color->g * 0xff) << 8) |
			((guint32)(color->b * 0xff) << 16) |
			((guint32)(alpha_coeff * 0xff) << 24)
		    )
		);
	    DSPViewSetFadeCoeff(
		dv,
		YC_GET_CFG_F(cp, YC_CFG_PARM_DSPVIEW_FADE_COEFF)
	    );
	}	/* Update connection bar's DSPView. */

	/* Update audio widget colors. */
	rcstyle = YAudioAudioDisplayRCStyleNew(ya);
	if(rcstyle != NULL)
	{
	    w = YAUDIO_MODES_CLIST(ya);
	    if(w != NULL)
		gtk_widget_modify_style(w, rcstyle);

	    w = ya->sample_rate_combo;
	    if(w != NULL)
		gtk_widget_modify_style(GTK_COMBO(w)->entry, rcstyle);
	    w = ya->sample_size_combo;
	    if(w != NULL)
		gtk_widget_modify_style(GTK_COMBO(w)->entry, rcstyle);
	    w = ya->channels_combo;
	    if(w != NULL)
		gtk_widget_modify_style(GTK_COMBO(w)->entry, rcstyle);
	    w = ya->fragment_size_combo;
	    if(w != NULL)
		gtk_widget_modify_style(GTK_COMBO(w)->entry, rcstyle);
	    w = ya->cycle_spin;
	    if(w != NULL)
		gtk_widget_modify_style(w, rcstyle);


	    if(rcstyle != NULL)
		gtk_rc_style_unref(rcstyle);
	}

}
