#include <stdlib.h>
#include <string.h>

#include <Y2/Y.h>
#include <Y2/Ylib.h>

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include "guiutils.h"

#include "cfg.h"
#include "conbar.h"
#include "yconsolewin.h"
#include "yconsolecb.h"
#include "yc.h"
#include "ycop.h"
#include "ycanalyze.h"
#include "yccfglist.h"
#include "config.h"


gint YConsoleDeleteEventCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
);
void YConsoleDestroyCB(GtkObject *object, gpointer data);

gint YConsoleDSPViewButtonCB(
	GtkWidget *widget, GdkEventButton *button, gpointer data
);
gint YConsoleConBarAudioButtonCB(
	GtkWidget *widget, GdkEventButton *button, gpointer data
);

void YConsoleConnectCB(
	GtkWidget *widget, const gchar *s, gpointer data
);
void YConsoleYServerListCB(GtkWidget *widget, gpointer data);
void YConsoleCloseCB(GtkWidget *widget, gpointer data);
void YConsoleNewYConsoleCB(GtkWidget *widget, gpointer data);
void YConsoleNewYMixerCB(GtkWidget *widget, gpointer data);
void YConsoleNewYAudioCB(GtkWidget *widget, gpointer data);
void YConsoleNewYPlayerCB(GtkWidget *widget, gpointer data);
void YConsoleUptionsCB(GtkWidget *widget, gpointer data);
void YConsoleRefreshCB(GtkWidget *widget, gpointer data);
gboolean YConsoleShelloutCB(
	GtkWidget *widget, gboolean shellout, gpointer data
);
void YConsoleShutdownCB(GtkWidget *widget, gpointer data);
void YConsoleManageCB(yconsole_struct *yc);
void YConsoleManage2CB(yconsole_struct *yc);

void YConsoleReconfiguredCB(
	yconsole_struct *yc, gpointer core_ptr
);


#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define ABSOLUTE(x)	(((x) < 0) ? ((x) * -1) : (x))


/*
 *	Toplevel "delete_event" signal callback.
 */
gint YConsoleDeleteEventCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
)
{
	YConsoleCloseCB(widget, data);
	return(TRUE);
}

/*
 *	Toplevel "destroy" signal callback.
 */
void YConsoleDestroyCB(GtkObject *object, gpointer data)
{

}


/*
 *	DSPView "button_press_event" signal callback.
 */
gint YConsoleDSPViewButtonCB(
	GtkWidget *widget, GdkEventButton *button, gpointer data
)
{
	gint status = FALSE;
	yconsole_struct *yc = YCONSOLE(data);
	if(yc == NULL)
	    return(status);

	switch(button->button)
	{
	  case 3:
	    switch((gint)button->type)
	    {
	      case GDK_BUTTON_PRESS:
		YCNewOptWinSelectPage(
		    YC_CORE(yc->core_ptr), YCONSOLE_TOPLEVEL(yc),
		    1
		);
		status = TRUE;
		break;

	      case GDK_BUTTON_RELEASE:
		status = TRUE;
		break;
	    }
	}

	return(status);
}

/*
 *	Connection bar "button_press_event" signal callback.
 */
gint YConsoleConBarAudioButtonCB(
	GtkWidget *widget, GdkEventButton *button, gpointer data
)
{
	gint status = FALSE;
	yconsole_struct *yc = YCONSOLE(data);
	if(yc == NULL)
	    return(status);

	switch(button->button)
	{
	  case 3:
	    switch((gint)button->type)
	    {
	      case GDK_BUTTON_PRESS:
		YCNewOptWinSelectPage(
		    YC_CORE(yc->core_ptr), YCONSOLE_TOPLEVEL(yc),
		    1
		);
		status = TRUE;
		break;

	      case GDK_BUTTON_RELEASE:
		status = TRUE;
		break;
	    }
	}

	return(status);
}


/*
 *	Connect callback.
 */
void YConsoleConnectCB(
	GtkWidget *widget, const gchar *s, gpointer data
)
{
	gchar *strptr;
	gchar *s2 = (s != NULL) ? g_strdup(s) : NULL;
	yconsole_struct *yc = YCONSOLE(data);
	if(yc == NULL)
	    return;

	YConsoleSetBusy(yc, TRUE);

	/* Empty string? Then use default location. */
	if(*s2 == '\0')
	{
	    g_free(s2);
	    s2 = g_strdup("127.0.0.1:9433");
	}

	/* No port number? Then append it. */
	strptr = strchr(s2, ':');
	if(strptr == NULL)
	{
	    strptr = g_strdup_printf("%s:9433", s2);
	    g_free(s2);
	    s2 = strptr;
	}

	/* Do connect. */
	YConsoleDoConnect(yc, s2, NULL);

	/* Deallocate coppied location. */
	g_free(s2);

	YConsoleSetBusy(yc, FALSE);

}

/*
 *	Y server list callback.
 */
void YConsoleYServerListCB(GtkWidget *widget, gpointer data)
{
	yconsole_struct *yc = YCONSOLE(data);
	if(yc == NULL)
	    return;

	YCNewYServerList(YC_CORE(yc->core_ptr), YCONSOLE_TOPLEVEL(yc));
}

/*
 *	Close callback.
 */
void YConsoleCloseCB(GtkWidget *widget, gpointer data)
{
	yconsole_struct *yc = YCONSOLE(data);
	YConsoleSavePosition(yc);
	YConsoleUnmap(yc);
}

/*
 *	New YConsole window callback.
 */
void YConsoleNewYConsoleCB(GtkWidget *widget, gpointer data)
{
	yconsole_struct *yc = YCONSOLE(data);
	if(yc != NULL)
	    YCNewYConsole(
		YC_CORE(yc->core_ptr),
		YConsoleGetLocation(yc)
	    );
}

/*
 *	New YMixer window callback.
 */
void YConsoleNewYMixerCB(GtkWidget *widget, gpointer data)
{
	yconsole_struct *yc = YCONSOLE(data);
	if(yc != NULL)
	    YCNewYMixer(
		YC_CORE(yc->core_ptr),
		YConsoleGetLocation(yc)
	    );
}

/*
 *	New YAudio window callback.
 */
void YConsoleNewYAudioCB(GtkWidget *widget, gpointer data)
{
	yconsole_struct *yc = YCONSOLE(data);
	if(yc != NULL)
	    YCNewYAudio(
		YC_CORE(yc->core_ptr),
		YConsoleGetLocation(yc)
	    );
}

/*
 *	New YPlayer window callback.
 */
void YConsoleNewYPlayerCB(GtkWidget *widget, gpointer data)
{
	yconsole_struct *yc = YCONSOLE(data);
	if(yc != NULL)
	    YCNewYPlayer(
		YC_CORE(yc->core_ptr),
		YConsoleGetLocation(yc)
	    );
}

/*
 *	Map options window callback.
 */
void YConsoleUptionsCB(GtkWidget *widget, gpointer data)
{
	yconsole_struct *yc = YCONSOLE(data);
	if(yc != NULL)
	    YCNewOptWin(
		YC_CORE(yc->core_ptr),
		YCONSOLE_TOPLEVEL(yc)
	    );
}

/*
 *	Refresh callback.
 */
void YConsoleRefreshCB(GtkWidget *widget, gpointer data)
{
	gchar *s;
	yconsole_struct *yc = YCONSOLE(data);
	if(yc == NULL)
	    return;

	YConsoleDoDisconnect(yc, FALSE);
	s = ConBarGetLocation(YCONSOLE_CONBAR(yc));
	s = (s != NULL) ? g_strdup(s) : NULL;
	YConsoleDoConnect(yc, s, NULL);
	g_free(s);

}

/*
 *	Shellout callback.
 *
 *	Returns actual shellout state.
 */
gboolean YConsoleShelloutCB(
	GtkWidget *widget, gboolean shellout, gpointer data
)
{
	YConnection *con;
	yconsole_struct *yc = YCONSOLE(data);
	if(yc == NULL)
	    return(yc->shellout);

	con = YCONSOLE_CONNECTION(yc);
	if(con == NULL)
	    return(yc->shellout);

	if(shellout == yc->shellout)
	    return(yc->shellout);

	if(shellout)
	{
	    if(YSetAudioModeValues(
		con,
		8,      /* Sample size in bits. */
		1,      /* Number of channels. */
		0,      /* Sample rate set to 0 implies shellout. */
		0,      /* Direction, 0 = play, 1 = record. */
		True,
		2,
		1024
	    ))
		yc->shellout = FALSE;
	    else
		yc->shellout = TRUE;
	}
	else
	{
	    if(YChangeAudioModePreset(
		con,
		(yc->audio_mode_name != NULL) ?
		    yc->audio_mode_name : "Default"
	    ))
		yc->shellout = TRUE;
	    else
		yc->shellout = FALSE;
	}

	YCONSOLE_SET_CONBAR_AUDIO(yc);

	return(yc->shellout);
}

/*
 *	Shutdown Y server callback.
 */
void YConsoleShutdownCB(GtkWidget *widget, gpointer data)
{
	yconsole_struct *yc = YCONSOLE(data);
	if(yc == NULL)
	    return;

	YConsoleDoDisconnect(yc, TRUE);
}

/*
 *	YConsole manage callback.
 */
void YConsoleManageCB(yconsole_struct *yc)
{
	YConnection *con;
	YEvent yevent;
	dspview_struct *dv;
	connection_bar_struct *cb;


	if(yc == NULL)
	    return;

	cb = YCONSOLE_CONBAR(yc);
	dv = ConBarGetDSPView(cb);
	if(dv == NULL)
	    return;

	/* Manage Y events. */
	con = YCONSOLE_CONNECTION(yc);
	while((con != NULL) && (YGetNextEvent(con, &yevent, False) > 0))
	{
	    switch(yevent.type)
	    {
	      case YAudioChange:
		/* Update Audio values on conbar and DSPView. */
		if(yevent.audio.preset)
		{
		    YEventAudioStats stat_buf;

		    g_free(yc->audio_mode_name);
		    yc->audio_mode_name = g_strdup(yevent.audio.mode_name);

		    if(!YGetAudioStats(con, &stat_buf))
		    {
			yc->sample_rate = stat_buf.sample_rate;
			yc->sample_size = stat_buf.sample_size;
			yc->channels = stat_buf.channels;
			yc->cycle_us = stat_buf.cycle_us;
			yc->num_fragments = stat_buf.num_fragments;
			yc->fragment_size = stat_buf.fragment_size;
			yc->flip_stereo = stat_buf.flip_stereo;
			yc->shellout = FALSE;
		    }
		}
		else
		{
		    YEventAudioChange *audio = &yevent.audio;
		    yc->sample_rate = audio->sample_rate;
		    yc->sample_size = audio->sample_size;
		    yc->channels = audio->channels;
		    yc->num_fragments = audio->num_fragments;
		    yc->fragment_size = audio->fragment_size_bytes;
		    yc->flip_stereo = FALSE;
		    yc->shellout = FALSE;
		}
		YCONSOLE_SET_CONBAR_AUDIO(yc);
		ConBarSetShellOut(cb, yc->shellout);
		DSPViewSetChannels(dv, yevent.audio.channels);
		break;

	      case YCycleChange:
		/* Update audio values. */
		yc->cycle_us = yevent.cycle.cycle_us;
		YCONSOLE_SET_CONBAR_AUDIO(yc);
		break;

	      case YDisconnect: case YShutdown:
		YConsoleDoDisconnect(yc, FALSE);
		con = NULL;	/* So we break out of this loop. */
		break;

	      case YSHMSound:
		if(yevent.yshmsound.minor_op_code == YSHMSoundClose)
		{
		    /* Need to reopen YSHM sound buffer when Audio
		     * mode changes.
		     */
		    YSHMCloseSound(con, yc->yshm_buf);
		    yc->yshm_buf = (guint8 *)YSHMOpenSound(
			con, &yc->yshm_buf_len, &yc->yshm_id
		    );
		}
		break;
	    }
	}

	/* Redraw DSP view. */
	ConBarDraw(cb, yc->yshm_buf, yc->yshm_buf_len);
}

/*
 *      YConsole manage 2 callback.
 *
 *	This function is called in longer successions.
 */
void YConsoleManage2CB(yconsole_struct *yc)
{
	YConnection *con;


	if(yc == NULL)
	    return;

	con = YCONSOLE_CONNECTION(yc);
	if(con != NULL)
	{
	    YEventServerStats stat_buf;

	    YCAnalyzeAverageChannelVolume(
		yc->yshm_buf, yc->yshm_buf_len,
		yc->channel_volume_avg,
		CLIP(yc->channels, 1, 4)
	    );

	    if(!YGetServerStats(con, &stat_buf))
	    {
		yc->start_time = stat_buf.start_time;
		yc->cycle_load = stat_buf.cycle_load;
		yc->version_major = stat_buf.protocol_version_major;
		yc->version_minor = stat_buf.protocol_version_minor;
		g_free(yc->server_name);
		yc->server_name = g_strdup(stat_buf.server_name);
		g_free(yc->sound_name);
		yc->sound_name = g_strdup(stat_buf.sound_name);
		YCONSOLE_SET_CONBAR_AUDIO(yc);
	    }
	}
}


/*
 *	YConsole reconfigured signal callback.
 */
void YConsoleReconfiguredCB(
	yconsole_struct *yc, gpointer core_ptr
)
{
	dspview_struct *dv;
	yc_core_struct *cp = YC_CORE(core_ptr);
	if((yc == NULL) || (cp == NULL))
	    return;

	/* Update connection bar's audio display. */
	if(TRUE)
	{
	    GdkColor *c, cs[3];
	    const cfg_color_struct *color;

#define DO_CONVERT_COLOR	\
{ if((c != NULL) && (color != NULL)) { \
 c->red		= (gushort)-1 * color->r; \
 c->green	= (gushort)-1 * color->g; \
 c->blue	= (gushort)-1 * color->b; \
} }
	    color = YC_GET_CFG_COLOR(
		cp, YC_CFG_PARM_AUDIO_DISPLAY_COLOR_BG
	    );
	    c = &cs[0];
	    DO_CONVERT_COLOR

	    color = YC_GET_CFG_COLOR(
		cp, YC_CFG_PARM_AUDIO_DISPLAY_COLOR_FG
	    );
	    c = &cs[1];
	    DO_CONVERT_COLOR

	    color = YC_GET_CFG_COLOR(
		cp, YC_CFG_PARM_AUDIO_DISPLAY_COLOR_FG_DIM
	    );
	    c = &cs[2];
	    DO_CONVERT_COLOR

	    ConBarSetAudioColors(
		YCONSOLE_CONBAR(yc), &cs[0], &cs[1], &cs[2]
	    );
#undef DO_CONVERT_COLOR

	    ConBarSetAudioFonts(
		YCONSOLE_CONBAR(yc),
		YC_GET_CFG_S(cp, YC_CFG_PARM_AUDIO_DISPLAY_FONT_LARGE),
		YC_GET_CFG_S(cp, YC_CFG_PARM_AUDIO_DISPLAY_FONT_MEDIUM),
		YC_GET_CFG_S(cp, YC_CFG_PARM_AUDIO_DISPLAY_FONT_SMALL)
	    );
	}

	/* Update connection bar's DSPView. */
	dv = ConBarGetDSPView(YCONSOLE_CONBAR(yc));
	if(dv != NULL)
	{
	    const cfg_color_struct *color;
	    gfloat alpha_coeff;

	    DSPViewSetDisplayStyle(
		dv,
		YC_GET_CFG_I(cp, YC_CFG_PARM_DSPVIEW_DISPLAY_STYLE)
	    );
	    DSPViewSetMergeMode(
		dv,
		YC_GET_CFG_I(cp, YC_CFG_PARM_DSPVIEW_MERGE_MODE)
	    );
	    color = YC_GET_CFG_COLOR(cp, YC_CFG_PARM_DSPVIEW_COLOR);
	    alpha_coeff = YC_GET_CFG_F(
		cp, YC_CFG_PARM_DSPVIEW_TRANSPARENCY
	    );
	    if(color != NULL)
		DSPViewSetDSPColor(
		    dv,
		    (guint32)(
			((guint32)(color->r * 0xff) << 0) |
			((guint32)(color->g * 0xff) << 8) |
			((guint32)(color->b * 0xff) << 16) |
			((guint32)(alpha_coeff * 0xff) << 24)
		    )
		);
	    DSPViewSetFadeCoeff(
		dv,
		YC_GET_CFG_F(cp, YC_CFG_PARM_DSPVIEW_FADE_COEFF)
	    );
	}	/* Update connection bar's DSPView. */



}
