/*
			     YConsole Window
 */

#ifndef YCONSOLE_H
#define YCONSOLE_H

#include <Y2/Y.h>
#include <Y2/Ylib.h>
#include <gtk/gtk.h>

#include "conbar.h"

/*
 *	YConsole:
 */
typedef struct {

	gboolean	initialized;
	gboolean	map_state;
	gint		busy_count;
	GtkAccelGroup	*accelgrp;
	gpointer	core_ptr;

	GtkWidget *toplevel;
	connection_bar_struct *conbar;

	/* Y server values. */
	YConnection *con;

	guint8 *yshm_buf;
	gint yshm_buf_len, yshm_id;

	gfloat channel_volume_avg[4];

	gulong start_time;
	gchar *audio_mode_name;
	gint sample_rate, sample_size, channels;
	glong cycle_us;
	gfloat cycle_load;
	gint num_fragments, fragment_size;
	gboolean flip_stereo, shellout;
	gint version_major, version_minor;
	gchar *server_name, *sound_name;

} yconsole_struct;

#define YCONSOLE(p)	((yconsole_struct *)(p))
#define YCONSOLE_TOPLEVEL(p)	(((p) != NULL) ? \
 (YCONSOLE(p)->toplevel) : NULL)
#define YCONSOLE_CONBAR(p)	(((p) != NULL) ? \
 (YCONSOLE(p)->conbar) : NULL)

#define YCONSOLE_IS_CONNECTED(p)	(((p) != NULL) ? \
 ((YCONSOLE(p)->con) != NULL) : FALSE)
#define YCONSOLE_CONNECTION(p)		(((p) != NULL) ? \
 (YCONSOLE(p)->con) : NULL)

#define YCONSOLE_SET_CONBAR_AUDIO(p)	\
{ if((p) != NULL) \
  ConBarSetAudio( \
   YCONSOLE_CONBAR(p), \
   (p)->start_time, (p)->audio_mode_name, \
   (p)->sample_rate, (p)->sample_size, (p)->channels, \
   (p)->cycle_us, (p)->cycle_load, \
   (p)->num_fragments, (p)->fragment_size, \
   (p)->flip_stereo, (p)->shellout, \
   (p)->yshm_buf, (p)->yshm_buf_len, \
   (p)->channel_volume_avg, \
   (p)->version_major, (p)->version_minor, \
   (p)->server_name, (p)->sound_name \
  ); }


extern void YConsoleSetTitle(yconsole_struct *yc, const gchar *s);
extern gchar *YConsoleGetLocation(yconsole_struct *yc);
extern gint YConsoleDoConnect(
	yconsole_struct *yc,
	const gchar *address, const gchar *start_command
);
extern void YConsoleDoDisconnect(yconsole_struct *yc, gboolean force_shutdown);
extern void YConsoleSavePosition(yconsole_struct *yc);

extern yconsole_struct *YConsoleNew(gpointer core_ptr);
extern void YConsoleUpdateMenus(yconsole_struct *yc);
extern void YConsoleSetBusy(yconsole_struct *yc, gboolean is_busy);
extern void YConsoleMap(yconsole_struct *yc);
extern void YConsoleUnmap(yconsole_struct *yc);
extern void YConsoleDelete(yconsole_struct *yc);



#endif	/* YCONSOLE_H */
