#include <stdlib.h>
#include <string.h>

#include <gtk/gtk.h>

#include "guiutils.h"
#include "cdialog.h"

#include "cfg.h"
#include "optwin.h"
#include "optwinop.h"
#include "ycoptwin.h"

#include "yconsolewin.h"
#include "yconsolecb.h"
#include "yaudio.h"
#include "yaudiocb.h"
#include "ymixer.h"
#include "ymixercb.h"
#include "yplayer.h"
#include "yplayercb.h"
#include "yserverlist.h"
#include "yserverlistcb.h"
#include "startupdialog.h"
#include "yc.h"
#include "ycop.h"
#include "yccfglist.h"
#include "config.h"


static void YCCenterWindowToWindow(GtkWidget *w1, GtkWidget *w2);

gint YCNewYConsole(
	yc_core_struct *core_ptr, const gchar *address
);
gint YCNewYAudio(
	yc_core_struct *core_ptr, const gchar *address
);
gint YCNewYMixer(
	yc_core_struct *core_ptr, const gchar *address
);
gint YCNewYPlayerAudioCD(
	yc_core_struct *core_ptr, const gchar *address,
	gint audio_cd_track_number
);
gint YCNewYPlayerSoundObject(
	yc_core_struct *core_ptr, const gchar *address,
	const gchar *sndobj_filename
);
gint YCNewYPlayer(
	yc_core_struct *core_ptr, const gchar *address
);
gint YCNewYServerList(yc_core_struct *core_ptr, GtkWidget *rel_toplevel);
gint YCNewOptWinSelectPage(
	yc_core_struct *core_ptr, GtkWidget *rel_toplevel,
	gint page_num
);
gint YCNewOptWin(yc_core_struct *core_ptr, GtkWidget *rel_toplevel);
gint YCNewStartupDialog(
	yc_core_struct *core_ptr, GtkWidget *rel_toplevel,
	const gchar *cmd, gboolean block
);

void YCReclaimMemory(yc_core_struct *core_ptr);


#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define ABSOLUTE(x)     (((x) < 0) ? ((x) * -1) : (x))


/*
 *      Moves the window w2 relative to window w1, both w1 and w2
 *      must be GtkWindows.
 */
void YCCenterWindowToWindow(GtkWidget *w1, GtkWidget *w2)
{
	gint x, y;
	gint width1, height1, width2, height2;
	GdkWindow *window1, *window2;


	if((w1 == NULL) || (w2 == NULL))
	    return;

	if(!GTK_IS_WINDOW(w1) || !GTK_IS_WINDOW(w2))
	    return;

	window1 = w1->window;
	window2 = w2->window;
	if((window1 == NULL) || (window2 == NULL))
	    return;

	gdk_window_get_root_origin(window1, &x, &y);
	gdk_window_get_size(window1, &width1, &height1);
	gdk_window_get_size(window2, &width2, &height2);

	gtk_widget_set_uposition(
	    w2,
	    (width2 > 0) ?
		x + (width1 / 2) - (width2 / 2) : x + 20,
	    (height2 > 0) ?
		y + (height1 / 2) - (height2 / 2) : y + 20
	);
}


/*
 *	Creates a new YConsole window on the core structure, returning
 *	the index number on success or -1 on error.
 *
 *	If address is not NULL then the window will be instructed to
 *	connect to the given address.
 */
gint YCNewYConsole(
	yc_core_struct *core_ptr, const gchar *address
)
{
	gint i;
	yconsole_struct *yc;


	if(core_ptr == NULL)
	    return(-1);

	/* Look for an available pointer in the array. */
	for(i = 0; i < core_ptr->total_yconsoles; i++)
	{
	    if(core_ptr->yconsole[i] == NULL)
		break;
	}
	/* Need to allocate more pointers? */
	if(i >= core_ptr->total_yconsoles)
	{
	    /* Append a new pointer. */
	    i = core_ptr->total_yconsoles;
	    core_ptr->total_yconsoles = i + 1;
	    core_ptr->yconsole = (yconsole_struct **)g_realloc(
		core_ptr->yconsole,
		core_ptr->total_yconsoles * sizeof(yconsole_struct *)
	    );
	    if(core_ptr->yconsole == NULL)
	    {
		core_ptr->total_yconsoles = 0;
		return(-1);
	    }
	}

	/* Create a new window at the newly allocated pointer index i. */
	core_ptr->yconsole[i] = yc = YConsoleNew(core_ptr);
	if(yc == NULL)
	    return(-1);

	YConsoleReconfiguredCB(yc, core_ptr);

	/* Map new window and connect as needed. */
	YConsoleMap(yc);
	if(address != NULL)
	{
	    gchar *s = g_strdup(address);
	    YConsoleDoConnect(yc, s, NULL);
	    g_free(s);
	}

	/* Since the window is mapped we can flush any GTK+ events. */
	while(gtk_events_pending() > 0)
	    gtk_main_iteration();

	return(i);
}

/*
 *      Creates a new YAudio window on the core structure, returning
 *      the index number on success or -1 on error.
 *
 *      If address is not NULL then the window will be instructed to
 *      connect to the given address.
 */
gint YCNewYAudio(
	yc_core_struct *core_ptr, const gchar *address
)
{
	gint i;
	yaudio_struct *ya;


	if(core_ptr == NULL)
	    return(-1);

	/* Look for an available pointer in the array. */
	for(i = 0; i < core_ptr->total_yaudios; i++)
	{
	    if(core_ptr->yaudio[i] == NULL)
		break;
	}
	/* Need to allocate more pointers? */
	if(i >= core_ptr->total_yaudios)
	{
	    /* Append a new pointer. */
	    i = core_ptr->total_yaudios;
	    core_ptr->total_yaudios = i + 1;
	    core_ptr->yaudio = (yaudio_struct **)g_realloc(
		core_ptr->yaudio,
		core_ptr->total_yaudios * sizeof(yaudio_struct *)
	    );
	    if(core_ptr->yaudio == NULL)
	    {
		core_ptr->total_yaudios = 0;
		return(-1);
	    }
	}

	/* Create a new window at the newly allocated pointer index i. */
	core_ptr->yaudio[i] = ya = YAudioNew(core_ptr);
	if(ya == NULL)
	    return(-1);

	YAudioReconfiguredCB(ya, core_ptr);

	/* Map new window and connect as needed. */
	YAudioMap(ya);
	if(address != NULL)
	{
	    gchar *s = g_strdup(address);
	    YAudioDoConnect(ya, s, NULL);
	    g_free(s);
	}

	/* Since the window is mapped we can flush any GTK+ events. */
	while(gtk_events_pending() > 0)
	    gtk_main_iteration();

	return(i);
}

/*
 *      Creates a new YMixer window on the core structure, returning
 *      the index number on success or -1 on error.
 *
 *      If address is not NULL then the window will be instructed to
 *      connect to the given address.
 */
gint YCNewYMixer(
	yc_core_struct *core_ptr, const gchar *address
)
{
	gint i;
	ymixer_struct *ym;


	if(core_ptr == NULL)
	    return(-1);

	/* Look for an available pointer in the array. */
	for(i = 0; i < core_ptr->total_ymixers; i++)
	{
	    if(core_ptr->ymixer[i] == NULL)
		break;
	}
	/* Need to allocate more pointers? */
	if(i >= core_ptr->total_ymixers)
	{
	    /* Append a new pointer. */
	    i = core_ptr->total_ymixers;
	    core_ptr->total_ymixers = i + 1;
	    core_ptr->ymixer = (ymixer_struct **)g_realloc(
		core_ptr->ymixer,
		core_ptr->total_ymixers * sizeof(ymixer_struct *)
	    );
	    if(core_ptr->ymixer == NULL)
	    {
		core_ptr->total_ymixers = 0;
		return(-1);
	    }
	}

	/* Create a new window at the newly allocated pointer index i. */
	core_ptr->ymixer[i] = ym = YMixerNew(core_ptr);
	if(ym == NULL)
	    return(-1);

	YMixerReconfiguredCB(ym, core_ptr);

	/* Map new window and connect as needed. */
	YMixerMap(ym);
	if(address != NULL)
	{
	    gchar *s = g_strdup(address);
	    YMixerDoConnect(ym, s, NULL);
	    g_free(s);
	}

	/* Since the window is mapped we can flush any GTK+ events. */
	while(gtk_events_pending() > 0)
	    gtk_main_iteration();

	return(i);
}

/*
 *      Creates a new YPlayer window on the core structure, returning
 *      the index number on success or -1 on error.
 *
 *      If address is not NULL then the window will be instructed to
 *      connect to the given address.
 */
static gint YCNewYPlayerNexus(
	yc_core_struct *core_ptr, const gchar *address,
	const gchar *sndobj_filename, gint audio_cd_track_number
)
{
	gint i;
	yplayer_struct *yp;


	if(core_ptr == NULL)
	    return(-1);

	/* Look for an available pointer in the array. */
	for(i = 0; i < core_ptr->total_yplayers; i++)
	{
	    if(core_ptr->yplayer[i] == NULL)
		break;
	}
	/* Need to allocate more pointers? */
	if(i >= core_ptr->total_yplayers)
	{
	    /* Append a new pointer. */
	    i = core_ptr->total_yplayers;
	    core_ptr->total_yplayers = i + 1;
	    core_ptr->yplayer = (yplayer_struct **)g_realloc(
		core_ptr->yplayer,
		core_ptr->total_yplayers * sizeof(yplayer_struct *)
	    );
	    if(core_ptr->yplayer == NULL)
	    {
		core_ptr->total_yplayers = 0;
		return(-1);
	    }
	}

	/* Create a new window at the newly allocated pointer index i. */
	core_ptr->yplayer[i] = yp = YPlayerNew(core_ptr);
	if(yp == NULL)
	    return(-1);

	YPlayerReconfiguredCB(yp, core_ptr);

	/* Map new window and connect as needed. */
	YPlayerMap(yp);
	YPlayerSetBusy(yp, TRUE);
	if(address != NULL)
	{
	    gchar *s = g_strdup(address);
	    YPlayerDoConnect(yp, s, NULL);
	    g_free(s);
	}

	/* Since the window is mapped we can flush any GTK+ events. */
	while(gtk_events_pending() > 0)
	    gtk_main_iteration();

	/* Load default playlist and last sound object. */
	if(TRUE)
	{
	    const gchar *cstrptr = YC_GET_CFG_S(
		core_ptr,
		YC_CFG_PARM_PLAYER_SNDOBJ_DEFAULT_PLAYLIST_FILE
	    );
	    if(cstrptr != NULL)
		YPlayerLoadPlayListFile(yp, cstrptr, FALSE);

	    cstrptr = YC_GET_CFG_S(
		core_ptr,
		YC_CFG_PARM_PLAYER_SNDOBJADV_LAST_FILE
	    );
	    if(cstrptr != NULL)
		YPlayerDoOpenSoundObjectAdv(yp, cstrptr);
	}

	while(gtk_events_pending() > 0)
	    gtk_main_iteration();

	/* Play sound object as needed. */
	if(sndobj_filename != NULL)
	    YPlayerDoPlaySoundObject(yp, sndobj_filename, TRUE);
	/* Play audio CD track as needed. */
	else if(audio_cd_track_number > 0)
	    YPlayerDoPlayAudioCDTrack(yp, audio_cd_track_number);

	YPlayerSetBusy(yp, FALSE);

	return(i);
}
gint YCNewYPlayerAudioCD(
	yc_core_struct *core_ptr, const gchar *address,
	gint audio_cd_track_number
)
{
	return(YCNewYPlayerNexus(
	    core_ptr, address, NULL, audio_cd_track_number
	));
}
gint YCNewYPlayerSoundObject(
	yc_core_struct *core_ptr, const gchar *address,
	const gchar *sndobj_filename
)
{
	return(YCNewYPlayerNexus(
	    core_ptr, address, sndobj_filename, 0
	));
}
gint YCNewYPlayer(yc_core_struct *core_ptr, const gchar *address)
{
	return(YCNewYPlayerNexus(
	    core_ptr, address, NULL, 0
	));
}


/*
 *	Creates a new YServer list window on the core structure,
 *	returning the index number on success or -1 on error.
 */
gint YCNewYServerList(yc_core_struct *core_ptr, GtkWidget *rel_toplevel)
{
	yserver_list_struct *sl;


	if(core_ptr == NULL)
	    return(-1);

	/* Get YServer list window, create it as needed. */
	sl = core_ptr->yserver_list;
	if(sl == NULL)
	{
	    core_ptr->yserver_list = sl = YServerListNew(core_ptr);
	    if(sl == NULL)
		return(-1);

	    YServerListFileOpen(
		sl,
		YC_GET_CFG_S(core_ptr, YC_CFG_PARM_FILE_SERVERS_LIST)
	    );
	}

	YServerListReconfiguredCB(sl, core_ptr);

	/* Record relative toplevel on YServer list window. */
	sl->rel_toplevel = rel_toplevel;

	/* Map YServer list window. */
	YServerListMap(sl);

	return(0);
}

/*
 *      Creates a new Options window on the core structure,
 *      returning the index number on success or -1 on error.
 */
gint YCNewOptWinSelectPage(
	yc_core_struct *core_ptr, GtkWidget *rel_toplevel,
	gint page_num
)
{
	optwin_struct *ow;


	if(core_ptr == NULL)
	    return(-1);

	/* Get Options window, create it as needed. */
	ow = core_ptr->optwin;
	if(ow == NULL)
	{
	    core_ptr->optwin = ow = YCOptWinNew(core_ptr);
	    if(ow == NULL)
		return(-1);
	}

	/* Select page? */
	if(page_num > -1)
	{
	    GtkWidget *w = OPTWIN_NOTEBOOK(ow);
	    if((w != NULL) && (page_num < OPTWIN_PAGES(ow)))
		gtk_notebook_set_page(
		    GTK_NOTEBOOK(w), page_num
		);
	}

	/* Fetch configuration values and map the Options window. */
	OptWinDoFetch(ow);
	YCCenterWindowToWindow(rel_toplevel, OPTWIN_TOPLEVEL(ow));
	OptWinMap(ow);

	return(0);
}
gint YCNewOptWin(yc_core_struct *core_ptr, GtkWidget *rel_toplevel)
{
	return(YCNewOptWinSelectPage(
	    core_ptr, rel_toplevel, -1
	));
}

/*
 *      Creates a new Startup Dialog on the core structure, returning
 *      the index number on success or -1 on error.
 *
 *      If cmd is not NULL then the window will be instructed to
 *      execute the given command.
 */
gint YCNewStartupDialog(
	yc_core_struct *core_ptr, GtkWidget *rel_toplevel,
	const gchar *cmd, gboolean block
)
{
	gint i;
	startup_dialog_struct *sd;


	if(core_ptr == NULL)
	    return(-1);

	/* Look for an available pointer in the array. */
	for(i = 0; i < core_ptr->total_startup_dialogs; i++)
	{
	    if(core_ptr->startup_dialog[i] == NULL)
		break;
	}
	/* Need to allocate more pointers? */
	if(i >= core_ptr->total_startup_dialogs)
	{
	    /* Append a new pointer. */
	    i = core_ptr->total_startup_dialogs;
	    core_ptr->total_startup_dialogs = i + 1;
	    core_ptr->startup_dialog = (startup_dialog_struct **)g_realloc(
		core_ptr->startup_dialog,
		core_ptr->total_startup_dialogs * sizeof(ymixer_struct *)
	    );
	    if(core_ptr->startup_dialog == NULL)
	    {
		core_ptr->total_startup_dialogs = 0;
		return(-1);
	    }
	}

	/* Create a new window at the newly allocated pointer index i. */
	core_ptr->startup_dialog[i] = sd = StartupDialogNew(
	    NULL,	/* Title string. */
	    NULL,	/* WM Icon xpm data. */
	    NULL,	/* Heading xpm data. */
	    core_ptr
	);
	if(sd == NULL)
	    return(-1);
/*
	StartupDialogReconfiguredCB(sd, core_ptr);
 */

	/* Map new window and connect as needed. */
	YCCenterWindowToWindow(rel_toplevel, STARTUP_DIALOG_TOPLEVEL(sd));
	StartupDialogMap(sd);
	if(cmd != NULL)
	{
	    gchar *s = g_strdup(cmd);
	    StartupDialogExecute(sd, s);
	    g_free(s);
	}

	/* Block execution until window is unmapped? */
	if(block)
	{
	    while(TRUE)
	    {
		sd = (i < core_ptr->total_startup_dialogs) ?
		    core_ptr->startup_dialog[i] : NULL;
		if((sd != NULL) ? !sd->map_state : TRUE)
		    break;

		while(gtk_events_pending() > 0)
		    gtk_main_iteration();
	    }
	}

	return(i);
}


/*
 *	Checks for and then reclaims/deallocates any unused memory on 
 *	the core structure.
 */
void YCReclaimMemory(yc_core_struct *core_ptr)
{
	gint *total, psize;
	gpointer **pa;

	if(core_ptr == NULL)
	    return;


#define DO_RECLAIM_POINTER_ARRAY				\
if(*total > 0) {						\
  /* Seek to highest index h. */				\
  gint i = 0, h = -1;						\
  for(; i < *total; i++) {					\
   if((*pa)[i] != NULL) h = i;					\
  }								\
								\
  /* Update total to highest index h + 1. */			\
  *total = h + 1;						\
								\
  /* If total is positive then reallocate pointer array,	\
   * otherwise deallocate the pointer array. */			\
  if(*total > 0) {						\
   *pa = (gpointer *)g_realloc(*pa, *total * psize);		\
   if(*pa == NULL) *total = 0;					\
  } else {							\
   g_free(*pa); *pa = NULL;					\
   *total = 0;							\
  }								\
}
	pa = (gpointer **)&core_ptr->yconsole;
	total = &core_ptr->total_yconsoles;
	psize = sizeof(yconsole_struct *);
	DO_RECLAIM_POINTER_ARRAY

	pa = (gpointer **)&core_ptr->yaudio;
	total = &core_ptr->total_yaudios;
	psize = sizeof(yaudio_struct *);
	DO_RECLAIM_POINTER_ARRAY

	pa = (gpointer **)&core_ptr->ymixer;
	total = &core_ptr->total_ymixers;
	psize = sizeof(ymixer_struct *);
	DO_RECLAIM_POINTER_ARRAY

	pa = (gpointer **)&core_ptr->yplayer;
	total = &core_ptr->total_yplayers;
	psize = sizeof(yplayer_struct *);
	DO_RECLAIM_POINTER_ARRAY

	pa = (gpointer **)&core_ptr->startup_dialog;
	total = &core_ptr->total_startup_dialogs;
	psize = sizeof(startup_dialog_struct *);
	DO_RECLAIM_POINTER_ARRAY




#undef DO_RECLAIM_POINTER_ARRAY
}
