#include <stdlib.h>
#include <string.h>

#include <Y2/Y.h>
#include <Y2/Ylib.h>
#include <Y2/Ymixercodes.h>

#include <gtk/gtk.h>

#include "../include/string.h"
#include "../include/disk.h"

#include "guiutils.h"
#include "cdialog.h"
#include "fb.h"
#include "csd.h"
#include "fsd.h"
#include "stacklist.h"

#include "cfg.h"
#include "optwin.h"

#include "dspview.h"
#include "ymixer.h"
#include "ycoptwin.h"
#include "yc.h"
#include "ycimg.h"
#include "yccfglist.h"
#include "config.h"

#include "images/icon_options2_48x48.xpm"

#include "images/icon_yconsole_48x48.xpm"
#include "images/icon_yaudio_48x48.xpm"
#include "images/icon_ymixer_48x48.xpm"
#include "images/icon_yplayer_48x48.xpm"

#include "images/icon_ymixer_20x20.xpm"
#include "images/icon_yplayer_20x20.xpm"

#include "images/icon_sound_32x32.xpm"
#include "images/icon_cd_32x32.xpm"


void YCOptWinManageCB(optwin_struct *ow, gpointer core_ptr);

static optwin_cfgref_struct *YCOptWinColorButtonCreate(
	optwin_struct *optwin,
	const gchar *label, gint width, gint id, const gchar *cfg_parm,
	GtkWidget *parent,
	void (*clicked_cb)(GtkWidget *, gpointer)
);
static optwin_cfgref_struct *YCOptWinFontButtonCreate(
	optwin_struct *optwin,
	const gchar *label, const gchar *browse_label, gint width,
	const gchar *cfg_parm,
	GtkWidget *parent,
	void (*clicked_cb)(GtkWidget *, gpointer)
);

static void YCOptWinBrowsePathCB(gpointer button, gpointer data);
static void YCOptWinColorChangedCB(gpointer data, csd_color_struct *c);
static void YCOptWinColorCB(GtkWidget *widget, gpointer data);
static void YCOptWinFontCB(GtkWidget *widget, gpointer data);

static void YCOptWinCreatePageYPlayer(
	yc_core_struct *core_ptr, optwin_struct *optwin,
	GtkWidget *parent
);

static void YCOptWinMixerScaleColorUseDefaultToggleCB(
	GtkWidget *widget, gpointer data
);
static void YCOptWinCreatePageYMixer(
	yc_core_struct *core_ptr, optwin_struct *optwin,
	GtkWidget *parent
);

static void YCOptWinDSPViewDestroyCB(gpointer cfgref, gpointer data);
static void YCOptWinDSPViewDisplayStyleWaveCB(
	GtkWidget *widget, gpointer data
);
static void YCOptWinDSPViewDisplayStyleSineGraphCB(
	GtkWidget *widget, gpointer data
);
static void YCOptWinDSPViewDisplayStyleAnalyzerCB(
	GtkWidget *widget, gpointer data
);
static void YCOptWinDSPViewDisplayStyleLighteningCB(
	GtkWidget *widget, gpointer data
);
static void YCOptWinDSPViewDisplayStylePlasmaCB(
	GtkWidget *widget, gpointer data
);
static void YCOptWinDSPViewDisplayStyleNovaCB(
	GtkWidget *widget, gpointer data
);
static void YCOptWinDSPViewMergeModeSetCB(
	GtkWidget *widget, gpointer data
);
static void YCOptWinDSPViewMergeModeSetAlphaCB(
	GtkWidget *widget, gpointer data
);
static void YCOptWinDSPViewMergeModeAddCB(
	GtkWidget *widget, gpointer data
);
static void YCOptWinDSPViewMergeModeSubtractCB(
	GtkWidget *widget, gpointer data
);
static void YCOptWinDSPViewMergeModeBlendCB(
	GtkWidget *widget, gpointer data
);
static void YCOptWinDSPViewMergeModeMultiplyCB(
	GtkWidget *widget, gpointer data
);
static void YCOptWinDSPViewColorChangedCB(
	gpointer data, csd_color_struct *c
);
static void YCOptWinDSPViewColorCB(
	GtkWidget *widget, gpointer data
);
static void YCOptWinDSPViewTransparencyChangedCB(
	GtkAdjustment *adj, gpointer data
);
static void YCOptWinDSPViewFadeCoeffChangedCB(
	GtkAdjustment *adj, gpointer data
);
static void YCOptWinDisplayLoadImagesCB(GtkWidget *widget, gpointer data);
static void YCOptWinCreatePageDisplay(
	yc_core_struct *core_ptr, optwin_struct *optwin,
	GtkWidget *parent
);
static void YCOptWinCreatePageGeneral(
	yc_core_struct *core_ptr, optwin_struct *optwin,
	GtkWidget *parent
);

optwin_struct *YCOptWinNew(gpointer core_ptr);



/*
 *	Options window manage callback.
 */
void YCOptWinManageCB(optwin_struct *ow, gpointer core_ptr)
{
	dspview_struct *dv;
	static guint32 rand_seed_count = 0;
#define buf_len	256
	guint8 buf[buf_len], *buf_ptr, *buf_end;
	yc_core_struct *cp = YC_CORE(core_ptr);
	if((ow == NULL) || (cp == NULL))
	    return;

	dv = cp->optwin_dspview;
	if(dv == NULL)
	    return;

	if(GTK_WIDGET_VISIBLE(dv->view_da))
	    return;

	/* Generate random buffer */
	srand(rand_seed_count);
	for(buf_ptr = buf,
	    buf_end = buf_ptr + buf_len;
	    buf_ptr < buf_end;
	    buf_ptr++
	)
	{
	    *buf_ptr = rand() % (guint8)-1;
	}
	rand_seed_count++;


	/* Redraw DSP view */
	DSPViewDraw(dv, buf, buf_len);

#undef buf_len
}


/*
 *	Creates a new color button.
 */
static optwin_cfgref_struct *YCOptWinColorButtonCreate(
	optwin_struct *optwin,
	const gchar *label, gint width, gint id, const gchar *cfg_parm,
	GtkWidget *parent,
	void (*clicked_cb)(GtkWidget *, gpointer)
)
{
	gint border_minor = 2;
	GtkWidget *w, *parent2, *parent3, *toplevel, *button, *da;
	optwin_cfgref_struct *cfgref;


	/* Main hbox to hold label and button */
	toplevel = parent2 = w = gtk_hbox_new(FALSE, border_minor);
	if(width > 0)
	    gtk_widget_set_usize(w, width, -1);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	/* Label */
	if(label != NULL)
	{
	    w = gtk_vbox_new(TRUE, 0);
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent3 = w;
	    w = gtk_label_new(label);
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	}

	/* Color button */
	button = w = gtk_button_new();
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;

	/* GtkDrawingArea to display color */
	da = w = gtk_drawing_area_new();
	gtk_widget_set_usize(w, 20, 20);
	gtk_widget_add_events(
	    w,
	    GDK_EXPOSURE_MASK | GDK_STRUCTURE_MASK |
	    GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
	    GDK_ENTER_NOTIFY_MASK | GDK_LEAVE_NOTIFY_MASK
	);
	gtk_container_add(GTK_CONTAINER(parent3), w);
	GUISetWidgetTip(w, "Click to select color");
	gtk_widget_show(w);

	/* Set cfgref for this GtkDrawingArea */
	cfgref = OptWinCfgRefAppend(optwin, id, da, cfg_parm);
	cfgref->parent = toplevel;

	/* Set signal callback for the GtkButton */
	if((button != NULL) && (clicked_cb != NULL))
	    gtk_signal_connect(
		GTK_OBJECT(button), "clicked",
		GTK_SIGNAL_FUNC(clicked_cb), cfgref
	    );

	return(cfgref);
}

/*
 *      Creates a new font selection button.
 */
static optwin_cfgref_struct *YCOptWinFontButtonCreate(
	optwin_struct *optwin,
	const gchar *label, const gchar *browse_label, gint width,
	const gchar *cfg_parm,
	GtkWidget *parent,
	void (*clicked_cb)(GtkWidget *, gpointer)
)
{
	gint border_minor = 2;
	GtkWidget *w, *parent2, *toplevel, *entry, *button;
	optwin_cfgref_struct *cfgref;

	/* Main hbox to hold label and button */
	toplevel = parent2 = w = gtk_hbox_new(FALSE, border_minor);
	if(width > 0)
	    gtk_widget_set_usize(w, width, -1);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	/* Label */
	if(label != NULL)
	{
	    w = gtk_label_new(label);
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	}

	/* Entry */
	entry = w = gtk_entry_new();
	if(width < 0)
	    gtk_widget_set_usize(w, 250, -1);
	gtk_entry_set_editable(GTK_ENTRY(w), FALSE);
	gtk_box_pack_start(
	    GTK_BOX(parent2), w,
	    (width < 0) ? TRUE : FALSE,
	    (width < 0) ? TRUE : FALSE,
	    0
	);
	gtk_widget_show(w);

	/* Select button */
	button = w = gtk_button_new_with_label(
	    (browse_label != NULL) ? browse_label : "Select..."
	);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	GUISetWidgetTip(w, "Select font");
	gtk_widget_show(w);

	/* Set cfgref for the GtkEntry */
	cfgref = OptWinCfgRefAppend(optwin, -1, entry, cfg_parm);
	cfgref->parent = toplevel;

	/* Set signal callback for the GtkButton */
	if((button != NULL) && (clicked_cb != NULL))
	    gtk_signal_connect(
		GTK_OBJECT(button), "clicked",
		GTK_SIGNAL_FUNC(clicked_cb), cfgref
	    );

	return(cfgref);
}


/*
 *	Path browse button "clicked" signal callback.
 */
static void YCOptWinBrowsePathCB(gpointer button, gpointer data)
{
	gboolean status;
	gchar *startup_path;
	GtkWidget *w, *toplevel;
	GtkEntry *entry;
	fb_type_struct **ftype = NULL, *ftype_rtn = NULL;
	gint total_ftypes = 0;
	gchar **path_rtn = NULL;
	gint total_path_rtns = 0;
	optwin_cfgref_struct *cfgref = OPTWIN_CFGREF(data);
	if(cfgref == NULL)
	    return;

	w = cfgref->w;
	if((w != NULL) ? GTK_IS_ENTRY(w) : FALSE)
	    entry = GTK_ENTRY(w);
	else
	    return;

	if(FileBrowserIsQuery())
	    return;

	toplevel = gtk_widget_get_toplevel(w);

	startup_path = gtk_entry_get_text(entry);
	startup_path = (ISPATHDIR(startup_path)) ?
	    g_strdup(startup_path) : g_strdup(GetParentDir(startup_path));

	FileBrowserTypeListNew(
	    &ftype, &total_ftypes,
	    "*.*", "All files"
	);

	/* Query user for program path */
	FileBrowserSetTransientFor(toplevel);
	status = FileBrowserGetResponse(
	    "Select Object",
	    "Select", "Cancel",
	    startup_path,
	    ftype, total_ftypes,
	    &path_rtn, &total_path_rtns,
	    &ftype_rtn
	);
	FileBrowserSetTransientFor(NULL);

	/* Got user response? */
	if(status)
	{
	    const gchar *new_path = (total_path_rtns > 0) ?
		path_rtn[0] : NULL;
	    if((new_path != NULL) ? (*new_path != '\0') : FALSE)
		gtk_entry_set_text(entry, new_path);
	}

	/* Deallocate file types list */
	FileBrowserDeleteTypeList(ftype, total_ftypes);
	g_free(startup_path);
}

/*
 *      Color selection dialog color changed signal callback.
 */
static void YCOptWinColorChangedCB(
	gpointer data, csd_color_struct *c
)
{
	GdkColor tc;
	GdkWindow *window;
	GtkWidget *w;
	optwin_cfgref_struct *cfgref = OPTWIN_CFGREF(data);
	if((c == NULL) || (cfgref == NULL))
	    return;

	tc.red          = (gushort)-1 * c->r;
	tc.green        = (gushort)-1 * c->g;
	tc.blue         = (gushort)-1 * c->b;

	/* Reallocate the cfgref's color */
	OptWinCfgRefSetColor(cfgref, &tc, &tc);

	w = cfgref->w;
	window = (w != NULL) ? w->window : NULL;
	if(window != NULL)
	{
	    gdk_window_set_background(window, &cfgref->fg_color);
	    gdk_window_clear(window);
	}
}
/*
 *	Color button "clicked" signal callback.
 */
static void YCOptWinColorCB(
	GtkWidget *widget, gpointer data
)
{
	gboolean status;
	optwin_cfgref_struct *cfgref = OPTWIN_CFGREF(data);
	GdkColor *c = &cfgref->fg_color;
	GtkWidget *w = cfgref->w;
	GtkWidget *toplevel = gtk_widget_get_toplevel(w);
	csd_color_struct cs, *cr = NULL;

	if(CSDIsQuery())
	    return;

	cs.r = (gfloat)c->red   / (gfloat)((gushort)-1);
	cs.g = (gfloat)c->green / (gfloat)((gushort)-1);
	cs.b = (gfloat)c->blue  / (gfloat)((gushort)-1);

	CSDSetTransientFor(toplevel);
	status = CSDGetResponse(
	    "Select Color",
	    "Select", "Cancel",
	    &cs, &cr,
	    cfgref, YCOptWinColorChangedCB
	);
	CSDSetTransientFor(NULL);

	/* If user canceled then revert to original color */
	if(!status)
	    YCOptWinColorChangedCB(cfgref, &cs);
}

/*
 *	Font select button "clicked" signal callback.
 */
static void YCOptWinFontCB(GtkWidget *widget, gpointer data)
{
	gboolean status;
	optwin_cfgref_struct *cfgref = OPTWIN_CFGREF(data);
	GtkWidget *w = cfgref->w;
	GtkWidget *toplevel = gtk_widget_get_toplevel(w);
	gchar *font_s, *font_t = NULL;

	if(FSDIsQuery())
	    return;

	if(w == NULL)
	    return;

	font_s = gtk_entry_get_text(GTK_ENTRY(w));

	FSDSetTransientFor(toplevel);
	status = FSDGetResponse(
	    "Select Font", "Select", "Cancel",
	    font_s, &font_t
	);
	FSDSetTransientFor(NULL);

	if(status && (font_t != NULL))
	    gtk_entry_set_text(GTK_ENTRY(w), font_t);
}


/*
 *      Create YPlayer page.
 */
static void YCOptWinCreatePageYPlayer(
	yc_core_struct *core_ptr, optwin_struct *optwin,
	GtkWidget *parent
)
{
	gint border_minor = 2, border_major = 5;
	gpointer entry_rtn, browse_rtn;
	GtkWidget *w, *parent2, *parent3;
	optwin_cfgref_struct *cfgref;
/*      yc_core_struct *cp = YC_CORE(core_ptr); */


	/* Sound Objects frame */
	w = gtk_frame_new("Sound Objects");
	gtk_box_pack_start(GTK_BOX(parent), w, TRUE, TRUE, 0);
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_hbox_new(FALSE, border_major);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_vbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;

	w = GUICreateMenuItemIcon((guint8 **)icon_sound_32x32_xpm);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	w = gtk_vbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent3 = w;

	/* Default playlist */
	w = GUIPromptBarWithBrowse(
	    NULL, "Default PlayList:",
	    NULL, &entry_rtn, &browse_rtn,
	    NULL, NULL
	);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	w = (GtkWidget *)entry_rtn;
	GUISetWidgetTip(
	    w,
"Enter the path of the default playlist file"
	);
	cfgref = OptWinCfgRefAppend(
	    optwin, -1, w,
	    YC_CFG_PARM_PLAYER_SNDOBJ_DEFAULT_PLAYLIST_FILE
	);
	w = (GtkWidget *)browse_rtn;
	GUISetWidgetTip(w, "Browse");
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(YCOptWinBrowsePathCB), cfgref
	);




	/* Audio CD frame */
	w = gtk_frame_new("Audio CD");
	gtk_box_pack_start(GTK_BOX(parent), w, TRUE, TRUE, 0);
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_hbox_new(FALSE, border_major);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_vbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;

	w = GUICreateMenuItemIcon((guint8 **)icon_cd_32x32_xpm);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	w = gtk_vbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent3 = w;



}


/*
 *	Mixer scale color use default "toggled" signal callback.
 */
static void YCOptWinMixerScaleColorUseDefaultToggleCB(
	GtkWidget *widget, gpointer data
)
{
	gboolean sensitive;
	optwin_cfgref_struct *cfgref;
	optwin_struct *optwin = OPTWIN(data);
	if((widget == NULL) || (optwin == NULL))
	    return;

	sensitive = !GTK_TOGGLE_BUTTON(widget)->active;

#define DO_SET_SENSITIVE			\
{ GtkWidget *w = OPTWIN_CFGREF_PARENT(cfgref);	\
  if(w != NULL)					\
   gtk_widget_set_sensitive(w, sensitive);	\
}

	cfgref = OptWinCfgRefMatchID(
	    optwin, (gint)YC_CFG_PARM_MIXER_COLOR_FG, NULL
	);
	DO_SET_SENSITIVE
	cfgref = OptWinCfgRefMatchID(
	    optwin, (gint)YC_CFG_PARM_MIXER_COLOR_BG, NULL
	);
	DO_SET_SENSITIVE
	cfgref = OptWinCfgRefMatchID(
	    optwin, (gint)YC_CFG_PARM_MIXER_COLOR_BG_PRELIGHT, NULL
	);
	DO_SET_SENSITIVE
	cfgref = OptWinCfgRefMatchID(
	    optwin, (gint)YC_CFG_PARM_MIXER_COLOR_TEXT, NULL
	);
	DO_SET_SENSITIVE
	cfgref = OptWinCfgRefMatchID(
	    optwin, (gint)YC_CFG_PARM_MIXER_COLOR_BASE, NULL
	);
	DO_SET_SENSITIVE
#undef DO_SET_SENSITIVE
}

/*
 *      Create YMixer page.
 */
static void YCOptWinCreatePageYMixer(
	yc_core_struct *core_ptr, optwin_struct *optwin,
	GtkWidget *parent
)
{
	gint border_minor = 2, border_major = 5;
	gint i;
	const gchar *cstrptr;
	guint8 **icon_data;
	GSList *gslist;
	GtkWidget *w, *parent2, *parent3;
	optwin_cfgref_struct *cfgref;
	stack_list_struct *slist;
/*	yc_core_struct *cp = YC_CORE(core_ptr); */
	gchar *mixer_name[] = YMixerNames;


	w = gtk_hbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	/* Label display type */
	w = gtk_frame_new("Display Mixer Labels As");
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, TRUE, 0);
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_widget_show(w);
	parent3 = w;

	w = gtk_hbox_new(FALSE, border_minor);
	gtk_container_add(GTK_CONTAINER(parent3), w);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_widget_show(w);
	parent3 = w;

	/* Text radio button */
	gslist = NULL;
	w = gtk_radio_button_new_with_label(gslist, "Text");
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	cfgref = OptWinCfgRefAppend(
	    optwin, -1, w,
	    YC_CFG_PARM_MIXER_MIXER_SCALE_LABEL_TYPE
	);
	cfgref->radio_value = YMIXER_MIXER_SCALE_LABEL_TYPE_NAME;
	gslist = gtk_radio_button_group(GTK_RADIO_BUTTON(w));

	/* Short text radio button */
	w = gtk_radio_button_new_with_label(gslist, "Short Text");
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	cfgref = OptWinCfgRefAppend(
	    optwin, -1, w,
	    YC_CFG_PARM_MIXER_MIXER_SCALE_LABEL_TYPE
	);
	cfgref->radio_value = YMIXER_MIXER_SCALE_LABEL_TYPE_CONICAL_NAME;
	gslist = gtk_radio_button_group(GTK_RADIO_BUTTON(w));

	/* Picture radio button */
	w = gtk_radio_button_new_with_label(gslist, "Pictures");
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	cfgref = OptWinCfgRefAppend(
	    optwin, -1, w,
	    YC_CFG_PARM_MIXER_MIXER_SCALE_LABEL_TYPE
	);
	cfgref->radio_value = YMIXER_MIXER_SCALE_LABEL_TYPE_ICON;


	/* Show */
	w = gtk_frame_new("Show");
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, TRUE, 0);
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_widget_show(w);
	parent3 = w;

	w = gtk_hbox_new(FALSE, border_major);
	gtk_container_add(GTK_CONTAINER(parent3), w);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_widget_show(w);
	parent3 = w;

	/* Show value toggle */
	w = gtk_check_button_new_with_label("Value");
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	cfgref = OptWinCfgRefAppend(
	    optwin, -1, w,
	    YC_CFG_PARM_MIXER_SHOW_MIXER_VALUE
	);


	w = gtk_hbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	/* Colors */
	w = gtk_frame_new("Colors");
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, TRUE, 0);
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_widget_show(w);
	parent3 = w;

	w = gtk_hbox_new(FALSE, border_major);
	gtk_container_add(GTK_CONTAINER(parent3), w);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_widget_show(w);
	parent3 = w;

	/* Color buttons */
	cfgref = YCOptWinColorButtonCreate(
	    optwin, "Foreground:", -1,
	    (gint)YC_CFG_PARM_MIXER_COLOR_FG, YC_CFG_PARM_MIXER_COLOR_FG,
	    parent3, YCOptWinColorCB
	);
	cfgref = YCOptWinColorButtonCreate(
	    optwin, "Background:", -1,
	    (gint)YC_CFG_PARM_MIXER_COLOR_BG, YC_CFG_PARM_MIXER_COLOR_BG,
	    parent3, YCOptWinColorCB
	);
	cfgref = YCOptWinColorButtonCreate(
	    optwin, "Background Prelight:", -1,
	    (gint)YC_CFG_PARM_MIXER_COLOR_BG_PRELIGHT,
	    YC_CFG_PARM_MIXER_COLOR_BG_PRELIGHT,
	    parent3, YCOptWinColorCB
	);
	cfgref = YCOptWinColorButtonCreate(
	    optwin, "Text:", -1,
	    (gint)YC_CFG_PARM_MIXER_COLOR_TEXT, YC_CFG_PARM_MIXER_COLOR_TEXT,
	    parent3, YCOptWinColorCB
	);
	cfgref = YCOptWinColorButtonCreate(
	    optwin, "Base:", -1,
	    (gint)YC_CFG_PARM_MIXER_COLOR_BASE, YC_CFG_PARM_MIXER_COLOR_BASE,
	    parent3, YCOptWinColorCB
	);

	/* Use default toggle */
	w = gtk_check_button_new_with_label("Use Default");
	gtk_box_pack_end(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "toggled",
	    YCOptWinMixerScaleColorUseDefaultToggleCB, optwin
	);
	gtk_widget_show(w);
	cfgref = OptWinCfgRefAppend(
	    optwin, -1, w,
	    YC_CFG_PARM_MIXER_COLOR_USE_DEFAULT
	);


	/* Arrangement frame */
	w = gtk_frame_new(
#ifdef PROG_LANGUAGE_ENGLISH
	    "Arrangement"
#endif
#ifdef PROG_LANGUAGE_SPANISH
	    "El Arreglo"
#endif
#ifdef PROG_LANGUAGE_FRENCH
	    "Arrangement"
#endif
	);
	gtk_box_pack_start(GTK_BOX(parent), w, TRUE, TRUE, 0);
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_vbox_new(FALSE, border_minor);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_widget_show(w);
	parent2 = w;


	/* Create contents list stack list */
	slist = StackListNew(
	    parent2,
#ifdef PROG_LANGUAGE_ENGLISH
	    "Available", "Current"
#endif
#ifdef PROG_LANGUAGE_SPANISH
	    "Disponible", "La Corriente"
#endif
#ifdef PROG_LANGUAGE_FRENCH
	    "Disponible", "Courant"
#endif
	);
	StackListMap(slist);
	cfgref = OptWinCfgRefAppend(
	    optwin, -1, NULL,
	    YC_CFG_PARM_MIXER_MIXER_SCALES
	);
	cfgref->slist = slist;

	for(i = 0; i < YTotalMixers; i++)
	{
	    cstrptr = mixer_name[i];
	    icon_data = YMixerScaleGetXpmDataFromMixerName(cstrptr);
	    StackListAppend(
		slist,
		cstrptr,	/* Name */
		cstrptr,	/* Description */
		icon_data,	/* Icon data */
		NULL,		/* No client data */
		i + YMixerCodeBaseOffset,	/* Mixer code as the ID */
		FALSE, FALSE	/* No multiple, no stay on target */
	    );
	}
	StackListUpdateMenus(slist);

}


/*
 *	DSPView destroy callback.
 */
static void YCOptWinDSPViewDestroyCB(gpointer cfgref, gpointer data)
{
/* When core structure is deleted, this will be deleted.
	DSPViewDelete(DSPVIEW(data)); */
}
/*
 *	DSPView display style callbacks.
 */
static void YCOptWinDSPViewDisplayStyleWaveCB(
	GtkWidget *widget, gpointer data
)
{
	DSPViewSetDisplayStyle(
	    YC_CORE(data)->optwin_dspview, DSPVIEW_DISPLAY_STYLE_WAVE
	);
}
static void YCOptWinDSPViewDisplayStyleSineGraphCB(
	GtkWidget *widget, gpointer data
)
{
	DSPViewSetDisplayStyle(
	    YC_CORE(data)->optwin_dspview, DSPVIEW_DISPLAY_STYLE_SGRAPH
	);
}
static void YCOptWinDSPViewDisplayStyleAnalyzerCB(
	GtkWidget *widget, gpointer data
)
{
	DSPViewSetDisplayStyle(
	    YC_CORE(data)->optwin_dspview, DSPVIEW_DISPLAY_STYLE_ANALYZER
	);
}
static void YCOptWinDSPViewDisplayStyleLighteningCB(
	GtkWidget *widget, gpointer data
)
{
	DSPViewSetDisplayStyle(
	    YC_CORE(data)->optwin_dspview, DSPVIEW_DISPLAY_STYLE_LIGHTENING
	);
}
static void YCOptWinDSPViewDisplayStylePlasmaCB(
	GtkWidget *widget, gpointer data
)
{
	DSPViewSetDisplayStyle(
	    YC_CORE(data)->optwin_dspview, DSPVIEW_DISPLAY_STYLE_PLASMA
	);
}
static void YCOptWinDSPViewDisplayStyleNovaCB(
	GtkWidget *widget, gpointer data
)
{
	DSPViewSetDisplayStyle(
	    YC_CORE(data)->optwin_dspview, DSPVIEW_DISPLAY_STYLE_NOVA
	);
}
/*
 *	DSPView merge mode callbacks.
 */
static void YCOptWinDSPViewMergeModeSetCB(
	GtkWidget *widget, gpointer data
)
{
	DSPViewSetMergeMode(
	    YC_CORE(data)->optwin_dspview, DSPVIEW_MERGE_MODE_SET
	);
}
static void YCOptWinDSPViewMergeModeSetAlphaCB(
	GtkWidget *widget, gpointer data
)
{
	DSPViewSetMergeMode(
	    YC_CORE(data)->optwin_dspview, DSPVIEW_MERGE_MODE_SET_ALPHA
	);
}
static void YCOptWinDSPViewMergeModeAddCB(
	GtkWidget *widget, gpointer data
)
{
	DSPViewSetMergeMode(
	    YC_CORE(data)->optwin_dspview, DSPVIEW_MERGE_MODE_ADD
	);
}
static void YCOptWinDSPViewMergeModeSubtractCB(
	GtkWidget *widget, gpointer data
)
{
	DSPViewSetMergeMode(
	    YC_CORE(data)->optwin_dspview, DSPVIEW_MERGE_MODE_SUBTRACT
	);
}
static void YCOptWinDSPViewMergeModeBlendCB(
	GtkWidget *widget, gpointer data
)
{
	DSPViewSetMergeMode(
	    YC_CORE(data)->optwin_dspview, DSPVIEW_MERGE_MODE_BLEND
	);
}
static void YCOptWinDSPViewMergeModeMultiplyCB(
	GtkWidget *widget, gpointer data
)
{
	DSPViewSetMergeMode(
	    YC_CORE(data)->optwin_dspview, DSPVIEW_MERGE_MODE_MULTIPLY
	);
}
/*
 *	Color selection dialog color changed signal callback.
 */
static void YCOptWinDSPViewColorChangedCB(
	gpointer data, csd_color_struct *c
)
{
	GdkColor tc;
	GdkWindow *window;
	GtkWidget *w;
	dspview_struct *dv;
	optwin_cfgref_struct *cfgref = OPTWIN_CFGREF(data);
	if((c == NULL) || (cfgref == NULL))
	    return;

	tc.red		= (gushort)-1 * c->r;
	tc.green	= (gushort)-1 * c->g;
	tc.blue		= (gushort)-1 * c->b;

	/* Reallocate the cfgref's color */
	OptWinCfgRefSetColor(cfgref, &tc, &tc);

	w = cfgref->w;
	window = (w != NULL) ? w->window : NULL;
	if(window != NULL)
	{
	    gdk_window_set_background(window, &cfgref->fg_color);
	    gdk_window_clear(window);
	}

	dv = DSPVIEW(cfgref->client_data);
	if(dv != NULL)
	{
	    guint32 a = (dv->dsp_color & 0xff000000);
	    guint32 v = (
		((guint32)(0xff * c->r) << 0) |
		((guint32)(0xff * c->g) << 8) |
		((guint32)(0xff * c->b) << 16) |
		a
	    );
	    DSPViewSetDSPColor(dv, v);
	}
}
/*
 *      DSPView color button "clicked" signal callback.
 */
static void YCOptWinDSPViewColorCB(
	GtkWidget *widget, gpointer data
)
{
	gboolean status;
	optwin_cfgref_struct *cfgref = OPTWIN_CFGREF(data);
	GdkColor *c = &cfgref->fg_color;
	GtkWidget *w = cfgref->w;
	GtkWidget *toplevel = gtk_widget_get_toplevel(w);
	csd_color_struct cs, *cr;

	if(CSDIsQuery())
	    return;

	cs.r = (gfloat)c->red	/ (gfloat)((gushort)-1);
	cs.g = (gfloat)c->green	/ (gfloat)((gushort)-1);
	cs.b = (gfloat)c->blue	/ (gfloat)((gushort)-1);

	CSDSetTransientFor(toplevel);
	status = CSDGetResponse(
	    "DSPView Color",
	    "Select", "Cancel",
	    &cs, &cr,
	    cfgref, YCOptWinDSPViewColorChangedCB
	);
	CSDSetTransientFor(NULL);

	/* If user canceled then revert to original color */
	if(!status)
	    YCOptWinDSPViewColorChangedCB(cfgref, &cs);
}
/*
 *	Transparency adjustment "value_changed" signal callback.
 */
static void YCOptWinDSPViewTransparencyChangedCB(
	GtkAdjustment *adj, gpointer data
)
{
	guint32 a, v;
	dspview_struct *dv;
	optwin_cfgref_struct *cfgref = OPTWIN_CFGREF(data);
	if((adj == NULL) || (cfgref == NULL))
	    return;

	dv = DSPVIEW(cfgref->client_data);
	if(dv == NULL)
	    return;

	a = (guint32)(adj->value * 0xff);
	v = dv->dsp_color;
	DSPViewSetDSPColor(
	    dv,
	    (guint32)(
		(v & 0x00ffffff) |
		(a << 24)
	    )
	);
}
/*
 *	Fade coefficient adjustment "value_changed" signal callback.
 */
static void YCOptWinDSPViewFadeCoeffChangedCB(
	GtkAdjustment *adj, gpointer data
)
{
	dspview_struct *dv;
	optwin_cfgref_struct *cfgref = OPTWIN_CFGREF(data);
	if((adj == NULL) || (cfgref == NULL))
	    return;

	dv = DSPVIEW(cfgref->client_data);
	if(dv == NULL)
	    return;

	DSPViewSetFadeCoeff(dv, adj->value);
}

/*
 *	Load images callback.
 */
static void YCOptWinDisplayLoadImagesCB(GtkWidget *widget, gpointer data)
{
	GtkEntry *entry;
	optwin_struct *optwin;
	optwin_cfgref_struct *cfgref[2];
	yc_core_struct *core_ptr = YC_CORE(data);
	if((widget == NULL) || (core_ptr == NULL))
	    return;

	if(!GTK_IS_ENTRY(widget))
	    return;
	else
	    entry = GTK_ENTRY(widget);

	optwin = core_ptr->optwin;
	if(optwin == NULL)
	    return;

	cfgref[0] = OptWinCfgRefMatchID(
	    optwin,
	    (gint)YC_CFG_PARM_DSPVIEW_FILE_IMAGE_BG_LARGE,
	    NULL
	);
	cfgref[1] = OptWinCfgRefMatchID(
	    optwin,
	    (gint)YC_CFG_PARM_DSPVIEW_FILE_IMAGE_BG_SMALL,
	    NULL
	);


/*
	YCImgLoad(
	yc_core_struct *core_ptr,
	const gchar *large_img_path,
	gint large_img_fit_method,
	const gchar *small_img_path,
	gint small_img_fit_method
);
 */


}

/*
 *      Create Display page.
 */
static void YCOptWinCreatePageDisplay(
	yc_core_struct *core_ptr, optwin_struct *optwin,
	GtkWidget *parent
)
{
	gint border_minor = 2, border_major = 5;
	gpointer entry_rtn, browse_rtn;
	GSList *group;
	GtkAdjustment *adj;
	GtkWidget *w, *parent2, *parent3, *parent4, *parent5;
	optwin_cfgref_struct *cfgref;
	dspview_struct *dv = NULL;
	yc_core_struct *cp = YC_CORE(core_ptr);
	gint display_style = YC_GET_CFG_I(
	    cp, YC_CFG_PARM_DSPVIEW_DISPLAY_STYLE
	);
	gint merge_mode = YC_GET_CFG_I(
	    cp, YC_CFG_PARM_DSPVIEW_MERGE_MODE
	);
	cfg_color_struct *color = YC_GET_CFG_COLOR(
	    cp, YC_CFG_PARM_DSPVIEW_COLOR
	);
	gfloat transparency = YC_GET_CFG_F(
	    cp, YC_CFG_PARM_DSPVIEW_TRANSPARENCY
	);


	/* DSPView frame */
	w = gtk_frame_new("DSPView");
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_vbox_new(FALSE, border_major);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_widget_show(w);
	parent2 = w;

	/* First row hbox displaying display style radios and DSPView */
	w = gtk_hbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;

	w = gtk_label_new("Style:");
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	/* Display style radio buttons */
	w = gtk_radio_button_new_with_label(NULL, "Wave");
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "toggled",
	    YCOptWinDSPViewDisplayStyleWaveCB, cp
	);
	GUISetWidgetTip(
	    w,
"Wave is a simple and literal amplitude display of the Sound's volume"
	);
	gtk_widget_show(w);
	group = gtk_radio_button_group(GTK_RADIO_BUTTON(w));
	cfgref = OptWinCfgRefAppend(
	    optwin, -1, w,
	    YC_CFG_PARM_DSPVIEW_DISPLAY_STYLE
	);
	cfgref->radio_value = DSPVIEW_DISPLAY_STYLE_WAVE;

	w = gtk_radio_button_new_with_label(group, "Sine Graph");
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "toggled",
	    YCOptWinDSPViewDisplayStyleSineGraphCB, cp
	);
	GUISetWidgetTip(
	    w,
"Sine Graph is a simple plotted display of the Sound using the Sine\
 equation"
	);
	gtk_widget_show(w);
	group = gtk_radio_button_group(GTK_RADIO_BUTTON(w));
	cfgref = OptWinCfgRefAppend(
	    optwin, -1, w,
	    YC_CFG_PARM_DSPVIEW_DISPLAY_STYLE
	);
	cfgref->radio_value = DSPVIEW_DISPLAY_STYLE_SGRAPH;

	w = gtk_radio_button_new_with_label(group, "Analyzer");
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "toggled",
	    YCOptWinDSPViewDisplayStyleAnalyzerCB, cp
	);
	GUISetWidgetTip(
	    w,
"Analyzer is a series of bars that increase in height depending on\
 the characteristic of the Sound"
	);
	gtk_widget_show(w);
	group = gtk_radio_button_group(GTK_RADIO_BUTTON(w));
	cfgref = OptWinCfgRefAppend(
	    optwin, -1, w,
	    YC_CFG_PARM_DSPVIEW_DISPLAY_STYLE
	);
	cfgref->radio_value = DSPVIEW_DISPLAY_STYLE_ANALYZER;

	w = gtk_radio_button_new_with_label(group, "Lightening");
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "toggled",
	    YCOptWinDSPViewDisplayStyleLighteningCB, cp
	);
	GUISetWidgetTip(
	    w,
"Lightening simulates a lightening bolt that thrashes direction\
 depending on the characteristic of the Sound"
	);
	gtk_widget_show(w);
	group = gtk_radio_button_group(GTK_RADIO_BUTTON(w));
	cfgref = OptWinCfgRefAppend(
	    optwin, -1, w,
	    YC_CFG_PARM_DSPVIEW_DISPLAY_STYLE
	);
	cfgref->radio_value = DSPVIEW_DISPLAY_STYLE_LIGHTENING;

	w = gtk_radio_button_new_with_label(group, "Plasma");
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "toggled",
	    YCOptWinDSPViewDisplayStylePlasmaCB, cp
	);
	GUISetWidgetTip(
	    w,
"Plasma is a series of blob-like particals that distort in shape\
 depending on the characteristic of the Sound"
	);
	gtk_widget_show(w);
	group = gtk_radio_button_group(GTK_RADIO_BUTTON(w));
	cfgref = OptWinCfgRefAppend(
	    optwin, -1, w,
	    YC_CFG_PARM_DSPVIEW_DISPLAY_STYLE
	);
	cfgref->radio_value = DSPVIEW_DISPLAY_STYLE_PLASMA;

	w = gtk_radio_button_new_with_label(group, "Nova");
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "toggled",
	    YCOptWinDSPViewDisplayStyleNovaCB, cp
	);
	GUISetWidgetTip(
	    w,
"Nova simulates an exploding star as seen through a hazy event\
 horizon.  The star jitters depending on the characteristic of\
 the Sound (Color and Fade values are very sensitive with this\
 display style)"
	);
	gtk_widget_show(w);
	cfgref = OptWinCfgRefAppend(
	    optwin, -1, w,
	    YC_CFG_PARM_DSPVIEW_DISPLAY_STYLE
	);
	cfgref->radio_value = DSPVIEW_DISPLAY_STYLE_NOVA;


	/* DSPView */
	if(cp->optwin_dspview == NULL)
	{
	    guint32 v = DSPVIEW_DEF_DSP_COLOR;

	    if(color != NULL)
		v = (guint32)(
		    ((guint32)(color->r * 0xff) << 0) |
		    ((guint32)(color->g * 0xff) << 8) |
		    ((guint32)(color->b * 0xff) << 16) |
		    ((guint32)(transparency * 0xff) << 24)
		);

	    w = gtk_vbox_new(TRUE, 0);
	    gtk_box_pack_end(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent4 = w;

	    cp->optwin_dspview = dv = DSPViewNew(
		parent4,
		cp->dsp_bg_img_width + 2, cp->dsp_bg_img_height + 2
	    );
	    DSPViewSetBG(
		dv, cp->dsp_bg_img,
		cp->dsp_bg_img_width, cp->dsp_bg_img_height,
		cp->dsp_bg_img_bpl, 4
	    );
	    DSPViewSetDisplayStyle(dv, display_style);
	    DSPViewSetMergeMode(dv, merge_mode);
	    DSPViewSetDSPColor(dv, v);
	    DSPViewMap(dv);
	}


	/* Second row hbox displaying merge mode radios */
	w = gtk_hbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;

	w = gtk_label_new("Merge Mode:");
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	/* Merge mode radio buttons */
	w = gtk_radio_button_new_with_label(NULL, "Set");
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "toggled",
	    YCOptWinDSPViewMergeModeSetCB, cp
	);
	gtk_widget_show(w);
	group = gtk_radio_button_group(GTK_RADIO_BUTTON(w));
	cfgref = OptWinCfgRefAppend(
	    optwin, -1, w,
	    YC_CFG_PARM_DSPVIEW_MERGE_MODE
	);
	cfgref->radio_value = DSPVIEW_MERGE_MODE_SET;

	w = gtk_radio_button_new_with_label(group, "Set Alpha");
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "toggled",
	    YCOptWinDSPViewMergeModeSetAlphaCB, cp
	);
	gtk_widget_show(w);
	group = gtk_radio_button_group(GTK_RADIO_BUTTON(w));
	cfgref = OptWinCfgRefAppend(
	    optwin, -1, w,
	    YC_CFG_PARM_DSPVIEW_MERGE_MODE
	);
	cfgref->radio_value = DSPVIEW_MERGE_MODE_SET_ALPHA;

	w = gtk_radio_button_new_with_label(group, "Add");
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "toggled",
	    YCOptWinDSPViewMergeModeAddCB, cp
	);
	gtk_widget_show(w);
	group = gtk_radio_button_group(GTK_RADIO_BUTTON(w));
	cfgref = OptWinCfgRefAppend(
	    optwin, -1, w,
	    YC_CFG_PARM_DSPVIEW_MERGE_MODE
	);
	cfgref->radio_value = DSPVIEW_MERGE_MODE_ADD;

	w = gtk_radio_button_new_with_label(group, "Subtract");
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "toggled",
	    YCOptWinDSPViewMergeModeSubtractCB, cp
	);
	gtk_widget_show(w);
	group = gtk_radio_button_group(GTK_RADIO_BUTTON(w));
	cfgref = OptWinCfgRefAppend(
	    optwin, -1, w,
	    YC_CFG_PARM_DSPVIEW_MERGE_MODE
	);
	cfgref->radio_value = DSPVIEW_MERGE_MODE_SUBTRACT;

	w = gtk_radio_button_new_with_label(group, "Blend");
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "toggled",
	    YCOptWinDSPViewMergeModeBlendCB, cp
	);
	gtk_widget_show(w);
	group = gtk_radio_button_group(GTK_RADIO_BUTTON(w));
	cfgref = OptWinCfgRefAppend(
	    optwin, -1, w,
	    YC_CFG_PARM_DSPVIEW_MERGE_MODE
	);
	cfgref->radio_value = DSPVIEW_MERGE_MODE_BLEND;

	w = gtk_radio_button_new_with_label(group, "Multiply");
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "toggled",
	    YCOptWinDSPViewMergeModeMultiplyCB, cp
	);
	gtk_widget_show(w);
	group = gtk_radio_button_group(GTK_RADIO_BUTTON(w));
	cfgref = OptWinCfgRefAppend(
	    optwin, -1, w,
	    YC_CFG_PARM_DSPVIEW_MERGE_MODE
	);
	cfgref->radio_value = DSPVIEW_MERGE_MODE_MULTIPLY;


	/* Third row hbox displaying color and transparency */
	w = gtk_hbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;

	/* DSPView color */
	cfgref = YCOptWinColorButtonCreate(
	    optwin, "Color:", -1, -1, YC_CFG_PARM_DSPVIEW_COLOR,
	    parent3, YCOptWinDSPViewColorCB
	);
/* Note, this cfgref for YC_CFG_PARM_DSPVIEW_COLOR holds the destroy
 * callback for the dspview.
 */
	cfgref->client_data = dv;
	cfgref->destroy_cb = YCOptWinDSPViewDestroyCB;


	/* Hbox for transparency label and scale */
	w = gtk_hbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent4 = w;

	w = gtk_vbox_new(TRUE, 0);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent5 = w;
	w = gtk_label_new("Transparency:");
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	/* Transparency scale */
	adj = (GtkAdjustment *)gtk_adjustment_new(
	    0.0f, 0.0f, 1.0f,
	    0.025f, 0.05f, 0.0f
	);
	w = gtk_hscale_new(adj);
	gtk_widget_set_usize(w, 100, -1);
	gtk_scale_set_draw_value(GTK_SCALE(w), FALSE);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	GUISetWidgetTip(
	    w,
"Adjust the transparency of the color (not available on all\
 display styles)"
	);
	gtk_widget_show(w);
	cfgref = OptWinCfgRefAppend(
	    optwin, -1, w,
	    YC_CFG_PARM_DSPVIEW_TRANSPARENCY
	);
	cfgref->client_data = dv;
	gtk_signal_connect(
	    GTK_OBJECT(adj), "value_changed",
	    GTK_SIGNAL_FUNC(YCOptWinDSPViewTransparencyChangedCB), cfgref
	);


	/* Hbox for fade coefficient label and scale */
	w = gtk_hbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent4 = w;

	w = gtk_vbox_new(TRUE, 0);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent5 = w;
	w = gtk_label_new("Fade:");
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	/* Fade coefficient scale */
	adj = (GtkAdjustment *)gtk_adjustment_new(
	    0.0f, 0.0f, 1.0f,
	    0.025f, 0.05f, 0.0f
	);
	w = gtk_hscale_new(adj);
	gtk_widget_set_usize(w, 100, -1);
	gtk_scale_set_draw_value(GTK_SCALE(w), FALSE);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	GUISetWidgetTip(
	    w,
"Adjust the fading power (has different affect depending on the\
 display style)"
	);
	gtk_widget_show(w);
	cfgref = OptWinCfgRefAppend(
	    optwin, -1, w,
	    YC_CFG_PARM_DSPVIEW_FADE_COEFF
	);
	cfgref->client_data = dv;
	gtk_signal_connect(
	    GTK_OBJECT(adj), "value_changed",
	    GTK_SIGNAL_FUNC(YCOptWinDSPViewFadeCoeffChangedCB), cfgref
	);


	/* Vbox for large background image prompts */
	w = gtk_vbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;

	/* Background image prompt */
	w = GUIPromptBarWithBrowse(
	    NULL, "Large Background Image:",
	    NULL, &entry_rtn, &browse_rtn,
	    NULL, NULL
	);
	gtk_box_pack_start(GTK_BOX(parent3), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	w = (GtkWidget *)entry_rtn;
	GUISetWidgetTip(
	    w,
"Enter the path to the large background image file (size 80x32)\
 or leave blank to use the default image"
	);
	cfgref = OptWinCfgRefAppend(
	    optwin, (gint)YC_CFG_PARM_DSPVIEW_FILE_IMAGE_BG_LARGE, w,
	    YC_CFG_PARM_DSPVIEW_FILE_IMAGE_BG_LARGE
	);
	cfgref->client_data = dv;
	w = (GtkWidget *)browse_rtn;
	GUISetWidgetTip(w, "Browse");
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(YCOptWinBrowsePathCB), cfgref
	);

	/* Fit method radios */
	w = gtk_hbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent4 = w;

	w = gtk_label_new("Fit Method:");
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	w = gtk_radio_button_new_with_label(NULL, "Tile");
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	group = gtk_radio_button_group(GTK_RADIO_BUTTON(w));
	cfgref = OptWinCfgRefAppend(
	    optwin, -1, w,
	    YC_CFG_PARM_DSPVIEW_IMAGE_BG_LARGE_FIT_METHOD
	);
	cfgref->radio_value = YC_IMG_FIT_METHOD_TILE;

	w = gtk_radio_button_new_with_label(group, "Resize");
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	group = gtk_radio_button_group(GTK_RADIO_BUTTON(w));
	cfgref = OptWinCfgRefAppend(
	    optwin, -1, w,
	    YC_CFG_PARM_DSPVIEW_IMAGE_BG_LARGE_FIT_METHOD
	);
	cfgref->radio_value = YC_IMG_FIT_METHOD_SCALE;


	/* Vbox for small background image prompts */
	w = gtk_vbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;

	/* Background image prompt */
	w = GUIPromptBarWithBrowse(
	    NULL, "Small Background Image:",
	    NULL, &entry_rtn, &browse_rtn,
	    NULL, NULL
	);
	gtk_box_pack_start(GTK_BOX(parent3), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	w = (GtkWidget *)entry_rtn;
	GUISetWidgetTip(
	    w,
"Enter the path to the small background image file (size 50x20)\
 or leave blank to use the default image"
	);
	cfgref = OptWinCfgRefAppend(
	    optwin, (gint)YC_CFG_PARM_DSPVIEW_FILE_IMAGE_BG_SMALL, w,
	    YC_CFG_PARM_DSPVIEW_FILE_IMAGE_BG_SMALL
	);
	cfgref->client_data = dv;
	w = (GtkWidget *)browse_rtn;
	GUISetWidgetTip(w, "Browse");
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(YCOptWinBrowsePathCB), cfgref
	);

	/* Fit method radios */
	w = gtk_hbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent4 = w;

	w = gtk_label_new("Fit Method:");
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	w = gtk_radio_button_new_with_label(NULL, "Tile");
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	group = gtk_radio_button_group(GTK_RADIO_BUTTON(w));
	cfgref = OptWinCfgRefAppend(
	    optwin, -1, w,
	    YC_CFG_PARM_DSPVIEW_IMAGE_BG_SMALL_FIT_METHOD
	);
	cfgref->radio_value = YC_IMG_FIT_METHOD_TILE;

	w = gtk_radio_button_new_with_label(group, "Resize");
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	group = gtk_radio_button_group(GTK_RADIO_BUTTON(w));
	cfgref = OptWinCfgRefAppend(
	    optwin, -1, w,
	    YC_CFG_PARM_DSPVIEW_IMAGE_BG_SMALL_FIT_METHOD
	);
	cfgref->radio_value = YC_IMG_FIT_METHOD_SCALE;


	/* Audio display frame */
	w = gtk_frame_new("Audio Display");
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_vbox_new(FALSE, border_major);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_widget_show(w);
	parent2 = w;

	/* First row hbox displaying colors */
	w = gtk_hbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;

	/* Color buttons */
	cfgref = YCOptWinColorButtonCreate(
	    optwin, "Foreground Color:", -1, -1,
	    YC_CFG_PARM_AUDIO_DISPLAY_COLOR_FG,
	    parent3, YCOptWinColorCB
	);
	cfgref = YCOptWinColorButtonCreate(
	    optwin, "Foreground Color Dim:", -1, -1,
	    YC_CFG_PARM_AUDIO_DISPLAY_COLOR_FG_DIM,
	    parent3, YCOptWinColorCB
	);
	cfgref = YCOptWinColorButtonCreate(
	    optwin, "Background Color:", -1, -1,
	    YC_CFG_PARM_AUDIO_DISPLAY_COLOR_BG,
	    parent3, YCOptWinColorCB
	);


	/* Vbox for fonts */
	w = gtk_vbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;

	/* Font entry and buttons */
	cfgref = YCOptWinFontButtonCreate(
	    optwin, "Large Font:", NULL, -1,
	    YC_CFG_PARM_AUDIO_DISPLAY_FONT_LARGE,
	    parent3, YCOptWinFontCB
	);
	cfgref = YCOptWinFontButtonCreate(
	    optwin, "Medium Font:", NULL, -1,
	    YC_CFG_PARM_AUDIO_DISPLAY_FONT_MEDIUM,
	    parent3, YCOptWinFontCB
	);
	cfgref = YCOptWinFontButtonCreate(
	    optwin, "Small Font:", NULL, -1,
	    YC_CFG_PARM_AUDIO_DISPLAY_FONT_SMALL,
	    parent3, YCOptWinFontCB
	);
}


/*
 *      Create General page.
 */
static void YCOptWinCreatePageGeneral(
	yc_core_struct *core_ptr, optwin_struct *optwin,
	GtkWidget *parent
)
{
	gint border_minor = 2, border_major = 5;
	GtkWidget *w, *parent2, *parent3;
	optwin_cfgref_struct *cfgref;


	/* Startup with frame */
	w = gtk_frame_new(
#ifdef PROG_LANGUAGE_ENGLISH
	    "Startup With"
#endif
#ifdef PROG_LANGUAGE_SPANISH
	    "Arranque Con"
#endif
#ifdef PROG_LANGUAGE_FRENCH
	    "Mettre En Route Avec"
#endif
	);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_vbox_new(FALSE, border_minor);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_widget_show(w);
	parent2 = w;


	/* First row hbox */
	w = gtk_hbox_new(TRUE, 0);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;

	/* Console */
	w = (GtkWidget *)GUIToggleButtonPixmapLabelV(
	    (u_int8_t **)icon_yconsole_48x48_xpm,
	    "Console",
	    NULL
	);
	gtk_widget_set_usize(w, 80, 80);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	cfgref = OptWinCfgRefAppend(
	    optwin, -1, w,
	    YC_CFG_PARM_STARTUP_CONSOLE
	);

	/* Audio */
	w = (GtkWidget *)GUIToggleButtonPixmapLabelV(
	    (u_int8_t **)icon_yaudio_48x48_xpm,
	    "Audio",
	    NULL
	);
	gtk_widget_set_usize(w, 80, 80);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	cfgref = OptWinCfgRefAppend(
	    optwin, -1, w,
	    YC_CFG_PARM_STARTUP_AUDIO
	);

	/* Mixer */
	w = (GtkWidget *)GUIToggleButtonPixmapLabelV(
	    (u_int8_t **)icon_ymixer_48x48_xpm,
	    "Mixer",
	    NULL
	);
	gtk_widget_set_usize(w, 80, 80);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	cfgref = OptWinCfgRefAppend(
	    optwin, -1, w,
	    YC_CFG_PARM_STARTUP_MIXER
	);

	/* Player */
	w = (GtkWidget *)GUIToggleButtonPixmapLabelV(
	    (u_int8_t **)icon_yplayer_48x48_xpm,
	    "Player",
	    NULL
	);
	gtk_widget_set_usize(w, 80, 80);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	cfgref = OptWinCfgRefAppend(
	    optwin, -1, w,
	    YC_CFG_PARM_STARTUP_PLAYER
	);


	/* Startup location frame */
	w = gtk_frame_new(
#ifdef PROG_LANGUAGE_ENGLISH
	    "Startup Location"
#endif
#ifdef PROG_LANGUAGE_SPANISH
	    "La Ubicacin De Startup"
#endif
#ifdef PROG_LANGUAGE_FRENCH
	    "Emplacement De Dmarrage"
#endif
	);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_vbox_new(FALSE, border_minor);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_widget_show(w);
	parent2 = w;

	/* Start up location entry */
	w = gtk_entry_new();
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	GUISetWidgetTip(
	    w,
"Enter the address of the Y server that you want to connect to\
 each time you run this application (addresses specified from the\
 command line will override this value)"
	);
	gtk_widget_show(w);
	cfgref = OptWinCfgRefAppend(
	    optwin, -1, w,
	    YC_CFG_PARM_STARTUP_LOCATION
	);
}


/*
 *	Creates a new options window.
 */
optwin_struct *YCOptWinNew(gpointer core_ptr)
{
	gint border_minor = 2, border_major = 5;
	GtkWidget *w, *parent, *parent2, *lbox;
	optwin_struct *optwin = OptWinNew(
	    core_ptr,
#ifdef PROG_LANGUAGE_ENGLISH
	    "Options",
#endif
#ifdef PROG_LANGUAGE_SPANISH
	    "Las Opciones",
#endif
#ifdef PROG_LANGUAGE_FRENCH
	    "Options",
#endif
	    (guint8 **)icon_options2_48x48_xpm);
	if(optwin == NULL)
	    return(optwin);


	/* Create General page */
	w = gtk_label_new(
#ifdef PROG_LANGUAGE_ENGLISH
	    "General"
#endif
#ifdef PROG_LANGUAGE_SPANISH
	    "General"
#endif
#ifdef PROG_LANGUAGE_FRENCH
	    "Gnral"
#endif
	);
	parent = OptWinPageAppend(optwin, w);
	if(parent != NULL)
	{
	    gtk_container_border_width(GTK_CONTAINER(parent), border_major);
	    gtk_box_set_spacing(GTK_BOX(parent), border_major);
	    parent2 = parent;

	    YCOptWinCreatePageGeneral(YC_CORE(core_ptr), optwin, parent2);
	}

	/* Create Display page */
	w = gtk_label_new(
#ifdef PROG_LANGUAGE_ENGLISH
	    "Display"
#endif
#ifdef PROG_LANGUAGE_SPANISH
	    "La Grfica"
#endif
#ifdef PROG_LANGUAGE_FRENCH
	    "Les Graphiques"
#endif
	);
	parent = OptWinPageAppend(optwin, w);
	if(parent != NULL)
	{
	    gtk_container_border_width(GTK_CONTAINER(parent), border_major);
	    gtk_box_set_spacing(GTK_BOX(parent), border_major);
	    parent2 = parent;

	    YCOptWinCreatePageDisplay(YC_CORE(core_ptr), optwin, parent2);
	}

	/* Create YMixer page */
	lbox = gtk_hbox_new(FALSE, border_minor);
	gtk_widget_show(lbox);
	w = GUICreateMenuItemIcon((guint8 **)icon_ymixer_20x20_xpm);
	gtk_box_pack_start(GTK_BOX(lbox), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	w = gtk_label_new("YMixer");
	gtk_box_pack_start(GTK_BOX(lbox), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent = OptWinPageAppend(optwin, lbox);
	if(parent != NULL)
	{
	    gtk_container_border_width(GTK_CONTAINER(parent), border_major);
	    gtk_box_set_spacing(GTK_BOX(parent), border_major);
	    parent2 = parent;

	    YCOptWinCreatePageYMixer(YC_CORE(core_ptr), optwin, parent2);
	}

	/* Create YPlayer page */
	lbox = gtk_hbox_new(FALSE, border_minor);
	gtk_widget_show(lbox);
	w = GUICreateMenuItemIcon((guint8 **)icon_yplayer_20x20_xpm);
	gtk_box_pack_start(GTK_BOX(lbox), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	w = gtk_label_new("YPlayer");
	gtk_box_pack_start(GTK_BOX(lbox), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent = OptWinPageAppend(optwin, lbox);
	if(parent != NULL)
	{
	    gtk_container_border_width(GTK_CONTAINER(parent), border_major);
	    gtk_box_set_spacing(GTK_BOX(parent), border_major);
	    parent2 = parent;

	    YCOptWinCreatePageYPlayer(YC_CORE(core_ptr), optwin, parent2);
	}


	return(optwin);
}
