#include <stdlib.h>
#include <string.h>

#include <Y2/Y.h>
#include <Y2/Ylib.h>
#include <Y2/Ymixercodes.h>

#include <Imlib.h>

#include <gtk/gtk.h>
#include <gdk/gdkx.h>
#include <gdk/gdkrgb.h>

#include "../include/string.h"
#include "../include/fio.h"
#include "../include/disk.h"
#include "../include/imlibio.h"

#include "guiutils.h"
#include "cdialog.h"
#include "csd.h"
#include "dspview.h"

#include "images/icon_y_48x48.xpm"
#include "images/icon_sound_20x20.xpm"
#include "images/icon_speaker_left_20x20.xpm"
#include "images/icon_speaker_right_20x20.xpm"

#include "images/dspview_bg1.xpm"
#include "images/dspview_bg2.xpm"
#include "images/dspview_bg3.xpm"
#include "images/dspview_bg4.xpm"
#include "images/dspview_bg5.xpm"




typedef struct {

	GtkWidget *toplevel;
	GtkWidget *audio_label;
	GtkAdjustment	*transparency_adj,
			*fade_speed_adj,
			*volume_adj,
			*balance_adj;
	dspview_struct *dv;

	YConnection *con;
	guint8 *yshm_buf;
	gint yshm_buf_len;
	gint yshm_id;

	gint bg_num;
	guint8 *bg_img;
	gint bg_width, bg_height, bg_bpl, bg_bpp;

} win_struct;


static gint delete_cb(GtkWidget *widget, GdkEvent *event, gpointer data);

static void refresh_cb(GtkWidget *widget, gpointer data);

static void toggle_dspview_cb(GtkWidget *widget, gpointer data);

static void bg_image_previous_cb(GtkWidget *widget, gpointer data);
static void bg_image_next_cb(GtkWidget *widget, gpointer data);

static void toggle_display_style_wave(
	GtkWidget *widget, gpointer data
);
static void toggle_display_style_sgraph(
	GtkWidget *widget, gpointer data
);
static void toggle_display_style_analyzer(
	GtkWidget *widget, gpointer data
);
static void toggle_display_style_lightening(
	GtkWidget *widget, gpointer data
);
static void toggle_display_style_plasma(
	GtkWidget *widget, gpointer data
);
static void toggle_display_style_nova(
	GtkWidget *widget, gpointer data
);

static void toggle_merge_mode_set(
	GtkWidget *widget, gpointer data
);
static void toggle_merge_mode_set_alpha(
	GtkWidget *widget, gpointer data
);
static void toggle_merge_mode_add(
	GtkWidget *widget, gpointer data
);
static void toggle_merge_mode_blend(
	GtkWidget *widget, gpointer data
);
static void toggle_merge_mode_multiply(
	GtkWidget *widget, gpointer data
);

static void change_color_cb(GtkWidget *widget, gpointer data);
static void scale_transparency_cb(GtkAdjustment *adj, gpointer data);
static void scale_fade_speed_cb(GtkAdjustment *adj, gpointer data);
static void scale_volume_cb(GtkAdjustment *adj, gpointer data);
static void scale_balance_cb(GtkAdjustment *adj, gpointer data);

static gint timeout_cb(gpointer data);


static gpointer imlib_handle;


#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define ABSOLUTE(x)	(((x) < 0) ? ((x) * -1) : (x))


/*
 *	Window toplevel "delete_event" signal callback.
 */
static gint delete_cb(GtkWidget *widget, GdkEvent *event, gpointer data)
{
	win_struct *win = (win_struct *)data;
	if(win == NULL)
	    return(FALSE);

	YSHMCloseSound(
	    win->con,
	    win->yshm_buf
	);
	win->yshm_buf = NULL;
	win->yshm_buf_len = 0;
	win->yshm_id = -1;

	YCloseConnection(win->con, False);
	win->con = NULL;

	return(TRUE);
}


/*
 *	Refresh callback.
 */
static void refresh_cb(GtkWidget *widget, gpointer data)
{
	win_struct *win = (win_struct *)data;
	if(win == NULL)
	    return;

	if(win->con != NULL)
	{
	    YSHMCloseSound(win->con, win->yshm_buf);
	    win->yshm_buf = (guint8 *)YSHMOpenSound(
		win->con, &win->yshm_buf_len, &win->yshm_id
	    );
	}
}

/*
 *	Toggles the DSP view sensitive or insensitive.
 */
static void toggle_dspview_cb(GtkWidget *widget, gpointer data)
{
	GtkWidget *w;
	dspview_struct *dv;
	win_struct *win = (win_struct *)data;
	if((widget == NULL) || (win == NULL))
	    return;

	dv = win->dv;
	if(dv == NULL)
	    return;

	w = dv->view_da;
	gtk_widget_set_sensitive(
	    w,
	    GTK_TOGGLE_BUTTON(widget)->active
	);
}

#define DO_LOAD_BG_IMAGE_BY_NUM		\
{ \
 gchar **data = NULL; \
 switch(win->bg_num) \
 { \
  case 0: data = NULL; break; \
  case 1: data = dspview_bg1_xpm; break; \
  case 2: data = dspview_bg2_xpm; break; \
  case 3: data = dspview_bg3_xpm; break; \
  case 4: data = dspview_bg4_xpm; break; \
  case 5: data = dspview_bg5_xpm; break; \
 } \
\
 DSPViewSetBG(dv, NULL, 0, 0, 0, 0); \
 g_free(win->bg_img); \
 win->bg_img = NULL; \
\
 if(data != NULL) \
 { \
  win->bg_img = ImlibIOLoadXPMDataRGBA( \
   imlib_handle, (const gchar **)data, \
   &win->bg_width, &win->bg_height, &win->bg_bpl \
  ); \
  win->bg_bpp = 4; \
  DSPViewSetBG( \
   dv, win->bg_img, \
   win->bg_width, win->bg_height, \
   win->bg_bpl, win->bg_bpp \
  ); \
 } \
}

/*
 *	Previous background image callback
 */
static void bg_image_previous_cb(GtkWidget *widget, gpointer data)
{
	dspview_struct *dv;
	win_struct *win = (win_struct *)data;
	if((widget == NULL) || (win == NULL))
	    return;

	dv = win->dv;
	if(dv == NULL)
	    return;

	win->bg_num--;
	if(win->bg_num < 0)
	    win->bg_num = 5;

	DO_LOAD_BG_IMAGE_BY_NUM
}

/*
 *      Next background image callback
 */
static void bg_image_next_cb(GtkWidget *widget, gpointer data)
{
	dspview_struct *dv;
	win_struct *win = (win_struct *)data;
	if((widget == NULL) || (win == NULL))
	    return;

	dv = win->dv;
	if(dv == NULL)
	    return;

	win->bg_num++;
	if(win->bg_num > 5)
	    win->bg_num = 0;

	DO_LOAD_BG_IMAGE_BY_NUM
}



/*
 *	Display style toggle button "toggled" signal callbacks.
 */
static void toggle_display_style_wave(
	GtkWidget *widget, gpointer data
)
{
	win_struct *win = (win_struct *)data;
	if(win != NULL)
	    DSPViewSetDisplayStyle(win->dv, DSPVIEW_DISPLAY_STYLE_WAVE);
}
static void toggle_display_style_sgraph(
	GtkWidget *widget, gpointer data
)
{
	win_struct *win = (win_struct *)data;
	if(win != NULL)
	    DSPViewSetDisplayStyle(win->dv, DSPVIEW_DISPLAY_STYLE_SGRAPH);
}
static void toggle_display_style_analyzer(
	GtkWidget *widget, gpointer data
)
{
	win_struct *win = (win_struct *)data;
	if(win != NULL)
	    DSPViewSetDisplayStyle(win->dv, DSPVIEW_DISPLAY_STYLE_ANALYZER);
}
static void toggle_display_style_lightening(
	GtkWidget *widget, gpointer data
)
{
	win_struct *win = (win_struct *)data;
	if(win != NULL)
	    DSPViewSetDisplayStyle(win->dv, DSPVIEW_DISPLAY_STYLE_LIGHTENING);
}
static void toggle_display_style_plasma(
	GtkWidget *widget, gpointer data
)
{
	win_struct *win = (win_struct *)data;
	if(win != NULL)
	    DSPViewSetDisplayStyle(win->dv, DSPVIEW_DISPLAY_STYLE_PLASMA);
}
static void toggle_display_style_nova(
	GtkWidget *widget, gpointer data
)
{
	win_struct *win = (win_struct *)data;
	if(win != NULL)
	    DSPViewSetDisplayStyle(win->dv, DSPVIEW_DISPLAY_STYLE_NOVA);
}

/*
 *	Merge mode toggle button "toggled" signal callbacks.
 */
static void toggle_merge_mode_set(
	GtkWidget *widget, gpointer data
)
{
	win_struct *win = (win_struct *)data;
	if(win != NULL)
	    DSPViewSetMergeMode(win->dv, DSPVIEW_MERGE_MODE_SET);
}
static void toggle_merge_mode_set_alpha(
	GtkWidget *widget, gpointer data
)
{
	win_struct *win = (win_struct *)data;
	if(win != NULL)
	    DSPViewSetMergeMode(win->dv, DSPVIEW_MERGE_MODE_SET_ALPHA);
}
static void toggle_merge_mode_add(
	GtkWidget *widget, gpointer data
)
{
	win_struct *win = (win_struct *)data;
	if(win != NULL)
	    DSPViewSetMergeMode(win->dv, DSPVIEW_MERGE_MODE_ADD);
}
static void toggle_merge_mode_blend(
	GtkWidget *widget, gpointer data
)
{
	win_struct *win = (win_struct *)data;
	if(win != NULL)
	    DSPViewSetMergeMode(win->dv, DSPVIEW_MERGE_MODE_BLEND);
}
static void toggle_merge_mode_multiply(
	GtkWidget *widget, gpointer data
)
{
	win_struct *win = (win_struct *)data;
	if(win != NULL)
	    DSPViewSetMergeMode(win->dv, DSPVIEW_MERGE_MODE_MULTIPLY);
}

/* Color changed callback, used in change_color_cb(). */
static void color_changed_cb(gpointer data, csd_color_struct *tc)
{
	win_struct *win = (win_struct *)data;
	if((win != NULL) && (tc != NULL))
	{
	    guint32 dsp_color = (win->dv != NULL) ?
		win->dv->dsp_color : 0xffffffff;
	    dsp_color = (
		((guint32)(tc->r * 0xff) << 0) |
		((guint32)(tc->g * 0xff) << 8) |
		((guint32)(tc->b * 0xff) << 16) |
		(dsp_color & 0xff000000)
	    );
	    DSPViewSetDSPColor(win->dv, dsp_color);
	}
}

/*
 *      Change color button "clicked" signal callback.
 */
static void change_color_cb(GtkWidget *widget, gpointer data)
{
	csd_color_struct sc, *tc;
	guint32 dsp_color;
	win_struct *win = (win_struct *)data;
	if(win == NULL)
	    return;

	if(win->dv == NULL)
	    return;

	memset(&sc, 0x00, sizeof(csd_color_struct));
	dsp_color = win->dv->dsp_color;
	sc.a = (gdouble)((dsp_color & 0xff000000) >> 24) / (gdouble)0xff;
	sc.r = (gdouble)((dsp_color & 0x000000ff) >> 0) / (gdouble)0xff;
	sc.g = (gdouble)((dsp_color & 0x0000ff00) >> 8) / (gdouble)0xff;
	sc.b = (gdouble)((dsp_color & 0x00ff0000) >> 16) / (gdouble)0xff;

	CSDSetTransientFor(win->toplevel);
	if(!CSDGetResponse(
	    "Select DSP Color",
	    "Select", "Cancel",
	    &sc, &tc,
	    win, color_changed_cb
	))
	{
	    /* Canceled, restore original DSP color. */
	    DSPViewSetDSPColor(win->dv, dsp_color);
	}
	CSDSetTransientFor(NULL);
}

/*
 *	Transparency adjustment "value_changed" signal callback.
 */
static void scale_transparency_cb(GtkAdjustment *adj, gpointer data)
{
	guint8 a;
	guint32 rgb;
	win_struct *win = (win_struct *)data;
	if((adj == NULL) || (win == NULL))
	    return;

	if(win->dv == NULL)
	    return;

	a = (guint8)(adj->value * 0xff);
	rgb = (win->dv->dsp_color & 0x00ffffff);
	DSPViewSetDSPColor(win->dv, ((guint32)a << 24) | rgb);
}

/*
 *	Fade speed adjustment "value_changed" signal callback.
 */
static void scale_fade_speed_cb(GtkAdjustment *adj, gpointer data)
{
	win_struct *win = (win_struct *)data;
	if((adj == NULL) || (win == NULL))
	    return;

	if(win->dv == NULL)
	    return;

	DSPViewSetFadeCoeff(win->dv, adj->value);
}

/*
 *	Volume adjustment "value_changed" signal callback.
 */
static void scale_volume_cb(GtkAdjustment *adj, gpointer data)
{
	GtkAdjustment *volume_adj, *balance_adj;
	win_struct *win = (win_struct *)data;
	if((adj == NULL) || (win == NULL))
	    return;

	volume_adj = win->volume_adj;
	balance_adj = win->balance_adj;
	if((volume_adj != NULL) && (balance_adj != NULL) &&
	   (win->con != NULL)
	)
	{
	    Coefficient mv[2];

	    mv[0] = volume_adj->value *
		CLIP(-balance_adj->value + 1.0f, 0.0, 1.0);
	    mv[1] = volume_adj->value *
		CLIP(balance_adj->value + 1.0f, 0.0, 1.0);
	    YSetMixerChannel(
		win->con,
		YMixerCodeVolume,
		mv[0], mv[1]
	    );
	}
}

/*
 *	Balance adjustment "value_changed" signal callback.
 */
static void scale_balance_cb(GtkAdjustment *adj, gpointer data)
{
	scale_volume_cb(adj, data);
}


/*
 *	Main timeout callback.
 */
static gint timeout_cb(gpointer data)
{
	YConnection *con;
	YEvent yevent;
	dspview_struct *dv;
	win_struct *win = (win_struct *)data;
	if(win == NULL)
	    return(FALSE);

#define QUIT_GTK_MAIN_RETURN	\
{ \
 gtk_main_quit(); \
 return(FALSE); \
}

	con = win->con;
	if(con == NULL)
	    QUIT_GTK_MAIN_RETURN

	dv = win->dv;
	if(dv == NULL)
	    QUIT_GTK_MAIN_RETURN

	/* Manage Y events. */
	while(YGetNextEvent(win->con, &yevent, False) > 0)
	{
	    switch(yevent.type)
	    {
	      case YAudioChange:
		/* Need to reopen YSHM sound buffer when Audio
		 * mode changes.
		 */
		YSHMCloseSound(con, win->yshm_buf);
		win->yshm_buf = (guint8 *)YSHMOpenSound(
		    con, &win->yshm_buf_len, &win->yshm_id
		);
		/* Get audio stats and update displays. */
		if(1)
		{
		    YEventAudioStats stat_buf;
		    if(!YGetAudioStats(con, &stat_buf))
		    {
			gchar *buf = g_strdup_printf(
			    "%i Hz  %i bits  %i channels",
			    stat_buf.sample_rate,
			    stat_buf.sample_size,
			    stat_buf.channels
			);
			gtk_label_set_text(
			    GTK_LABEL(win->audio_label), buf
			);
			g_free(buf);

			DSPViewSetChannels(
			    dv, stat_buf.channels
			);

			gtk_widget_queue_resize(win->toplevel);
		    }
		}
		break;

	      case YDisconnect: case YShutdown:
		/* Close YSHM sound buffer. */
		YSHMCloseSound(con, win->yshm_buf);
		win->yshm_buf = NULL;
		win->yshm_buf_len = 0;
		win->yshm_id = -1;
		/* Close connection. */
		YCloseConnection(con, False);
		win->con = con = NULL;
		break;
	    }
	}

	/* Redraw DSP view. */
	DSPViewDraw(dv, win->yshm_buf, win->yshm_buf_len);

	/* Don't hog the cpu. */
	usleep(30 * 1000);

#undef QUIT_GTK_MAIN_RETURN

	return(TRUE);
}

int main(int argc, char *argv[])
{
	gbool initialized_gtk = FALSE;
	const gchar *y_server_address = "127.0.0.1:9433";
	gint i;
	const gchar *arg_ptr;
	guint toid;
	GSList *group;
	GdkWindow *window;
	GtkWidget *toplevel, *w, *parent, *parent2, *parent3, *parent4;
	win_struct *win;
	dspview_struct *dv;
	YConnection *con;


	for(i = 1; i < argc; i++)
	{
	    arg_ptr = argv[i];
	    if(arg_ptr == NULL)
		continue;

	    if(!strcmp(arg_ptr, "--help") ||
	       !strcmp(arg_ptr, "-help") ||
	       !strcmp(arg_ptr, "--h") ||
	       !strcmp(arg_ptr, "-h") ||
	       !strcmp(arg_ptr, "-?")
	    )
	    {
		printf(
"\
Usage: ydspview [recorder]\n\
\n"
		);
		return(0);
	    }
	}

	/* Initialize GTK+ as needed. */
	if(!initialized_gtk)
	{
	    if(!gtk_init_check(&argc, &argv))
	    {
		fprintf(
		    stderr,
 "This program requires X.\n"
		);
		return(1);
	    }
	    initialized_gtk = TRUE;
	}

	gdk_rgb_init();

	imlib_handle = Imlib_init(GDK_DISPLAY());

	CDialogInit();
	CSDInit();


	/* Begin creating window. */
	win = (win_struct *)g_malloc0(sizeof(win_struct));

	/* Toplevel. */
	win->toplevel = toplevel = parent = w =
	    gtk_window_new(GTK_WINDOW_DIALOG);
	gtk_window_set_title(GTK_WINDOW(w), "YDSPView");
	gtk_window_set_policy(
	    GTK_WINDOW(w),
	    TRUE, TRUE, TRUE
	);
	gtk_widget_realize(w);
	window = w->window;
	if(window != NULL)
	{
	    gdk_window_set_decorations(
		window,
		GDK_DECOR_BORDER | GDK_DECOR_TITLE | GDK_DECOR_MENU |
		GDK_DECOR_MINIMIZE
	    );
	    gdk_window_set_functions(
		window,
		GDK_FUNC_MOVE | GDK_FUNC_MINIMIZE | GDK_FUNC_CLOSE
	    );
	    GUISetWMIcon(window, (u_int8_t **)icon_y_48x48_xpm);
	}
	gtk_signal_connect(
	    GTK_OBJECT(w), "delete_event",
	    GTK_SIGNAL_FUNC(delete_cb), win
	);

	w = gtk_vbox_new(FALSE, 5);
	gtk_container_add(GTK_CONTAINER(parent), w);
	gtk_container_border_width(GTK_CONTAINER(w), 5);
	gtk_widget_show(w);
	parent = w;


	/* Hbox to hold location and DSP view widgets. */
	w = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_vbox_new(TRUE, 0);
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent3 = w;

	w = gtk_hbox_new(FALSE, 5);
	gtk_box_pack_start(GTK_BOX(parent3), w, TRUE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;

	/* Audio label. */
	win->audio_label = w = gtk_label_new(y_server_address);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	/* Refresh button. */
	w = gtk_button_new_with_label("Refresh");
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    refresh_cb, win
	);
	gtk_widget_show(w);

	/* DSP active check button. */
	w = gtk_check_button_new_with_label("DSP Active");
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), TRUE);
	gtk_signal_connect(
	    GTK_OBJECT(w), "toggled",
	    toggle_dspview_cb, win
	);
	gtk_widget_show(w);

	/* Background buttons. */
	parent4 = w = gtk_button_new();
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    bg_image_previous_cb, win
	);
	gtk_widget_show(w);
	w = gtk_arrow_new(GTK_ARROW_LEFT, GTK_SHADOW_OUT);
	gtk_container_add(GTK_CONTAINER(parent4), w);
	gtk_widget_set_usize(w, 15, 15);
	gtk_widget_show(w);

	parent4 = w = gtk_button_new();
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    bg_image_next_cb, win
	);
	gtk_widget_show(w);
	w = gtk_arrow_new(GTK_ARROW_RIGHT, GTK_SHADOW_OUT);
	gtk_container_add(GTK_CONTAINER(parent4), w);
	gtk_widget_set_usize(w, 15, 15);
	gtk_widget_show(w);


	/* End vbox to hold DSP view. */
	w = gtk_vbox_new(FALSE, 0);
	gtk_box_pack_end(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;

	/* DSP view. */
	win->dv = dv = DSPViewNew(
	    parent3,
	    80 + 2, 32 + 2
	);
	DSPViewSetDisplayStyle(dv, DSPVIEW_DISPLAY_STYLE_WAVE);
	DSPViewSetMergeMode(dv, DSPVIEW_MERGE_MODE_SET);
	DSPViewSetDSPColor(dv, DSPVIEW_DEF_DSP_COLOR);	/* 0xAABBGGRR. */
	DSPViewMap(dv);


	w = gtk_hseparator_new();
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);


	w = gtk_hbox_new(FALSE, 2);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_label_new("Display Style:");
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	/* Display style radio buttons. */
	w = gtk_radio_button_new_with_label(NULL, "Wave");
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), TRUE);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "toggled",
	    toggle_display_style_wave, win
	);
	gtk_widget_show(w);
	group = gtk_radio_button_group(GTK_RADIO_BUTTON(w));

	w = gtk_radio_button_new_with_label(group, "Sine Graph");
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "toggled",
	    toggle_display_style_sgraph, win
	);
	gtk_widget_show(w);
	group = gtk_radio_button_group(GTK_RADIO_BUTTON(w));

	w = gtk_radio_button_new_with_label(group, "Analyzer");
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "toggled",
	    toggle_display_style_analyzer, win
	);
	gtk_widget_show(w);
	group = gtk_radio_button_group(GTK_RADIO_BUTTON(w));

	w = gtk_radio_button_new_with_label(group, "Lightening");
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "toggled",
	    toggle_display_style_lightening, win
	);
	gtk_widget_show(w);
	group = gtk_radio_button_group(GTK_RADIO_BUTTON(w));

	w = gtk_radio_button_new_with_label(group, "Plasma");
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "toggled",
	    toggle_display_style_plasma, win
	);
	gtk_widget_show(w);
	group = gtk_radio_button_group(GTK_RADIO_BUTTON(w));

	w = gtk_radio_button_new_with_label(group, "Nova");
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "toggled",
	    toggle_display_style_nova, win
	);
	gtk_widget_show(w);


	w = gtk_hbox_new(FALSE, 2);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_label_new("Merge Mode:");
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	/* Merge mode radio buttons. */
	w = gtk_radio_button_new_with_label(NULL, "Set");
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), TRUE);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "toggled",
	    toggle_merge_mode_set, win
	);
	gtk_widget_show(w);
	group = gtk_radio_button_group(GTK_RADIO_BUTTON(w));

	w = gtk_radio_button_new_with_label(group, "Set Alpha");
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "toggled",
	    toggle_merge_mode_set_alpha, win
	);
	gtk_widget_show(w);
	group = gtk_radio_button_group(GTK_RADIO_BUTTON(w));

	w = gtk_radio_button_new_with_label(group, "Add");
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "toggled",
	    toggle_merge_mode_add, win
	);
	gtk_widget_show(w);
	group = gtk_radio_button_group(GTK_RADIO_BUTTON(w));

	w = gtk_radio_button_new_with_label(group, "Blend");
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "toggled",
	    toggle_merge_mode_blend, win
	);
	gtk_widget_show(w);
	group = gtk_radio_button_group(GTK_RADIO_BUTTON(w));

	w = gtk_radio_button_new_with_label(group, "Multiply");
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "toggled",
	    toggle_merge_mode_multiply, win
	);
	gtk_widget_show(w);


	if(1)
	{
	    GtkAdjustment *adj;

	    w = gtk_hbox_new(FALSE, 2);
	    gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent2 = w;


	    /* Color. */
	    w = gtk_button_new_with_label("Color...");
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_signal_connect(
		GTK_OBJECT(w), "clicked",
		change_color_cb, win
	    );
	    gtk_widget_show(w);

	    /* Transparency scale. */
	    w = gtk_label_new("Transparency:");
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);

	    win->transparency_adj = adj = (GtkAdjustment *)gtk_adjustment_new(
		1.0f, 0.0f, 1.0f,
		0.1f, 0.25f, 0.0f
	    );
	    gtk_signal_connect(
		GTK_OBJECT(adj), "value_changed",
		scale_transparency_cb, win
	    );
	    gtk_object_ref(GTK_OBJECT(adj));
	    w = gtk_hscale_new(adj);
	    gtk_widget_set_usize(w, 100, -1);
	    gtk_scale_set_draw_value(GTK_SCALE(w), FALSE);
/*
	    gtk_scale_set_digits(GTK_SCALE(w), 2);
	    gtk_scale_set_value_pos(GTK_SCALE(w), GTK_POS_RIGHT);
 */
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);

	    /* Fade speed scale. */
	    w = gtk_label_new("Fade Speed:");
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);

	    win->fade_speed_adj = adj = (GtkAdjustment *)gtk_adjustment_new(
		DSPVIEW_DEF_FADE_COEFF, 0.0f, 1.0f,
		0.1f, 0.25f, 0.0f
	    );
	    gtk_signal_connect(
		GTK_OBJECT(adj), "value_changed",
		scale_fade_speed_cb, win
	    );
	    gtk_object_ref(GTK_OBJECT(adj));
	    w = gtk_hscale_new(adj);
	    gtk_widget_set_usize(w, 100, -1);
	    gtk_scale_set_draw_value(GTK_SCALE(w), FALSE);
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);

	}


	if(1)
	{
	    GtkAdjustment *adj;


	    w = gtk_hbox_new(FALSE, 5);
	    gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent2 = w;


	    /* Hbox for volume scale. */
	    w = gtk_hbox_new(FALSE, 2);
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent3 = w;

	    w = GUICreateMenuItemIcon(
		(u_int8_t **)icon_sound_20x20_xpm
	    );
	    gtk_widget_set_usize(w, 20, 20);
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);

	    /* Volume scale. */
	    win->volume_adj = adj = (GtkAdjustment *)gtk_adjustment_new(
		0.5f, 0.0f, 1.0f,
		0.1f, 0.25f, 0.0f
	    );
	    gtk_signal_connect(
		GTK_OBJECT(adj), "value_changed",
		scale_volume_cb, win
	    );
	    gtk_object_ref(GTK_OBJECT(adj));
	    w = gtk_hscale_new(adj);
	    gtk_widget_set_usize(w, 100, -1);
	    gtk_scale_set_draw_value(GTK_SCALE(w), FALSE);
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);


	    /* Hbox for balance scale. */
	    w = gtk_hbox_new(FALSE, 2);
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent3 = w;

	    w = GUICreateMenuItemIcon(
		(u_int8_t **)icon_speaker_left_20x20_xpm
	    );
	    gtk_widget_set_usize(w, 20, 20);
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);

	    /* Balance scale. */
	    win->balance_adj = adj = (GtkAdjustment *)gtk_adjustment_new(
		0.0f, -1.0f, 1.0f,
		0.1f, 0.25f, 0.0f
	    );
	    gtk_signal_connect(
		GTK_OBJECT(adj), "value_changed",
		scale_balance_cb, win
	    );
	    gtk_object_ref(GTK_OBJECT(adj));
	    w = gtk_hscale_new(adj);
	    gtk_widget_set_usize(w, 80, -1);
	    gtk_scale_set_draw_value(GTK_SCALE(w), FALSE);
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);

	    w = GUICreateMenuItemIcon(
		(u_int8_t **)icon_speaker_right_20x20_xpm
	    );
	    gtk_widget_set_usize(w, 20, 20);
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);

	}

	/* Load background image. */
	win->bg_num = 1;
	DO_LOAD_BG_IMAGE_BY_NUM

	/* Connect to Y server. */
	win->yshm_buf = NULL;
	win->yshm_buf_len = 0;
	win->yshm_id = -1;
	win->con = con = YOpenConnection(
	    NULL,
	    y_server_address
	);
	if(con != NULL)
	{
	    /* Connected to Y server, now get some initial values. */
	    Coefficient mv[2];
	    YEventAudioStats stat_buf;

	    /* YSHM sound buffer pointer. */
	    win->yshm_buf = (guint8 *)YSHMOpenSound(
		con, &win->yshm_buf_len, &win->yshm_id
	    );

	    /* Audio stats. */
	    if(!YGetAudioStats(con, &stat_buf))
	    {
		gchar *buf = g_strdup_printf(
		    "%i Hz  %i bits  %i channels",
		    stat_buf.sample_rate,
		    stat_buf.sample_size,
		    stat_buf.channels
		);
		gtk_label_set_text(
		    GTK_LABEL(win->audio_label), buf
		);
		g_free(buf);

		DSPViewSetChannels(
		    dv, stat_buf.channels
		);

		gtk_widget_queue_resize(win->toplevel);
	    }

	    /* Mixer values. */
	    if(!YGetMixerChannel(
		con,
		YMixerCodeVolume,
		&mv[0], &mv[1]) &&
		(win->volume_adj != NULL) &&
		(win->balance_adj != NULL)
	    )
	    {
		if((mv[0] <= mv[1]) && (mv[1] > 0.0))
		{
		    win->volume_adj->value = (gfloat)(mv[1]);
		    win->balance_adj->value = 1.0f - (gfloat)(mv[0] / mv[1]);
		}
		else if(mv[0] > 0.0)
		{
		    win->volume_adj->value = (gfloat)(mv[0]);
		    win->balance_adj->value = (gfloat)(mv[1] / mv[0]) - 1.0f;
		}
		else
		{
		    win->volume_adj->value = 0.0f;
		    win->balance_adj->value = 0.0f;
		}
		gtk_signal_emit_by_name(
		    GTK_OBJECT(win->volume_adj), "value_changed"
		);
		gtk_signal_emit_by_name(
		    GTK_OBJECT(win->balance_adj), "value_changed"
		);
	    }
	}
	else
	{
	    /* Unable to connect to Y server. */

	}

	/* Show toplevel window. */
	gtk_widget_show(toplevel);

	/* Set main timeout callback. */
	toid = gtk_idle_add_priority(
/*	    G_PRIORITY_DEFAULT, */
/*	    G_PRIORITY_HIGH_IDLE, */
/*	    G_PRIORITY_DEFAULT_IDLE, */
	    G_PRIORITY_LOW,
	    (GtkFunction)timeout_cb, win
	);


	/* Enter main GTK+ loop. */
	gtk_main();


	/* Destroy DSP view. */
	DSPViewDelete(win->dv);
	/* Window. */
	gtk_widget_destroy(win->toplevel);
	gtk_object_unref(GTK_OBJECT(win->transparency_adj));
	gtk_object_unref(GTK_OBJECT(win->fade_speed_adj));
	gtk_object_unref(GTK_OBJECT(win->volume_adj));
	gtk_object_unref(GTK_OBJECT(win->balance_adj));
	g_free(win->bg_img);
	g_free(win);

	CDialogShutdown();
	CSDShutdown();


	return(0);
}

#undef DO_LOAD_BG_IMAGE_BY_NUM
