#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <Y2/Y.h>
#include <Y2/Ylib.h>
#include <Y2/Ymixercodes.h>

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include "guiutils.h"

#include "cfg.h"
#include "conbar.h"
#include "ymixer.h"
#include "ymixercb.h"
#include "yc.h"
#include "ycop.h"
#include "yccfglist.h"
#include "config.h"

#include "images/icon_ymixer_48x48.xpm"
#include "images/icon_speaker_left_20x20.xpm"
#include "images/icon_speaker_right_20x20.xpm"

#include "images/icon_mixer_bass_20x20.xpm"
#include "images/icon_mixer_cd_20x20.xpm"
#include "images/icon_mixer_igain_20x20.xpm"
#include "images/icon_mixer_line_20x20.xpm"
#include "images/icon_mixer_mic_20x20.xpm"
#include "images/icon_mixer_mix_20x20.xpm"
#include "images/icon_mixer_monitor_20x20.xpm"
#include "images/icon_mixer_ogain_20x20.xpm"
#include "images/icon_mixer_pcm_20x20.xpm"
#include "images/icon_mixer_phonein_20x20.xpm"
#include "images/icon_mixer_phoneout_20x20.xpm"
#include "images/icon_mixer_radio_20x20.xpm"
#include "images/icon_mixer_recorder_20x20.xpm"
#include "images/icon_mixer_speaker_20x20.xpm"
#include "images/icon_mixer_synth_20x20.xpm"
#include "images/icon_mixer_treble_20x20.xpm"
#include "images/icon_mixer_video_20x20.xpm"
#include "images/icon_mixer_volume_20x20.xpm"


void YMixerScaleCalculateBATo2C(
	gfloat bal, gfloat amp, gfloat *left, gfloat *right
);
void YMixerScaleCalculate2CToBA(
	gfloat left, gfloat right, gfloat *bal, gfloat *amp
);
GtkRcStyle *YMixerMixerScaleRCStyleNew(ymixer_struct *ym);
guchar **YMixerScaleGetXpmDataFromMixerName(const gchar *name);

void YMixerSetTitle(ymixer_struct *ym, const gchar *s);
gchar *YMixerGetLocation(ymixer_struct *ym);
gint YMixerDoConnect(
	ymixer_struct *ym,
	const gchar *address, const gchar *start_command
);
void YMixerDoDisconnect(ymixer_struct *ym, gboolean force_shutdown);
void YMixerSavePosition(ymixer_struct *ym);

void YMixerScaleSetValue2C(
	ymixer_struct *ym, gint mixer_code,
	gfloat left, gfloat right
);
ymixer_scale_struct *YMixerScaleNew(
	ymixer_struct *ym,
	gint mixer_code, const gchar *name, const gchar *short_name
);
void YMixerScaleDelete(ymixer_scale_struct *ms);
void YMixerScaleCreateAll(ymixer_struct *ym);

ymixer_struct *YMixerNew(gpointer core_ptr);
void YMixerUpdateMenus(ymixer_struct *ym);
void YMixerSetBusy(ymixer_struct *ym, gboolean is_busy);
void YMixerMap(ymixer_struct *ym);
void YMixerUnmap(ymixer_struct *ym);
void YMixerDelete(ymixer_struct *ym);


#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define ABSOLUTE(x)     (((x) < 0) ? ((x) * -1) : (x))


/*
 *	Convert balance and amplitude values to left and right channel
 *	values.
 */
void YMixerScaleCalculateBATo2C(
	gfloat bal, gfloat amp, gfloat *left, gfloat *right
)
{
	if(left != NULL)
	    *left = amp * CLIP(-bal + 1.0f, 0.0f, 1.0f);
	if(right != NULL)
	    *right = amp * CLIP(bal + 1.0f, 0.0f, 1.0f);
}

/*
 *	Convert left and right channel values to balance and amplitude
 *	values.
 */
void YMixerScaleCalculate2CToBA(
	gfloat left, gfloat right, gfloat *bal, gfloat *amp
)
{
	if((left <= right) && (right > 0.0))
	{
	    if(amp != NULL)
		*amp = right;
	    if(bal != NULL)
		*bal = 1.0f - (left / right);
	}
	else if(left > 0.0)
	{
	    if(amp != NULL)
		*amp = left;
	    if(bal != NULL)
		*bal = (right / left) - 1.0f;
	}
	else
	{
	    if(amp != NULL)
		*amp = 0.0f;
	    if(bal != NULL)
		*bal = 0.0f;
	}
}

/*
 *	Returns the pointer to the XPM data based on the given mixer
 *	name (can be either mixer name or conical mixer name).
 *
 *	Can return NULL if there is no match.
 */
guchar **YMixerScaleGetXpmDataFromMixerName(const gchar *name)
{
	if(name == NULL)
	    return(NULL);

	if(!strcasecmp(name, "Volume") ||
	   !strcasecmp(name, "vol")
	)
	    return((guint8 **)icon_mixer_volume_20x20_xpm);
	else if(!strcasecmp(name, "Synthesizer") ||
	        !strcasecmp(name, "synth")
	)
	    return((guint8 **)icon_mixer_synth_20x20_xpm);
	else if(!strcasecmp(name, "PCM") ||
		!strcasecmp(name, "PCM2")
	)
	    return((guint8 **)icon_mixer_pcm_20x20_xpm);
	else if(!strcasecmp(name, "Input Gain") ||
		!strcasecmp(name, "igain")
	)
	    return((guint8 **)icon_mixer_igain_20x20_xpm);
	else if(!strcasecmp(name, "Output Gain") ||
		!strcasecmp(name, "ogain")
	)
	    return((guint8 **)icon_mixer_ogain_20x20_xpm);
	else if(!strcasecmp(name, "Bass"))
	    return((guint8 **)icon_mixer_bass_20x20_xpm);
	else if(!strcasecmp(name, "Treble"))
	    return((guint8 **)icon_mixer_treble_20x20_xpm);
	else if(!strcasecmp(name, "CD"))
	    return((guint8 **)icon_mixer_cd_20x20_xpm);
	else if(!strcasecmp(name, "PC Speaker") ||
		!strcasecmp(name, "speaker")
	)
	    return((guint8 **)icon_mixer_speaker_20x20_xpm);
	else if(!strcasecmp(name, "Microphone") ||
		!strcasecmp(name, "mic")
	)
	    return((guint8 **)icon_mixer_mic_20x20_xpm);
	else if(!strcasecmp(name, "Recorder") ||
		!strcasecmp(name, "rec")
	)
	    return((guint8 **)icon_mixer_recorder_20x20_xpm);
	else if(!strcasecmp(name, "Mix"))
	    return((guint8 **)icon_mixer_mix_20x20_xpm);
	else if(!strcasecmp(name, "Line") ||
		!strcasecmp(name, "Line1") ||
		!strcasecmp(name, "Line2") ||
		!strcasecmp(name, "Line3")
	)
	    return((guint8 **)icon_mixer_line_20x20_xpm);
	else if(!strcasecmp(name, "Digital") ||
		!strcasecmp(name, "Digital1") ||
		!strcasecmp(name, "Digital2") ||
		!strcasecmp(name, "Digital3") ||
		!strcasecmp(name, "dig") ||
		!strcasecmp(name, "dig1") ||
		!strcasecmp(name, "dig2") ||
		!strcasecmp(name, "dig3")
	)
	    return((guint8 **)icon_mixer_pcm_20x20_xpm);
	else if(!strcasecmp(name, "Phone In") ||
		!strcasecmp(name, "phin")
	)
	    return((guint8 **)icon_mixer_phonein_20x20_xpm);
	else if(!strcasecmp(name, "Phone Out") ||
		!strcasecmp(name, "phout")
	)
	    return((guint8 **)icon_mixer_phoneout_20x20_xpm);
	else if(!strcasecmp(name, "Video"))
	    return((guint8 **)icon_mixer_video_20x20_xpm);
	else if(!strcasecmp(name, "Radio"))
	    return((guint8 **)icon_mixer_radio_20x20_xpm);
	else if(!strcasecmp(name, "Monitor"))
	    return((guint8 **)icon_mixer_monitor_20x20_xpm);
	else
	    return(NULL);
}

/*
 *	Returns a rcstyle specifying the style for a mixer scale.
 */
GtkRcStyle *YMixerMixerScaleRCStyleNew(ymixer_struct *ym)
{
	gint state;
	GdkColor *ct;
	GtkRcStyle *rcstyle;
	const cfg_color_struct *cs;
	cfg_color_struct c_insensitive;
	yc_core_struct *core_ptr;


	if(ym == NULL)
	    return(NULL);

	core_ptr = YC_CORE(ym->core_ptr);
	if(core_ptr == NULL)
	    return(NULL);

	rcstyle = gtk_rc_style_new();

#define GET_CFG_COLOR(p)	(YC_GET_CFG_COLOR(core_ptr, (p)))
#define DO_CONVERT_COLOR        \
{ if((cs != NULL) && (ct != NULL)) { \
 ct->red        = (gushort)-1 * cs->r; \
 ct->green      = (gushort)-1 * cs->g; \
 ct->blue       = (gushort)-1 * cs->b; \
} }
	state = GTK_STATE_NORMAL;
	rcstyle->color_flags[state] = GTK_RC_FG | GTK_RC_BG |
	    GTK_RC_TEXT | GTK_RC_BASE;
	cs = GET_CFG_COLOR(YC_CFG_PARM_MIXER_COLOR_FG);
	ct = &rcstyle->fg[state];
	DO_CONVERT_COLOR
	cs = GET_CFG_COLOR(YC_CFG_PARM_MIXER_COLOR_BG);
	ct = &rcstyle->bg[state];
	DO_CONVERT_COLOR
	cs = GET_CFG_COLOR(YC_CFG_PARM_MIXER_COLOR_TEXT);
	ct = &rcstyle->text[state];
	DO_CONVERT_COLOR
	cs = GET_CFG_COLOR(YC_CFG_PARM_MIXER_COLOR_BASE);
	ct = &rcstyle->base[state];
	DO_CONVERT_COLOR

	state = GTK_STATE_PRELIGHT;
	rcstyle->color_flags[state] = GTK_RC_FG | GTK_RC_BG |
	    GTK_RC_TEXT | GTK_RC_BASE;
	cs = GET_CFG_COLOR(YC_CFG_PARM_MIXER_COLOR_FG);
	ct = &rcstyle->fg[state];
	DO_CONVERT_COLOR
	cs = GET_CFG_COLOR(YC_CFG_PARM_MIXER_COLOR_BG_PRELIGHT);
	ct = &rcstyle->bg[state];
	DO_CONVERT_COLOR
	cs = GET_CFG_COLOR(YC_CFG_PARM_MIXER_COLOR_TEXT);
	ct = &rcstyle->text[state];
	DO_CONVERT_COLOR
	cs = GET_CFG_COLOR(YC_CFG_PARM_MIXER_COLOR_BASE);
	ct = &rcstyle->base[state];
	DO_CONVERT_COLOR

	state = GTK_STATE_ACTIVE;
	rcstyle->color_flags[state] = GTK_RC_FG | GTK_RC_BG |
	    GTK_RC_TEXT | GTK_RC_BASE;
	cs = GET_CFG_COLOR(YC_CFG_PARM_MIXER_COLOR_FG);
	ct = &rcstyle->fg[state];
	DO_CONVERT_COLOR
	cs = GET_CFG_COLOR(YC_CFG_PARM_MIXER_COLOR_BASE);
	ct = &rcstyle->bg[state];
	DO_CONVERT_COLOR
	cs = GET_CFG_COLOR(YC_CFG_PARM_MIXER_COLOR_TEXT);
	ct = &rcstyle->text[state];
	DO_CONVERT_COLOR
	cs = GET_CFG_COLOR(YC_CFG_PARM_MIXER_COLOR_BASE);
	ct = &rcstyle->base[state];
	DO_CONVERT_COLOR

	state = GTK_STATE_INSENSITIVE;
	rcstyle->color_flags[state] = GTK_RC_FG | GTK_RC_BG |
	    GTK_RC_TEXT | GTK_RC_BASE;
	cs = GET_CFG_COLOR(YC_CFG_PARM_MIXER_COLOR_FG);
	if(cs != NULL)
	{
	    memcpy(&c_insensitive, cs, sizeof(cfg_color_struct));
	    c_insensitive.r *= 0.5;
	    c_insensitive.g *= 0.5;
	    c_insensitive.b *= 0.5;
	    cs = &c_insensitive;
	}
	ct = &rcstyle->fg[state];
	DO_CONVERT_COLOR
	cs = GET_CFG_COLOR(YC_CFG_PARM_MIXER_COLOR_BG);
	ct = &rcstyle->bg[state];
	DO_CONVERT_COLOR
	cs = GET_CFG_COLOR(YC_CFG_PARM_MIXER_COLOR_TEXT);
	if(cs != NULL)
	{
	    memcpy(&c_insensitive, cs, sizeof(cfg_color_struct));
	    c_insensitive.r *= 0.5;
	    c_insensitive.g *= 0.5;
	    c_insensitive.b *= 0.5;
	    cs = &c_insensitive;
	}
	ct = &rcstyle->text[state];
	DO_CONVERT_COLOR
	cs = GET_CFG_COLOR(YC_CFG_PARM_MIXER_COLOR_BASE);
	ct = &rcstyle->base[state];
	DO_CONVERT_COLOR

	state = GTK_STATE_SELECTED;
	rcstyle->color_flags[state] = GTK_RC_FG | GTK_RC_BG |
	    GTK_RC_TEXT | GTK_RC_BASE;
	cs = GET_CFG_COLOR(YC_CFG_PARM_MIXER_COLOR_BG);
	ct = &rcstyle->fg[state];
	DO_CONVERT_COLOR
	cs = GET_CFG_COLOR(YC_CFG_PARM_MIXER_COLOR_FG);
	ct = &rcstyle->bg[state];
	DO_CONVERT_COLOR
	cs = GET_CFG_COLOR(YC_CFG_PARM_MIXER_COLOR_BASE);
	ct = &rcstyle->text[state];
	DO_CONVERT_COLOR
	cs = GET_CFG_COLOR(YC_CFG_PARM_MIXER_COLOR_TEXT);
	ct = &rcstyle->base[state];
	DO_CONVERT_COLOR

#undef DO_CONVERT_COLOR
#undef GET_CFG_COLOR

	return(rcstyle);
}

/* 
 *	Sets the title string to s, if s is NULL then values from the
 *	ymixer will be used to form the title.
 */
void YMixerSetTitle(ymixer_struct *ym, const gchar *s)
{
	gchar *title;
	GtkWidget *w;


	if(ym == NULL)
	    return;

	w = YMIXER_TOPLEVEL(ym);
	if(w == NULL)
	    return;

	/* Title string given? */
	if(s != NULL)
	{
	    gtk_window_set_title(GTK_WINDOW(w), s);
	    return;
	}

	/* Format title string */
	if(YMIXER_IS_CONNECTED(ym))
	{
	    s = ConBarGetLocation(YMIXER_CONBAR(ym));
	    title = g_strdup_printf("YMixer: %s", s);
	}
	else
	{
	    title = g_strdup("YMixer");
	}
	gtk_window_set_title(GTK_WINDOW(w), title);

	g_free(title);
}


/*
 *	Returns the current location.
 */
gchar *YMixerGetLocation(ymixer_struct *ym)
{
	return((ym != NULL) ? ConBarGetLocation(YMIXER_CONBAR(ym)) : NULL);
}

/*
 *	Connects the ymixer to the Y server at the addres s.  If the
 *	ymixer is already connected then it will be disconneted first.
 *
 *	Returns non-zero on error and 0 on success.
 */
gint YMixerDoConnect(
	ymixer_struct *ym,
	const gchar *address, const gchar *start_command
)
{
	YConnection *con;


	if(ym == NULL)
	    return(-1);

	/* Disconnect from Y server as needed */
	YMixerDoDisconnect(ym, FALSE);

	/* If no address is given then only perform disconnect */
	if(address == NULL)
	    return(0);

	/* Connect to Y server */
	ym->con = con = YOpenConnection(NULL, address);
	/* If not able to connect on first try and a server start
	 * command is given, then execute that command and try
	 * connecting again.
	 */
	if((con == NULL) && (start_command != NULL))
	{
	    /* Start server (with dialog) and try connecting again */
	    YCNewStartupDialog(
		YC_CORE(ym->core_ptr), YMIXER_TOPLEVEL(ym),
		start_command, TRUE
	    );
	    ym->con = con = YOpenConnection(NULL, address);
	}
	if(con != NULL)
	{
	    /* Connected to Y server, now get some initial values */
	    gint i;
	    Coefficient vc[2];
	    YEventAudioStats stat_buf;
	    dspview_struct *dv = ConBarGetDSPView(YMIXER_CONBAR(ym));

	    /* Update connection bar connected state */
	    ConBarSetConnectState(YMIXER_CONBAR(ym), TRUE);

	    /* YSHM sound buffer pointer */
	    ym->yshm_buf = (guint8 *)YSHMOpenSound(
		con, &ym->yshm_buf_len, &ym->yshm_id
	    );

	    /* Audio stats */
	    if(!YGetAudioStats(con, &stat_buf))
	    {
		g_free(ym->audio_mode_name);
		ym->audio_mode_name = g_strdup(stat_buf.mode_name);
		ym->sample_rate = stat_buf.sample_rate;
		ym->sample_size = stat_buf.sample_size;
		ym->channels = stat_buf.channels;
		ym->cycle_us = stat_buf.cycle_us;
		ym->cycle_load = 0.0;
		ym->num_fragments = stat_buf.num_fragments;
		ym->fragment_size = stat_buf.fragment_size;
		ym->shellout = FALSE;

		/* Update audio values on connection bar */
		YMIXER_SET_CONBAR_AUDIO(ym);

		ConBarSetShellOut(YMIXER_CONBAR(ym), ym->shellout);
		DSPViewSetChannels(dv, ym->channels);
	    }

	    /* Record location history on connection bar */
	    ConBarSetLocation(YMIXER_CONBAR(ym), address, FALSE, TRUE);

	    /* Update mixers */
	    for(i = 0; i < ym->total_mixer_scales; i++)
	    {
		YGetMixerChannel(
		    con, i + YMixerCodeBaseOffset,
		    &vc[0], &vc[1]
		);
		YMixerScaleSetSensitive(
		    ym, i + YMixerCodeBaseOffset, TRUE
		);
		YMixerScaleSetValue2C(
		    ym, i + YMixerCodeBaseOffset,
		    (gfloat)vc[0], (gfloat)vc[1]
		);
	    }

	    /* Update title */
	    YMixerSetTitle(ym, NULL);
	    YMixerUpdateMenus(ym);
	}

	return(0);
}

/*
 *	Disconnects the ymixer from the Y server.
 */
void YMixerDoDisconnect(ymixer_struct *ym, gboolean force_shutdown)
{
	gint i;
	dspview_struct *dv;


	if(ym == NULL)
	    return;


	/* Update connection bar connected state */
	ConBarSetConnectState(YMIXER_CONBAR(ym), FALSE);


	/* Close YSHM sound buffer */
	YSHMCloseSound(YMIXER_CONNECTION(ym), ym->yshm_buf);
	ym->yshm_buf = NULL;
	ym->yshm_buf_len = 0;
	ym->yshm_id = -1;

	/* Close connection or shutdown? */
	if(force_shutdown)
	    YShutdownServer(YMIXER_CONNECTION(ym));
	else
	    YCloseConnection(YMIXER_CONNECTION(ym), False);
	ym->con = NULL;


	/* Reset Y audio values */
	memset(ym->channel_volume_avg, 0x00, 4 * sizeof(gfloat));
	ym->start_time = 0;
	g_free(ym->audio_mode_name);
	ym->audio_mode_name = NULL;
	ym->sample_rate = 0;
	ym->sample_size = 0;
	ym->channels = 0;
	ym->cycle_us = 0;
	ym->cycle_load = 0.0f;
	ym->num_fragments = 0;
	ym->fragment_size = 0;
	ym->flip_stereo = FALSE;
	ym->shellout = FALSE;
	ym->version_major = ym->version_minor = 0;
	g_free(ym->server_name);
	ym->server_name = NULL;
	g_free(ym->sound_name);
	ym->sound_name = NULL;

	/* Update audio values on connection bar */
	YMIXER_SET_CONBAR_AUDIO(ym);

	dv = ConBarGetDSPView(YMIXER_CONBAR(ym));
	DSPViewSetChannels(dv, ym->channels);

	/* Update mixers */
	for(i = 0; i < ym->total_mixer_scales; i++)
	    YMixerScaleSetSensitive(
		ym, i + YMixerCodeBaseOffset, FALSE
	    );

	/* Update title */
	YMixerSetTitle(ym, NULL);
	YMixerUpdateMenus(ym);
}


/*
 *	Save current position value of window to configuration list on
 *	core structure.
 */
void YMixerSavePosition(ymixer_struct *ym)
{
	GtkWidget *w;
	yc_core_struct *core_ptr;


	if(ym == NULL)
	    return;

	core_ptr = YC_CORE(ym->core_ptr);
	if(core_ptr == NULL)
	    return;

	/* Do not save positions? */
	if(!YC_GET_CFG_I(
	    core_ptr, YC_CFG_PARM_MIXER_SAVE_POSITIONS
	))
	    return;

	/* Save toplevel position */
	w = YMIXER_TOPLEVEL(ym);
	if(w != NULL)
	{
	    GdkWindow *window = w->window;
	    gint x = 0, y = 0, width, height;


	    if(window != NULL)
		gdk_window_get_root_origin(window, &x, &y);
	    YC_SET_CFG_I(core_ptr, YC_CFG_PARM_MIXER_X, x);
	    YC_SET_CFG_I(core_ptr, YC_CFG_PARM_MIXER_Y, y);

	    width = w->allocation.width;
	    height = w->allocation.height;
	    YC_SET_CFG_I(core_ptr, YC_CFG_PARM_MIXER_WIDTH, width);
	    YC_SET_CFG_I(core_ptr, YC_CFG_PARM_MIXER_HEIGHT, height);
	}

	if(TRUE)
	{
	    const gint n = ym->total_mixer_scales;
	    gint i;
	    GList *glist = NULL;
	    cfg_intlist_struct *intlist;
	    ymixer_scale_struct *ms;

	    for(i = 0; i < n; i++)
	    {
		ms = ym->mixer_scale[i];
		glist = g_list_append(
		    glist,
		    (gpointer)((ms != NULL) ?
			ms->mixer_code : YMixerCodeVolume
		    )
		);
	    }
	    intlist = CFGIntListNew(glist);
	    g_list_free(glist);
	    YC_SET_CFG_INTLIST(
		core_ptr, YC_CFG_PARM_MIXER_MIXER_SCALES,
		intlist
	    );
	    CFGIntListDelete(intlist);
	}
}


/*
 *	Returns the mixer scale that matches mixer_code.
 */
ymixer_scale_struct *YMixerScaleMatch(
	ymixer_struct *ym, gint mixer_code, gint *n
)
{
	gint i;
	ymixer_scale_struct *ms;


	if(n != NULL)
	    *n = -1;

	if(ym == NULL)
	    return(NULL);

	for(i = 0; i < ym->total_mixer_scales; i++)
	{
	    ms = ym->mixer_scale[i];
	    if(ms == NULL)
		continue;

	    if(ms->mixer_code == mixer_code)
	    {
		if(n != NULL)
		    *n = i;
		return(ms);
	    }
	}

	return(NULL);
}

/*
 *	Sets the mixer scale sensitive or insensitive.
 */
void YMixerScaleSetSensitive(
	ymixer_struct *ym, gint mixer_code,
	gboolean sensitive
)
{
	GtkWidget *w;
	ymixer_scale_struct *ms = YMixerScaleMatch(
	    ym, mixer_code, NULL
	);
	if(ms == NULL)
	    return;

#define DO_SET_SENSITIVE	\
{ if(w != NULL) gtk_widget_set_sensitive(w, sensitive); }

	w = ms->balance_scale;
	DO_SET_SENSITIVE
	w = ms->amplitude_scale;
	DO_SET_SENSITIVE
	w = ms->channel_values_da;
	DO_SET_SENSITIVE

#undef DO_SET_SENSITIVE
}

/*
 *	Sets the mixer scale's value specified by mixer_code.
 */
void YMixerScaleSetValue2C(
	ymixer_struct *ym, gint mixer_code,
	gfloat left, gfloat right
)
{
	gfloat bal, amp;
	ymixer_scale_struct *ms = YMixerScaleMatch(
	    ym, mixer_code, NULL
	);
	if(ms == NULL)
	    return;

	if(ms->freeze_count > 0)
	    return;

	YMixerScaleCalculate2CToBA(left, right, &bal, &amp);
	ms->freeze_count++;
	gtk_adjustment_set_value(ms->balance_adj, bal);
	gtk_adjustment_set_value(ms->amplitude_adj, 1.0f - amp);
	ms->freeze_count--;
}

/*
 *	Create a new mixer scale.
 */
ymixer_scale_struct *YMixerScaleNew(
	ymixer_struct *ym,
	gint mixer_code, const gchar *name, const gchar *short_name
)
{
	gint state, border_minor = 2;
	gboolean show_value;
	gint label_type;
	GtkAdjustment *adj;
	GtkRcStyle *rcstyle;
	GtkWidget *w, *parent, *parent2, *parent3;
	ymixer_scale_struct *ms;
	yc_core_struct *core_ptr;

	if(ym == NULL)
	    return(NULL);

	core_ptr = YC_CORE(ym->core_ptr);
	if(core_ptr == NULL)
	    return(NULL);

	/* Get configuration values that we will need to use in the
	 * creation of this mixer scale.
	 */
	label_type = YC_GET_CFG_I(
	    core_ptr, YC_CFG_PARM_MIXER_MIXER_SCALE_LABEL_TYPE
	);
	show_value = YC_GET_CFG_I(
	    core_ptr, YC_CFG_PARM_MIXER_SHOW_MIXER_VALUE
	);


	/* Get parent for mixer scale */
	parent = ym->mixers_hbox_client;
	if(parent == NULL)
	    return(NULL);

	/* Allocate mixer scale structure */
	ms = YMIXER_SCALE(g_malloc0(
	    sizeof(ymixer_scale_struct)
	));
	if(ms == NULL)
	    return(NULL);

	ms->mixer_code = mixer_code;
	ms->freeze_count = 0;
	ms->ym = ym;
	ms->balance_last = 0.0f;
	ms->amplitude_last = 0.0f;

	/* Create rcstyle for mixer scales only if we are not to use
	 * default style for the mixers scales.
	 */
	rcstyle = (YC_GET_CFG_I(core_ptr, YC_CFG_PARM_MIXER_COLOR_USE_DEFAULT)) ?
	    NULL : YMixerMixerScaleRCStyleNew(ym);


	/* Create adjustments */
	ms->balance_adj = adj = (GtkAdjustment *)gtk_adjustment_new(
	    0.0f, -1.0f, 1.0f,
	    0.025f, 0.05f, 0.0f
	);
	gtk_signal_connect(
	    GTK_OBJECT(adj), "value_changed",
	    GTK_SIGNAL_FUNC(YMixerScaleValueChangeCB), ms
	);

	ms->amplitude_adj = adj = (GtkAdjustment *)gtk_adjustment_new(
	    0.5f, 0.0f, 1.0f,
	    0.025f, 0.05f, 0.0f
	);
	gtk_signal_connect(
	    GTK_OBJECT(adj), "value_changed",
	    GTK_SIGNAL_FUNC(YMixerScaleValueChangeCB), ms
	);

	/* Create toplevel */
	ms->toplevel = w = gtk_frame_new(NULL);
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_OUT);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	if(rcstyle != NULL)
	    gtk_widget_modify_style(w, rcstyle);
	gtk_widget_show(w);
	parent = w;

	w = gtk_event_box_new();
	gtk_widget_add_events(w, GDK_EXPOSURE_MASK | GDK_STRUCTURE_MASK);
	gtk_container_add(GTK_CONTAINER(parent), w);
	if(rcstyle != NULL)
	    gtk_widget_modify_style(w, rcstyle);
	gtk_widget_show(w);
	parent = w;

	w = gtk_vbox_new(FALSE, border_minor);
	gtk_container_border_width(GTK_CONTAINER(w), border_minor);
	gtk_container_add(GTK_CONTAINER(parent), w);
	gtk_widget_show(w);
	parent = w;


	/* Hbox for name */
	w = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	/* Left speaker icon */
	w = GUICreateMenuItemIcon(
	    (guint8 **)icon_speaker_left_20x20_xpm
	);
	gtk_widget_set_usize(w, 20, 20);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	/* Mixer icon or name */
	w = NULL;
	if(label_type == YMIXER_MIXER_SCALE_LABEL_TYPE_ICON)
	{
	    parent3 = w = gtk_event_box_new();
	    gtk_widget_add_events(
		w,
		GDK_ENTER_NOTIFY_MASK | GDK_LEAVE_NOTIFY_MASK
	    );
	    GUISetWidgetTip(w, name);

	    w = GUICreateMenuItemIcon(
		YMixerScaleGetXpmDataFromMixerName(name)
	    );
	    gtk_widget_set_usize(w, 20, 20);
	    gtk_container_add(GTK_CONTAINER(parent3), w);
	    gtk_widget_show(w);
	    w = parent3;
	}
	if(w == NULL)
	    w = gtk_label_new((name != NULL) ? name : "Untitled");
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, FALSE, 0);
	if(rcstyle != NULL)
	    gtk_widget_modify_style(w, rcstyle);
	gtk_widget_show(w);

	/* Right speaker icon */
	w = GUICreateMenuItemIcon(
	    (guint8 **)icon_speaker_right_20x20_xpm
	);
	gtk_widget_set_usize(w, 20, 20);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);



	/* Hbox for balance widgets */
	w = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	/* Balance scale */
	gtk_object_ref(GTK_OBJECT(ms->balance_adj));
	ms->balance_scale = w = gtk_hscale_new(ms->balance_adj);
	gtk_widget_set_usize(w, 90, -1);
	gtk_scale_set_draw_value(GTK_SCALE(w), FALSE);
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, TRUE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "key_press_event",
	    GTK_SIGNAL_FUNC(YMixerScaleKeyEventCB), ms
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "key_release_event",
	    GTK_SIGNAL_FUNC(YMixerScaleKeyEventCB), ms
	);
	if(rcstyle != NULL)
	    gtk_widget_modify_style(w, rcstyle);
	gtk_widget_show(w);


	
	/* Hbox for amplitude widgets */
	w = gtk_hbox_new(TRUE, 0);
	gtk_box_pack_start(GTK_BOX(parent), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_hbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;

	/* Left ruler drawing area */
	ms->amplitude_ruler_left_da = w = gtk_drawing_area_new();
	gtk_widget_add_events(
	    w,
	    GDK_EXPOSURE_MASK | GDK_STRUCTURE_MASK |
	    GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "expose_event",
	    GTK_SIGNAL_FUNC(YMixerScaleRulerEventCB), ms
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "configure_event",
	    GTK_SIGNAL_FUNC(YMixerScaleRulerEventCB), ms
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "button_press_event",
	    GTK_SIGNAL_FUNC(YMixerScaleRulerEventCB), ms
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "button_release_event",
	    GTK_SIGNAL_FUNC(YMixerScaleRulerEventCB), ms
	);
	gtk_widget_set_usize(w, 20, -1);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	if(rcstyle != NULL)
	    gtk_widget_modify_style(w, rcstyle);
	gtk_widget_show(w);

	/* Amplitude scale */
	gtk_object_ref(GTK_OBJECT(ms->amplitude_adj));
	ms->amplitude_scale = w = gtk_vscale_new(ms->amplitude_adj);
	gtk_scale_set_draw_value(GTK_SCALE(w), FALSE);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "key_press_event",
	    GTK_SIGNAL_FUNC(YMixerScaleKeyEventCB), ms
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "key_release_event",
	    GTK_SIGNAL_FUNC(YMixerScaleKeyEventCB), ms
	);
	if(rcstyle != NULL)
	    gtk_widget_modify_style(w, rcstyle);
	gtk_widget_show(w);

	/* Right ruler drawing area */
	ms->amplitude_ruler_right_da = w = gtk_drawing_area_new();
	gtk_widget_add_events(
	    w,
	    GDK_EXPOSURE_MASK | GDK_STRUCTURE_MASK |
	    GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "expose_event",
	    GTK_SIGNAL_FUNC(YMixerScaleRulerEventCB), ms
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "configure_event",
	    GTK_SIGNAL_FUNC(YMixerScaleRulerEventCB), ms
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "button_press_event",
	    GTK_SIGNAL_FUNC(YMixerScaleRulerEventCB), ms
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "button_release_event",
	    GTK_SIGNAL_FUNC(YMixerScaleRulerEventCB), ms
	);
	gtk_widget_set_usize(w, 20, -1);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	if(rcstyle != NULL)
	    gtk_widget_modify_style(w, rcstyle);
	gtk_widget_show(w);


	/* Create values display widgets? */
	if(show_value)
	{
	    const cfg_color_struct *cs;
	    GdkColor *ct;
	    GtkRcStyle *rcstyle2 = gtk_rc_style_new();

#define DO_GET_COLOR(p)		YC_GET_CFG_COLOR(core_ptr, (p))
#define DO_CONVERT_COLOR		\
{ if((cs != NULL) && (ct != NULL)) {	\
 ct->red        = (gushort)-1 * cs->r;	\
 ct->green      = (gushort)-1 * cs->g;	\
 ct->blue       = (gushort)-1 * cs->b;	\
} }

	    state = GTK_STATE_NORMAL;
	    rcstyle2->color_flags[state] = GTK_RC_FG | GTK_RC_BG |
		GTK_RC_TEXT | GTK_RC_BASE;
	    cs = DO_GET_COLOR(YC_CFG_PARM_AUDIO_DISPLAY_COLOR_FG);
	    ct = &rcstyle2->fg[state];
	    DO_CONVERT_COLOR
	    ct = &rcstyle2->text[state];
	    DO_CONVERT_COLOR
	    cs = DO_GET_COLOR(YC_CFG_PARM_AUDIO_DISPLAY_COLOR_BG);
	    ct = &rcstyle2->bg[state];
	    DO_CONVERT_COLOR
	    ct = &rcstyle2->base[state];
	    DO_CONVERT_COLOR
	    state = GTK_STATE_PRELIGHT;
	    rcstyle2->color_flags[state] = GTK_RC_FG | GTK_RC_BG |
		GTK_RC_TEXT | GTK_RC_BASE;
	    cs = DO_GET_COLOR(YC_CFG_PARM_AUDIO_DISPLAY_COLOR_FG);
	    ct = &rcstyle2->fg[state];
	    DO_CONVERT_COLOR
	    ct = &rcstyle2->text[state];
	    DO_CONVERT_COLOR
	    cs = DO_GET_COLOR(YC_CFG_PARM_AUDIO_DISPLAY_COLOR_BG);
	    ct = &rcstyle2->bg[state];
	    DO_CONVERT_COLOR
	    ct = &rcstyle2->base[state];
	    DO_CONVERT_COLOR
	    state = GTK_STATE_INSENSITIVE;
	    rcstyle2->color_flags[state] = GTK_RC_FG | GTK_RC_BG |
		GTK_RC_TEXT | GTK_RC_BASE;
	    cs = DO_GET_COLOR(YC_CFG_PARM_AUDIO_DISPLAY_COLOR_FG_DIM);
	    ct = &rcstyle2->fg[state];
	    DO_CONVERT_COLOR
	    ct = &rcstyle2->text[state];
	    DO_CONVERT_COLOR
	    cs = DO_GET_COLOR(YC_CFG_PARM_AUDIO_DISPLAY_COLOR_BG);
	    ct = &rcstyle2->bg[state];
	    DO_CONVERT_COLOR
	    ct = &rcstyle2->base[state];
	    DO_CONVERT_COLOR
#undef DO_GET_COLOR
#undef DO_CONVERT_COLOR


	    /* Hbox for values drawing area */
	    w = gtk_hbox_new(TRUE, 0);
	    gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent2 = w;

	    w = gtk_frame_new(NULL);
	    gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_IN);
	    gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, TRUE, 0);
	    gtk_widget_modify_style(w, rcstyle);
	    gtk_widget_show(w);
	    parent3 = w;

	    ms->channel_values_da = w = gtk_drawing_area_new();
	    gtk_widget_add_events(
		w,
		GDK_EXPOSURE_MASK | GDK_STRUCTURE_MASK |
		GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "expose_event",
		GTK_SIGNAL_FUNC(YMixerScaleRulerEventCB), ms
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "configure_event",
		GTK_SIGNAL_FUNC(YMixerScaleRulerEventCB), ms
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "button_press_event",
		GTK_SIGNAL_FUNC(YMixerScaleRulerEventCB), ms
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "button_release_event",
		GTK_SIGNAL_FUNC(YMixerScaleRulerEventCB), ms
	    );
	    gtk_widget_set_usize(w, -1, 15);
	    gtk_container_add(GTK_CONTAINER(parent3), w);
	    gtk_widget_modify_style(w, rcstyle2);
	    gtk_widget_show(w);


	    gtk_rc_style_unref(rcstyle2);
	}

	if(rcstyle != NULL)
	    gtk_rc_style_unref(rcstyle);

	return(ms);
}

/*
 *	Deletes the mixer scale.
 */
void YMixerScaleDelete(ymixer_scale_struct *ms)
{
	GtkWidget **w;
	GtkAdjustment **adj;


	if(ms == NULL)
	    return;

#define DO_DESTROY_WIDGET       \
{ if(*w != NULL) { gtk_widget_destroy(*w); *w = NULL; } }
#define DO_UNREF_ADJUSTMENT	\
{ if(*adj != NULL) { gtk_object_unref(GTK_OBJECT(*adj)); *adj = NULL; } }

	w = &ms->amplitude_ruler_left_da;
	DO_DESTROY_WIDGET
	w = &ms->amplitude_ruler_right_da;
	DO_DESTROY_WIDGET
	w = &ms->mute_btn;
	DO_DESTROY_WIDGET
	w = &ms->center_btn;
	DO_DESTROY_WIDGET
	w = &ms->undo_btn;
	DO_DESTROY_WIDGET
	w = &ms->channel_values_da;
	DO_DESTROY_WIDGET
	w = &ms->balance_scale;
	DO_DESTROY_WIDGET
	w = &ms->amplitude_scale;
	DO_DESTROY_WIDGET
	w = &ms->toplevel;
	DO_DESTROY_WIDGET

	adj = &ms->balance_adj;
	DO_UNREF_ADJUSTMENT
	adj = &ms->amplitude_adj;
	DO_UNREF_ADJUSTMENT

#undef DO_UNREF_ADJUSTMENT
#undef DO_DESTROY_WIDGET

	g_free(ms);
}

/*
 *	Creates all mixer scales defined by Y2.
 *
 *	Existing mixer scales will be deleted.
 */
void YMixerScaleCreateAll(ymixer_struct *ym)
{
	gint i, mixer_num, mixer_code;
	gchar *s;
	GtkWidget *w, *parent;
	const cfg_intlist_struct *intlist;
	ymixer_scale_struct *ms;
	yc_core_struct *core_ptr;
	gchar *mixer_name[] = YMixerNames;


	if(ym == NULL)
	    return;

	core_ptr = YC_CORE(ym->core_ptr);
	if(core_ptr == NULL)
	    return;


	/* Delete all mixer scales */
	for(i = 0; i < ym->total_mixer_scales; i++)
	    YMixerScaleDelete(ym->mixer_scale[i]);
	g_free(ym->mixer_scale);
	ym->mixer_scale = NULL;
	ym->total_mixer_scales = 0;

	/* Destroy old mixer scale client hbox */
	w = ym->mixers_hbox_client;
	if(w != NULL)
	{
	    gtk_widget_destroy(w);
	    ym->mixers_hbox_client = w = NULL;
	}


	/* Get configuration specified list of YMixer scales */
	intlist = YC_GET_CFG_INTLIST(
	    core_ptr, YC_CFG_PARM_MIXER_MIXER_SCALES
	);
	if(intlist == NULL)
	    return;


	/* (Re)create client hbox to hold mixer scales */
	parent = ym->mixers_hbox;
	if(parent == NULL)
	    return;
	ym->mixers_hbox_client = w = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(parent), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent = w;

	/* Create mixer scales */
	ym->total_mixer_scales = g_list_length(intlist->list);
	if(ym->total_mixer_scales > 0)
	{
	    GList *glist = intlist->list;

	    ym->mixer_scale = (ymixer_scale_struct **)g_malloc0(
		ym->total_mixer_scales * sizeof(ymixer_scale_struct *)
	    );

	    for(i = 0; i < ym->total_mixer_scales; i++)
	    {
		if(glist == NULL)
		    break;

		/* Get mixer number and mixer code */
		mixer_num = CLIP(
		    (gint)glist->data - YMixerCodeBaseOffset,
		    0, YTotalMixers - 1
		);
		mixer_code = mixer_num + YMixerCodeBaseOffset;

		s = g_strdup(mixer_name[mixer_num]);
		*s = toupper(*s);		/* Capitalize */
		ym->mixer_scale[i] = ms = YMixerScaleNew(
		    ym, mixer_code, s, s
		);
		g_free(s);

		glist = g_list_next(glist);
	    }
	}
}

/*
 *	Create a new ymixer.
 */
ymixer_struct *YMixerNew(gpointer core_ptr)
{
	GdkWindow *window;
	GtkAccelGroup *accelgrp;
	GtkWidget *w, *parent, *parent2, *toplevel;
	GtkScrolledWindow *sw;
	connection_bar_struct *cb;
	dspview_struct *dv;
	yc_core_struct *cp = YC_CORE(core_ptr);
	ymixer_struct *ym = (ymixer_struct *)g_malloc0(
	    sizeof(ymixer_struct)
	);
	if(ym == NULL)
	    return(NULL);


	/* Reset values */
	ym->initialized = TRUE;
	ym->map_state = FALSE;
	ym->busy_count = 0;
	ym->accelgrp = accelgrp = gtk_accel_group_new();
	ym->core_ptr = core_ptr;

	ym->mixer_scale = NULL;
	ym->total_mixer_scales = 0;

	ym->con = NULL;
	ym->yshm_buf = NULL;
	ym->yshm_buf_len = 0;
	ym->yshm_id = -1;

	ym->start_time = 0;
	ym->audio_mode_name = NULL;
	ym->sample_rate = 0;
	ym->sample_size = 0;
	ym->channels = 0;
	ym->cycle_us = 0;
	ym->cycle_load = 0.0f;
	ym->num_fragments = 0;
	ym->fragment_size = 0;
	ym->flip_stereo = FALSE;
	ym->shellout = FALSE;
	ym->version_major = ym->version_minor = 0;
	ym->server_name = ym->sound_name = NULL;


	/* Create toplevel */
	ym->toplevel = toplevel = w = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_widget_set_name(w, "ymixer window");
	if(cp != NULL)
	{
	    if(YC_GET_CFG_I(cp, YC_CFG_PARM_MIXER_SAVE_POSITIONS))
	    {
		gtk_widget_set_uposition(
		    w,
		    YC_GET_CFG_I(cp, YC_CFG_PARM_MIXER_X),
		    YC_GET_CFG_I(cp, YC_CFG_PARM_MIXER_Y)
		);
		gtk_widget_set_usize(
		    w,
		    YC_GET_CFG_I(cp, YC_CFG_PARM_MIXER_WIDTH),
		    YC_GET_CFG_I(cp, YC_CFG_PARM_MIXER_HEIGHT)
		);
	    }
	}
	gtk_window_set_policy(GTK_WINDOW(w), TRUE, TRUE, TRUE);
	gtk_widget_realize(w);
	window = w->window;
	if(window != NULL)
	{
	    gdk_window_set_decorations(
		window,
		GDK_DECOR_BORDER | GDK_DECOR_RESIZEH |
		GDK_DECOR_TITLE | GDK_DECOR_MENU |
		GDK_DECOR_MINIMIZE | GDK_DECOR_MAXIMIZE
	    );
	    gdk_window_set_functions(
		window,
		GDK_FUNC_MOVE | GDK_FUNC_RESIZE | GDK_FUNC_MINIMIZE |
		GDK_FUNC_MAXIMIZE | GDK_FUNC_CLOSE
	    );
	    GUISetWMIcon(window, (guint8 **)icon_ymixer_48x48_xpm);
	}
	gtk_accel_group_attach(accelgrp, GTK_OBJECT(w));
	gtk_signal_connect(
	    GTK_OBJECT(w), "delete_event",
	    GTK_SIGNAL_FUNC(YMixerDeleteEventCB), ym
	);
	parent = w;


	/* Create main vbox */
	w = gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(parent), w);
	gtk_widget_show(w);
	parent = w;


	/* Create connection bar */
	ym->conbar = cb = ConBarNew(
	    CONBAR_DISPLAY_FLAG_WINDOW_MENU | CONBAR_DISPLAY_FLAG_LOCATION |
	    CONBAR_DISPLAY_FLAG_DSPVIEW,
	    cp->dsp_bg_img_small_width + 2, cp->dsp_bg_img_small_height + 2,
	    ym,
	    YMixerConnectCB, YMixerYServerListCB, YMixerCloseCB,
	    YMixerNewYConsoleCB, YMixerNewYAudioCB, YMixerNewYMixerCB,
	    YMixerNewYPlayerCB, YMixerOptionsCB,
	    YMixerRefreshCB, YMixerShelloutCB, YMixerShutdownCB,
	    parent, accelgrp
	);
	w = CONBAR_AUDIO(cb);
	dv = ConBarGetDSPView(cb);
	ConBarMap(cb);

	if(cp != NULL)
	    DSPViewSetBG(
		dv, cp->dsp_bg_img_small,
		cp->dsp_bg_img_small_width, cp->dsp_bg_img_small_height,
		cp->dsp_bg_img_small_bpl, 4
	    );
	DSPViewSetDisplayStyle(dv, DSPVIEW_DISPLAY_STYLE_NOVA);
	DSPViewSetMergeMode(dv, DSPVIEW_MERGE_MODE_SET);
	DSPViewSetDSPColor(dv, DSPVIEW_DEF_DSP_COLOR);  /* 0xAABBGGRR */
	w = DSPVIEW_VIEW(dv);
	gtk_widget_add_events(
	    w, GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "button_press_event",
	    GTK_SIGNAL_FUNC(YMixerDSPViewButtonCB), ym
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "button_release_event",
	    GTK_SIGNAL_FUNC(YMixerDSPViewButtonCB), ym
	);

	/* Event box to hold scrolled window so that style colors come
	 * out right (no gaps between scroll bars).
	 */
	ym->mixers_ebox = w = gtk_event_box_new();
	gtk_widget_add_events(w, GDK_EXPOSURE_MASK | GDK_STRUCTURE_MASK);
	gtk_box_pack_start(GTK_BOX(parent), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent2 = w;

	/* Create scrolled window for mixers */
	ym->mixers_sw = w = gtk_scrolled_window_new(NULL, NULL);
	sw = GTK_SCROLLED_WINDOW(w);
	gtk_scrolled_window_set_policy(
	    GTK_SCROLLED_WINDOW(w),
	    GTK_POLICY_AUTOMATIC,
	    GTK_POLICY_AUTOMATIC
	);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_widget_show(w);
	parent2 = w;

	/* Main hbox for mixers */
	ym->mixers_hbox = w = gtk_hbox_new(FALSE, 0);
	gtk_scrolled_window_add_with_viewport(
	    GTK_SCROLLED_WINDOW(parent2), w
	);
	gtk_widget_show(w);
	parent2 = w;


	/* Update title */
	YMixerSetTitle(ym, NULL);
	YMixerUpdateMenus(ym);

	return(ym);
}

/*
 *	Update menus on ymixer.
 */
void YMixerUpdateMenus(ymixer_struct *ym)
{
	gboolean sensitive;
	gint i;
	GtkWidget *w;
	ymixer_scale_struct *ms;


	if(ym == NULL)
	    return;

#define DO_SET_SENSITIVE	\
{ if(w != NULL) gtk_widget_set_sensitive(w, sensitive); }

	sensitive = (YMIXER_CONNECTION(ym) != NULL) ? TRUE : FALSE;

	for(i = 0; i < ym->total_mixer_scales; i++)
	{
	    ms = ym->mixer_scale[i];
	    if(ms == NULL)
		continue;

	    w = ms->balance_scale;
	    DO_SET_SENSITIVE
	    w = ms->amplitude_scale;
	    DO_SET_SENSITIVE
	}


#undef DO_SET_SENSITIVE
}

/*
 *	Mark ymixer as busy or not.
 */
void YMixerSetBusy(ymixer_struct *ym, gboolean is_busy)
{
	GdkCursor *cursor;
	GtkWidget *w;
	yc_core_struct *core_ptr;


	if(ym == NULL)
	    return;

	core_ptr = YC_CORE(ym->core_ptr);
	if(core_ptr == NULL)
	    return;

	w = YMIXER_TOPLEVEL(ym);
	if(w != NULL)
	{
	    if(is_busy)
	    {
		/* Increase busy count */
		ym->busy_count++;

		/* If already busy then don't change anything */
		if(ym->busy_count > 1)
		    return;

		cursor = core_ptr->cursor_busy;
	    }
	    else
	    {
		/* Reduce busy count */
		ym->busy_count--;
		if(ym->busy_count < 0)
		    ym->busy_count = 0;

		/* If still busy do not change anything */
		if(ym->busy_count > 0)
		    return;

		cursor = NULL;  /* Use default cursor */
	    }

	    /* Update toplevel window's cursor */
	    if(w->window != NULL)
	    {
		gdk_window_set_cursor(w->window, cursor);
		gdk_flush();
	    }
	}
}

/*
 *	Map ymixer.
 */
void YMixerMap(ymixer_struct *ym)
{
	if(ym == NULL)
	    return;

	if(TRUE)
	{
	    GtkWidget *w = YMIXER_TOPLEVEL(ym);
	    if(w != NULL)
		gtk_widget_show_raise(w);
	    ym->map_state = TRUE;
	}
}

/*
 *	Unmap ymixer.
 */
void YMixerUnmap(ymixer_struct *ym)
{
	if(ym == NULL)
	    return;

	if(TRUE)
	{
	    GtkWidget *w = YMIXER_TOPLEVEL(ym);
	    if(w != NULL)
		gtk_widget_hide(w);
	    ym->map_state = FALSE;
	}
}

/*
 *	Delete the ymixer.
 */
void YMixerDelete(ymixer_struct *ym)
{
	gint i;
	GtkWidget **w;


	if(ym == NULL)
	    return;

	if(ym->initialized)
	{
#define DO_DESTROY_WIDGET       \
{ if(*w != NULL) { gtk_widget_destroy(*w); *w = NULL; } }

	    /* Disconnect from Y server as needed, this will reset
	     * all connection values.
	     */
	    YMixerDoDisconnect(ym, FALSE);

	    /* Delete all mixer scales */
	    for(i = 0; i < ym->total_mixer_scales; i++)
		YMixerScaleDelete(ym->mixer_scale[i]);
	    g_free(ym->mixer_scale);
	    ym->mixer_scale = NULL;
	    ym->total_mixer_scales = 0;

	    w = &ym->mixers_hbox_client;
	    DO_DESTROY_WIDGET
	    w = &ym->mixers_hbox;
	    DO_DESTROY_WIDGET
	    w = &ym->mixers_sw;
	    DO_DESTROY_WIDGET
	    w = &ym->mixers_ebox;
	    DO_DESTROY_WIDGET

	    /* Delete connection bar */
	    ConBarDelete(YMIXER_CONBAR(ym));
	    ym->conbar = NULL;

	    /* Destroy widgets */
	    w = &ym->toplevel;
	    DO_DESTROY_WIDGET

	    if(ym->accelgrp != NULL)
	    {
		gtk_accel_group_unref(ym->accelgrp);
		ym->accelgrp = NULL;
	    }

	}

	/* Deallocate structure itself */
	g_free(ym);
}
