/*
			     YMixer Window
 */

#ifndef YMIXER_H
#define YMIXER_H

#include <Y2/Y.h>
#include <Y2/Ylib.h>
#include <gtk/gtk.h>

#include "conbar.h"

/*
 *	YMixer scale label type:
 */
#define YMIXER_MIXER_SCALE_LABEL_TYPE_NAME		0
#define YMIXER_MIXER_SCALE_LABEL_TYPE_CONICAL_NAME	1
#define YMIXER_MIXER_SCALE_LABEL_TYPE_ICON		2

/*
 *	YMixer scale structure:
 */
typedef struct {

	GtkWidget	*toplevel,		/* A GtkFrame. */
			*balance_scale, *amplitude_scale,
			*amplitude_ruler_left_da,
			*amplitude_ruler_right_da,
			*mute_btn, *center_btn, *undo_btn,
			*channel_values_da;

	gint mixer_code;		/* One of YMixerCode*. */

	GtkAdjustment	*balance_adj, *amplitude_adj;
	gfloat		balance_last, amplitude_last;
	gint freeze_count;
	gpointer ym;	/* Pointer back to ymixer_struct. */

} ymixer_scale_struct;

#define YMIXER_SCALE(p)	((ymixer_scale_struct *)(p))

/*
 *	YMixer window structure:
 */
typedef struct {

	gboolean initialized;
	gboolean map_state;
	gint busy_count;
	GtkAccelGroup *accelgrp;
	gpointer core_ptr;

	GtkWidget 	*toplevel;
	connection_bar_struct *conbar;
	GtkWidget	*mixers_ebox,
			*mixers_sw,
			*mixers_hbox,
			*mixers_hbox_client;

	ymixer_scale_struct **mixer_scale;
	gint total_mixer_scales;


	/* Y server values. */
	YConnection *con;

	guint8 *yshm_buf;
	gint yshm_buf_len, yshm_id;

	gulong start_time;
	gchar *audio_mode_name;
	gint sample_rate, sample_size, channels;
	glong cycle_us;
	gfloat cycle_load;
	gint num_fragments, fragment_size;
	gboolean flip_stereo, shellout;
	gint version_major, version_minor;
	gchar *server_name, *sound_name;

	gfloat channel_volume_avg[4];

} ymixer_struct;

#define YMIXER(p)	((ymixer_struct *)(p))
#define YMIXER_TOPLEVEL(p)		(((p) != NULL) ? \
 (YMIXER(p)->toplevel) : NULL)
#define YMIXER_MIXERS_SCROLLED_WINDOW(p)	(((p) != NULL) ? \
 (YMIXER(p)->mixers_sw) : NULL)
#define YMIXER_MIXERS_HBOX(p)			(((p) != NULL) ? \
 (YMIXER(p)->mixers_hbox) : NULL)
#define YMIXER_MIXERS_HBOX_CLIENT(p)		(((p) != NULL) ? \
 (YMIXER(p)->mixers_hbox_client) : NULL)
#define YMIXER_CONBAR(p)	(((p) != NULL) ? \
 (YMIXER(p)->conbar) : NULL)

#define YMIXER_IS_CONNECTED(p)	(((p) != NULL) ? \
 ((YMIXER(p)->con) != NULL) : FALSE)
#define YMIXER_CONNECTION(p)	(((p) != NULL) ? \
 (YMIXER(p)->con) : NULL)

#define YMIXER_MIXER_SCALE(p)	(((p) != NULL) ? \
 (YMIXER(p)->mixer_scale) : NULL)
#define YMIXER_MIXER_SCALES(p)	(((p) != NULL) ? \
 (YMIXER(p)->total_mixer_scales) : 0)

#define YMIXER_SET_CONBAR_AUDIO(p)	\
{ if((p) != NULL) \
  ConBarSetAudio( \
   YMIXER_CONBAR(p), \
   (p)->start_time, (p)->audio_mode_name, \
   (p)->sample_rate, (p)->sample_size, (p)->channels, \
   (p)->cycle_us, (p)->cycle_load, \
   (p)->num_fragments, (p)->fragment_size, \
   (p)->flip_stereo, (p)->shellout, \
   (p)->yshm_buf, (p)->yshm_buf_len, \
   (p)->channel_volume_avg, \
   (p)->version_major, (p)->version_minor, \
   (p)->server_name, (p)->sound_name \
  ); }


#define YMIXER_SCALE(p)	((ymixer_scale_struct *)(p))

#define YMIXER_SCALE_TOPLEVEL(p)	(((p) != NULL) ? \
 (YMIXER_SCALE(p)->toplevel) : NULL)


extern void YMixerScaleCalculateBATo2C(
	gfloat bal, gfloat amp, gfloat *left, gfloat *right
);
extern void YMixerScaleCalculate2CToBA(
	gfloat left, gfloat right, gfloat *bal, gfloat *amp
);
extern GtkRcStyle *YMixerMixerScaleRCStyleNew(ymixer_struct *ym);
extern guchar **YMixerScaleGetXpmDataFromMixerName(const gchar *name);

extern void YMixerSetTitle(ymixer_struct *ym, const gchar *s);
extern gchar *YMixerGetLocation(ymixer_struct *ym);
extern gint YMixerDoConnect(
	ymixer_struct *ym,
	const gchar *address, const gchar *start_command
);
extern void YMixerDoDisconnect(ymixer_struct *ym, gboolean force_shutdown);
extern void YMixerSavePosition(ymixer_struct *ym);

extern ymixer_scale_struct *YMixerScaleMatch(
	ymixer_struct *ym, gint mixer_code, gint *n
);
extern void YMixerScaleSetSensitive(
	ymixer_struct *ym, gint mixer_code,
	gboolean sensitive
);
extern void YMixerScaleSetValue2C(
	ymixer_struct *ym, gint mixer_code,
	gfloat left, gfloat right
);
extern ymixer_scale_struct *YMixerScaleNew(
	ymixer_struct *ym,
	gint mixer_code, const gchar *name, const gchar *short_name
);
extern void YMixerScaleDelete(ymixer_scale_struct *ms);
extern void YMixerScaleCreateAll(ymixer_struct *ym);

extern ymixer_struct *YMixerNew(gpointer core_ptr);
extern void YMixerUpdateMenus(ymixer_struct *ym);
extern void YMixerSetBusy(ymixer_struct *ym, gboolean is_busy);
extern void YMixerMap(ymixer_struct *ym);
extern void YMixerUnmap(ymixer_struct *ym);
extern void YMixerDelete(ymixer_struct *ym);



#endif	/* YMIXER_H */
