#include <stdlib.h>
#include <string.h>

#include <Y2/Y.h>
#include <Y2/Ylib.h>
#include <Y2/Ymixercodes.h>

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include "guiutils.h"

#include "cfg.h"
#include "conbar.h"
#include "ymixer.h"
#include "ymixercb.h"
#include "yc.h"
#include "ycop.h"
#include "yccfglist.h"
#include "config.h"


gint YMixerDeleteEventCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
);
void YMixerDestroyCB(GtkObject *object, gpointer data);

gint YMixerDSPViewButtonCB(
	GtkWidget *widget, GdkEventButton *button, gpointer data
);

gint YMixerScaleKeyEventCB(
	GtkWidget *widget, GdkEventKey *key, gpointer data
);
void YMixerScaleValueChangeCB(GtkAdjustment *adj, gpointer data);
gint YMixerScaleRulerEventCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
);

void YMixerConnectCB(
	GtkWidget *widget, const gchar *s, gpointer data
);
void YMixerYServerListCB(GtkWidget *widget, gpointer data);
void YMixerCloseCB(GtkWidget *widget, gpointer data);
void YMixerNewYConsoleCB(GtkWidget *widget, gpointer data);
void YMixerNewYMixerCB(GtkWidget *widget, gpointer data);
void YMixerNewYAudioCB(GtkWidget *widget, gpointer data);
void YMixerNewYPlayerCB(GtkWidget *widget, gpointer data);
void YMixerOptionsCB(GtkWidget *widget, gpointer data);
void YMixerRefreshCB(GtkWidget *widget, gpointer data);
gboolean YMixerShelloutCB(
	GtkWidget *widget, gboolean shellout, gpointer data
);
void YMixerShutdownCB(GtkWidget *widget, gpointer data);
void YMixerManageCB(ymixer_struct *ym);
void YMixerManage2CB(ymixer_struct *ym);

void YMixerReconfiguredCB(
	ymixer_struct *ym, gpointer core_ptr
);


#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define ABSOLUTE(x)	(((x) < 0) ? ((x) * -1) : (x))


/*
 *	Toplevel "delete_event" signal callback.
 */
gint YMixerDeleteEventCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
)
{
	YMixerCloseCB(widget, data);
	return(TRUE);
}

/*
 *	Toplevel "destroy" signal callback.
 */
void YMixerDestroyCB(GtkObject *object, gpointer data)
{

}


/*
 *      DSPView "button_press_event" signal callback.
 */
gint YMixerDSPViewButtonCB(
	GtkWidget *widget, GdkEventButton *button, gpointer data
)
{
	gint status = FALSE;
	ymixer_struct *ym = YMIXER(data);
	if(ym == NULL)
	    return(status);

	switch(button->button)
	{
	  case 3:
	    switch((gint)button->type)
	    {
	      case GDK_BUTTON_PRESS:
		YCNewOptWinSelectPage(
		    YC_CORE(ym->core_ptr), YMIXER_TOPLEVEL(ym),
		    1
		);
		status = TRUE;
		break;

	      case GDK_BUTTON_RELEASE:
		status = TRUE;
		break;
	    }
	}

	return(status);
}


/*
 *	Mixer scale "key_press_event" signal callback.
 */
gint YMixerScaleKeyEventCB(
	GtkWidget *widget, GdkEventKey *key, gpointer data
)
{
	gint status = FALSE;
	gboolean keystate;
	GtkAdjustment *hadj, *vadj;
	GtkWidget *w;
	GtkScrolledWindow *sw;
	ymixer_struct *ym;
	ymixer_scale_struct *ms = YMIXER_SCALE(data);
	if((key == NULL) || (ms == NULL))
	    return(status);

	ym = YMIXER(ms->ym);
	if(ym == NULL)
	    return(status);

	keystate = (key->type == GDK_KEY_PRESS) ? TRUE : FALSE;
	w = YMIXER_MIXERS_SCROLLED_WINDOW(ym);
	if(w == NULL)
	    return(status);
	else
	    sw = GTK_SCROLLED_WINDOW(w);
	hadj = (GtkAdjustment *)GTK_RANGE(sw->hscrollbar)->adjustment;
	vadj = (GtkAdjustment *)GTK_RANGE(sw->vscrollbar)->adjustment;

#define DO_STOP_KEY_SIGNAL      \
{ if(widget != NULL) \
 gtk_signal_emit_stop_by_name( \
  GTK_OBJECT(widget), \
  keystate ? "key_press_event" : "key_release_event" \
 ); \
}
#define DO_CLIP_HADJ		\
{ \
 if(hadj->value > (hadj->upper - hadj->page_size)) \
  hadj->value = hadj->upper - hadj->page_size; \
 if(hadj->value < hadj->lower) \
  hadj->value = hadj->lower; \
}

	switch(key->keyval)
	{
	  case GDK_Left:
	  case GDK_KP_Left:
	  case GDK_Up:
	  case GDK_KP_Up:
	    if(key->state & GDK_CONTROL_MASK)
	    {
		if(keystate)
		{
		    hadj->value -= (key->state & GDK_SHIFT_MASK) ?
			hadj->page_increment : hadj->step_increment;
		    if(hadj->value < hadj->lower)
			hadj->value = hadj->lower;
		    gtk_signal_emit_by_name(
			GTK_OBJECT(hadj), "value_changed"
		    );
		}
		DO_STOP_KEY_SIGNAL
		status = TRUE;
	    }
	    break;

	  case GDK_Right:
	  case GDK_KP_Right:
	  case GDK_Down:
	  case GDK_KP_Down:
	    if(key->state & GDK_CONTROL_MASK)
	    {
		if(keystate)
		{
		    hadj->value += (key->state & GDK_SHIFT_MASK) ?
			hadj->page_increment : hadj->step_increment;
		    if(hadj->value > (hadj->upper - hadj->page_size))
			hadj->value = hadj->upper - hadj->page_size;
		    gtk_signal_emit_by_name(
			GTK_OBJECT(hadj), "value_changed"
		    );
		}
		DO_STOP_KEY_SIGNAL
		status = TRUE;
	    }
	    break;

	  case '-':
	  case 'b':
	  case 'p':
	  case GDK_Page_Up:
	  case GDK_KP_Page_Up:
	    if(TRUE)
	    {
		if(keystate)
		{
		    hadj->value -= hadj->page_increment;
		    if(hadj->value < hadj->lower)
			hadj->value = hadj->lower;
		    gtk_signal_emit_by_name(
			GTK_OBJECT(hadj), "value_changed"
		    );
		}
		DO_STOP_KEY_SIGNAL
		status = TRUE;
	    }
	    break;

	  case '+':
	  case 'n':
	  case GDK_space:
	  case GDK_Page_Down:
	  case GDK_KP_Page_Down:
	    if(TRUE)
	    {
		if(keystate)
		{
		    hadj->value += hadj->page_increment;
		    if(hadj->value > (hadj->upper - hadj->page_size))
			hadj->value = hadj->upper - hadj->page_size;
		    gtk_signal_emit_by_name(
			GTK_OBJECT(hadj), "value_changed"
		    );
		}
		DO_STOP_KEY_SIGNAL
		status = TRUE;
	    }
	    break;

	  case '1':
	  case GDK_Home:
	  case GDK_KP_Home:
	    if(TRUE)
	    {
		if(keystate)
		{
		    hadj->value = hadj->lower;
		    gtk_signal_emit_by_name(
			GTK_OBJECT(hadj), "value_changed"
		    );
		}
		DO_STOP_KEY_SIGNAL
		status = TRUE;
	    }
	    break;

	  case '0':
	  case GDK_End:
	  case GDK_KP_End:
	    if(TRUE)
	    {
		if(keystate)
		{
		    hadj->value = hadj->upper - hadj->page_size;
		    gtk_signal_emit_by_name(
			GTK_OBJECT(hadj), "value_changed"
		    );
		}
		DO_STOP_KEY_SIGNAL
		status = TRUE;
	    }
	    break;

	  case '2':
	    if(keystate)
	    {
		hadj->value = (0.2f * (hadj->upper - hadj->lower - hadj->page_size))
		    + hadj->lower;
		DO_CLIP_HADJ
		gtk_signal_emit_by_name(
		    GTK_OBJECT(hadj), "value_changed"
		);
	    }
	    DO_STOP_KEY_SIGNAL
	    status = TRUE;
	    break;

	  case '3':
	    if(keystate)
	    {
		hadj->value = (0.3f * (hadj->upper - hadj->lower - hadj->page_size))
		    + hadj->lower;
		DO_CLIP_HADJ
		gtk_signal_emit_by_name(
		    GTK_OBJECT(hadj), "value_changed"
		);
	    }
	    DO_STOP_KEY_SIGNAL
	    status = TRUE;
	    break;

	  case '4':
	    if(keystate)
	    {
		hadj->value = (0.4f * (hadj->upper - hadj->lower - hadj->page_size))
		    + hadj->lower;
		DO_CLIP_HADJ
		gtk_signal_emit_by_name(
		    GTK_OBJECT(hadj), "value_changed"
		);
	    }
	    DO_STOP_KEY_SIGNAL
	    status = TRUE;
	    break;

	  case '5':
	    if(keystate)
	    {
		hadj->value = (0.5f * (hadj->upper - hadj->lower - hadj->page_size))
		    + hadj->lower;
		DO_CLIP_HADJ
		gtk_signal_emit_by_name(
		    GTK_OBJECT(hadj), "value_changed"
		);
	    }
	    DO_STOP_KEY_SIGNAL
	    status = TRUE;
	    break;

	  case '6':
	    if(keystate)
	    {
		hadj->value = (0.6f * (hadj->upper - hadj->lower - hadj->page_size))
		    + hadj->lower;
		DO_CLIP_HADJ
		gtk_signal_emit_by_name(
		    GTK_OBJECT(hadj), "value_changed"
		);
	    }
	    DO_STOP_KEY_SIGNAL
	    status = TRUE;
	    break;

	  case '7':
	    if(keystate)
	    {
		hadj->value = (0.7f * (hadj->upper - hadj->lower - hadj->page_size))
		    + hadj->lower;
		DO_CLIP_HADJ
		gtk_signal_emit_by_name(
		    GTK_OBJECT(hadj), "value_changed"
		);
	    }
	    DO_STOP_KEY_SIGNAL
	    status = TRUE;
	    break;

	  case '8':
	    if(keystate)
	    {
		hadj->value = (0.8f * (hadj->upper - hadj->lower - hadj->page_size))
		    + hadj->lower;
		DO_CLIP_HADJ
		gtk_signal_emit_by_name(
		    GTK_OBJECT(hadj), "value_changed"
		);
	    }
	    DO_STOP_KEY_SIGNAL
	    status = TRUE;
	    break;

	  case '9':
	    if(keystate)
	    {
		hadj->value = (0.8f * (hadj->upper - hadj->lower - hadj->page_size))
		    + hadj->lower;
		DO_CLIP_HADJ
		gtk_signal_emit_by_name(
		    GTK_OBJECT(hadj), "value_changed"
		);
	    }
	    DO_STOP_KEY_SIGNAL
	    status = TRUE;
	    break;

	}

#undef DO_CLIP_HADJ
#undef DO_STOP_KEY_SIGNAL

	return(status);
}

/*
 *	Mixer scale GtkAdjustment "value_changed" signal callback.
 */
void YMixerScaleValueChangeCB(GtkAdjustment *adj, gpointer data)
{
	gfloat left, right;
	GdkEventExpose e;
	GtkAdjustment *balance_adj, *amplitude_adj;
	GtkWidget *w;
	ymixer_struct *ym;
	ymixer_scale_struct *ms = YMIXER_SCALE(data);
	if(ms == NULL)
	    return;

	/* Frozen? Implies values being changed implicitly and not
	 * to update values to Y server.
	 */
	if(ms->freeze_count > 0)
	    return;

	ym = YMIXER(ms->ym);
	if(ym == NULL)
	    return;

	balance_adj = ms->balance_adj;
	amplitude_adj = ms->amplitude_adj;
	if((balance_adj == NULL) || (amplitude_adj == NULL))
	    return;

	ms->freeze_count++;
	YMixerScaleCalculateBATo2C(
	    balance_adj->value, 1.0f - amplitude_adj->value,
	    &left, &right
	);
	YSetMixerChannel(
	    YMIXER_CONNECTION(ym), ms->mixer_code, left, right
	);
	ms->freeze_count--;

	w = ms->channel_values_da;
	memset(&e, 0x00, sizeof(GdkEventExpose));
	e.type = GDK_EXPOSE;
	e.window = (w != NULL) ? w->window : NULL;
	e.send_event = TRUE;
	YMixerScaleRulerEventCB(w, (GdkEvent *)&e, ms);
}

/*
 *	Mixer scale ruler drawing area event signal callback.
 */
gint YMixerScaleRulerEventCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
)
{
	gint status = FALSE;
	ymixer_scale_struct *ms = YMIXER_SCALE(data);
	if((widget == NULL) || (event == NULL) || (ms == NULL))
	    return(status);

	switch((gint)event->type)
	{
	  case GDK_EXPOSE:
	  case GDK_CONFIGURE:
	    if(widget == ms->amplitude_ruler_left_da)
	    {
		GtkWidget *w = widget;
		gint state = GTK_WIDGET_STATE(w);
		GdkWindow *window = w->window;
		GtkStyle *style = gtk_widget_get_style(w);
		gint	width = w->allocation.width,
			height = w->allocation.height;
		if((window != NULL) && (width > 0) && (height > 0))
		{
		    gint y, yo = 20, theight = height - (2 * yo);

#define DO_DRAW_LINE1    \
{ gdk_draw_line(window, style->dark_gc[state], \
   width * 0.0, yo + y - 1, width, yo + y - 1); \
  gdk_draw_line(window, style->fg_gc[state], \
   width * 0.0, yo + y, width, yo + y); \
  gdk_draw_line(window, style->light_gc[state], \
   width * 0.0, yo + y + 1, width, yo + y + 1); \
}
#define DO_DRAW_LINE2	\
{ gdk_draw_line(window, style->dark_gc[state], \
   width * 0.5, yo + y - 1, width, yo + y - 1); \
  gdk_draw_line(window, style->fg_gc[state], \
   width * 0.5, yo + y, width, yo + y); \
  gdk_draw_line(window, style->light_gc[state], \
   width * 0.5, yo + y + 1, width, yo + y + 1); \
}

		    gdk_window_clear(window);
		    y = theight * 0.0;
		    DO_DRAW_LINE1
		    y = theight * 0.25;
		    DO_DRAW_LINE2
		    y = theight * 0.5;
		    DO_DRAW_LINE1
		    y = theight * 0.75;
		    DO_DRAW_LINE2
		    y = theight - 1;
		    DO_DRAW_LINE1

#undef DO_DRAW_LINE1
#undef DO_DRAW_LINE2
		}
		status = TRUE;
	    }
	    else if(widget == ms->amplitude_ruler_right_da)
	    {
		GtkWidget *w = widget;
		gint state = GTK_WIDGET_STATE(w);
		GdkWindow *window = w->window;
		GtkStyle *style = gtk_widget_get_style(w);
		gint    width = w->allocation.width,
			height = w->allocation.height;
		if((window != NULL) && (width > 0) && (height > 0))
		{
		    gint y, yo = 20, theight = height - (2 * yo);

#define DO_DRAW_LINE1	\
{ gdk_draw_line(window, style->dark_gc[state], \
   0, yo + y - 1, width * 1.0, yo + y - 1); \
  gdk_draw_line(window, style->fg_gc[state], \
   0, yo + y, width * 1.0, yo + y); \
  gdk_draw_line(window, style->light_gc[state], \
   0, yo + y + 1, width * 1.0, yo + y + 1); \
}
#define DO_DRAW_LINE2	\
{ gdk_draw_line(window, style->dark_gc[state], \
   0, yo + y - 1, width * 0.5, yo + y - 1); \
  gdk_draw_line(window, style->fg_gc[state], \
   0, yo + y, width * 0.5, yo + y); \
  gdk_draw_line(window, style->light_gc[state], \
   0, yo + y + 1, width * 0.5, yo + y + 1); \
}

		    gdk_window_clear(window);
		    y = theight * 0.0;
		    DO_DRAW_LINE1
		    y = theight * 0.25;
		    DO_DRAW_LINE2
		    y = theight * 0.5;
		    DO_DRAW_LINE1
		    y = theight * 0.75;
		    DO_DRAW_LINE2
		    y = theight - 1;
		    DO_DRAW_LINE1

#undef DO_DRAW_LINE1
#undef DO_DRAW_LINE2
		}
		status = TRUE;
	    }
	    else if(widget == ms->channel_values_da)
	    {
		GtkWidget *w = widget;
		GtkAdjustment	*bal_adj = ms->balance_adj,
				*amp_adj = ms->amplitude_adj; 
		gint state = GTK_WIDGET_STATE(w);
		GdkWindow *window = w->window;
		GtkStyle *style = gtk_widget_get_style(w);
		gint    width = w->allocation.width,
			height = w->allocation.height;
		if((window != NULL) && (width > 0) && (height > 0) &&
		   (bal_adj != NULL) && (amp_adj != NULL) && (style->font != NULL)
		)
		{
		    GdkFont *font = style->font;
		    gchar s[80];
		    gint x, y;
		    gint lbearing, rbearing, swidth, ascent, descent;
		    gfloat left, right;


		    YMixerScaleCalculateBATo2C(
			bal_adj->value, 1.0f - amp_adj->value, &left, &right
		    );

		    /* Clear background. */
		    gdk_draw_rectangle(
			window, style->base_gc[state], TRUE,
			0, 0, width, height
		    );

		    /* Draw left channel value. */
		    sprintf(s, "%.0f%%", left * 100);
		    gdk_string_extents(
			font, s,
			&lbearing, &rbearing, &swidth, &ascent, &descent
		    );
		    x = (0.25 * width) - (swidth / 2);
		    y = (height / 2) - ((ascent + descent) / 2);
		    gdk_draw_string(
			window, font, style->text_gc[state],
			x + lbearing, y + ascent,
			s
		    );

		    /* Draw right channel value. */
		    sprintf(s, "%.0f%%", right * 100);
		    gdk_string_extents(
			font, s,
			&lbearing, &rbearing, &swidth, &ascent, &descent
		    );
		    x = (0.75 * width)  - (swidth / 2);
		    y = (height / 2) - ((ascent + descent) / 2);
		    gdk_draw_string(
			window, font, style->text_gc[state],
			x + lbearing, y + ascent,
			s
		    );

		}
		status = TRUE;
	    }
	    break;

	  case GDK_BUTTON_PRESS:
	    if(TRUE)
	    {
		GtkWidget *w = ms->amplitude_scale;
		if((w != NULL) ? !GTK_WIDGET_HAS_FOCUS(w) : FALSE)
		    gtk_widget_grab_focus(w);
	    }
	    status = TRUE;
	    break;

	  case GDK_BUTTON_RELEASE:
	    status = TRUE;
	    break;
	}

	return(status);
}


/*
 *	Connect callback.
 */
void YMixerConnectCB(
	GtkWidget *widget, const gchar *s, gpointer data
)
{
	gchar *strptr;
	gchar *s2 = (s != NULL) ? g_strdup(s) : NULL;
	ymixer_struct *ym = YMIXER(data);
	if(ym == NULL)
	    return;

	YMixerSetBusy(ym, TRUE);

	/* Empty string? Then use default location. */
	if(*s2 == '\0')
	{
	    g_free(s2);
	    s2 = g_strdup("127.0.0.1:9433");
	}

	/* No port number? Then append it. */
	strptr = strchr(s2, ':');
	if(strptr == NULL)
	{
	    strptr = g_strdup_printf("%s:9433", s2);
	    g_free(s2);
	    s2 = strptr;
	}

	/* Do connect. */
	YMixerDoConnect(ym, s2, NULL);

	/* Deallocate coppied location. */
	g_free(s2);

	YMixerSetBusy(ym, FALSE);

}

/*
 *	Y server list callback.
 */
void YMixerYServerListCB(GtkWidget *widget, gpointer data)
{
	ymixer_struct *ym = YMIXER(data);
	if(ym == NULL)
	    return;

	YCNewYServerList(YC_CORE(ym->core_ptr), YMIXER_TOPLEVEL(ym));
}

/*
 *	Close callback.
 */
void YMixerCloseCB(GtkWidget *widget, gpointer data)
{
	ymixer_struct *ym = YMIXER(data);
	YMixerSavePosition(ym);
	YMixerUnmap(ym);
}

/*
 *	New YConsole window callback.
 */
void YMixerNewYConsoleCB(GtkWidget *widget, gpointer data)
{
	ymixer_struct *ym = YMIXER(data);
	if(ym != NULL)
	    YCNewYConsole(
		YC_CORE(ym->core_ptr),
		YMixerGetLocation(ym)
	    );
}

/*
 *	New YMixer window callback.
 */
void YMixerNewYMixerCB(GtkWidget *widget, gpointer data)
{
	ymixer_struct *ym = YMIXER(data);
	if(ym != NULL)
	    YCNewYMixer(
		YC_CORE(ym->core_ptr),
		YMixerGetLocation(ym)
	    );
}

/*
 *	New YAudio window callback.
 */
void YMixerNewYAudioCB(GtkWidget *widget, gpointer data)
{
	ymixer_struct *ym = YMIXER(data);
	if(ym != NULL)
	    YCNewYAudio(
		YC_CORE(ym->core_ptr),
		YMixerGetLocation(ym)
	    );
}

/*
 *	New YPlayer window callback.
 */
void YMixerNewYPlayerCB(GtkWidget *widget, gpointer data)
{
	ymixer_struct *ym = YMIXER(data);
	if(ym != NULL)
	    YCNewYPlayer(
		YC_CORE(ym->core_ptr),
		YMixerGetLocation(ym)
	    );
}

/*
 *	Map options window callback.
 */
void YMixerOptionsCB(GtkWidget *widget, gpointer data)
{
	ymixer_struct *ym = YMIXER(data);
	if(ym != NULL)
	    YCNewOptWin(
		YC_CORE(ym->core_ptr),
		YMIXER_TOPLEVEL(ym)
	    );
}

/*
 *	Refresh callback.
 */
void YMixerRefreshCB(GtkWidget *widget, gpointer data)
{
	gchar *s;
	ymixer_struct *ym = YMIXER(data);
	if(ym == NULL)
	    return;

	YMixerDoDisconnect(ym, FALSE);
	s = ConBarGetLocation(YMIXER_CONBAR(ym));
	s = (s != NULL) ? g_strdup(s) : NULL;
	YMixerDoConnect(ym, s, NULL);
	g_free(s);

}

/*
 *	Shellout callback.
 *
 *	Returns actual shellout state.
 */
gboolean YMixerShelloutCB(
	GtkWidget *widget, gboolean shellout, gpointer data
)
{
	YConnection *con;
	ymixer_struct *ym = YMIXER(data);
	if(ym == NULL)
	    return(ym->shellout);

	con = YMIXER_CONNECTION(ym);
	if(con == NULL)
	    return(ym->shellout);

	if(shellout == ym->shellout)
	    return(ym->shellout);

	if(shellout)
	{
	    if(YSetAudioModeValues(
		con,
		8,      /* Sample size in bits. */
		1,      /* Number of channels. */
		0,      /* Sample rate set to 0 implies shellout. */
		0,      /* Direction, 0 = play, 1 = record. */
		True,
		2,
		1024
	    ))
		ym->shellout = FALSE;
	    else
		ym->shellout = TRUE;
	}
	else
	{
	    if(YChangeAudioModePreset(
		con,
		(ym->audio_mode_name != NULL) ?
		    ym->audio_mode_name : "Default"
	    ))
		ym->shellout = TRUE;
	    else
		ym->shellout = FALSE;
	}

	YMIXER_SET_CONBAR_AUDIO(ym);

	return(ym->shellout);
}

/*
 *	Shutdown Y server callback.
 */
void YMixerShutdownCB(GtkWidget *widget, gpointer data)
{
	ymixer_struct *ym = YMIXER(data);
	if(ym == NULL)
	    return;

	YMixerDoDisconnect(ym, TRUE);
}

/*
 *	YMixer manage callback.
 */
void YMixerManageCB(ymixer_struct *ym)
{
	YConnection *con;
	YEvent yevent;
	dspview_struct *dv;
	connection_bar_struct *cb;


	if(ym == NULL)
	    return;

	cb = YMIXER_CONBAR(ym);
	dv = ConBarGetDSPView(cb);
	if(dv == NULL)
	    return;

	/* Manage Y events. */
	con = YMIXER_CONNECTION(ym);
	while((con != NULL) && (YGetNextEvent(con, &yevent, False) > 0))
	{
	    switch(yevent.type)
	    {
	      case YAudioChange:
		/* Update Audio values on conbar and DSPView. */
		if(yevent.audio.preset)
		{
		    YEventAudioStats stat_buf;

		    g_free(ym->audio_mode_name);
		    ym->audio_mode_name = g_strdup(yevent.audio.mode_name);

		    if(!YGetAudioStats(con, &stat_buf))
		    {
			ym->sample_rate = stat_buf.sample_rate;
			ym->sample_size = stat_buf.sample_size;
			ym->channels = stat_buf.channels;
			ym->cycle_us = stat_buf.cycle_us;
			ym->num_fragments = stat_buf.num_fragments;
			ym->fragment_size = stat_buf.fragment_size;
			ym->flip_stereo = stat_buf.flip_stereo;
			ym->shellout = FALSE;
		    }
		}
		else
		{
		    YEventAudioChange *audio = &yevent.audio;
		    ym->sample_rate = audio->sample_rate;
		    ym->sample_size = audio->sample_size;
		    ym->channels = audio->channels;
		    ym->num_fragments = audio->num_fragments;
		    ym->fragment_size = audio->fragment_size_bytes;
		    ym->flip_stereo = FALSE;
		    ym->shellout = FALSE;
		}
		YMIXER_SET_CONBAR_AUDIO(ym);
		ConBarSetShellOut(cb, ym->shellout);
		DSPViewSetChannels(dv, yevent.audio.channels);
		break;

	      case YCycleChange:
		/* Update audio values. */
		ym->cycle_us = yevent.cycle.cycle_us;
		YMIXER_SET_CONBAR_AUDIO(ym);
		break;

	      case YDisconnect: case YShutdown:
		YMixerDoDisconnect(ym, FALSE);
		con = NULL;	/* So we break out of this loop. */
		break;

#if 0
	      case YMixerChannel:
		YMixerScaleSetValue2C(
		    ym, yevent.mixer.code,
		    (YMixerValues > 0) ? yevent.mixer.value[0] : 0.0f,
		    (YMixerValues > 1) ? yevent.mixer.value[1] : 0.0f
		);
		break;
#endif
	      case YSHMSound:
		if(yevent.yshmsound.minor_op_code == YSHMSoundClose)
		{
		    /* Need to reopen YSHM sound buffer when Audio
		     * mode changes.
		     */
		    YSHMCloseSound(con, ym->yshm_buf);
		    ym->yshm_buf = (guint8 *)YSHMOpenSound(
			con, &ym->yshm_buf_len, &ym->yshm_id
		    );
		}
		break;
	    }
	}

	/* Redraw DSP view. */
	ConBarDraw(cb, ym->yshm_buf, ym->yshm_buf_len);
}

/*
 *      YMixer manage 2 callback.
 *
 *	This function is called in longer successions.
 */
void YMixerManage2CB(ymixer_struct *ym)
{
	YConnection *con;


	if(ym == NULL)
	    return;

	con = YMIXER_CONNECTION(ym);
	if(con != NULL)
	{
	    YEventServerStats stat_buf;

	    if(!YGetServerStats(con, &stat_buf))
	    {
		ym->start_time = stat_buf.start_time;
		ym->cycle_load = stat_buf.cycle_load;
		ym->version_major = stat_buf.protocol_version_major;
		ym->version_minor = stat_buf.protocol_version_minor;
		g_free(ym->server_name);
		ym->server_name = g_strdup(stat_buf.server_name);
		g_free(ym->sound_name);
		ym->sound_name = g_strdup(stat_buf.sound_name);
		YMIXER_SET_CONBAR_AUDIO(ym);
	    }
	}
}


/*
 *	YMixer reconfigured signal callback.
 */
void YMixerReconfiguredCB(
	ymixer_struct *ym, gpointer core_ptr
)
{
	GtkWidget *w;
	dspview_struct *dv;
	yc_core_struct *cp = YC_CORE(core_ptr);
	if((ym == NULL) || (cp == NULL))
	    return;

	/* Update connection bar's audio display. */
	if(TRUE)
	{
	    GdkColor *c, cs[3];
	    const cfg_color_struct *color;

#define DO_CONVERT_COLOR        \
{ if((c != NULL) && (color != NULL)) { \
 c->red         = (gushort)-1 * color->r; \
 c->green       = (gushort)-1 * color->g; \
 c->blue        = (gushort)-1 * color->b; \
} }
	    color = YC_GET_CFG_COLOR(
		cp, YC_CFG_PARM_AUDIO_DISPLAY_COLOR_BG
	    );
	    c = &cs[0];
	    DO_CONVERT_COLOR

	    color = YC_GET_CFG_COLOR(
		cp, YC_CFG_PARM_AUDIO_DISPLAY_COLOR_FG
	    );
	    c = &cs[1];
	    DO_CONVERT_COLOR

	    color = YC_GET_CFG_COLOR(
		cp, YC_CFG_PARM_AUDIO_DISPLAY_COLOR_FG_DIM
	    );
	    c = &cs[2];
	    DO_CONVERT_COLOR

	    ConBarSetAudioColors(
		YMIXER_CONBAR(ym), &cs[0], &cs[1], &cs[2]
	    );
#undef DO_CONVERT_COLOR

	    ConBarSetAudioFonts(
		YMIXER_CONBAR(ym),
		YC_GET_CFG_S(cp, YC_CFG_PARM_AUDIO_DISPLAY_FONT_LARGE),
		YC_GET_CFG_S(cp, YC_CFG_PARM_AUDIO_DISPLAY_FONT_MEDIUM),
		YC_GET_CFG_S(cp, YC_CFG_PARM_AUDIO_DISPLAY_FONT_SMALL)
	    );
	}

	/* Update connection bar's DSPView. */
	dv = ConBarGetDSPView(YMIXER_CONBAR(ym));
	if(dv != NULL)
	{
	    const cfg_color_struct *color;
	    gfloat alpha_coeff;

	    DSPViewSetDisplayStyle(
		dv,
		YC_GET_CFG_I(cp, YC_CFG_PARM_DSPVIEW_DISPLAY_STYLE)
	    );
	    DSPViewSetMergeMode(
		dv,
		YC_GET_CFG_I(cp, YC_CFG_PARM_DSPVIEW_MERGE_MODE)
	    );
	    color = YC_GET_CFG_COLOR(
		cp, YC_CFG_PARM_DSPVIEW_COLOR
	    );
	    alpha_coeff = YC_GET_CFG_F(
		cp, YC_CFG_PARM_DSPVIEW_TRANSPARENCY
	    );
	    if(color != NULL)
		DSPViewSetDSPColor(
		    dv,
		    (guint32)(
			((guint32)(color->r * 0xff) << 0) |
			((guint32)(color->g * 0xff) << 8) |
			((guint32)(color->b * 0xff) << 16) |
			((guint32)(alpha_coeff * 0xff) << 24)
		    )
		);
	    DSPViewSetFadeCoeff(
		dv,
		YC_GET_CFG_F(cp, YC_CFG_PARM_DSPVIEW_FADE_COEFF)
	    );
	}       /* Update connection bar's DSPView. */


	/* Recreate YMixer scales. */
	YMixerScaleCreateAll(ym);

	/* Update mixer values. */
	if(ym->total_mixer_scales > 0)
	{
	    gint i;
	    YConnection *con = YMIXER_CONNECTION(ym);
	    gboolean is_connected = (con != NULL) ? TRUE : FALSE;
	    Coefficient vc[4];
	    for(i = 0; i < ym->total_mixer_scales; i++)
	    {
		YMixerScaleSetSensitive(
		    ym, i + YMixerCodeBaseOffset, is_connected
		);
		if(is_connected)
		{
		    YGetMixerChannel(
			con, i + YMixerCodeBaseOffset,
			&vc[0], &vc[1]
		    );
		    YMixerScaleSetValue2C(
			ym, i + YMixerCodeBaseOffset,
			(gfloat)vc[0], (gfloat)vc[1]
		    );
		}
	    }
	}

	/* Update mixer scales scrolled window. */
	w = YMIXER_MIXERS_SCROLLED_WINDOW(ym);
	if(w != NULL)
	{
	    /* Create rcstyle for mixer scales only if we are not to use
	     * default style for the mixers scales.
	     */
	    GtkRcStyle *rcstyle = (YC_GET_CFG_I(
		cp, YC_CFG_PARM_MIXER_COLOR_USE_DEFAULT)) ?
		NULL : YMixerMixerScaleRCStyleNew(ym);
	    GtkScrolledWindow *sw = GTK_SCROLLED_WINDOW(w);

	    if(rcstyle != NULL)
	    {
		if(ym->mixers_ebox != NULL)
		    gtk_widget_modify_style(ym->mixers_ebox, rcstyle);
		gtk_widget_modify_style(w, rcstyle);
		gtk_widget_modify_style(sw->hscrollbar, rcstyle);
		gtk_widget_modify_style(sw->vscrollbar, rcstyle);

		gtk_rc_style_unref(rcstyle);
	    }
	    else
	    {
		if(ym->mixers_ebox != NULL)
		    gtk_widget_restore_default_style(ym->mixers_ebox);
		gtk_widget_restore_default_style(w);
		gtk_widget_restore_default_style(sw->hscrollbar);
		gtk_widget_restore_default_style(sw->vscrollbar);
	    }
	}

}

