#include <stdlib.h>
#include <string.h>

#include <Y2/Y.h>
#include <Y2/Ylib.h>

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include "../include/string.h"
#include "../include/fio.h"
#include "../include/disk.h"

#include "guiutils.h"
#include "menubutton.h"
#include "cdialog.h"

#include "cfg.h"
#include "conbar.h"
#include "yplayer.h"
#include "yplayercb.h"
#include "yc.h"
#include "ycop.h"
#include "yccfglist.h"
#include "config.h"

#include "images/icon_yplayer_48x48.xpm"
#include "images/icon_sounds_20x20.xpm"
#include "images/icon_sound_20x20.xpm"
#include "images/icon_mixer_cd_20x20.xpm"
#include "images/icon_file_music_20x20.xpm"
#include "images/icon_file_sound_20x20.xpm"

#include "images/icon_yplayer_file_14x14.xpm"
#include "images/icon_yplayer_play_14x14.xpm"
#include "images/icon_yplayer_stop_14x14.xpm"
#include "images/icon_yplayer_eject_14x14.xpm"
#include "images/icon_yplayer_remove_14x14.xpm"
#include "images/icon_yplayer_refresh_14x14.xpm"
#include "images/icon_yplayer_pause_14x14.xpm"
#include "images/icon_yplayer_repeat_14x14.xpm"
#include "images/icon_yplayer_rewind_14x14.xpm"
#include "images/icon_yplayer_mode_14x14.xpm"
#include "images/icon_yplayer_up_14x14.xpm"
#include "images/icon_yplayer_down_14x14.xpm"
#include "images/icon_yplayer_open_file_menu_14x14.xpm"
#include "images/icon_yplayer_save_file_menu_14x14.xpm"
#include "images/icon_yplayer_sound_source_14x14.xpm"
#include "images/icon_yplayer_ear_14x14.xpm"

#include "images/icon_yplayer_file_20x20.xpm"
#include "images/icon_yplayer_play_20x20.xpm"
#include "images/icon_yplayer_random_20x20.xpm"
#include "images/icon_yplayer_stop_20x20.xpm"
#include "images/icon_yplayer_eject_20x20.xpm"
#include "images/icon_yplayer_remove_20x20.xpm"
#include "images/icon_yplayer_refresh_20x20.xpm"
#include "images/icon_yplayer_pause_20x20.xpm"
#include "images/icon_yplayer_rewind_20x20.xpm"
#include "images/icon_yplayer_mode_20x20.xpm"
#include "images/icon_yplayer_up_20x20.xpm"
#include "images/icon_yplayer_down_20x20.xpm"
#include "images/icon_yplayer_open_file_20x20.xpm"
#include "images/icon_yplayer_save_file_20x20.xpm"


/* Utilities */
void YPlayerVolumeXYToQuad(
	gfloat x, gfloat y, gfloat a,
	gfloat *l, gfloat *r, gfloat *bl, gfloat *br
);
gchar *YPlayerDataLengthToTimeString(
	gulong length,
	gint sample_rate, gint sample_size, gint channels
);
GtkRcStyle *YPlayerAudioDisplayRCStyleNew(yplayer_struct *yp);
void YPlayerSetTitle(yplayer_struct *yp, const gchar *s);
gchar *YPlayerGetLocation(yplayer_struct *yp);

/* Procedures */
gint YPlayerDoConnect(
	yplayer_struct *yp,
	const gchar *address, const gchar *start_command
);
void YPlayerDoDisconnect(yplayer_struct *yp, gboolean force_shutdown);
void YPlayerDoPlaySoundObject(
	yplayer_struct *yp, const gchar *path, gboolean add_as_needed
);
void YPlayerDoOpenSoundObjectAdv(yplayer_struct *yp, const gchar *path);
void YPlayerDoCloseSoundObjectAdv(yplayer_struct *yp);
void YPlayerDoPlaySoundObjectAdv(yplayer_struct *yp, const gchar *path);
void YPlayerDoPlayAudioCDTrack(yplayer_struct *yp, gint track_number);

/* List IO */
gint YPlayerSndObjListAdd(
	yplayer_struct *yp, gint row,
	const gchar *filename, gint format,
	gint sample_rate, gint sample_size, gint channels,
	gulong length
);
gint YPlayerSndObjPlayingListAdd(
	yplayer_struct *yp, gint row,
	const gchar *filename, gint format, YID yid
);
void YPlayerSndObjPlayingListClear(yplayer_struct *yp);
void YPlayerAudioCDTracksListUpdate(yplayer_struct *yp);

/* Play list file IO */
void YPlayerLoadPlayListFile(
	yplayer_struct *yp, const gchar *path, gboolean insert
);
void YPlayerSavePlayListFile(
	yplayer_struct *yp, const gchar *path, gboolean selected_only
);

/* Drawing */
void YPlayerSndObjAdvDisplayDraw(yplayer_struct *yp);
void YPlayerSndObjAdvVolumeDraw(yplayer_struct *yp);

/* Basic */
void YPlayerSavePosition(yplayer_struct *yp);
yplayer_struct *YPlayerNew(gpointer core_ptr);
void YPlayerUpdateMenus(yplayer_struct *yp);
void YPlayerSetBusy(yplayer_struct *yp, gboolean is_busy);
void YPlayerMap(yplayer_struct *yp);
void YPlayerUnmap(yplayer_struct *yp);
void YPlayerDelete(yplayer_struct *yp);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)


#define YPLAYER_BTN_WIDTH	60
#define YPLAYER_BTN_HEIGHT	20


/*
 *	Calculates the volume (from 0.0 to 1.0) in quad based on the
 *	given position of the source (left hand rule).
 *
 *	x and y can be from -1.0 to 1.0 and a (amplitude) can be from
 *	0.0 to 1.0.
 */
void YPlayerVolumeXYToQuad(
	gfloat x, gfloat y, gfloat a,
	gfloat *l, gfloat *r, gfloat *bl, gfloat *br
)
{
	gfloat	lc = MIN(1.0f - x, 1.0f),
		rc = MIN(x + 1.0f, 1.0f),
		fc = MIN(y + 1.0f, 1.0f),
		bc = MIN(1.0f - y, 1.0f);

	if(l != NULL)
	    *l = a * lc * fc;
	if(r != NULL)
	    *r = a * rc * fc;
	if(bl != NULL)
	    *bl = a * lc * bc;
	if(br != NULL)
	    *br = a * rc * bc;
}

/*
 *	Returns a time string in the format hh:mm:ss based on the
 *	given inputs.
 *
 *	The returned string is statically allocated and must not be
 *	freed by the calling function, this function never returns NULL.
 */
gchar *YPlayerDataLengthToTimeString(
	gulong length,
	gint sample_rate, gint sample_size, gint channels
)
{
	static gchar s[40];
	gulong seconds, samples;

	if((length == 0) || (sample_rate <= 0) || (sample_size < 8) ||
	   (channels <= 0)
	)
	    return("00:00:00");

	/* Calculate length in samples */
	samples = length / (sample_size / 8) / channels;
	/* Calculate number of seconds */
	seconds = samples / sample_rate;

	sprintf(
	    s,
	    "%.2i:%.2i:%.2i",
	    (gint)(seconds / 3600),
	    (gint)((seconds / 60) % 60),
	    (gint)(seconds % 60)
	);
	return(s);
}

/*
 *      Returns a rcstyle specifying the style for the audio displays.
 */
GtkRcStyle *YPlayerAudioDisplayRCStyleNew(yplayer_struct *yp)
{
	gint state;
	GdkColor *ct;
	GtkRcStyle *rcstyle;
	const cfg_color_struct *cs;
	yc_core_struct *core_ptr;


	if(yp == NULL)
	    return(NULL);

	core_ptr = YC_CORE(yp->core_ptr);
	if(core_ptr == NULL)
	    return(NULL);

	rcstyle = gtk_rc_style_new();

#define GET_CFG_COLOR(p)        (YC_GET_CFG_COLOR(core_ptr, (p)))
#define DO_CONVERT_COLOR        \
{ if((cs != NULL) && (ct != NULL)) { \
 ct->red        = (gushort)-1 * cs->r; \
 ct->green      = (gushort)-1 * cs->g; \
 ct->blue       = (gushort)-1 * cs->b; \
} }
	state = GTK_STATE_NORMAL;
	rcstyle->color_flags[state] = GTK_RC_FG |
	    GTK_RC_TEXT | GTK_RC_BASE;
	cs = GET_CFG_COLOR(YC_CFG_PARM_AUDIO_DISPLAY_COLOR_FG);
	ct = &rcstyle->fg[state];
	DO_CONVERT_COLOR
	ct = &rcstyle->text[state];
	DO_CONVERT_COLOR
	cs = GET_CFG_COLOR(YC_CFG_PARM_AUDIO_DISPLAY_COLOR_BG);
	ct = &rcstyle->base[state];
	DO_CONVERT_COLOR

	state = GTK_STATE_PRELIGHT;
	rcstyle->color_flags[state] = GTK_RC_FG |
	    GTK_RC_TEXT | GTK_RC_BASE;
	cs = GET_CFG_COLOR(YC_CFG_PARM_AUDIO_DISPLAY_COLOR_FG);
	ct = &rcstyle->fg[state];
	DO_CONVERT_COLOR
	ct = &rcstyle->text[state];
	DO_CONVERT_COLOR
	cs = GET_CFG_COLOR(YC_CFG_PARM_AUDIO_DISPLAY_COLOR_BG);
	ct = &rcstyle->base[state];
	DO_CONVERT_COLOR

	state = GTK_STATE_ACTIVE;
	rcstyle->color_flags[state] = GTK_RC_FG |
	    GTK_RC_TEXT | GTK_RC_BASE;
	cs = GET_CFG_COLOR(YC_CFG_PARM_AUDIO_DISPLAY_COLOR_FG);
	ct = &rcstyle->fg[state];
	DO_CONVERT_COLOR
	ct = &rcstyle->text[state];
	DO_CONVERT_COLOR
	cs = GET_CFG_COLOR(YC_CFG_PARM_AUDIO_DISPLAY_COLOR_FG_DIM);
	ct = &rcstyle->base[state];
	DO_CONVERT_COLOR

	state = GTK_STATE_INSENSITIVE;
	rcstyle->color_flags[state] = GTK_RC_FG |
	    GTK_RC_TEXT | GTK_RC_BASE;
	cs = GET_CFG_COLOR(YC_CFG_PARM_AUDIO_DISPLAY_COLOR_FG_DIM);
	ct = &rcstyle->fg[state];
	DO_CONVERT_COLOR
	ct = &rcstyle->text[state];
	DO_CONVERT_COLOR
	cs = GET_CFG_COLOR(YC_CFG_PARM_AUDIO_DISPLAY_COLOR_BG);
	ct = &rcstyle->base[state];
	DO_CONVERT_COLOR

	state = GTK_STATE_SELECTED;
	rcstyle->color_flags[state] = GTK_RC_FG | GTK_RC_BG |
	    GTK_RC_TEXT | GTK_RC_BASE;
	cs = GET_CFG_COLOR(YC_CFG_PARM_AUDIO_DISPLAY_COLOR_BG);
	ct = &rcstyle->fg[state];
	DO_CONVERT_COLOR
	ct = &rcstyle->text[state];
	DO_CONVERT_COLOR
	cs = GET_CFG_COLOR(YC_CFG_PARM_AUDIO_DISPLAY_COLOR_FG);
	ct = &rcstyle->bg[state];
	DO_CONVERT_COLOR
	ct = &rcstyle->base[state];
	DO_CONVERT_COLOR

	return(rcstyle);
}

/* 
 *	Sets the title string to s, if s is NULL then values from the
 *	yconsole will be used to form the title.
 */
void YPlayerSetTitle(yplayer_struct *yp, const gchar *s)
{
	gchar *title;
	GtkWidget *w;


	if(yp == NULL)
	    return;

	w = YPLAYER_TOPLEVEL(yp);
	if(w == NULL)
	    return;

	/* Title string given? */
	if(s != NULL)
	{
	    gtk_window_set_title(GTK_WINDOW(w), s);
	    return;
	}

	/* Format title string */
	if(YPLAYER_IS_CONNECTED(yp))
	{
	    s = ConBarGetLocation(YPLAYER_CONBAR(yp));
	    title = g_strdup_printf("YPlayer: %s", s);
	}
	else
	{
	    title = g_strdup("YPlayer");
	}
	gtk_window_set_title(GTK_WINDOW(w), title);

	g_free(title);
}


/*
 *	Returns the current location.
 */
gchar *YPlayerGetLocation(yplayer_struct *yp)
{
	return((yp != NULL) ? ConBarGetLocation(YPLAYER_CONBAR(yp)) : NULL);
}

/*
 *	Connects the yconsole to the Y server at the addres s.  If the
 *	yconsole is already connected then it will be disconneted first.
 *
 *	Returns non-zero on error and 0 on success.
 */
gint YPlayerDoConnect(
	yplayer_struct *yp,
	const gchar *address, const gchar *start_command
)
{
	YConnection *con;


	if(yp == NULL)
	    return(-1);

	/* Disconnect from Y server as needed */
	YPlayerDoDisconnect(yp, FALSE);

	/* If no address is given then only perform disconnect */
	if(address == NULL)
	    return(0);

	/* Connect to Y server */
	yp->con = con = YOpenConnection(NULL, address);
	/* If not able to connect on first try and a server start
	 * command is given, then execute that command and try
	 * connecting again.
	 */
	if((con == NULL) && (start_command != NULL))
	{
	    /* Start server (with dialog) and try connecting again */
	    YCNewStartupDialog(
		YC_CORE(yp->core_ptr), YPLAYER_TOPLEVEL(yp),
		start_command, TRUE
	    );
	    yp->con = con = YOpenConnection(NULL, address);
	}
	if(con != NULL)
	{
	    /* Connected to Y server, now get some initial values */
	    YEventAudioStats stat_buf;
	    dspview_struct *dv = ConBarGetDSPView(YPLAYER_CONBAR(yp));

	    /* Update connection bar connected state */
	    ConBarSetConnectState(YPLAYER_CONBAR(yp), TRUE);

	    /* YSHM sound buffer pointer */
	    yp->yshm_buf = (guint8 *)YSHMOpenSound(
		con, &yp->yshm_buf_len, &yp->yshm_id
	    );

	    /* Audio stats */
	    if(!YGetAudioStats(con, &stat_buf))
	    {
		g_free(yp->audio_mode_name);
		yp->audio_mode_name = g_strdup(stat_buf.mode_name);
		yp->sample_rate = stat_buf.sample_rate;
		yp->sample_size = stat_buf.sample_size;
		yp->channels = stat_buf.channels;
		yp->cycle_us = stat_buf.cycle_us;
		yp->cycle_load = 0.0;
		yp->num_fragments = stat_buf.num_fragments;
		yp->fragment_size = stat_buf.fragment_size;
		yp->shellout = FALSE;

		/* Update audio values on connection bar */
		YPLAYER_SET_CONBAR_AUDIO(yp);

		ConBarSetShellOut(YPLAYER_CONBAR(yp), yp->shellout);
		DSPViewSetChannels(dv, yp->channels);
	    }

	    /* Record location history on connection bar */
	    ConBarSetLocation(YPLAYER_CONBAR(yp), address, FALSE, TRUE);


	    /* Reget audio CD tracks list */
	    YPlayerAudioCDTracksListUpdate(yp);

	    /* Update title and menus */
	    YPlayerSetTitle(yp, NULL);
	    YPlayerUpdateMenus(yp);
	}

	return(0);
}

/*
 *	Disconnects the yconsole from the Y server.
 */
void YPlayerDoDisconnect(yplayer_struct *yp, gboolean force_shutdown)
{
	dspview_struct *dv;


	if(yp == NULL)
	    return;

	/* Stop sound object from playing */
	YPlayerDoCloseSoundObjectAdv(yp);

	/* Update connection bar connected state */
	ConBarSetConnectState(YPLAYER_CONBAR(yp), FALSE);


	/* Close YSHM sound buffer */
	YSHMCloseSound(YPLAYER_CONNECTION(yp), yp->yshm_buf);
	yp->yshm_buf = NULL;
	yp->yshm_buf_len = 0;
	yp->yshm_id = -1;

	/* Close connection or shutdown? */
	if(force_shutdown)
	    YShutdownServer(YPLAYER_CONNECTION(yp));
	else
	    YCloseConnection(YPLAYER_CONNECTION(yp), False);
	yp->con = NULL;


	/* Reset Y audio values */
	memset(yp->channel_volume_avg, 0x00, 4 * sizeof(gfloat));
	yp->start_time = 0;
	g_free(yp->audio_mode_name);
	yp->audio_mode_name = NULL;
	yp->sample_rate = 0;
	yp->sample_size = 0;
	yp->channels = 0;
	yp->cycle_us = 0;
	yp->cycle_load = 0.0;
	yp->num_fragments = 0;
	yp->fragment_size = 0;
	yp->flip_stereo = FALSE;
	yp->shellout = FALSE;
	yp->version_major = yp->version_minor = 0;
	g_free(yp->server_name);
	yp->server_name = NULL;
	g_free(yp->sound_name);
	yp->sound_name = NULL;

	/* Update audio values on connection bar */
	YPLAYER_SET_CONBAR_AUDIO(yp);

	dv = ConBarGetDSPView(YPLAYER_CONBAR(yp));
	DSPViewSetChannels(dv, yp->channels);


	/* Clear sound object playing list (each referenced sound object
	 * has already been killed so the row data destroy callback will
	 * note that and not kill it again).
	 */
	YPlayerSndObjPlayingListClear(yp);

	/* Clear audio CD tracks list */
	YPlayerAudioCDTracksListUpdate(yp);

	/* Update title and menus */
	YPlayerSetTitle(yp, NULL);
	YPlayerUpdateMenus(yp);
}

/*
 *	Switches to the sound objects page and selects the sound
 *	object specified by path.  If the sound object is not in the
 *	playlist and add_as_needed is TRUE then the sound object will
 *	be appended to the playlist and then played.
 *
 *      Must be already connected.
 */
void YPlayerDoPlaySoundObject(
	yplayer_struct *yp, const gchar *path, gboolean add_as_needed
)
{
	const gchar *cstrptr;
	gint insert_row = -1;
	gint row;
	GtkCList *clist = (GtkCList *)YPLAYER_SNDOBJ_CLIST(yp);
	GtkNotebook *notebook = (GtkNotebook *)YPLAYER_NOTEBOOK(yp);
	YConnection *con = YPLAYER_CONNECTION(yp);
	if((con == NULL) || (path == NULL))
	    return;

	/* Switch to the sound object playing page */
	if(notebook != NULL)
	    gtk_notebook_set_page(notebook, 0);

	/* Check a sound object is already loaded */
	for(row = 0; row < clist->rows; row++)
	{
	    cstrptr = (const gchar *)gtk_clist_get_row_data(
		clist, row
	    );
	    if((cstrptr != NULL) ? !strcmp(cstrptr, path) : FALSE)
		break;
	}
	/* Not found in list? */
	if(row >= clist->rows)
	{
	    YEventSoundObjectAttributes soa;

	    /* Do not add if not in list? */
	    if(!add_as_needed)
		return;

	    /* Get sound object attributes and add to list */
	    if(YGetSoundObjectAttributes(con, path, &soa))
		return;

	    /* Add this sound object to the list */
	    row = YPlayerSndObjListAdd(
		yp, insert_row,
		path, soa.format,
		soa.sample_rate, soa.sample_size, soa.channels,
		soa.length
	    );
	    if(row < 0)
		return;
	}

	/* At this point the row indicates the sound object in the list
	 * that we want to play.
	 */

	/* Select row */
	gtk_clist_freeze(clist);
	gtk_clist_unselect_all(clist);
	gtk_clist_select_row(clist, row, 0);
	gtk_clist_thaw(clist);

	/* Play */
	YPlayerSndObjPlayCB(yp->sndobj_play_btn, yp);
}

/*
 *	Opens a new sound object for the sound object advanced page,
 *	the existing sound object will be closed.
 */
void YPlayerDoOpenSoundObjectAdv(yplayer_struct *yp, const gchar *path)
{
	YEventSoundObjectAttributes *soa;
	GtkWidget *w;
	YConnection *con = YPLAYER_CONNECTION(yp);


	if(yp == NULL)
	    return;

	/* Close previous sound object (if any) and stop it from playing */
	YPlayerDoCloseSoundObjectAdv(yp);

	/* Must be connected */
	if(con == NULL)
	    return;

	/* if no sound object path is given then do not open */
	if((path != NULL) ? (*path == '\0') : TRUE)
	    return;

	/* Get sound object attributes */
	soa = &yp->sndobjadv_soa;
	if(YGetSoundObjectAttributes(con, path, soa))
	    return;


	YPLAYER_SNDOBJADV_FREEZE(yp);

	/* Record sound object path, indicating that a sound object is
	 * loaded.
	 */
	yp->sndobjadv_path = g_strdup(path);

	/* Position */
	w = yp->sndobjadv_position_scale;
	if(w != NULL)
	{
	    if(soa->format == SndObjTypeDSP)
	    {
		GtkAdjustment *adj = GTK_RANGE(w)->adjustment;
		if(adj != NULL)
		{
		    adj->value = 0.0f;
		    adj->lower = 0.0f;
		    adj->upper = (gfloat)soa->length;
		    adj->step_increment = (gfloat)soa->length / 500.0f;
		    adj->page_increment = (gfloat)soa->length / 100.0f;
		    adj->page_size = 0.0f;
		    gtk_signal_emit_by_name(GTK_OBJECT(adj), "changed");
		    gtk_signal_emit_by_name(GTK_OBJECT(adj), "value_changed");
		}
	    }
	    else if(soa->format == SndObjTypeMIDI)
	    {
		GtkAdjustment *adj = GTK_RANGE(w)->adjustment;
		if(adj != NULL)
		{
		    adj->value = 0.0f;
		    adj->lower = 0.0f;
		    adj->upper = 0.0f;
		    adj->step_increment = 0.1f;
		    adj->page_increment = 0.1f;
		    adj->page_size = 0.0f;
		    gtk_signal_emit_by_name(GTK_OBJECT(adj), "changed");
		    gtk_signal_emit_by_name(GTK_OBJECT(adj), "value_changed");
		}
	    }
	}

	/* Applied sample rate */
	w = yp->sndobjadv_sample_rate_spin;
	if(w != NULL)
	{
	    if(soa->format == SndObjTypeDSP)
	    {
		GtkAdjustment *adj = GTK_SPIN_BUTTON(w)->adjustment;
		if(adj != NULL)
		{
		    adj->value = (gfloat)soa->sample_rate;
		    adj->lower = 8000.0f;
		    adj->upper = (gfloat)soa->sample_rate * 100.0;
		    adj->step_increment = 11025.0f / 10.0f;
		    adj->page_increment = 11025.0f / 4.0f;
		    adj->page_size = 0.0f;
		    gtk_signal_emit_by_name(GTK_OBJECT(adj), "changed");
		    gtk_signal_emit_by_name(GTK_OBJECT(adj), "value_changed");
		}
	    }
	    else if(soa->format == SndObjTypeMIDI)
	    {
		GtkAdjustment *adj = GTK_SPIN_BUTTON(w)->adjustment;
		if(adj != NULL)
		{
		    adj->value = 0.0f;
		    adj->lower = 0.0f;
		    adj->upper = 0.0f;
		    adj->step_increment = 0.1f;
		    adj->page_increment = 0.1f;
		    adj->page_size = 0.0f;
		    gtk_signal_emit_by_name(GTK_OBJECT(adj), "changed");
		    gtk_signal_emit_by_name(GTK_OBJECT(adj), "value_changed");
		}
	    }
	}

	YPLAYER_SNDOBJADV_THAW(yp);

	/* Update sound object advanced display */
	YPlayerSndObjAdvDisplayDraw(yp);
	YPlayerUpdateMenus(yp);
}

/*
 *	Closes the sound object on the sound object advanced page.
 */
void YPlayerDoCloseSoundObjectAdv(yplayer_struct *yp)
{
	YConnection *con = YPLAYER_CONNECTION(yp);
	if(yp == NULL)
	    return;

	/* Stop sound object (if any) from playing (do this only if
	 * connected).
	 */
	if((con != NULL) && (yp->sndobjadv_yid != YIDNULL))
	    YDestroyPlaySoundObject(con, yp->sndobjadv_yid);
	yp->sndobjadv_yid = YIDNULL;

	/* Delete sound object path, indicating nothing is loaded */
	g_free(yp->sndobjadv_path);
	yp->sndobjadv_path = NULL;

	/* Clear sound object attributes */
	memset(&yp->sndobjadv_soa, 0x00, sizeof(YEventSoundObjectAttributes));

	/* Update sound object advanced display */
	YPlayerSndObjAdvDisplayDraw(yp);
	YPlayerUpdateMenus(yp);
}

/*
 *	Switches to the sound object advanced page and begins playing
 *	the sound object specified by path.
 *
 *	Must be already connected.
 */
void YPlayerDoPlaySoundObjectAdv(yplayer_struct *yp, const gchar *path)
{
	GtkNotebook *notebook = (GtkNotebook *)YPLAYER_NOTEBOOK(yp);
	YConnection *con = YPLAYER_CONNECTION(yp);
	if((con == NULL) || (path == NULL))
	    return;

	/* Switch to the sound object advanced page */
	if(notebook != NULL)
	    gtk_notebook_set_page(notebook, 1);

/* TODO */

}

/*
 *	Switches to the audio CD tracks page and begins playing the
 *	track specified by track_number.
 *
 *	Must be already connected.
 */
void YPlayerDoPlayAudioCDTrack(yplayer_struct *yp, gint track_number)
{
	gint row;
	GtkCList *clist = (GtkCList *)YPLAYER_AUDIOCD_TRACKS_CLIST(yp);
	GtkNotebook *notebook = (GtkNotebook *)YPLAYER_NOTEBOOK(yp);
	YConnection *con = YPLAYER_CONNECTION(yp);
	if(con == NULL)
	    return;

	/* Switch to the audio CD page */
	if(notebook != NULL)
	    gtk_notebook_set_page(notebook, 2);

	/* No tracks? Implies no audio CD */
	if(clist->rows <= 0)
	    return;

	/* Check if the given track_number is in the list */
	if((track_number >= 1) && (track_number <= clist->rows))
	    row = track_number - 1;
	else
	    row = -1;

	/* Not found in list? */
	if(row < 0)
	    return;

	/* At this point the row indicates the track number in the list
	 * that we want to play.
	 */

	/* Select row */
	gtk_clist_freeze(clist);
	gtk_clist_unselect_all(clist);
	gtk_clist_select_row(clist, row, 0);
	gtk_clist_thaw(clist);

	/* Play audio CD track */
	YPlayerAudioCDPlayCB(yp->audiocd_play_mbtn, yp);
}


/*
 *	Adds sound object to list.
 */
gint YPlayerSndObjListAdd(
	yplayer_struct *yp, gint row,
	const gchar *filename, gint format,
	gint sample_rate, gint sample_size, gint channels,
	gulong length
)
{
	const gchar *cstrptr;
	gchar *v[5], *s;
	GtkCList *clist = (GtkCList *)YPLAYER_SNDOBJ_CLIST(yp);
	if(clist == NULL)
	    return(-1);

	/* Format cell values */
	cstrptr = (filename != NULL) ? strrchr(filename, DIR_DELIMINATOR) : NULL;
	if(cstrptr != NULL)
	    cstrptr++;
	v[0] = g_strdup((cstrptr != NULL) ? cstrptr : "(null)");
	s = strrchr(v[0], '.');
	if(s != NULL)
	    *s = '\0';
	v[1] = (sample_rate > 0) ?
	    g_strdup_printf("%i", sample_rate) :
	    g_strdup("-");
	v[2] = (sample_size > 0) ?
	    g_strdup_printf("%i", sample_size) :
	    g_strdup("-");
	v[3] = (channels > 0) ?
	    g_strdup(
	        (channels == 4) ? "Quad" : ((channels == 2) ? "Stereo" : "Mono")
	    ) :
	    g_strdup("-");
	v[4] = (length > 0) ?
	    g_strdup(YPlayerDataLengthToTimeString(
		length, sample_rate, sample_size, channels
	    )) :
	    g_strdup("-");


	gtk_clist_freeze(clist);

	/* Append or insert */
	if(row < 0)
	    row = gtk_clist_append(clist, v);
	else
	    row = gtk_clist_insert(clist, row, v);

	if(row >= 0)
	{
	    GdkPixmap *pixmap = NULL;
	    GdkBitmap *mask = NULL;

	    /* Get icon and update first cell */
	    switch(format)
	    {
	      case SndObjTypeDSP:
		pixmap = yp->pixmap_sndobj_dsp;
		mask = yp->mask_sndobj_dsp;
		break;
	      case SndObjTypeMIDI:
		pixmap = yp->pixmap_sndobj_midi;
		mask = yp->mask_sndobj_midi;
		break;
	    }
	    if(pixmap != NULL)
		gtk_clist_set_pixtext(
		    clist, row, 0,
		    v[0], YC_LIST_PIXMAP_TEXT_SPACING,
		    pixmap, mask
		);

	    /* Set row data */
	    if(filename != NULL)
		gtk_clist_set_row_data_full(
		    clist, row,
		    g_strdup(filename),
		    YPlayerSndObjListItemDestroyCB
		);
	}

	gtk_clist_thaw(clist);

	/* Delete cell values */
	g_free(v[0]);
	g_free(v[1]);
	g_free(v[2]);
	g_free(v[3]);
	g_free(v[4]);

	return(row);
}

/*
 *      Adds sound object playing to list.
 */
gint YPlayerSndObjPlayingListAdd(
	yplayer_struct *yp, gint row,
	const gchar *filename, gint format, YID yid
)
{
	const gchar *cstrptr;
	gchar *v[1], *s;
	GtkCList *clist = (GtkCList *)YPLAYER_SNDOBJ_PLAYING_CLIST(yp);
	if(clist == NULL)
	    return(-1);

	/* Format cell values */
	cstrptr = (filename != NULL) ? strrchr(filename, DIR_DELIMINATOR) : NULL;
	if(cstrptr != NULL)
	    cstrptr++;
	v[0] = g_strdup((cstrptr != NULL) ? cstrptr : "(null)");
	s = strrchr(v[0], '.');
	if(s != NULL)
	    *s = '\0';

	gtk_clist_freeze(clist);

	/* Append or insert */
	if(row < 0)
	    row = gtk_clist_append(clist, v);
	else
	    row = gtk_clist_insert(clist, row, v);

	if(row > -1)
	{
	    GdkPixmap *pixmap = NULL;
	    GdkBitmap *mask = NULL;
	    yplayer_sndobj_playing_struct *sop;

	    /* Get icon and update first cell */
	    switch(format)
	    {
	      case SndObjTypeDSP:
		pixmap = yp->pixmap_sndobj_dsp;
		mask = yp->mask_sndobj_dsp;
		break;
	      case SndObjTypeMIDI:
		pixmap = yp->pixmap_sndobj_midi;
		mask = yp->mask_sndobj_midi;
		break;
	    }
	    if(pixmap != NULL)
		gtk_clist_set_pixtext(
		    clist, row, 0,
		    v[0], YC_LIST_PIXMAP_TEXT_SPACING,
		    pixmap, mask
		);

	    /* Set row data */
	    sop = YPLAYER_SNDOBJ_PLAYING(
		g_malloc0(sizeof(yplayer_sndobj_playing_struct))
	    );
	    if(sop != NULL)
	    {
		sop->yp = yp;
		sop->filename = g_strdup(
		    (filename != NULL) ? filename : "(null)"
		);
		sop->yid = yid;
		gtk_clist_set_row_data_full(
		    clist, row,
		    sop,
		    YPlayerSndObjPlayingListItemDestroyCB
		);
	    }
	}

	gtk_clist_thaw(clist);

	/* Delete cell values */
	g_free(v[0]);

	return(row);
}

/*
 *	Clears sound object playing list.
 */
void YPlayerSndObjPlayingListClear(yplayer_struct *yp)
{
	GtkCList *clist = (GtkCList *)YPLAYER_SNDOBJ_PLAYING_CLIST(yp);
	if(clist == NULL)
	    return;

	/* Delete each row and the row data.  In the row data destroy
	 * callback the sop structure will be deleted and the sound
	 * object will be killed (if connected).
	 */
	gtk_clist_freeze(clist);
	gtk_clist_clear(clist);
	gtk_clist_thaw(clist);
}

/*
 *	Regets all items for the audio CD tracks list.
 */
void YPlayerAudioCDTracksListUpdate(yplayer_struct *yp)
{
	YConnection *con = YPLAYER_CONNECTION(yp);
	GtkCList *clist = (GtkCList *)YPLAYER_AUDIOCD_TRACKS_CLIST(yp);
	if(clist == NULL)
	    return;

	/* Clear list */
	gtk_clist_freeze(clist);
	gtk_clist_clear(clist);
	gtk_clist_thaw(clist);

	/* If connected, get new list */
	if(con != NULL)
	{
	    gint total;
	    YAudioCDTrackStruct *t, **track = YGetAudioCDTracks(
		con, &total
	    );
	    if(track != NULL)
	    {
		gint i, row;
		gchar *v[4];

		/* Iterate through tracks list */
		gtk_clist_freeze(clist);
		for(i = 0; i < total; i++)
		{
		    t = track[i];
		    if(t == NULL)
			continue;

		    /* Format cell values */
		    v[0] = g_strdup_printf(
			"%i", t->number
		    );
		    v[1] = g_strdup_printf(
			"%.2i:%.2i:%.2i",
			(gint)(t->start_time / 3600),
			(gint)(t->start_time / 60),
			(gint)(t->start_time % 60)
		    );
		    v[2] = g_strdup_printf(
			"%.2i:%.2i",
			(int)(t->length / 60),
			(int)(t->length % 60)
		    );
		    v[3] = g_strdup(t->name);

		    /* Append new row */
		    row = gtk_clist_append(clist, v);
		    if(row > -1)
		    {

		    }

		    /* Delete cell values */
		    g_free(v[0]);
		    g_free(v[1]);
		    g_free(v[2]);
		    g_free(v[3]);
		}
		gtk_clist_thaw(clist);

		YFreeAudioCDTracksList(track, total);

		if(clist->rows > 0)
		    gtk_clist_select_row(clist, 0, 0);
	    }
	}
}


/*
 *	Loads play list items into the sound objects list.
 */
void YPlayerLoadPlayListFile(
	yplayer_struct *yp, const gchar *path, gboolean insert
)
{
	gchar *title = g_strdup("Untitled");
	gfloat	hscroll = 0.0f,
		vscroll = 0.0f;
	gint line_num, fline_num, insert_row;
	FILE *fp;
	gchar *buf, *line_ptr, **strv;
	GtkCList *clist = (GtkCList *)YPLAYER_SNDOBJ_CLIST(yp);
	YConnection *con = YPLAYER_CONNECTION(yp);


	if(clist == NULL)
	    return;

	/* Inserting rows? */
	if(insert)
	{
	    /* Get last selected row on the clist as the insert row */
	    GList *glist = clist->selection_end;
	    insert_row = (glist != NULL) ? (gint)glist->data : -1;
	}
	else
	{
	    /* Set the insert row to -1 to indicate that we should
	     * append rows.
	     */
	    insert_row = -1;

	    /* Since we are not inserting, we need to delete all
	     * existing rows in the clist.
	     */
	    gtk_clist_freeze(clist);
	    gtk_clist_clear(clist);
	    gtk_clist_thaw(clist);
	}

	/* Open file for reading */
	fp = fopen(path, "rb");
	if(fp == NULL)
	    return;

	/* Begin reading file */
	gtk_clist_freeze(clist);
	fline_num = line_num = 0;
	buf = NULL;
	while(TRUE)
	{
#define DO_CONTINUE	{	\
 fline_num++;			\
 continue;			\
}

	    /* Delete the previous line and get the next line */
	    g_free(buf);
	    buf = FGetStringLiteral(fp);
	    if(buf == NULL)
		break;

	    /* Get pointer to new line, seek past spaces and check if
	     * it is a comment.
	     */
	    line_ptr = buf;
	    while(ISBLANK(*line_ptr))
		line_ptr++;

	    /* Comment? */
	    if(*line_ptr == '#')
		DO_CONTINUE

	    /* Empty line? */
	    if(*line_ptr == '\0')
		DO_CONTINUE

	    /* Explode line using ';' as the deliminator character */
	    strv = g_strsplit(line_ptr, ";", -1);
	    if(strv == NULL)
		DO_CONTINUE

	    /* Begin handling line by checking which logical line number
	     * we are currently on
	     */

	    /* Header: "<title>;<hadj>;<vadj>" */
	    if(line_num == 0)
	    {
		int n = 0;

		g_free(title);
		title = STRDUP(strv[n]);
		if(strv[n] != NULL)
		    n++;

		hscroll = ATOF(strv[n]);
		if(strv[n] != NULL)
		    n++;

		vscroll = ATOF(strv[n]);
	    }
	    /* Subsequent lines: "<path>:<selected?>" */
	    else
	    {
		/* Get the values from exploded line */
		gint n = 0;
		gboolean is_selected;
		const gchar *sndobj_path;

		sndobj_path = strv[n];
		if(strv[n] != NULL) 
		    n++;

		is_selected = ATOI(strv[n]) ? TRUE : FALSE;
		if(strv[n] != NULL) 
		    n++;

		/* Load this line as one file path */
		if(sndobj_path != NULL)
		{
		    gint new_row;
		    YEventSoundObjectAttributes soa;

		    /* Get sound object's attributes */
		    if((con != NULL) ? YGetSoundObjectAttributes(
			con, sndobj_path, &soa
		    ) : TRUE)
		    {
			soa.format = 0;
			soa.sample_rate = 0;
			soa.sample_size = 0;
			soa.channels = 0;
			soa.length = 0;
			*soa.path = '\0';
		    }

		    /* Add this sound object to the list */
		    new_row = YPlayerSndObjListAdd(
			yp, insert_row,
			sndobj_path, soa.format,
			soa.sample_rate, soa.sample_size, soa.channels,
			soa.length
		    );
		    if(new_row > -1)
		    {
		        if(is_selected)
			    gtk_clist_select_row(clist, new_row, 0);
		        if(insert_row > -1)
			    insert_row++;
		    }
		}
	    }

	    /* Delete exploded line */
	    g_strfreev(strv);

	    /* Increment logical and file line numbers */
	    line_num++;
	    fline_num++;
	}
	gtk_clist_thaw(clist);

	/* Close the file */
	fclose(fp);

	/* Apply header values to clist */
	if(TRUE)
	{
	    GtkAdjustment *adj;

	    if(!insert)
	    {
		adj = clist->hadjustment;
		if(adj != NULL)
		{
		    adj->value = hscroll;
		    gtk_signal_emit_by_name(GTK_OBJECT(adj), "value_changed");
		}
		adj = clist->vadjustment;
		if(adj != NULL)
		{
		    adj->value = vscroll;
		    gtk_signal_emit_by_name(GTK_OBJECT(adj), "value_changed");
		}
	    }
	}

	g_free(title);

	YPlayerUpdateMenus(yp);
}

/*
 *	Saves play list items into the sound objects list.
 */
void YPlayerSavePlayListFile(
	yplayer_struct *yp, const gchar *path, gboolean selected_only
)
{
	const gchar *sndobj_path;
	gboolean is_selected;
	FILE *fp;
	GtkCList *clist = (GtkCList *)YPLAYER_SNDOBJ_CLIST(yp);


	if(clist == NULL)
	    return;

	/* Open file for writing */
	fp = fopen(path, "wb");
	if(fp == NULL)
	    return;

	/* Write header */
	if(TRUE)
	{
	    GtkAdjustment	*hadj = clist->hadjustment,
				*vadj = clist->vadjustment;

	    fprintf(
		fp,
		"%s;%.0f;%.0f\n",
		"Untitled",
		(hadj != NULL) ? hadj->value : 0.0f,
		(vadj != NULL) ? vadj->value : 0.0f
	    );
	}

#define WRITE_SNDOBJ	\
{ if(sndobj_path != NULL) \
   fprintf(fp, "%s;%i\n", sndobj_path, is_selected); \
}

	/* Save selected rows (or else save all rows)? */
	if(selected_only)
	{
	    gint row;
	    GList *glist = clist->selection;
	    while(glist != NULL)
	    {
		row = (gint)glist->data;
		sndobj_path = (const gchar *)gtk_clist_get_row_data(
		    clist, row
		);
		is_selected = TRUE;
		WRITE_SNDOBJ
		glist = glist->next;
	    }
	}
	else
	{
	    gint row;
	    for(row = 0; row < clist->rows; row++)
	    {
		sndobj_path = (const gchar *)gtk_clist_get_row_data(
		    clist, row
		);
		is_selected = (clist->selection != NULL) ?
		    (g_list_find(clist->selection, (gpointer)row) != NULL) :
		    FALSE;
		WRITE_SNDOBJ
	    }
	}
#undef WRITE_SNDOBJ

	fclose(fp);

	YPlayerUpdateMenus(yp);
}


/*
 *	Redraws the sound object advanced display.
 */
void YPlayerSndObjAdvDisplayDraw(yplayer_struct *yp)
{
	gint state, width, height;
	GdkPixmap *pixmap;
	GdkWindow *window;
	GdkFont *font;
	GtkStyle *style;
	GtkWidget *w;
	const gchar *path;
	const YEventSoundObjectAttributes *soa;
	YID yid;

	if(yp == NULL)
	    return;

	pixmap = yp->sndobjadv_display_pm;
	w = yp->sndobjadv_display_da;
	window = (w != NULL) ? w->window : NULL;

	if((pixmap == NULL) || (window == NULL))
	    return;

	state = GTK_WIDGET_STATE(w);
	style = gtk_widget_get_style(w);
	gdk_window_get_size(pixmap, &width, &height);
	if((style == NULL) || (width <= 0) || (height <= 0))
	    return;

	/* Clear background */
	gdk_draw_rectangle(
	    pixmap, style->base_gc[state], TRUE,
	    0, 0, width, height
	);

	/* Get sound object values */
	path = yp->sndobjadv_path;
	soa = &yp->sndobjadv_soa;
	yid = yp->sndobjadv_yid;

	font = style->font;

	/* No sound object loaded? */
	if((font != NULL) && (path == NULL))
	{
	    gint x, y;
	    gint lbearing, rbearing, swidth, ascent, descent;
	    const gchar *s = "No Sound Object Loaded";
	    GdkGC *gc = style->text_gc[state];

	    gdk_string_extents(
		font, s,
		&lbearing, &rbearing, &swidth, &ascent, &descent
	    );
	    x = (width / 2) - (swidth / 2);
	    y = (height / 2) - ((ascent + descent) / 2);
	    gdk_draw_string(
		pixmap, font, gc,
		x + lbearing, y + ascent,
		s
	    );
	}

	/* Sound object loaded? */
	if((font != NULL) && (path != NULL))
	{
	    gint x = 5, y = 5;
	    gint lbearing, rbearing, swidth, ascent, descent;
	    gchar *s, *s2, *name;
	    GdkPixmap *icon_pixmap = NULL;
	    GdkBitmap *icon_mask = NULL;
	    GdkGC *gc = style->fg_gc[state];

	    /* Draw icon */
	    switch(soa->format)
	    {
	      case SndObjTypeDSP:
		icon_pixmap = yp->pixmap_sndobj_dsp;
		icon_mask = yp->mask_sndobj_dsp;
		break;
	      case SndObjTypeMIDI:
		icon_pixmap = yp->pixmap_sndobj_midi;
		icon_mask = yp->mask_sndobj_midi;
		break;
	    }
	    gdk_gc_set_clip_mask(gc, icon_mask);
	    gdk_gc_set_clip_origin(gc, x, y);
	    if(icon_pixmap != NULL)
		gdk_draw_pixmap(
		    pixmap, gc, icon_pixmap,
		    0, 0, x, y, 20, 20
		);
	    gdk_gc_set_clip_mask(gc, NULL);

	    /* Get sound object path and simplify into just the name
	     * of the sound object.
	     */
	    s2 = strrchr(path, DIR_DELIMINATOR);
	    name = g_strdup((s2 != NULL) ? (s2 + 1) : path);
	    s2 = strrchr(name, '.');
	    if(s2 != NULL) *s2 = '\0';

	    /* Draw sound object name */
	    if(soa->format == SndObjTypeDSP)
		s = g_strdup_printf(
		    "%s   DSP   %s (%ld bytes)",
		    name,
		    YPlayerDataLengthToTimeString(
			soa->length,
			soa->sample_rate, soa->sample_size, soa->channels
		    ),
		    soa->length
		);
	    else if(soa->format == SndObjTypeMIDI)
		s = g_strdup_printf("%s   MIDI", name);
	    else
		s = g_strdup_printf("%s   *Unknown Type*", name);
	    gdk_string_extents(
		font, s,
		&lbearing, &rbearing, &swidth, &ascent, &descent
	    );
	    x = 5 + 20 + 2;
	    y = 5 + (20 / 2) - ((ascent + descent) / 2);
	    gc = style->text_gc[state];
	    gdk_draw_string(
		pixmap, font, gc,
		x + lbearing, y + ascent,
		s
	    );
	    g_free(s);
	    g_free(name);

	    /* Draw sound object attributes */
	    if(soa->format == SndObjTypeDSP)
	    {
		GtkWidget *w2;
		GtkRange *range = (GtkRange *)yp->sndobjadv_position_scale;
		GtkAdjustment *adj = (range != NULL) ? range->adjustment : NULL;

		/* Sample rate, sample size, and channels */
		s = g_strdup_printf(
		    "Sample Rate: %i Hz  Bits: %i  Channels: %s",
		    soa->sample_rate,
		    soa->sample_size,
		    (soa->channels == 4) ? "Quad" :
			((soa->channels == 2) ? "Stereo" : "Mono")
		);
		gdk_string_extents(
		    font, s,
		    &lbearing, &rbearing, &swidth, &ascent, &descent
		);
		x = 5;
		y = 5 + 20 + 5;
		gc = style->text_gc[state];
		gdk_draw_string(
		    pixmap, font, gc,
		    x + lbearing, y + ascent,
		    s
		);
		g_free(s);

		/* Position */
		x = 5;
		y += (ascent + descent) + 5;
		if(adj != NULL)
		{
		    if(yid != YIDNULL)
		      s = g_strdup_printf(
			"Position: %s (%ld bytes) %.1f%%   YID: %ld",
			YPlayerDataLengthToTimeString(
			    (gulong)adj->value,
			    soa->sample_rate, soa->sample_size, soa->channels
			),
			(gulong)adj->value,
			(gfloat)adj->value / (gfloat)soa->length * 100.0f,
			yid
		      );
		    else
		      s = g_strdup_printf(
			"Position: %s (%ld bytes) %.1f%%",
			YPlayerDataLengthToTimeString(
			    (gulong)adj->value,
			    soa->sample_rate, soa->sample_size, soa->channels
			),
			(gulong)adj->value,
			(gfloat)adj->value / (gfloat)soa->length * 100.0f
		      );

		    gdk_string_extents(
			font, s,
			&lbearing, &rbearing, &swidth, &ascent, &descent
		    );
		    gc = style->text_gc[state];
		    gdk_draw_string(
			pixmap, font, gc,
			x + lbearing, y + ascent,
			s
		    );
		    g_free(s);

		    x += swidth + (5 * 3);
		}

		/* Playing icon */
		if(yid != YIDNULL)
		{
		    GdkBitmap *mask = yp->mask_sndobjadv_playing;
		    gint mask_width = 14, mask_height = 14;
		    gint	x2 = x,
				y2 = y - 2;

		    gc = style->fg_gc[state];
		    gdk_gc_set_clip_mask(gc, mask);
		    gdk_gc_set_clip_origin(gc, x2, y2);
		    gdk_draw_rectangle(
			pixmap, gc, TRUE, x2, y2, mask_width, mask_height
		    );
		    gdk_gc_set_clip_mask(gc, NULL);

		    x += mask_width + 5;
		}

		/* Repeating */
		w2 = yp->sndobjadv_repeat_tb;
		if((yid != YIDNULL) &&
		   ((w2 != NULL) ? GTK_TOGGLE_BUTTON(w2)->active : FALSE)
		)
		{
		    GdkBitmap *mask = yp->mask_sndobjadv_repeating;
		    gint mask_width = 14, mask_height = 14;
		    gint        x2 = x,
				y2 = y - 2;

		    gc = style->fg_gc[state];
		    gdk_gc_set_clip_mask(gc, mask);
		    gdk_gc_set_clip_origin(gc, x2, y2);
		    gdk_draw_rectangle(
			pixmap, gc, TRUE, x2, y2, mask_width, mask_height
		    );
		    gdk_gc_set_clip_mask(gc, NULL);

		    x += mask_width + 5;
		}
	    }
	    else if(soa->format == SndObjTypeMIDI)
	    {
		GtkWidget *w2;

		x = 5;
		y = 5 + 20 + 5;

		/* Playing icon */
		if(yid != YIDNULL)
		{
		    GdkBitmap *mask = yp->mask_sndobjadv_playing;
		    gint mask_width = 14, mask_height = 14;
		    gint x2 = x, y2 = y;

		    gc = style->fg_gc[state];
		    gdk_gc_set_clip_mask(gc, mask);
		    gdk_gc_set_clip_origin(gc, x2, y2);
		    gdk_draw_rectangle(
			pixmap, gc, TRUE, x2, y2, mask_width, mask_height
		    );
		    gdk_gc_set_clip_mask(gc, NULL);

		    x += mask_width + 5;
		}
		/* Repeating */
		w2 = yp->sndobjadv_repeat_tb;
		if((yid != YIDNULL) &&
		   ((w2 != NULL) ? GTK_TOGGLE_BUTTON(w2)->active : FALSE)
		)
		{
		    GdkBitmap *mask = yp->mask_sndobjadv_repeating;
		    gint mask_width = 14, mask_height = 14;
		    gint x2 = x, y2 = y;

		    gc = style->fg_gc[state];
		    gdk_gc_set_clip_mask(gc, mask);
		    gdk_gc_set_clip_origin(gc, x2, y2);
		    gdk_draw_rectangle(
			pixmap, gc, TRUE, x2, y2, mask_width, mask_height
		    );
		    gdk_gc_set_clip_mask(gc, NULL);

		    x += mask_width + 5;
		}

	    }	/* Draw sound object attributes */
	}	/* Sound object loaded? */

	/* Draw focus? */
	if(GTK_WIDGET_HAS_FOCUS(w) && GTK_WIDGET_SENSITIVE(w))
	    gdk_draw_rectangle(
		pixmap, style->fg_gc[state], FALSE,
		0, 0, width - 1, height - 1
	    );

	gdk_window_clear(window);
}

/*
 *	Redraws the sound object advanced volume.
 */
void YPlayerSndObjAdvVolumeDraw(yplayer_struct *yp)
{
	gint state, width, height, center_x, center_y;
	gfloat center_xc, center_yc, volume_c;
	GdkFont *font;
	GdkPixmap *pixmap;
	GdkWindow *window;
	GtkAdjustment *adj = NULL;
	GtkStyle *style;
	GtkWidget *w;


	if(yp == NULL)
	    return;

	pixmap = yp->sndobjadv_volume_pm;
	w = yp->sndobjadv_volume_da;
	window = (w != NULL) ? w->window : NULL;

	if((pixmap == NULL) || (window == NULL))
	    return;

	state = GTK_WIDGET_STATE(w);
	style = gtk_widget_get_style(w);
	gdk_window_get_size(pixmap, &width, &height);
	if((style == NULL) || (width <= 0) || (height <= 0))
	    return;

	/* Clear background */
	gdk_draw_rectangle(
	    pixmap, style->base_gc[state], TRUE,
	    0, 0, width, height
	);


	if(yp->sndobjadv_volume_scale != NULL)
	    adj = GTK_RANGE(yp->sndobjadv_volume_scale)->adjustment;

	/* Get center of sound [-1.0, 1.0], left hand rule */
	center_xc = yp->sndobjadv_volume_x;
	center_yc = yp->sndobjadv_volume_y;
	volume_c = CLIP((adj != NULL) ? adj->value : 0.0f, 0.0f, 1.0f);

	/* Calculate center of sound in window coordinates */
	center_x = (gint)(((center_xc + 1.0f) / 2.0f) * width);
	center_y = (gint)(((1.0f - center_yc) / 2.0f) * height);

	font = style->font;
	if(font != NULL)
	{
	    gint x, y;
	    const gchar *s;
	    gint lbearing, rbearing, swidth, ascent, descent;
	    GdkGC *gc = style->fg_gc[state];
	    GdkGCValues gcv;
	    GdkBitmap *mask = yp->mask_sndobjadv_source;
	    gint mask_width = 14, mask_height = 14;


	    /* Draw center grid lines */
	    gdk_draw_line(
		pixmap, gc,
		0, height / 2, width, height / 2
	    );
	    gdk_draw_line(
		pixmap, gc,
		width / 2, 0, width / 2, height
	    );

	    /* Draw labels at each corner */
	    gc = style->text_gc[state];
	    s = "FL";
	    gdk_string_extents(
		font, s,
		&lbearing, &rbearing, &swidth, &ascent, &descent
	    );
	    x = 2;
	    y = 2;
	    gdk_draw_string(
		pixmap, font, gc,
		x + lbearing, y + ascent,
		s
	    );
	    s = "FR";
	    gdk_string_extents(
		font, s,
		&lbearing, &rbearing, &swidth, &ascent, &descent
	    );
	    x = width - 3 - swidth;
	    y = 2;
	    gdk_draw_string(
		pixmap, font, gc,
		x + lbearing, y + ascent,
		s
	    );
	    s = "BL";
	    gdk_string_extents(
		font, s,
		&lbearing, &rbearing, &swidth, &ascent, &descent
	    );
	    x = 2;
	    y = height - 3 - (ascent + descent);
	    gdk_draw_string(
		pixmap, font, gc,
		x + lbearing, y + ascent,
		s
	    );
	    s = "BR";
	    gdk_string_extents(
		font, s,
		&lbearing, &rbearing, &swidth, &ascent, &descent
	    );
	    x = width - 3 - swidth;
	    y = height - 3 - (ascent + descent);
	    gdk_draw_string(
		pixmap, font, gc,
		x + lbearing, y + ascent,
		s
	    );

	    /* Draw sound source icon */
	    x = center_x - (mask_width / 2);
	    y = center_y - (mask_height / 2);
	    gdk_gc_set_clip_mask(gc, mask);
	    gdk_gc_set_clip_origin(gc, x, y);
	    gdk_draw_rectangle(
		pixmap, gc, TRUE, x, y, mask_width, mask_height
	    );
	    gdk_gc_set_clip_mask(gc, NULL);
/*
	    gdk_draw_arc(
		pixmap, gc, TRUE,
		center_x - r, center_y - r,
		2 * r, 2 * r,
		0 * 64, 360 * 64
	    );
 */

	    /* Draw inversion cross hair */
	    gdk_gc_get_values(gc, &gcv);
	    gdk_gc_set_function(gc, GDK_INVERT);
	    gdk_draw_line(
		pixmap, gc,
		0, center_y, width, center_y
	    );
	    gdk_draw_line(
		pixmap, gc,
		center_x, 0, center_x, height
	    );
	    gdk_gc_set_function(gc, gcv.function);
	}

	/* Draw focus? */
	if(GTK_WIDGET_HAS_FOCUS(w) && GTK_WIDGET_SENSITIVE(w))
	    gdk_draw_rectangle(
		pixmap, style->fg_gc[state], FALSE,
		0, 0, width - 1, height - 1
	    );

	gdk_window_clear(window);
}


/*
 *	Save current position value of window to configuration list on
 *	core structure.
 */
void YPlayerSavePosition(yplayer_struct *yp)
{
	GtkWidget *w;
	yc_core_struct *core_ptr;


	if(yp == NULL)
	    return;

	core_ptr = YC_CORE(yp->core_ptr);
	if(core_ptr == NULL)
	    return;

	/* Do not save positions? */
	if(!YC_GET_CFG_I(
	    core_ptr, YC_CFG_PARM_PLAYER_SAVE_POSITIONS
	))
	    return;

	/* Save toplevel position */
	w = YPLAYER_TOPLEVEL(yp);
	if(w != NULL)
	{
	    GdkWindow *window = w->window;
	    gint x = 0, y = 0, width, height;


	    if(window != NULL)
		gdk_window_get_root_origin(window, &x, &y);
	    YC_SET_CFG_I(core_ptr, YC_CFG_PARM_PLAYER_X, x);
	    YC_SET_CFG_I(core_ptr, YC_CFG_PARM_PLAYER_Y, y);

	    width = w->allocation.width;
	    height = w->allocation.height;
	    YC_SET_CFG_I(core_ptr, YC_CFG_PARM_PLAYER_WIDTH, width);
	    YC_SET_CFG_I(core_ptr, YC_CFG_PARM_PLAYER_HEIGHT, height);
	}

	/* Save notebook page */	
	w = YPLAYER_NOTEBOOK(yp);
	if(w != NULL)
	{
	    GtkNotebook *notebook = GTK_NOTEBOOK(w);
	    YC_SET_CFG_I(
		core_ptr, YC_CFG_PARM_PLAYER_PAGE,
		gtk_notebook_get_current_page(notebook)
	    );
	}

	/* Save sound objects panned position */
	w = YPLAYER_SNDOBJ_PANED(yp);
	if(w != NULL)
	    YC_SET_CFG_I(
		core_ptr, YC_CFG_PARM_PLAYER_SNDOBJ_PANED_POSITION,
		GTK_PANED(w)->child1_size
	    );

	/* Save sound objects list column widths */
	w = YPLAYER_SNDOBJ_CLIST(yp);
	if(w != NULL)
	{
	    gint i;
	    GList *glist = NULL;
	    GtkCList *clist = GTK_CLIST(w);
	    const gint columns = clist->columns;
	    GtkCListColumn *column;
	    cfg_intlist_struct *intlist;
	    for(i = 0; i < columns; i++)
	    {
		column = &clist->column[i];
		glist = g_list_append(
		    glist, (gpointer)column->width
		);
	    }
	    intlist = CFGIntListNew(glist);
	    g_list_free(glist);
	    YC_SET_CFG_INTLIST(
		core_ptr, YC_CFG_PARM_PLAYER_SNDOBJ_LIST_COLUMN_WIDTHS,
		intlist
	    );
	    CFGIntListDelete(intlist);
	}

	/* Save sound object playing list column widths */
	w = YPLAYER_SNDOBJ_PLAYING_CLIST(yp);
	if(w != NULL)
	{

	}

	/* Save sound object play options */
	w = yp->sndobj_change_mode_tb;
	if(w != NULL)
	    YC_SET_CFG_I(
		core_ptr, YC_CFG_PARM_PLAYER_SNDOBJ_CHANGE_MODE,
		GTK_TOGGLE_BUTTON(w)->active
	    );
	w = yp->sndobj_repeat_tb;
	if(w != NULL)
	    YC_SET_CFG_I(
		core_ptr, YC_CFG_PARM_PLAYER_SNDOBJ_REPEAT,
		GTK_TOGGLE_BUTTON(w)->active
	    );

	/* Save sound object advanced play options */
	YC_SET_CFG_F(
	    core_ptr, YC_CFG_PARM_PLAYER_SNDOBJADV_VOLUME_X,
	    yp->sndobjadv_volume_x
	);
	YC_SET_CFG_F(
	    core_ptr, YC_CFG_PARM_PLAYER_SNDOBJADV_VOLUME_Y,
	    yp->sndobjadv_volume_y
	);
	w = yp->sndobjadv_volume_scale;
	if(w != NULL)
	{
	    GtkAdjustment *adj = GTK_RANGE(w)->adjustment;
	    YC_SET_CFG_F(
		core_ptr, YC_CFG_PARM_PLAYER_SNDOBJADV_VOLUME_A,
		1.0f - ((adj != NULL) ? adj->value : 0.0f)
	    );
	}
	YC_SET_CFG_S(
	    core_ptr, YC_CFG_PARM_PLAYER_SNDOBJADV_LAST_FILE,
	    (yp->sndobjadv_path != NULL) ? yp->sndobjadv_path : ""
	);
	w = yp->sndobjadv_repeat_tb;
	if(w != NULL)
	    YC_SET_CFG_I(
		core_ptr, YC_CFG_PARM_PLAYER_SNDOBJADV_REPEAT,
		GTK_TOGGLE_BUTTON(w)->active
	    );

	/* Save audio CD tracks list column widths */
	w = YPLAYER_AUDIOCD_TRACKS_CLIST(yp);
	if(w != NULL)
	{
	    gint i;
	    GList *glist = NULL;
	    GtkCList *clist = GTK_CLIST(w);
	    const gint columns = clist->columns;
	    GtkCListColumn *column;
	    cfg_intlist_struct *intlist;
	    for(i = 0; i < columns; i++)
	    {
		column = &clist->column[i];
		glist = g_list_append(
		    glist, (gpointer)column->width
		);
	    }
	    intlist = CFGIntListNew(glist);
	    g_list_free(glist);
	    YC_SET_CFG_INTLIST(
		core_ptr, YC_CFG_PARM_PLAYER_AUDIOCD_TRACKS_LIST_COLUMN_WIDTHS,
		intlist
	    );
	    CFGIntListDelete(intlist);
	}
}


/*
 *	Create a new yconsole.
 */
yplayer_struct *YPlayerNew(gpointer core_ptr)
{
	gint border_major = 5, border_minor = 2;
	GdkWindow *window;
	GtkAccelGroup *accelgrp;
	GtkWidget *w, *parent, *parent2, *parent3, *parent4, *toplevel;
	connection_bar_struct *cb;
	dspview_struct *dv;
	yc_core_struct *cp = YC_CORE(core_ptr);
	yplayer_struct *yp = (yplayer_struct *)g_malloc0(
	    sizeof(yplayer_struct)
	);
	if(yp == NULL)
	    return(NULL);


	/* Reset values */
	yp->initialized = TRUE;
	yp->map_state = FALSE;
	yp->busy_count = 0;
	yp->accelgrp = accelgrp = gtk_accel_group_new();
	yp->core_ptr = core_ptr;

	yp->con = NULL;
	yp->yshm_buf = NULL;
	yp->yshm_buf_len = 0;
	yp->yshm_id = -1;

	yp->start_time = 0;
	yp->audio_mode_name = NULL;
	yp->sample_rate = 0;
	yp->sample_size = 0;
	yp->channels = 0;
	yp->cycle_us = 0;
	yp->cycle_load = 0.0f;
	yp->num_fragments = 0;
	yp->fragment_size = 0;
	yp->flip_stereo = FALSE;
	yp->shellout = FALSE;
	yp->version_major = yp->version_minor = 0;
	yp->server_name = yp->sound_name = NULL;

	yp->sndobjadv_freeze_count = 0;
	yp->sndobjadv_button = 0;
	yp->sndobjadv_path = NULL;
	memset(&yp->sndobjadv_soa, 0x00, sizeof(YEventSoundObjectAttributes));
	yp->sndobjadv_yid = YIDNULL;
	yp->sndobjadv_volume_x = YC_GET_CFG_F(
	    cp, YC_CFG_PARM_PLAYER_SNDOBJADV_VOLUME_X
	);
	yp->sndobjadv_volume_y = YC_GET_CFG_F(
	    cp, YC_CFG_PARM_PLAYER_SNDOBJADV_VOLUME_Y
	);

	/* Create toplevel */
	yp->toplevel = toplevel = w = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_widget_set_name(w, "yplayer window");
	if(cp != NULL)
	{
	    if(YC_GET_CFG_I(cp, YC_CFG_PARM_PLAYER_SAVE_POSITIONS))
	    {
		gtk_widget_set_uposition(
		    w,
		    YC_GET_CFG_I(cp, YC_CFG_PARM_PLAYER_X),
		    YC_GET_CFG_I(cp, YC_CFG_PARM_PLAYER_Y)
		);
		gtk_widget_set_usize(
		    w,
		    YC_GET_CFG_I(cp, YC_CFG_PARM_PLAYER_WIDTH),
		    YC_GET_CFG_I(cp, YC_CFG_PARM_PLAYER_HEIGHT)
		);
	    }
	}
	gtk_window_set_policy(GTK_WINDOW(w), TRUE, TRUE, TRUE);
	gtk_widget_realize(w);
	window = w->window;
	if(window != NULL)
	{
	    gdk_window_set_decorations(
		window,
		GDK_DECOR_BORDER | GDK_DECOR_RESIZEH |
		GDK_DECOR_TITLE | GDK_DECOR_MENU |
		GDK_DECOR_MINIMIZE | GDK_DECOR_MAXIMIZE
	    );
	    gdk_window_set_functions(
		window,
		GDK_FUNC_MOVE | GDK_FUNC_RESIZE |
		GDK_FUNC_MINIMIZE | GDK_FUNC_MAXIMIZE |
		GDK_FUNC_CLOSE
	    );
	    GUISetWMIcon(window, (guint8 **)icon_yplayer_48x48_xpm);
	}
	gtk_accel_group_attach(accelgrp, GTK_OBJECT(w));
	gtk_signal_connect(
	    GTK_OBJECT(w), "delete_event",
	    GTK_SIGNAL_FUNC(YPlayerDeleteEventCB), yp
	);

	/* Create connection bar */
	yp->conbar = cb = ConBarNew(
	    CONBAR_DISPLAY_FLAG_WINDOW_MENU | CONBAR_DISPLAY_FLAG_LOCATION |
	    CONBAR_DISPLAY_FLAG_DSPVIEW | CONBAR_DISPLAY_FLAG_REFRESH |
	    CONBAR_DISPLAY_FLAG_SHELLOUT | CONBAR_DISPLAY_FLAG_SHUTDOWN,
	    cp->dsp_bg_img_width + 2, cp->dsp_bg_img_height + 2,
	    yp,
	    YPlayerConnectCB, YPlayerYServerListCB, YPlayerCloseCB,
	    YPlayerNewYConsoleCB, YPlayerNewYAudioCB, YPlayerNewYMixerCB,
	    YPlayerNewYPlayerCB, YPlayerUptionsCB,
	    YPlayerRefreshCB, YPlayerShelloutCB, YPlayerShutdownCB,
	    toplevel, accelgrp
	);
	dv = ConBarGetDSPView(cb);
	ConBarMap(cb);

	if(cp != NULL)
	    DSPViewSetBG(
		dv, cp->dsp_bg_img,
		cp->dsp_bg_img_width, cp->dsp_bg_img_height,
		cp->dsp_bg_img_bpl, 4
	    );
	DSPViewSetDisplayStyle(dv, DSPVIEW_DISPLAY_STYLE_NOVA);
	DSPViewSetMergeMode(dv, DSPVIEW_MERGE_MODE_SET);
	DSPViewSetDSPColor(dv, DSPVIEW_DEF_DSP_COLOR);  /* 0xAABBGGRR */
	w = DSPVIEW_VIEW(dv);
	gtk_widget_add_events(
	    w, GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "button_press_event",
	    GTK_SIGNAL_FUNC(YPlayerDSPViewButtonCB), yp
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "button_release_event",
	    GTK_SIGNAL_FUNC(YPlayerDSPViewButtonCB), yp
	);


	/* Get left pane vbox from connection bar and add the notebook
	 * to it.
	 */
	parent = cb->left_pane_vbox;
	if(parent != NULL)
	{
	    int i;
	    gchar *heading[5];
	    GtkAdjustment *adj;
	    GtkStyle *style;
	    GtkWidget *menu;
	    GtkNotebook *notebook;
	    GtkCList *clist;
	    const cfg_intlist_struct *intlist;

	    /* Frame for notebook */
	    w = gtk_frame_new(NULL);
	    gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_IN);
	    gtk_widget_set_usize(w, 250, 100);
	    gtk_box_pack_start(GTK_BOX(parent), w, TRUE, TRUE, 0);
	    gtk_widget_show(w);
	    parent2 = w;

	    /* Main notebook */
	    yp->notebook = w = gtk_notebook_new();
	    notebook = GTK_NOTEBOOK(w);
	    gtk_notebook_set_tab_pos(notebook, GTK_POS_LEFT);
	    gtk_notebook_set_scrollable(notebook, TRUE);
	    gtk_notebook_set_show_tabs(notebook, TRUE);
	    gtk_notebook_set_show_border(notebook, TRUE);
	    gtk_container_add(GTK_CONTAINER(parent2), w);
	    gtk_widget_show(w);


	    /* Create sound objects page */
	    w = GUICreateMenuItemIcon((guint8 **)icon_sounds_20x20_xpm);
	    gtk_widget_show(w);
	    parent2 = gtk_vbox_new(FALSE, border_major);
	    gtk_container_border_width(GTK_CONTAINER(parent2), border_major);
	    gtk_notebook_append_page(notebook, parent2, w);
	    gtk_widget_show(parent2);

	    /* Hpaned */
	    yp->sndobj_paned = w = gtk_hpaned_new();
	    gtk_paned_set_handle_size(
		GTK_PANED(w), YC_DEF_PANED_HANDLE_SIZE
	    );
	    gtk_paned_set_gutter_size(
		GTK_PANED(w), YC_DEF_PANED_GUTTER_SIZE
	    );
	    gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, TRUE, 0);	    
	    gtk_widget_show(w);
	    parent3 = w;

	    /* Scrolled window for clist */
	    w = gtk_scrolled_window_new(NULL, NULL);
	    gtk_scrolled_window_set_policy(
		GTK_SCROLLED_WINDOW(w),
		GTK_POLICY_AUTOMATIC,
		GTK_POLICY_AUTOMATIC
	    );
	    gtk_paned_add1(GTK_PANED(parent3), w);
	    gtk_widget_show(w);
	    parent4 = w;

	    /* Sound objects clist */
	    heading[0] = "Name";
	    heading[1] = "Sample Rate";
	    heading[2] = "Bits";
	    heading[3] = "Channels";
	    heading[4] = "Length";
	    yp->sndobj_clist = w = gtk_clist_new_with_titles(5, heading);
	    clist = GTK_CLIST(w);
	    gtk_clist_column_titles_passive(clist);
	    intlist = YC_GET_CFG_INTLIST(
		cp, YC_CFG_PARM_PLAYER_SNDOBJ_LIST_COLUMN_WIDTHS
	    );
	    if(intlist != NULL)
	    {
		GList *glist = intlist->list;
		for(i = 0; i < clist->columns; i++)
		{
		    if(glist == NULL)
			break;

		    gtk_clist_set_column_width(
			clist, i, (gint)glist->data
		    );
		    glist = g_list_next(glist);
		}
	    }
	    gtk_clist_set_row_height(clist, YC_LIST_ROW_SPACING);
	    gtk_widget_add_events(
		w,
		GDK_KEY_PRESS_MASK | GDK_KEY_RELEASE_MASK |
		GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "key_press_event",
		GTK_SIGNAL_FUNC(YPlayerKeyEventCB), yp
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "key_release_event",
		GTK_SIGNAL_FUNC(YPlayerKeyEventCB), yp
	    );
	    gtk_signal_connect_after(
		GTK_OBJECT(w), "button_press_event",
		GTK_SIGNAL_FUNC(YPlayerButtonEventCB), yp
	    );
	    gtk_signal_connect_after(
		GTK_OBJECT(w), "button_release_event",
		GTK_SIGNAL_FUNC(YPlayerButtonEventCB), yp
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "select_row",
		GTK_SIGNAL_FUNC(YPlayerCListSelectRowCB), yp
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "unselect_row",
		GTK_SIGNAL_FUNC(YPlayerCListUnselectRowCB), yp
	    );
	    gtk_clist_set_selection_mode(clist, GTK_SELECTION_EXTENDED);
	    gtk_container_add(GTK_CONTAINER(parent4), w);
	    if(TRUE)
	    {
		const GtkTargetEntry dnd_tar_types[] = {
{"text/plain",		0,	YC_DND_TYPE_INFO_TEXT_PLAIN},
{"text/uri-list",	0,	YC_DND_TYPE_INFO_TEXT_URI_LIST},
{"STRING",		0,	YC_DND_TYPE_INFO_STRING}
		};
		GUIDNDSetTar(
		    w,
		    dnd_tar_types,
		    sizeof(dnd_tar_types) / sizeof(GtkTargetEntry),
		    GDK_ACTION_COPY,		/* Actions */
		    GDK_ACTION_COPY,		/* Default action if same */
		    GDK_ACTION_COPY,		/* Default action */
		    YPlayerSndObjListDragDataReceivedCB,
		    yp
		);
		gtk_signal_connect_after(
		    GTK_OBJECT(w), "drag_motion",
		    GTK_SIGNAL_FUNC(YPlayerSndObjListDragMotionCB), yp
		);
	    }
	    gtk_widget_show(w);
	    /* Load pixmap and mask pairs for mode list items */
	    style = gtk_widget_get_style(w);
	    yp->pixmap_sndobj_midi = gdk_pixmap_create_from_xpm_d(
		window, &yp->mask_sndobj_midi,
		&style->bg[GTK_STATE_NORMAL],
		(gchar **)icon_file_music_20x20_xpm
	    );
	    yp->pixmap_sndobj_dsp = gdk_pixmap_create_from_xpm_d(
		window, &yp->mask_sndobj_dsp,
		&style->bg[GTK_STATE_NORMAL],
		(gchar **)icon_file_sound_20x20_xpm
	    );

	    /* Right click menu */
	    yp->sndobj_menu = menu = GUIMenuCreate();
	    if(menu != NULL)
	    {
		guchar **icon;
		const gchar *label;
		gint accel_key;
		guint accel_mods;
		gpointer mclient_data = yp;
		void (*func_cb)(GtkWidget *w, gpointer);
		gpointer fw;

#define DO_ADD_MENU_ITEM_LABEL  \
{ w = GUIMenuItemCreate( \
  menu, GUI_MENU_ITEM_TYPE_LABEL, accelgrp, \
  icon, label, accel_key, accel_mods, (void **)&fw, \
  mclient_data, func_cb \
 ); }
#define DO_ADD_MENU_ITEM_CHECK	\
{ w = GUIMenuItemCreate( \
   menu, GUI_MENU_ITEM_TYPE_CHECK, accelgrp, \
   icon, label, accel_key, accel_mods, (void **)&fw, \
   mclient_data, func_cb \
  ); }
#define DO_ADD_MENU_SEP         \
{ w = GUIMenuItemCreate( \
  menu, GUI_MENU_ITEM_TYPE_SEPARATOR, NULL, \
  NULL, NULL, 0, 0, NULL, \
  NULL, NULL \
 ); }
		icon = (guchar **)icon_yplayer_play_20x20_xpm;
		label = "Play";
		accel_key = 0;
		accel_mods = 0;
		func_cb = YPlayerSndObjPlayCB;
		DO_ADD_MENU_ITEM_LABEL
		yp->sndobj_play_mi = w;

		DO_ADD_MENU_SEP

		icon = (guchar **)icon_yplayer_file_20x20_xpm;
		label = "Add...";
		accel_key = 0;
		accel_mods = 0;
		func_cb = YPlayerSndObjAddCB;
		DO_ADD_MENU_ITEM_LABEL
		yp->sndobj_add_mi = w;

		icon = (guchar **)icon_yplayer_remove_20x20_xpm;
		label = "Remove";
		accel_key = 0;
		accel_mods = 0;
		func_cb = YPlayerSndObjRemoveCB;
		DO_ADD_MENU_ITEM_LABEL
		yp->sndobj_remove_mi = w;

		icon = (guchar **)icon_yplayer_up_20x20_xpm;
		label = "Shift Up";
		accel_key = 0;
		accel_mods = 0;
		func_cb = YPlayerSndObjShiftUpCB;
		DO_ADD_MENU_ITEM_LABEL
		yp->sndobj_shift_up_mi = w;

		icon = (guchar **)icon_yplayer_down_20x20_xpm;
		label = "Shift Down";
		accel_key = 0;
		accel_mods = 0;
		func_cb = YPlayerSndObjShiftDownCB;
		DO_ADD_MENU_ITEM_LABEL
		yp->sndobj_shift_down_mi = w;

		DO_ADD_MENU_SEP

		icon = NULL;
		label = "Change Audio Mode On Play";
		accel_key = 0;
		accel_mods = 0;
		func_cb = YPlayerSndObjChangeModeToggleCB;
		DO_ADD_MENU_ITEM_CHECK
		yp->sndobj_change_mode_micheck = w;

		icon = NULL;
		label = "Repeat";
		accel_key = 0;
		accel_mods = 0;
		func_cb = YPlayerSndObjRepeatToggleCB;
		DO_ADD_MENU_ITEM_CHECK
		yp->sndobj_repeat_micheck = w;

		DO_ADD_MENU_SEP

		icon = NULL;
		label = "Select All";
		accel_key = 0;
		accel_mods = 0;
		func_cb = YPlayerSndObjSelectAllCB;
		DO_ADD_MENU_ITEM_LABEL
		yp->sndobj_select_all_mi = w;

		icon = NULL;
		label = "Unselect All";
		accel_key = 0;
		accel_mods = 0;
		func_cb = YPlayerSndObjUnselectAllCB;
		DO_ADD_MENU_ITEM_LABEL
		yp->sndobj_unselect_all_mi = w;

		icon = NULL;
		label = "Invert Selection";
		accel_key = 0;
		accel_mods = 0;
		func_cb = YPlayerSndObjInvertSelectionCB;
		DO_ADD_MENU_ITEM_LABEL
		yp->sndobj_invert_selection_mi = w;

		DO_ADD_MENU_SEP

		icon = (guchar **)icon_yplayer_open_file_20x20_xpm;
		label = "Open PlayList...";
		accel_key = 0;
		accel_mods = 0;
		func_cb = YPlayerSndObjOpenPlayListCB;
		DO_ADD_MENU_ITEM_LABEL
		yp->sndobj_open_playlist_mi = w;

		icon = (guchar **)icon_yplayer_open_file_20x20_xpm;
		label = "Insert PlayList...";
		accel_key = 0;
		accel_mods = 0;
		func_cb = YPlayerSndObjInsertPlayListCB;
		DO_ADD_MENU_ITEM_LABEL
		yp->sndobj_insert_playlist_mi = w;

		icon = (guchar **)icon_yplayer_save_file_20x20_xpm;
		label = "Save PlayList...";
		accel_key = 0;
		accel_mods = 0;
		func_cb = YPlayerSndObjSavePlayListCB;
		DO_ADD_MENU_ITEM_LABEL
		yp->sndobj_save_playlist_mi = w;

		icon = (guchar **)icon_yplayer_save_file_20x20_xpm;
		label = "Save Selected...";
		accel_key = 0;
		accel_mods = 0;
		func_cb = YPlayerSndObjSaveSelectedCB;
		DO_ADD_MENU_ITEM_LABEL
		yp->sndobj_save_selected_mi = w;

#undef DO_ADD_MENU_SEP
#undef DO_ADD_MENU_ITEM_CHECK
#undef DO_ADD_MENU_ITEM_LABEL
	    }


	    /* Scrolled window for clist */
	    w = gtk_scrolled_window_new(NULL, NULL);
	    gtk_scrolled_window_set_policy(
		GTK_SCROLLED_WINDOW(w),
		GTK_POLICY_AUTOMATIC,
		GTK_POLICY_AUTOMATIC
	    );
	    gtk_paned_add2(GTK_PANED(parent3), w);
	    gtk_widget_show(w);
	    parent4 = w;

	    /* Sound object playing clist */
	    heading[0] = "Playing";
	    yp->sndobj_playing_clist = w = gtk_clist_new_with_titles(1, heading);
	    clist = GTK_CLIST(w);
	    gtk_clist_column_titles_passive(clist);
	    gtk_clist_set_column_width(clist, 0, 10);
	    gtk_clist_set_row_height(clist, YC_LIST_ROW_SPACING);
	    gtk_widget_add_events(
		w,
		GDK_KEY_PRESS_MASK | GDK_KEY_RELEASE_MASK |
		GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "key_press_event",
		GTK_SIGNAL_FUNC(YPlayerKeyEventCB), yp
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "key_release_event",
		GTK_SIGNAL_FUNC(YPlayerKeyEventCB), yp
	    );
	    gtk_signal_connect_after(
		GTK_OBJECT(w), "button_press_event",
		GTK_SIGNAL_FUNC(YPlayerButtonEventCB), yp
	    );
	    gtk_signal_connect_after(
		GTK_OBJECT(w), "button_release_event",
		GTK_SIGNAL_FUNC(YPlayerButtonEventCB), yp
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "select_row",
		GTK_SIGNAL_FUNC(YPlayerCListSelectRowCB), yp
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "unselect_row",
		GTK_SIGNAL_FUNC(YPlayerCListUnselectRowCB), yp
	    );
	    gtk_clist_set_selection_mode(clist, GTK_SELECTION_EXTENDED);
	    gtk_container_add(GTK_CONTAINER(parent4), w);
	    gtk_widget_show(w);

	    /* Right click menu */
	    yp->sndobj_playing_menu = menu = GUIMenuCreate();
	    if(menu != NULL)
	    {
		guchar **icon;
		const gchar *label;
		gint accel_key;
		guint accel_mods;
		gpointer mclient_data = yp;
		void (*func_cb)(GtkWidget *w, gpointer);
		gpointer fw;

#define DO_ADD_MENU_ITEM_LABEL  \
{ w = GUIMenuItemCreate( \
  menu, GUI_MENU_ITEM_TYPE_LABEL, accelgrp, \
  icon, label, accel_key, accel_mods, (void **)&fw, \
  mclient_data, func_cb \
 ); }
#define DO_ADD_MENU_SEP         \
{ w = GUIMenuItemCreate( \
  menu, GUI_MENU_ITEM_TYPE_SEPARATOR, NULL, \
  NULL, NULL, 0, 0, NULL, \
  NULL, NULL \
 ); }

		icon = (guchar **)icon_yplayer_stop_20x20_xpm;
		label = "Stop";
		accel_key = 0;
		accel_mods = 0;
		func_cb = YPlayerSndObjStopCB;
		DO_ADD_MENU_ITEM_LABEL
		yp->sndobj_playing_stop_mi = w;

#undef DO_ADD_MENU_SEP
#undef DO_ADD_MENU_ITEM_LABEL
	    }

	    /* Set panned position */
	    w = YPLAYER_SNDOBJ_PANED(yp);
	    if(w != NULL)
		gtk_paned_set_position(
		    GTK_PANED(w),
		    YC_GET_CFG_I(cp, YC_CFG_PARM_PLAYER_SNDOBJ_PANED_POSITION)
		);

	    /* Hbox for sound object buttons */
	    w = gtk_hbox_new(FALSE, border_major);
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent3 = w;

	    w = gtk_hbox_new(FALSE, 0);
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent4 = w;

	    /* Open playlist menu button */
	    yp->sndobj_open_playlist_mbtn = w = MenuButtonNewPixmap(
		(guint8 **)icon_yplayer_open_file_menu_14x14_xpm,
		&menu
	    );
	    MenuButtonSetMapTrigger(w, MENU_BUTTON_MAP_TYPE_PRESSED_DRAG);
	    gtk_widget_set_usize(
		w, YPLAYER_BTN_HEIGHT, YPLAYER_BTN_HEIGHT
	    );
	    gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	    gtk_signal_connect(
		GTK_OBJECT(w), "clicked",
		GTK_SIGNAL_FUNC(YPlayerSndObjOpenPlayListCB), yp
	    );
	    GUISetWidgetTip(w, "Open playlist");
	    gtk_widget_show(w);
	    if(menu != NULL)
	    {
		guchar **icon;
		const gchar *label;
		gint accel_key;
		guint accel_mods;
		gpointer mclient_data = yp;
		void (*func_cb)(GtkWidget *w, gpointer);
		gpointer fw;

#define DO_ADD_MENU_ITEM_LABEL  \
{ w = GUIMenuItemCreate( \
  menu, GUI_MENU_ITEM_TYPE_LABEL, accelgrp, \
  icon, label, accel_key, accel_mods, (void **)&fw, \
  mclient_data, func_cb \
 ); }
#define DO_ADD_MENU_SEP         \
{ w = GUIMenuItemCreate( \
  menu, GUI_MENU_ITEM_TYPE_SEPARATOR, NULL, \
  NULL, NULL, 0, 0, NULL, \
  NULL, NULL \
 ); }
		icon = (guchar **)icon_yplayer_open_file_20x20_xpm;
		label = "Open PlayList...";
		accel_key = 0;
		accel_mods = 0;
		func_cb = YPlayerSndObjOpenPlayListCB;
		DO_ADD_MENU_ITEM_LABEL

		icon = (guchar **)icon_yplayer_open_file_20x20_xpm;
		label = "Insert PlayList...";
		accel_key = 0;
		accel_mods = 0;
		func_cb = YPlayerSndObjInsertPlayListCB;
		DO_ADD_MENU_ITEM_LABEL

#undef DO_ADD_MENU_SEP
#undef DO_ADD_MENU_ITEM_LABEL
	    }

	    /* Save playlist menu button */
	    yp->sndobj_save_playlist_mbtn = w = MenuButtonNewPixmap(
		(guint8 **)icon_yplayer_save_file_menu_14x14_xpm,
		&menu
	    );
	    MenuButtonSetMapTrigger(w, MENU_BUTTON_MAP_TYPE_PRESSED_DRAG);
	    gtk_widget_set_usize(
		w, YPLAYER_BTN_HEIGHT, YPLAYER_BTN_HEIGHT
	    );
	    gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	    gtk_signal_connect(
		GTK_OBJECT(w), "clicked",
		GTK_SIGNAL_FUNC(YPlayerSndObjSavePlayListCB), yp
	    );
	    GUISetWidgetTip(w, "Save playlist");
	    gtk_widget_show(w);
	    if(menu != NULL)
	    {
		guchar **icon;
		const gchar *label;
		gint accel_key;
		guint accel_mods;
		gpointer mclient_data = yp;
		void (*func_cb)(GtkWidget *w, gpointer);
		gpointer fw;

#define DO_ADD_MENU_ITEM_LABEL  \
{ w = GUIMenuItemCreate( \
  menu, GUI_MENU_ITEM_TYPE_LABEL, accelgrp, \
  icon, label, accel_key, accel_mods, (void **)&fw, \
  mclient_data, func_cb \
 ); }
#define DO_ADD_MENU_SEP         \
{ w = GUIMenuItemCreate( \
  menu, GUI_MENU_ITEM_TYPE_SEPARATOR, NULL, \
  NULL, NULL, 0, 0, NULL, \
  NULL, NULL \
 ); }
		icon = (guchar **)icon_yplayer_save_file_20x20_xpm;
		label = "Save PlayList...";
		accel_key = 0;
		accel_mods = 0;
		func_cb = YPlayerSndObjSavePlayListCB;
		DO_ADD_MENU_ITEM_LABEL

		icon = (guchar **)icon_yplayer_save_file_20x20_xpm;
		label = "Save Selected...";
		accel_key = 0;
		accel_mods = 0;
		func_cb = YPlayerSndObjSaveSelectedCB;
		DO_ADD_MENU_ITEM_LABEL

#undef DO_ADD_MENU_SEP
#undef DO_ADD_MENU_ITEM_LABEL
	    }


	    w = gtk_hbox_new(TRUE, 0);
	    gtk_box_pack_start(GTK_BOX(parent3), w, TRUE, TRUE, 0);
	    gtk_widget_show(w);
	    parent4 = w;
	    w = gtk_hbox_new(FALSE, 0);
	    gtk_box_pack_start(GTK_BOX(parent4), w, TRUE, FALSE, 0);
	    gtk_widget_show(w);
	    parent4 = w;

	    /* Play button */
	    yp->sndobj_play_btn = w = GUIButtonPixmap(
		(guint8 **)icon_yplayer_play_14x14_xpm
	    );
	    gtk_widget_set_usize(
		w, YPLAYER_BTN_WIDTH, YPLAYER_BTN_HEIGHT
	    );
	    gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	    gtk_signal_connect(
		GTK_OBJECT(w), "clicked",
		GTK_SIGNAL_FUNC(YPlayerSndObjPlayCB), yp
	    );
	    GUISetWidgetTip(w, "Play");
	    gtk_widget_show(w);

	    /* Stop button */
	    yp->sndobj_playing_stop_btn = w = GUIButtonPixmap(
		(guint8 **)icon_yplayer_stop_14x14_xpm
	    );
	    gtk_widget_set_usize(
		w, YPLAYER_BTN_WIDTH, YPLAYER_BTN_HEIGHT
	    );
	    gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	    gtk_signal_connect(
		GTK_OBJECT(w), "clicked",
		GTK_SIGNAL_FUNC(YPlayerSndObjStopCB), yp
	    );
	    GUISetWidgetTip(w, "Stop");
	    gtk_widget_show(w);

	    /* Add button */
	    yp->sndobj_add_btn = w = GUIButtonPixmap(
		(guint8 **)icon_yplayer_file_14x14_xpm
	    );
	    gtk_widget_set_usize(
		w, YPLAYER_BTN_WIDTH, YPLAYER_BTN_HEIGHT
	    );
	    gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	    gtk_signal_connect(
		GTK_OBJECT(w), "clicked",
		GTK_SIGNAL_FUNC(YPlayerSndObjAddCB), yp
	    );
	    GUISetWidgetTip(w, "Add...");
	    gtk_widget_show(w);

	    /* Remove button */
	    yp->sndobj_remove_btn = w = GUIButtonPixmap(
		(guint8 **)icon_yplayer_remove_14x14_xpm
	    );
	    gtk_widget_set_usize(
		w, YPLAYER_BTN_WIDTH, YPLAYER_BTN_HEIGHT
	    );
	    gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	    gtk_signal_connect(
		GTK_OBJECT(w), "clicked",
		GTK_SIGNAL_FUNC(YPlayerSndObjRemoveCB), yp
	    );
	    GUISetWidgetTip(w, "Remove");
	    gtk_widget_show(w);

	    /* Shift up button */
	    yp->sndobj_shift_up_btn = w = GUIButtonPixmap(
		(guint8 **)icon_yplayer_up_14x14_xpm
	    );
	    gtk_widget_set_usize(
		w, YPLAYER_BTN_WIDTH, YPLAYER_BTN_HEIGHT
	    );
	    gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	    gtk_signal_connect(
		GTK_OBJECT(w), "clicked",
		GTK_SIGNAL_FUNC(YPlayerSndObjShiftUpCB), yp
	    );
	    GUISetWidgetTip(w, "Shift Up");
	    gtk_widget_show(w);

	    /* Shift down button */
	    yp->sndobj_shift_down_btn = w = GUIButtonPixmap(
		(guint8 **)icon_yplayer_down_14x14_xpm
	    );
	    gtk_widget_set_usize(
		w, YPLAYER_BTN_WIDTH, YPLAYER_BTN_HEIGHT
	    );
	    gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	    gtk_signal_connect(
		GTK_OBJECT(w), "clicked",
		GTK_SIGNAL_FUNC(YPlayerSndObjShiftDownCB), yp
	    );
	    GUISetWidgetTip(w, "Shift Down");
	    gtk_widget_show(w);


	    w = gtk_hbox_new(FALSE, 0);
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent4 = w;

	    yp->sndobj_change_mode_tb = w = GUIToggleButtonPixmap(
		(guint8 **)icon_yplayer_mode_14x14_xpm
	    );
	    gtk_widget_set_usize(
		w, YPLAYER_BTN_HEIGHT, YPLAYER_BTN_HEIGHT
	    );
	    gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	    gtk_signal_connect(
		GTK_OBJECT(w), "toggled",
		GTK_SIGNAL_FUNC(YPlayerSndObjChangeModeToggleCB), yp
	    );
	    gtk_toggle_button_set_active(
		GTK_TOGGLE_BUTTON(w),
		YC_GET_CFG_I(cp, YC_CFG_PARM_PLAYER_SNDOBJ_CHANGE_MODE)
	    );
	    GUISetWidgetTip(w, "Change audio mode on play");
	    gtk_widget_show(w);

	    yp->sndobj_repeat_tb = w = GUIToggleButtonPixmap(
		(guint8 **)icon_yplayer_repeat_14x14_xpm
	    );
	    gtk_widget_set_usize(
		w, YPLAYER_BTN_HEIGHT, YPLAYER_BTN_HEIGHT
	    );
	    gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	    gtk_signal_connect(
		GTK_OBJECT(w), "toggled",
		GTK_SIGNAL_FUNC(YPlayerSndObjRepeatToggleCB), yp
	    );
	    gtk_toggle_button_set_active(
		GTK_TOGGLE_BUTTON(w),
		YC_GET_CFG_I(cp, YC_CFG_PARM_PLAYER_SNDOBJ_REPEAT)
	    );
	    GUISetWidgetTip(w, "Repeat");
	    gtk_widget_show(w);


	    /* Create sound object advanced page */
	    w = GUICreateMenuItemIcon((guint8 **)icon_sound_20x20_xpm);
	    gtk_widget_show(w);
	    parent2 = gtk_vbox_new(FALSE, border_major);
	    gtk_container_border_width(GTK_CONTAINER(parent2), border_major);
	    gtk_notebook_append_page(notebook, parent2, w);
	    gtk_widget_show(parent2);

	    w = gtk_hbox_new(FALSE, border_major);
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent3 = w;

	    /* GtkFrame for display GtkDrawingArea */
	    w = gtk_frame_new(NULL);
	    gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_IN);
	    gtk_box_pack_start(GTK_BOX(parent3), w, TRUE, TRUE, 0);
	    gtk_widget_show(w);
	    parent4 = w;
	    /* Display GtkDrawingArea */
	    yp->sndobjadv_display_da = w = gtk_drawing_area_new();
	    gtk_widget_set_usize(w, -1, 70);
	    gtk_widget_add_events(
		w,
		GDK_STRUCTURE_MASK | GDK_EXPOSURE_MASK |
		GDK_KEY_PRESS_MASK | GDK_KEY_RELEASE_MASK |
		GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
		GDK_POINTER_MOTION_MASK | GDK_POINTER_MOTION_HINT_MASK |
		GDK_ENTER_NOTIFY_MASK | GDK_LEAVE_NOTIFY_MASK |
		GDK_FOCUS_CHANGE_MASK
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "configure_event",
		GTK_SIGNAL_FUNC(YPlayerSndObjAdvEventCB), yp
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "expose_event",
		GTK_SIGNAL_FUNC(YPlayerSndObjAdvEventCB), yp
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "focus_in_event",
		GTK_SIGNAL_FUNC(YPlayerSndObjAdvEventCB), yp
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "focus_out_event",
		GTK_SIGNAL_FUNC(YPlayerSndObjAdvEventCB), yp
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "key_press_event",
		GTK_SIGNAL_FUNC(YPlayerSndObjAdvEventCB), yp
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "key_release_event",
		GTK_SIGNAL_FUNC(YPlayerSndObjAdvEventCB), yp
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "button_press_event",
		GTK_SIGNAL_FUNC(YPlayerSndObjAdvEventCB), yp
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "button_release_event",
		GTK_SIGNAL_FUNC(YPlayerSndObjAdvEventCB), yp
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "motion_notify_event",
		GTK_SIGNAL_FUNC(YPlayerSndObjAdvEventCB), yp
	    );
	    gtk_container_add(GTK_CONTAINER(parent4), w);
	    GTK_WIDGET_SET_FLAGS(w, GTK_CAN_FOCUS | GTK_CAN_DEFAULT);
	    if(TRUE)
	    {
		const GtkTargetEntry dnd_tar_types[] = {
{"text/plain",          0,      YC_DND_TYPE_INFO_TEXT_PLAIN},
{"text/uri-list",       0,      YC_DND_TYPE_INFO_TEXT_URI_LIST},
{"STRING",              0,      YC_DND_TYPE_INFO_STRING}
		};
		GUIDNDSetTar(
		    w,
		    dnd_tar_types,
		    sizeof(dnd_tar_types) / sizeof(GtkTargetEntry),
		    GDK_ACTION_COPY,            /* Actions */
		    GDK_ACTION_COPY,            /* Default action if same */
		    GDK_ACTION_COPY,            /* Default action */
		    YPlayerSndObjListDragDataReceivedCB,
		    yp
		);
		gtk_signal_connect_after(
		    GTK_OBJECT(w), "drag_motion",
		    GTK_SIGNAL_FUNC(YPlayerSndObjListDragMotionCB), yp
		);
	    }
	    gtk_widget_show(w);
	    yp->sndobjadv_display_pm = NULL;
	    /* Load pixmap and mask pairs for display drawing area */
	    style = gtk_widget_get_style(w);
	    yp->pixmap_sndobjadv_playing = gdk_pixmap_create_from_xpm_d(
		window, &yp->mask_sndobjadv_playing,
		&style->bg[GTK_STATE_NORMAL],
		(gchar **)icon_yplayer_play_14x14_xpm
	    );
	    yp->pixmap_sndobjadv_repeating = gdk_pixmap_create_from_xpm_d(
		window, &yp->mask_sndobjadv_repeating,
		&style->bg[GTK_STATE_NORMAL],
		(gchar **)icon_yplayer_repeat_14x14_xpm
	    );

	    /* Display area menu */
	    yp->sndobjadv_display_menu = menu = GUIMenuCreate();
	    if(menu != NULL)
	    {
		guchar **icon;
		const gchar *label;
		gint accel_key;
		guint accel_mods;
		gpointer mclient_data = yp;
		void (*func_cb)(GtkWidget *w, gpointer);
		gpointer fw;

#define DO_ADD_MENU_ITEM_LABEL  \
{ w = GUIMenuItemCreate( \
  menu, GUI_MENU_ITEM_TYPE_LABEL, accelgrp, \
  icon, label, accel_key, accel_mods, (void **)&fw, \
  mclient_data, func_cb \
 ); }
#define DO_ADD_MENU_SEP         \
{ w = GUIMenuItemCreate( \
  menu, GUI_MENU_ITEM_TYPE_SEPARATOR, NULL, \
  NULL, NULL, 0, 0, NULL, \
  NULL, NULL \
 ); }
		icon = (guchar **)icon_yplayer_play_20x20_xpm;
		label = "Play";
		accel_key = 0;
		accel_mods = 0;
		func_cb = YPlayerSndObjAdvPlayCB;
		DO_ADD_MENU_ITEM_LABEL
		yp->sndobjadv_play_mi = w;

		icon = (guchar **)icon_yplayer_pause_20x20_xpm;
		label = "Pause";
		accel_key = 0;
		accel_mods = 0;
		func_cb = YPlayerSndObjAdvPauseCB;
		DO_ADD_MENU_ITEM_LABEL
		yp->sndobjadv_pause_mi = w;

		icon = (guchar **)icon_yplayer_rewind_20x20_xpm;
		label = "Rewind";
		accel_key = 0;
		accel_mods = 0;
		func_cb = YPlayerSndObjAdvRewindCB;
		DO_ADD_MENU_ITEM_LABEL
		yp->sndobjadv_rewind_mi = w;

		DO_ADD_MENU_SEP

		icon = (guchar **)icon_yplayer_file_20x20_xpm;
		label = "Open...";
		accel_key = 0;
		accel_mods = 0;
		func_cb = YPlayerSndObjAdvOpenCB;
		DO_ADD_MENU_ITEM_LABEL
		yp->sndobjadv_open_mi = w;

		icon = (guchar **)icon_yplayer_remove_20x20_xpm;
		label = "Close";
		accel_key = 0;
		accel_mods = 0;
		func_cb = YPlayerSndObjAdvCloseCB;
		DO_ADD_MENU_ITEM_LABEL
		yp->sndobjadv_close_mi = w;

		DO_ADD_MENU_SEP
		icon = (guchar **)icon_yplayer_mode_20x20_xpm;
		label = "Change Audio Mode";
		accel_key = 0;
		accel_mods = 0;
		func_cb = YPlayerSndObjAdvChangeModeCB;
		DO_ADD_MENU_ITEM_LABEL
		yp->sndobjadv_change_mode_mi = w;
	    }

	    /* GtkFrame for volume GtkDrawingArea */
	    w = gtk_frame_new(NULL);
	    gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_IN);
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent4 = w;
	    /* Volume GtkDrawingArea */
	    yp->sndobjadv_volume_da = w = gtk_drawing_area_new();
	    gtk_widget_set_usize(w, 70, 70);
	    gtk_widget_add_events(
		w,
		GDK_STRUCTURE_MASK | GDK_EXPOSURE_MASK |
		GDK_KEY_PRESS_MASK | GDK_KEY_RELEASE_MASK |
		GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
		GDK_POINTER_MOTION_MASK | GDK_POINTER_MOTION_HINT_MASK |
		GDK_ENTER_NOTIFY_MASK | GDK_LEAVE_NOTIFY_MASK |
		GDK_FOCUS_CHANGE_MASK
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "configure_event",
		GTK_SIGNAL_FUNC(YPlayerSndObjAdvEventCB), yp
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "expose_event",
		GTK_SIGNAL_FUNC(YPlayerSndObjAdvEventCB), yp
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "focus_in_event",
		GTK_SIGNAL_FUNC(YPlayerSndObjAdvEventCB), yp
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "focus_out_event",
		GTK_SIGNAL_FUNC(YPlayerSndObjAdvEventCB), yp
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "key_press_event",
		GTK_SIGNAL_FUNC(YPlayerSndObjAdvEventCB), yp
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "key_release_event",
		GTK_SIGNAL_FUNC(YPlayerSndObjAdvEventCB), yp
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "button_press_event",
		GTK_SIGNAL_FUNC(YPlayerSndObjAdvEventCB), yp
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "button_release_event",
		GTK_SIGNAL_FUNC(YPlayerSndObjAdvEventCB), yp
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "motion_notify_event",
		GTK_SIGNAL_FUNC(YPlayerSndObjAdvEventCB), yp
	    );
	    gtk_container_add(GTK_CONTAINER(parent4), w);
	    GTK_WIDGET_SET_FLAGS(w, GTK_CAN_FOCUS | GTK_CAN_DEFAULT);
	    GUISetWidgetTip(w, "Volume Position");
	    gtk_widget_show(w);
	    yp->sndobjadv_volume_pm = NULL;
	    /* Load pixmap and mask pairs for volume drawing area */
	    style = gtk_widget_get_style(w);
	    yp->pixmap_sndobjadv_source = gdk_pixmap_create_from_xpm_d(
		window, &yp->mask_sndobjadv_source,
		&style->bg[GTK_STATE_NORMAL],
		(gchar **)icon_yplayer_sound_source_14x14_xpm
	    );
	    yp->pixmap_sndobjadv_ear = gdk_pixmap_create_from_xpm_d(
		window, &yp->mask_sndobjadv_ear,
		&style->bg[GTK_STATE_NORMAL],
		(gchar **)icon_yplayer_ear_14x14_xpm
	    );

	    /* Vbox for volume scale */
	    w = gtk_vbox_new(FALSE, border_minor);
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent4 = w;
	    /* Volume scale */
	    adj = (GtkAdjustment *)gtk_adjustment_new(
		1.0f - YC_GET_CFG_F(
		    cp, YC_CFG_PARM_PLAYER_SNDOBJADV_VOLUME_A
		),
		0.0f, 1.0f, 0.01f, 0.1f, 0.0f
	    );
	    yp->sndobjadv_volume_scale = w = gtk_vscale_new(adj);
	    gtk_scale_set_draw_value(GTK_SCALE(w), FALSE);
	    gtk_box_pack_start(GTK_BOX(parent4), w, TRUE, TRUE, 0);
	    gtk_signal_connect(
		GTK_OBJECT(adj), "value_changed",
		GTK_SIGNAL_FUNC(YPlayerSndObjAdvVolumeChangedCB), yp
	    );
	    GUISetWidgetTip(w, "Volume Amplitude");
	    gtk_widget_show(w);

	    /* Hbox for position, repeat, and applied sample rate */
	    w = gtk_hbox_new(FALSE, border_major);
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent3 = w;
	    /* Position scale */
	    w = gtk_vbox_new(TRUE, 0);
	    gtk_box_pack_start(GTK_BOX(parent3), w, TRUE, TRUE, 0);
	    gtk_widget_show(w);
	    parent4 = w;
	    yp->sndobjadv_position_scale = w = gtk_hscale_new(NULL);
	    gtk_scale_set_draw_value(GTK_SCALE(w), FALSE);
	    gtk_box_pack_start(GTK_BOX(parent4), w, TRUE, FALSE, 0);
	    gtk_signal_connect(
		GTK_OBJECT(GTK_RANGE(w)->adjustment), "value_changed",
		GTK_SIGNAL_FUNC(YPlayerSndObjAdvPositionChangedCB), yp
	    );
	    GUISetWidgetTip(w, "Position");
	    gtk_widget_show(w);
	    /* Repeat toggle button */
	    w = gtk_vbox_new(TRUE, 0);
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent4 = w;
	    yp->sndobjadv_repeat_tb = w = GUIToggleButtonPixmap(
		(guint8 **)icon_yplayer_repeat_14x14_xpm
	    );
	    gtk_widget_set_usize(
		w, YPLAYER_BTN_HEIGHT, YPLAYER_BTN_HEIGHT
	    );
	    gtk_toggle_button_set_active(
		GTK_TOGGLE_BUTTON(w),
		YC_GET_CFG_I(cp, YC_CFG_PARM_PLAYER_SNDOBJADV_REPEAT)
	    );
	    gtk_box_pack_start(GTK_BOX(parent4), w, TRUE, FALSE, 0);
	    GUISetWidgetTip(w, "Toggle repeating/not repeating");
	    gtk_widget_show(w);
	    /* Applied sample rate spin button */
	    w = gtk_hbox_new(FALSE, border_minor);
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent4 = w;
	    yp->sndobjadv_sample_rate_spin = w = gtk_spin_button_new(
		NULL, 1.0, 0
	    );
	    gtk_widget_set_usize(w, 80, -1);
	    gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	    gtk_signal_connect(
		GTK_OBJECT(GTK_SPIN_BUTTON(w)->adjustment), "value_changed",
		GTK_SIGNAL_FUNC(YPlayerSndObjAdvSampleRateChangedCB), yp
	    );
	    GUISetWidgetTip(w, "Applied sample rate");
	    gtk_widget_show(w);

	    /* Hbox for buttons */
	    w = gtk_hbox_new(TRUE, 0);
	    gtk_box_pack_end(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent3 = w;
	    /* Hbox for buttons */
	    w = gtk_hbox_new(FALSE, 0);
	    gtk_box_pack_start(GTK_BOX(parent3), w, TRUE, FALSE, 0);
	    gtk_widget_show(w);
	    parent3 = w;

	    /* Open button */
	    yp->sndobjadv_open_btn = w = GUIButtonPixmap(
		(guint8 **)icon_yplayer_file_14x14_xpm
	    );
	    gtk_widget_set_usize(
		w, YPLAYER_BTN_WIDTH, YPLAYER_BTN_HEIGHT
	    );
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_signal_connect(
		GTK_OBJECT(w), "clicked",
		GTK_SIGNAL_FUNC(YPlayerSndObjAdvOpenCB), yp
	    );
	    GUISetWidgetTip(w, "Open Sound Object");
	    gtk_widget_show(w);

	    /* Change Audio mode button */
	    yp->sndobjadv_change_mode_btn = w = GUIButtonPixmap(
		(guint8 **)icon_yplayer_mode_14x14_xpm
	    );
	    gtk_widget_set_usize(
		w, YPLAYER_BTN_WIDTH, YPLAYER_BTN_HEIGHT
	    );
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_signal_connect(
		GTK_OBJECT(w), "clicked",
		GTK_SIGNAL_FUNC(YPlayerSndObjAdvChangeModeCB), yp
	    );
	    GUISetWidgetTip(w, "Change Audio Mode");
	    gtk_widget_show(w);

	    /* Play button */
	    yp->sndobjadv_play_btn = w = GUIButtonPixmap(
		(guint8 **)icon_yplayer_play_14x14_xpm
	    );
	    gtk_widget_set_usize(
		w, YPLAYER_BTN_WIDTH, YPLAYER_BTN_HEIGHT
	    );
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_signal_connect(
		GTK_OBJECT(w), "clicked",
		GTK_SIGNAL_FUNC(YPlayerSndObjAdvPlayCB), yp
	    );
	    GUISetWidgetTip(w, "Play");
	    gtk_widget_show(w);

	    /* Pause button */
	    yp->sndobjadv_pause_btn = w = GUIButtonPixmap(
		(guint8 **)icon_yplayer_pause_14x14_xpm
	    );
	    gtk_widget_set_usize(
		w, YPLAYER_BTN_WIDTH, YPLAYER_BTN_HEIGHT
	    );
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_signal_connect(
		GTK_OBJECT(w), "clicked",
		GTK_SIGNAL_FUNC(YPlayerSndObjAdvPauseCB), yp
	    );
	    GUISetWidgetTip(w, "Pause");
	    gtk_widget_show(w);

	    /* Rewind button */
	    yp->sndobjadv_rewind_btn = w = GUIButtonPixmap(
		(guint8 **)icon_yplayer_rewind_14x14_xpm
	    );
	    gtk_widget_set_usize(
		w, YPLAYER_BTN_WIDTH, YPLAYER_BTN_HEIGHT
	    );
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_signal_connect(
		GTK_OBJECT(w), "clicked",
		GTK_SIGNAL_FUNC(YPlayerSndObjAdvRewindCB), yp
	    );
	    GUISetWidgetTip(w, "Rewind");
	    gtk_widget_show(w);

	    /* Close button */
	    yp->sndobjadv_close_btn = w = GUIButtonPixmap(
		(guint8 **)icon_yplayer_remove_14x14_xpm
	    );
	    gtk_widget_set_usize(
		w, YPLAYER_BTN_WIDTH, YPLAYER_BTN_HEIGHT
	    );
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_signal_connect(
		GTK_OBJECT(w), "clicked",
		GTK_SIGNAL_FUNC(YPlayerSndObjAdvCloseCB), yp
	    );
	    GUISetWidgetTip(w, "Close");
	    gtk_widget_show(w);

/* TODO */


	    /* Create audio CD page */
	    w = GUICreateMenuItemIcon((guint8 **)icon_mixer_cd_20x20_xpm);
	    gtk_widget_show(w);
	    parent2 = gtk_vbox_new(FALSE, border_major);
	    gtk_container_border_width(GTK_CONTAINER(parent2), border_major);
	    gtk_notebook_append_page(notebook, parent2, w);
	    gtk_widget_show(parent2);

	    /* Scrolled window for clist */
	    w = gtk_scrolled_window_new(NULL, NULL);
	    gtk_scrolled_window_set_policy(
		GTK_SCROLLED_WINDOW(w),
		GTK_POLICY_AUTOMATIC,
		GTK_POLICY_AUTOMATIC
	    );
	    gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, TRUE, 0);
	    gtk_widget_show(w);
	    parent3 = w;

	    /* Audio CD tracks clist */
	    heading[0] = "Track";
	    heading[1] = "Start";
	    heading[2] = "Length";
	    heading[3] = "Name";
	    yp->audiocd_tracks_clist = w = gtk_clist_new_with_titles(4, heading);
	    clist = GTK_CLIST(w);
	    gtk_clist_column_titles_passive(clist);
	    intlist = YC_GET_CFG_INTLIST(
		cp, YC_CFG_PARM_PLAYER_AUDIOCD_TRACKS_LIST_COLUMN_WIDTHS
	    );
	    if(intlist != NULL)
	    {
		GList *glist = intlist->list;
		for(i = 0; i < clist->columns; i++)
		{
		    if(glist == NULL)
			break;

		    gtk_clist_set_column_width(
			clist, i, (gint)glist->data
		    );
		    glist = g_list_next(glist);
		}
	    }
	    gtk_clist_set_row_height(clist, YC_LIST_ROW_SPACING);
	    gtk_widget_add_events(
		w,
		GDK_KEY_PRESS_MASK | GDK_KEY_RELEASE_MASK |
		GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "key_press_event",
		GTK_SIGNAL_FUNC(YPlayerKeyEventCB), yp
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "key_release_event",
		GTK_SIGNAL_FUNC(YPlayerKeyEventCB), yp
	    );
	    gtk_signal_connect_after(
		GTK_OBJECT(w), "button_press_event",
		GTK_SIGNAL_FUNC(YPlayerButtonEventCB), yp
	    );
	    gtk_signal_connect_after(
		GTK_OBJECT(w), "button_release_event",
		GTK_SIGNAL_FUNC(YPlayerButtonEventCB), yp
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "select_row",
		GTK_SIGNAL_FUNC(YPlayerCListSelectRowCB), yp
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "unselect_row",
		GTK_SIGNAL_FUNC(YPlayerCListUnselectRowCB), yp
	    );
	    gtk_container_add(GTK_CONTAINER(parent3), w);
	    gtk_widget_show(w);

	    /* Right click menu */
	    yp->audiocd_tracks_menu = menu = GUIMenuCreate();
	    if(menu != NULL)
	    {
		guchar **icon;
		const gchar *label;
		gint accel_key;
		guint accel_mods;
		gpointer mclient_data = yp;
		void (*func_cb)(GtkWidget *w, gpointer);
		gpointer fw;

#define DO_ADD_MENU_ITEM_LABEL	\
{ \
 w = GUIMenuItemCreate( \
  menu, GUI_MENU_ITEM_TYPE_LABEL, accelgrp, \
  icon, label, accel_key, accel_mods, (void **)&fw, \
  mclient_data, func_cb \
 ); \
}
#define DO_ADD_MENU_SEP		\
{ \
 w = GUIMenuItemCreate( \
  menu, GUI_MENU_ITEM_TYPE_SEPARATOR, NULL, \
  NULL, NULL, 0, 0, NULL, \
  NULL, NULL \
 ); \
}
		icon = (guchar **)icon_yplayer_play_20x20_xpm;
		label = "Play";
		accel_key = 0;
		accel_mods = 0;
		func_cb = YPlayerAudioCDPlayCB;
		DO_ADD_MENU_ITEM_LABEL
		yp->audiocd_play_mi = w;

		icon = (guchar **)icon_yplayer_random_20x20_xpm;
		label = "Play Random";
		accel_key = 0;
		accel_mods = 0;
		func_cb = YPlayerAudioCDRandomCB;
		DO_ADD_MENU_ITEM_LABEL
		yp->audiocd_random_mi = w;

		icon = (guchar **)icon_yplayer_stop_20x20_xpm;
		label = "Stop";
		accel_key = 0;
		accel_mods = 0;
		func_cb = YPlayerAudioCDStopCB;
		DO_ADD_MENU_ITEM_LABEL
		yp->audiocd_stop_mi = w;

		icon = (guchar **)icon_yplayer_eject_20x20_xpm;
		label = "Eject";
		accel_key = 0;
		accel_mods = 0;
		func_cb = YPlayerAudioCDEjectCB;
		DO_ADD_MENU_ITEM_LABEL
		yp->audiocd_eject_mi = w;

		icon = (guchar **)icon_yplayer_refresh_20x20_xpm;
		label = "Refresh";
		accel_key = 0;
		accel_mods = 0;
		func_cb = YPlayerAudioCDRefreshCB;
		DO_ADD_MENU_ITEM_LABEL
		yp->audiocd_refresh_mi = w;

#undef DO_ADD_MENU_SEP
#undef DO_ADD_MENU_ITEM_LABEL
	    }

	    /* Hbox for audio CD buttons */
	    w = gtk_hbox_new(TRUE, 0);
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent3 = w;
	    w = gtk_hbox_new(FALSE, 0);
	    gtk_box_pack_start(GTK_BOX(parent3), w, TRUE, FALSE, 0);
	    gtk_widget_show(w);
	    parent3 = w;

	    /* Play menu button */
	    yp->audiocd_play_mbtn = w = MenuButtonNewPixmap(
		(guint8 **)icon_yplayer_play_14x14_xpm,
		&menu
	    );
	    MenuButtonSetMapTrigger(w, MENU_BUTTON_MAP_TYPE_PRESSED_DRAG);
	    gtk_widget_set_usize(
		w, YPLAYER_BTN_WIDTH, YPLAYER_BTN_HEIGHT
	    );
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_signal_connect(
		GTK_OBJECT(w), "clicked",
		GTK_SIGNAL_FUNC(YPlayerAudioCDPlayCB), yp
	    );
	    GUISetWidgetTip(w, "Play");
	    gtk_widget_show(w);
	    if(menu != NULL)
	    {
		guchar **icon;
		const gchar *label;
		gint accel_key;
		guint accel_mods;
		gpointer mclient_data = yp;
		void (*func_cb)(GtkWidget *w, gpointer);
		gpointer fw;

#define DO_ADD_MENU_ITEM_LABEL  \
{ w = GUIMenuItemCreate( \
  menu, GUI_MENU_ITEM_TYPE_LABEL, accelgrp, \
  icon, label, accel_key, accel_mods, (void **)&fw, \
  mclient_data, func_cb \
 ); }
#define DO_ADD_MENU_SEP         \
{ w = GUIMenuItemCreate( \
  menu, GUI_MENU_ITEM_TYPE_SEPARATOR, NULL, \
  NULL, NULL, 0, 0, NULL, \
  NULL, NULL \
 ); }
		icon = (guchar **)icon_yplayer_play_20x20_xpm;
		label = "Play";
		accel_key = 0;
		accel_mods = 0;
		func_cb = YPlayerAudioCDPlayCB;
		DO_ADD_MENU_ITEM_LABEL

		icon = (guchar **)icon_yplayer_random_20x20_xpm;
		label = "Random";
		accel_key = 0;
		accel_mods = 0;
		func_cb = YPlayerAudioCDRandomCB;
		DO_ADD_MENU_ITEM_LABEL

#undef DO_ADD_MENU_SEP
#undef DO_ADD_MENU_ITEM_LABEL
	    }

	    /* Stop button */
	    yp->audiocd_stop_btn = w = GUIButtonPixmap(
		(guint8 **)icon_yplayer_stop_14x14_xpm
	    );
	    gtk_widget_set_usize(
		w, YPLAYER_BTN_WIDTH, YPLAYER_BTN_HEIGHT
	    );
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_signal_connect(
		GTK_OBJECT(w), "clicked",
		GTK_SIGNAL_FUNC(YPlayerAudioCDStopCB), yp
	    );
	    GUISetWidgetTip(w, "Stop");
	    gtk_widget_show(w);

	    /* Eject button */
	    yp->audiocd_eject_btn = w = GUIButtonPixmap(
		(guint8 **)icon_yplayer_eject_14x14_xpm
	    );
	    gtk_widget_set_usize(
		w, YPLAYER_BTN_WIDTH, YPLAYER_BTN_HEIGHT
	    );
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_signal_connect(
		GTK_OBJECT(w), "clicked",
		GTK_SIGNAL_FUNC(YPlayerAudioCDEjectCB), yp
	    );
	    GUISetWidgetTip(w, "Eject");
	    gtk_widget_show(w);

	    /* Refresh button */
	    yp->audiocd_refresh_btn = w = GUIButtonPixmap(
		(guint8 **)icon_yplayer_refresh_14x14_xpm
	    );
	    gtk_widget_set_usize(
		w, YPLAYER_BTN_WIDTH, YPLAYER_BTN_HEIGHT
	    );
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_signal_connect(
		GTK_OBJECT(w), "clicked",
		GTK_SIGNAL_FUNC(YPlayerAudioCDRefreshCB), yp
	    );
	    GUISetWidgetTip(w, "Refresh");
	    gtk_widget_show(w);


	    /* Select notebook page */
	    gtk_notebook_set_page(
		notebook, YC_GET_CFG_I(cp, YC_CFG_PARM_PLAYER_PAGE)
	    );
	}


	/* Update title and menus */
	YPlayerSetTitle(yp, NULL);
	YPlayerUpdateMenus(yp);

	return(yp);
}

/*
 *	Update menus on yconsole.
 */
void YPlayerUpdateMenus(yplayer_struct *yp)
{
	gboolean sensitive;
	gint row;
	GList *glist;
	GtkWidget *w;
	GtkCList *clist;
	YConnection *con;
	const YEventSoundObjectAttributes *soa;

#define DO_SET_SENSITIVE	\
{ if(w != NULL) gtk_widget_set_sensitive(w, sensitive); }


	if(yp == NULL)
	    return;

	con = YPLAYER_CONNECTION(yp);
	soa = &yp->sndobjadv_soa;

	clist = (GtkCList *)YPLAYER_SNDOBJ_CLIST(yp);
	if(clist != NULL)
	{
	    glist = clist->selection_end;
	    row = (glist != NULL) ? (gint)glist->data : -1;
	    sensitive = (row >= 0) ? TRUE : FALSE;
	    if(sensitive && (con == NULL))
		sensitive = FALSE;

	    w = yp->sndobj_play_mi;
	    DO_SET_SENSITIVE
	    w = yp->sndobj_remove_mi;
	    DO_SET_SENSITIVE
	    w = yp->sndobj_play_btn;
	    DO_SET_SENSITIVE
	    w = yp->sndobj_remove_btn;
	    DO_SET_SENSITIVE

	    sensitive = (con != NULL) ? TRUE : FALSE;
	    w = GTK_WIDGET(clist);
	    DO_SET_SENSITIVE
	    w = yp->sndobj_add_mi;
	    DO_SET_SENSITIVE
	    w = yp->sndobj_add_btn;
	    DO_SET_SENSITIVE

	    sensitive = (row > 0) ? TRUE : FALSE;
	    w = yp->sndobj_shift_up_btn;
	    DO_SET_SENSITIVE
	    w = yp->sndobj_shift_up_mi;
	    DO_SET_SENSITIVE

	    sensitive = ((row >= 0) && (row < (clist->rows - 1))) ? TRUE : FALSE;
	    w = yp->sndobj_shift_down_btn;
	    DO_SET_SENSITIVE
	    w = yp->sndobj_shift_down_mi;
	    DO_SET_SENSITIVE

	    sensitive = (row >= 0) ? TRUE : FALSE;
	    w = yp->sndobj_save_selected_mi;
	    DO_SET_SENSITIVE
	}

	clist = (GtkCList *)YPLAYER_SNDOBJ_PLAYING_CLIST(yp);
	if(clist != NULL)
	{
	    glist = clist->selection_end;
	    row = (glist != NULL) ? (gint)glist->data : -1;
	    sensitive = (row >= 0) ? TRUE : FALSE;
	    if(sensitive && (con == NULL))
		sensitive = FALSE;

	    w = yp->sndobj_playing_stop_mi;
	    DO_SET_SENSITIVE
	    w = yp->sndobj_playing_stop_btn;
	    DO_SET_SENSITIVE

	    sensitive = (con != NULL) ? TRUE : FALSE;
	    w = GTK_WIDGET(clist);
	    DO_SET_SENSITIVE
	}

	sensitive = (con != NULL) ? TRUE : FALSE;
	w = yp->sndobjadv_open_mi;
	DO_SET_SENSITIVE
	w = yp->sndobjadv_open_btn;
	DO_SET_SENSITIVE
	sensitive = ((con != NULL) && YPLAYER_SNDOBJADV_IS_OPENED(yp) &&
	    !YPLAYER_SNDOBJADV_IS_PLAYING(yp)) ? TRUE : FALSE;
	w = yp->sndobjadv_play_mi;
	DO_SET_SENSITIVE
	w = yp->sndobjadv_play_btn;
	DO_SET_SENSITIVE
	sensitive = YPLAYER_SNDOBJADV_IS_OPENED(yp);
	w = yp->sndobjadv_rewind_mi;
	DO_SET_SENSITIVE
	w = yp->sndobjadv_rewind_btn;
	DO_SET_SENSITIVE
	w = yp->sndobjadv_close_mi;
	DO_SET_SENSITIVE
	w = yp->sndobjadv_close_btn;
	DO_SET_SENSITIVE
	sensitive = (YPLAYER_SNDOBJADV_IS_OPENED(yp) &&
	    (soa->format == SndObjTypeDSP)) ? TRUE : FALSE;
	w = yp->sndobjadv_change_mode_mi;
	DO_SET_SENSITIVE
	w = yp->sndobjadv_change_mode_btn;
	DO_SET_SENSITIVE
	w = yp->sndobjadv_position_scale;
	DO_SET_SENSITIVE
	w = yp->sndobjadv_sample_rate_spin;
	DO_SET_SENSITIVE
	sensitive = YPLAYER_SNDOBJADV_IS_PLAYING(yp);
	w = yp->sndobjadv_pause_mi;
	DO_SET_SENSITIVE
	w = yp->sndobjadv_pause_btn;
	DO_SET_SENSITIVE
	sensitive = !sensitive;
	w = yp->sndobjadv_repeat_tb;
	DO_SET_SENSITIVE

	clist = (GtkCList *)YPLAYER_AUDIOCD_TRACKS_CLIST(yp);
	if(clist != NULL)
	{
	    glist = clist->selection_end;
	    row = (glist != NULL) ? (gint)glist->data : -1;
	    sensitive = (row >= 0) ? TRUE : FALSE;
	    if(sensitive && (con == NULL))
		sensitive = FALSE;

	    w = yp->audiocd_play_mi;
	    DO_SET_SENSITIVE
	    w = yp->audiocd_play_mbtn;
	    DO_SET_SENSITIVE

	    sensitive = (con != NULL) ? TRUE : FALSE;
	    w = GTK_WIDGET(clist);
	    DO_SET_SENSITIVE
	    w = yp->audiocd_random_mi;
	    DO_SET_SENSITIVE
	    w = yp->audiocd_stop_mi;
	    DO_SET_SENSITIVE
	    w = yp->audiocd_eject_mi;
	    DO_SET_SENSITIVE
	    w = yp->audiocd_refresh_mi;
	    DO_SET_SENSITIVE
	    w = yp->audiocd_stop_btn;
	    DO_SET_SENSITIVE
	    w = yp->audiocd_eject_btn;
	    DO_SET_SENSITIVE
	    w = yp->audiocd_refresh_btn;
	    DO_SET_SENSITIVE
	}

#undef DO_SET_SENSITIVE
}

/*
 *	Mark yconsole as busy or not.
 */
void YPlayerSetBusy(yplayer_struct *yp, gboolean is_busy)
{
	GdkCursor *cursor;
	GtkWidget *w;
	yc_core_struct *core_ptr;


	if(yp == NULL)
	    return;

	core_ptr = YC_CORE(yp->core_ptr);
	if(core_ptr == NULL)
	    return;

	w = YPLAYER_TOPLEVEL(yp);
	if(w != NULL)
	{
	    if(is_busy)
	    {
		/* Increase busy count */
		yp->busy_count++;

		/* If already busy then don't change anything */
		if(yp->busy_count > 1)
		    return;

		cursor = core_ptr->cursor_busy;
	    }
	    else
	    {
		/* Reduce busy count */
		yp->busy_count--;
		if(yp->busy_count < 0)
		    yp->busy_count = 0;

		/* If still busy do not change anything */
		if(yp->busy_count > 0)
		    return;

		cursor = NULL;  /* Use default cursor */
	    }

	    /* Update toplevel window's cursor */
	    if(w->window != NULL)
	    {
		gdk_window_set_cursor(w->window, cursor);
		gdk_flush();
	    }
	}
}

/*
 *	Map yconsole.
 */
void YPlayerMap(yplayer_struct *yp)
{
	if(yp == NULL)
	    return;

	if(TRUE)
	{
	    GtkWidget *w = YPLAYER_TOPLEVEL(yp);
	    if(w != NULL)
		gtk_widget_show_raise(w);
	    yp->map_state = TRUE;
	}
}

/*
 *	Unmap yconsole.
 */
void YPlayerUnmap(yplayer_struct *yp)
{
	if(yp == NULL)
	    return;

	if(TRUE)
	{
	    GtkWidget *w = YPLAYER_TOPLEVEL(yp);
	    if(w != NULL)
		gtk_widget_hide(w);
	    yp->map_state = FALSE;
	}
}

/*
 *	Delete the yconsole.
 */
void YPlayerDelete(yplayer_struct *yp)
{
	GdkBitmap **mask;
	GdkPixmap **pixmap;
/*	GtkAdjustment *adj; */
	GtkWidget **w;


	if(yp == NULL)
	    return;

	if(yp->initialized)
	{
#define DO_UNREF_ADJ		\
{ if(*adj != NULL) { gtk_object_unref(GTK_OBJECT(*adj)); *adj = NULL; } }
#define DO_UNREF_PIXMAP		\
{ if(*pixmap != NULL) { gdk_pixmap_unref(*pixmap); *pixmap = NULL; } }
#define DO_UNREF_PIXMAP_MASK	\
{ if(*pixmap != NULL) { gdk_pixmap_unref(*pixmap); *pixmap = NULL; } \
  if(*mask != NULL) { gdk_bitmap_unref(*mask); *mask = NULL; } }
#define DO_DESTROY_WIDGET       \
{ if(*w != NULL) { gtk_widget_destroy(*w); *w = NULL; } }

	    /* Disconnect from Y server as needed, this will reset
	     * all connection values.
	     */
	    YPlayerDoDisconnect(yp, FALSE);

	    /* Destroy widgets */
	    w = &yp->sndobj_menu;
	    yp->sndobj_play_mi = NULL;
	    yp->sndobj_add_mi = NULL;
	    yp->sndobj_remove_mi = NULL;
	    yp->sndobj_shift_up_mi = NULL;
	    yp->sndobj_shift_down_mi = NULL;
	    yp->sndobj_change_mode_micheck = NULL;
	    yp->sndobj_repeat_micheck = NULL;
	    yp->sndobj_select_all_mi = NULL;
	    yp->sndobj_unselect_all_mi = NULL;
	    yp->sndobj_invert_selection_mi = NULL;
	    yp->sndobj_open_playlist_mi = NULL;
	    yp->sndobj_insert_playlist_mi = NULL;
	    yp->sndobj_save_playlist_mi = NULL;
	    yp->sndobj_save_selected_mi = NULL;
	    DO_DESTROY_WIDGET
	    w = &yp->sndobj_clist;
	    DO_DESTROY_WIDGET
	    w = &yp->sndobj_play_btn;
	    DO_DESTROY_WIDGET
	    w = &yp->sndobj_add_btn;
	    DO_DESTROY_WIDGET
	    w = &yp->sndobj_remove_btn;
	    DO_DESTROY_WIDGET
	    w = &yp->sndobj_shift_up_btn;
	    DO_DESTROY_WIDGET
	    w = &yp->sndobj_shift_down_btn;
	    DO_DESTROY_WIDGET
	    w = &yp->sndobj_change_mode_tb;
	    DO_DESTROY_WIDGET
	    w = &yp->sndobj_repeat_tb;
	    DO_DESTROY_WIDGET
	    w = &yp->sndobj_open_playlist_mbtn;
	    DO_DESTROY_WIDGET
	    w = &yp->sndobj_save_playlist_mbtn;
	    DO_DESTROY_WIDGET
	    w = &yp->sndobj_playing_menu;
	    yp->sndobj_playing_stop_mi = NULL;
	    DO_DESTROY_WIDGET
	    w = &yp->sndobj_playing_clist;
	    DO_DESTROY_WIDGET
	    w = &yp->sndobj_playing_stop_btn;
	    DO_DESTROY_WIDGET
	    w = &yp->sndobj_paned;
	    DO_DESTROY_WIDGET

	    w = &yp->sndobjadv_display_da;
	    DO_DESTROY_WIDGET
	    pixmap = &yp->sndobjadv_display_pm;
	    DO_UNREF_PIXMAP
	    w = &yp->sndobjadv_display_menu;
	    yp->sndobjadv_open_mi = NULL;
	    yp->sndobjadv_play_mi = NULL;
	    yp->sndobjadv_pause_mi = NULL;
	    yp->sndobjadv_rewind_mi = NULL;
	    yp->sndobjadv_change_mode_mi = NULL;
	    yp->sndobjadv_close_mi = NULL;
	    DO_DESTROY_WIDGET
	    w = &yp->sndobjadv_volume_da;
	    DO_DESTROY_WIDGET
	    pixmap = &yp->sndobjadv_volume_pm;
	    DO_UNREF_PIXMAP
	    w = &yp->sndobjadv_position_scale;
	    DO_DESTROY_WIDGET
	    w = &yp->sndobjadv_volume_scale;
	    DO_DESTROY_WIDGET
	    w = &yp->sndobjadv_repeat_tb;
	    DO_DESTROY_WIDGET
	    w = &yp->sndobjadv_sample_rate_spin;
	    DO_DESTROY_WIDGET
	    w = &yp->sndobjadv_open_btn;
	    DO_DESTROY_WIDGET
	    w = &yp->sndobjadv_play_btn;
	    DO_DESTROY_WIDGET
	    w = &yp->sndobjadv_pause_btn;
	    DO_DESTROY_WIDGET
	    w = &yp->sndobjadv_rewind_btn;
	    DO_DESTROY_WIDGET
	    w = &yp->sndobjadv_change_mode_btn;
	    DO_DESTROY_WIDGET
	    w = &yp->sndobjadv_close_btn;
	    DO_DESTROY_WIDGET

	    w = &yp->audiocd_tracks_menu;
	    yp->audiocd_play_mi = NULL;
	    yp->audiocd_random_mi = NULL;
	    yp->audiocd_stop_mi = NULL;
	    yp->audiocd_eject_mi = NULL;
	    yp->audiocd_refresh_mi = NULL;
	    DO_DESTROY_WIDGET
	    w = &yp->audiocd_tracks_clist;
	    DO_DESTROY_WIDGET
	    w = &yp->audiocd_play_mbtn;
	    DO_DESTROY_WIDGET
	    w = &yp->audiocd_stop_btn;
	    DO_DESTROY_WIDGET
	    w = &yp->audiocd_eject_btn;
	    DO_DESTROY_WIDGET
	    w = &yp->audiocd_refresh_btn;
	    DO_DESTROY_WIDGET

	    w = &yp->notebook;
	    DO_DESTROY_WIDGET

	    /* Delete connection bar */
	    ConBarDelete(YPLAYER_CONBAR(yp));
	    yp->conbar = NULL;

	    /* Destroy widgets */
	    w = &yp->toplevel;
	    DO_DESTROY_WIDGET

	    if(yp->accelgrp != NULL)
	    {
		gtk_accel_group_unref(yp->accelgrp);
		yp->accelgrp = NULL;
	    }

	    pixmap = &yp->pixmap_sndobj_dsp;
	    mask = &yp->mask_sndobj_dsp;
	    DO_UNREF_PIXMAP_MASK
	    pixmap = &yp->pixmap_sndobj_midi;
	    mask = &yp->mask_sndobj_midi;
	    DO_UNREF_PIXMAP_MASK

	    pixmap = &yp->pixmap_sndobj_dsp;
	    mask = &yp->mask_sndobj_dsp;
	    DO_UNREF_PIXMAP_MASK
	    pixmap = &yp->pixmap_sndobj_midi;
	    mask = &yp->mask_sndobj_midi;
	    DO_UNREF_PIXMAP_MASK
	    pixmap = &yp->pixmap_sndobjadv_playing;
	    mask = &yp->mask_sndobjadv_playing;
	    DO_UNREF_PIXMAP_MASK
	    pixmap = &yp->pixmap_sndobjadv_repeating;
	    mask = &yp->mask_sndobjadv_repeating;
	    DO_UNREF_PIXMAP_MASK
	    pixmap = &yp->pixmap_sndobjadv_source;
	    mask = &yp->mask_sndobjadv_source;
	    DO_UNREF_PIXMAP_MASK
	    pixmap = &yp->pixmap_sndobjadv_ear;
	    mask = &yp->mask_sndobjadv_ear;
	    DO_UNREF_PIXMAP_MASK

#undef DO_UNREF_ADJ
#undef DO_UNREF_PIXMAP_MASK
#undef DO_DESTROY_WIDGET
	}

	/* Deallocate structure itself */
	g_free(yp);
}
