/*
			     YPlayer Window
 */

#ifndef YPLAYER_H
#define YPLAYER_H

#include <Y2/Y.h>
#include <Y2/Ylib.h>
#include <gtk/gtk.h>

#include "conbar.h"

/*
 *	YPlayer window structure:
 */
typedef struct {

	gboolean	initialized;
	gboolean	map_state;
	gint		busy_count;
	GtkAccelGroup	*accelgrp;
	gpointer	core_ptr;

	GtkWidget	*toplevel;
	connection_bar_struct *conbar;
	GtkWidget	*notebook;

	/* Widgets on sound object page. */
	GtkWidget	*sndobj_paned,
			*sndobj_clist,
			*sndobj_menu,
			*sndobj_play_mi,
			*sndobj_add_mi,
			*sndobj_remove_mi,
			*sndobj_shift_up_mi,
			*sndobj_shift_down_mi,
			*sndobj_change_mode_micheck,
			*sndobj_repeat_micheck,
			*sndobj_select_all_mi,
			*sndobj_unselect_all_mi,
			*sndobj_invert_selection_mi,
			*sndobj_open_playlist_mi,
			*sndobj_insert_playlist_mi,
			*sndobj_save_playlist_mi,
			*sndobj_save_selected_mi,
			*sndobj_play_btn,
			*sndobj_add_btn,
			*sndobj_remove_btn,
			*sndobj_shift_up_btn,
			*sndobj_shift_down_btn,
			*sndobj_change_mode_tb,
			*sndobj_repeat_tb,
			*sndobj_open_playlist_mbtn,
			*sndobj_save_playlist_mbtn,
			*sndobj_playing_clist,
			*sndobj_playing_menu,
			*sndobj_playing_stop_mi,
			*sndobj_playing_stop_btn;
	GdkPixmap       *pixmap_sndobj_dsp;
	GdkBitmap       *mask_sndobj_dsp;
	GdkPixmap       *pixmap_sndobj_midi;
	GdkBitmap       *mask_sndobj_midi;

	/* Widgets on sound object advanced page. */
	GtkWidget	*sndobjadv_display_da,
			*sndobjadv_display_menu,
			*sndobjadv_volume_da,
			*sndobjadv_open_mi,
			*sndobjadv_play_mi,
			*sndobjadv_pause_mi,
			*sndobjadv_rewind_mi,
			*sndobjadv_change_mode_mi,
			*sndobjadv_close_mi,
			*sndobjadv_position_scale,
			*sndobjadv_volume_scale,	/* Amplitude. */
			*sndobjadv_repeat_tb,
			*sndobjadv_sample_rate_spin,
			*sndobjadv_open_btn,
			*sndobjadv_play_btn,
			*sndobjadv_pause_btn,
			*sndobjadv_rewind_btn,
			*sndobjadv_change_mode_btn,
			*sndobjadv_close_btn;
	GdkPixmap	*sndobjadv_display_pm,
			*sndobjadv_volume_pm;
	gint		sndobjadv_freeze_count;
	gint		sndobjadv_button;
	gchar		*sndobjadv_path;	/* If not NULL implies loaded. */
	YEventSoundObjectAttributes	sndobjadv_soa;
	YID		sndobjadv_yid;		/* If not YIDNULL implies playing. */
	gfloat		sndobjadv_volume_x,	/* Left hand rule. */
			sndobjadv_volume_y;
	GdkPixmap       *pixmap_sndobjadv_playing;
	GdkBitmap       *mask_sndobjadv_playing;
	GdkPixmap       *pixmap_sndobjadv_repeating;
	GdkBitmap       *mask_sndobjadv_repeating;
	GdkPixmap       *pixmap_sndobjadv_source;
	GdkBitmap       *mask_sndobjadv_source;
	GdkPixmap       *pixmap_sndobjadv_ear;
	GdkBitmap       *mask_sndobjadv_ear;

	/* Widgets on audio CD page. */
	GtkWidget	*audiocd_tracks_clist,
			*audiocd_tracks_menu,
			*audiocd_play_mi,
			*audiocd_random_mi,
			*audiocd_stop_mi,
			*audiocd_eject_mi,
			*audiocd_refresh_mi,
			*audiocd_play_mbtn,
			*audiocd_stop_btn,
			*audiocd_eject_btn,
			*audiocd_refresh_btn;

	/* Y server values. */
	YConnection *con;

	guint8 *yshm_buf;
	gint yshm_buf_len, yshm_id;

	gfloat channel_volume_avg[4];

	gulong start_time;
	gchar *audio_mode_name;
	gint sample_rate, sample_size, channels;
	glong cycle_us;
	gfloat cycle_load;
	gint num_fragments, fragment_size;
	gboolean flip_stereo, shellout;
	gint version_major, version_minor;
	gchar *server_name, *sound_name;

} yplayer_struct;

#define YPLAYER(p)	((yplayer_struct *)(p))
#define YPLAYER_TOPLEVEL(p)	(((p) != NULL) ? \
 (YPLAYER(p)->toplevel) : NULL)
#define YPLAYER_CONBAR(p)	(((p) != NULL) ? \
 (YPLAYER(p)->conbar) : NULL)
#define YPLAYER_NOTEBOOK(p)	(((p) != NULL) ? \
 (YPLAYER(p)->notebook) : NULL)

#define YPLAYER_SNDOBJ_PANED(p) (((p) != NULL) ? \
 (YPLAYER(p)->sndobj_paned) : NULL)
#define YPLAYER_SNDOBJ_CLIST(p)	(((p) != NULL) ? \
 (YPLAYER(p)->sndobj_clist) : NULL)
#define YPLAYER_SNDOBJ_MENU(p)	(((p) != NULL) ? \
 (YPLAYER(p)->sndobj_menu) : NULL)
#define YPLAYER_SNDOBJ_PLAYING_CLIST(p)  (((p) != NULL) ? \
 (YPLAYER(p)->sndobj_playing_clist) : NULL)
#define YPLAYER_SNDOBJ_PLAYING_MENU(p)  (((p) != NULL) ? \
 (YPLAYER(p)->sndobj_playing_menu) : NULL)

#define YPLAYER_SNDOBJADV_IS_OPENED(p)		(((p) != NULL) ? \
 ((YPLAYER(p)->sndobjadv_path) != NULL) : FALSE)
#define YPLAYER_SNDOBJADV_IS_PLAYING(p)		(((p) != NULL) ? \
 ((YPLAYER(p)->sndobjadv_yid) != YIDNULL) : FALSE)
#define YPLAYER_SNDOBJADV_FREEZE_COUNT(p)	(((p) != NULL) ? \
 (YPLAYER(p)->sndobjadv_freeze_count) : 0)
#define YPLAYER_SNDOBJADV_FREEZE(p)		{ if((p) != NULL) { \
 YPLAYER(p)->sndobjadv_freeze_count++; \
} }
#define YPLAYER_SNDOBJADV_THAW(p)		{ if((p) != NULL) { \
 YPLAYER(p)->sndobjadv_freeze_count--; \
 if(YPLAYER(p)->sndobjadv_freeze_count < 0) \
  YPLAYER(p)->sndobjadv_freeze_count = 0; \
} }

#define YPLAYER_AUDIOCD_TRACKS_CLIST(p)	(((p) != NULL) ? \
 (YPLAYER(p)->audiocd_tracks_clist) : NULL)
#define YPLAYER_AUDIOCD_TRACKS_MENU(p)	(((p) != NULL) ? \
 (YPLAYER(p)->audiocd_tracks_menu) : NULL)

#define YPLAYER_IS_CONNECTED(p)	(((p) != NULL) ? \
 ((YPLAYER(p)->con) != NULL) : FALSE)
#define YPLAYER_CONNECTION(p)		(((p) != NULL) ? \
 (YPLAYER(p)->con) : NULL)

#define YPLAYER_SET_CONBAR_AUDIO(p)	\
{ if((p) != NULL) \
  ConBarSetAudio( \
   YPLAYER_CONBAR(p), \
   (p)->start_time, (p)->audio_mode_name, \
   (p)->sample_rate, (p)->sample_size, (p)->channels, \
   (p)->cycle_us, (p)->cycle_load, \
   (p)->num_fragments, (p)->fragment_size, \
   (p)->flip_stereo, (p)->shellout, \
   (p)->yshm_buf, (p)->yshm_buf_len, \
   (p)->channel_volume_avg, \
   (p)->version_major, (p)->version_minor, \
   (p)->server_name, (p)->sound_name \
  ); }

/*
 *	YPlayer sound object playing structure:
 */
typedef struct {

	yplayer_struct *yp;	/* Shared. */
	gchar *filename;
	YID yid;		/* Sound object. */

} yplayer_sndobj_playing_struct;

#define YPLAYER_SNDOBJ_PLAYING(p)	((yplayer_sndobj_playing_struct *)(p))


extern void YPlayerVolumeXYToQuad(
	gfloat x, gfloat y, gfloat a,
	gfloat *l, gfloat *r, gfloat *bl, gfloat *br
);
extern gchar *YPlayerDataLengthToTimeString(
	gulong length,
	gint sample_rate, gint sample_size, gint channels
);
extern GtkRcStyle *YPlayerAudioDisplayRCStyleNew(yplayer_struct *yp);
extern void YPlayerSetTitle(yplayer_struct *yp, const gchar *s);
extern gchar *YPlayerGetLocation(yplayer_struct *yp);
extern gint YPlayerDoConnect(
	yplayer_struct *yp,
	const gchar *address, const gchar *start_command
);
extern void YPlayerDoDisconnect(yplayer_struct *yp, gboolean force_shutdown);
extern void YPlayerDoPlaySoundObject(
	yplayer_struct *yp, const gchar *path, gboolean add_as_needed
);
extern void YPlayerDoOpenSoundObjectAdv(yplayer_struct *yp, const gchar *path);
extern void YPlayerDoCloseSoundObjectAdv(yplayer_struct *yp);
extern void YPlayerDoPlaySoundObjectAdv(yplayer_struct *yp, const gchar *path);
extern void YPlayerDoPlayAudioCDTrack(yplayer_struct *yp, gint track_number);

extern gint YPlayerSndObjListAdd(
	yplayer_struct *yp, gint row,
	const gchar *filename, gint format,
	gint sample_rate, gint sample_size, gint channels,
	gulong length
);
extern gint YPlayerSndObjPlayingListAdd(
	yplayer_struct *yp, gint row,
	const gchar *filename, gint format, YID yid
);
extern void YPlayerSndObjPlayingListClear(yplayer_struct *yp);
extern void YPlayerAudioCDTracksListUpdate(yplayer_struct *yp);

extern void YPlayerLoadPlayListFile(
	yplayer_struct *yp, const gchar *path, gboolean insert
);
extern void YPlayerSavePlayListFile(
	yplayer_struct *yp, const gchar *path, gboolean selected_only
);

extern void YPlayerSndObjAdvDisplayDraw(yplayer_struct *yp);
extern void YPlayerSndObjAdvVolumeDraw(yplayer_struct *yp);

extern void YPlayerSavePosition(yplayer_struct *yp);

extern yplayer_struct *YPlayerNew(gpointer core_ptr);
extern void YPlayerUpdateMenus(yplayer_struct *yp);
extern void YPlayerSetBusy(yplayer_struct *yp, gboolean is_busy);
extern void YPlayerMap(yplayer_struct *yp);
extern void YPlayerUnmap(yplayer_struct *yp);
extern void YPlayerDelete(yplayer_struct *yp);



#endif	/* YPLAYER_H */
