#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/stat.h>

#include <Y2/Y.h>
#include <Y2/Ylib.h>

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include "../include/disk.h"

#include "guiutils.h"
#include "cdialog.h"
#include "fb.h"

#include "cfg.h"
#include "conbar.h"
#include "yplayer.h"
#include "yplayercb.h"
#include "yc.h"
#include "ycop.h"
#include "ycanalyze.h"
#include "yccfglist.h"
#include "config.h"


void YPlayerSndObjListItemDestroyCB(gpointer data);
void YPlayerSndObjPlayingListItemDestroyCB(gpointer data);

gint YPlayerDeleteEventCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
);
void YPlayerDestroyCB(GtkObject *object, gpointer data);

gint YPlayerDSPViewButtonCB(
	GtkWidget *widget, GdkEventButton *button, gpointer data
);
gint YPlayerKeyEventCB(
	GtkWidget *widget, GdkEventKey *key, gpointer data
);
gint YPlayerButtonEventCB(
	GtkWidget *widget, GdkEventButton *button, gpointer data
);
void YPlayerCListSelectRowCB(
	GtkCList *clist, gint row, gint column, GdkEvent *event,
	gpointer data
);
void YPlayerCListUnselectRowCB(
	GtkCList *clist, gint row, gint column, GdkEvent *event,
	gpointer data
);

void YPlayerConnectCB(
	GtkWidget *widget, const gchar *s, gpointer data
);
void YPlayerYServerListCB(GtkWidget *widget, gpointer data);
void YPlayerCloseCB(GtkWidget *widget, gpointer data);
void YPlayerNewYConsoleCB(GtkWidget *widget, gpointer data);
void YPlayerNewYMixerCB(GtkWidget *widget, gpointer data);
void YPlayerNewYAudioCB(GtkWidget *widget, gpointer data);
void YPlayerNewYPlayerCB(GtkWidget *widget, gpointer data);
void YPlayerUptionsCB(GtkWidget *widget, gpointer data);
void YPlayerRefreshCB(GtkWidget *widget, gpointer data);
gboolean YPlayerShelloutCB(
	GtkWidget *widget, gboolean shellout, gpointer data
);
void YPlayerShutdownCB(GtkWidget *widget, gpointer data);
void YPlayerManageCB(yplayer_struct *yp);
void YPlayerManage2CB(yplayer_struct *yp);

gboolean YPlayerSndObjListDragMotionCB(
	GtkWidget *widget, GdkDragContext *dc, gint x, gint y, guint t,
	gpointer data
);
void YPlayerSndObjListDragDataReceivedCB(
	GtkWidget *widget, GdkDragContext *dc, gint x, gint y,
	GtkSelectionData *selection_data, guint info, guint t,
	gpointer data
);

void YPlayerSndObjPlayCB(GtkWidget *widget, gpointer data);
void YPlayerSndObjStopCB(GtkWidget *widget, gpointer data);
void YPlayerSndObjAddCB(GtkWidget *widget, gpointer data);
void YPlayerSndObjRemoveCB(GtkWidget *widget, gpointer data);
void YPlayerSndObjShiftUpCB(GtkWidget *widget, gpointer data);
void YPlayerSndObjShiftDownCB(GtkWidget *widget, gpointer data);
void YPlayerSndObjChangeModeToggleCB(GtkWidget *widget, gpointer data);
void YPlayerSndObjRepeatToggleCB(GtkWidget *widget, gpointer data);
void YPlayerSndObjSelectAllCB(GtkWidget *widget, gpointer data);
void YPlayerSndObjUnselectAllCB(GtkWidget *widget, gpointer data);
void YPlayerSndObjInvertSelectionCB(GtkWidget *widget, gpointer data);
void YPlayerSndObjOpenPlayListCB(GtkWidget *widget, gpointer data);
void YPlayerSndObjInsertPlayListCB(GtkWidget *widget, gpointer data);
void YPlayerSndObjSavePlayListCB(GtkWidget *widget, gpointer data);
void YPlayerSndObjSaveSelectedCB(GtkWidget *widget, gpointer data);
void YPlayerSndObjPlayingKilledCB(yplayer_struct *yp, YID yid);

gint YPlayerSndObjAdvEventCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
);
void YPlayerSndObjAdvKilledCB(yplayer_struct *yp, YID yid);
void YPlayerSndObjAdvVolumeChangedCB(
	GtkAdjustment *adj, gpointer data
);
void YPlayerSndObjAdvPositionChangedCB(
	GtkAdjustment *adj, gpointer data
);
void YPlayerSndObjAdvSampleRateChangedCB(
	GtkAdjustment *adj, gpointer data
);
void YPlayerSndObjAdvOpenCB(GtkWidget *widget, gpointer data);
void YPlayerSndObjAdvPlayCB(GtkWidget *widget, gpointer data);
void YPlayerSndObjAdvPauseCB(GtkWidget *widget, gpointer data);
void YPlayerSndObjAdvRewindCB(GtkWidget *widget, gpointer data);
void YPlayerSndObjAdvChangeModeCB(GtkWidget *widget, gpointer data);
void YPlayerSndObjAdvCloseCB(GtkWidget *widget, gpointer data);

void YPlayerAudioCDPlayCB(GtkWidget *widget, gpointer data);
void YPlayerAudioCDRandomCB(GtkWidget *widget, gpointer data);
void YPlayerAudioCDStopCB(GtkWidget *widget, gpointer data);
void YPlayerAudioCDEjectCB(GtkWidget *widget, gpointer data);
void YPlayerAudioCDRefreshCB(GtkWidget *widget, gpointer data);

void YPlayerReconfiguredCB(
	yplayer_struct *yp, gpointer core_ptr
);


#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define ABSOLUTE(x)	(((x) < 0) ? ((x) * -1) : (x))


/*
 *	Sound object GtkCList item destroy signal callback.
 */
void YPlayerSndObjListItemDestroyCB(gpointer data)
{
	gchar *filename = (gchar *)data;
	g_free(filename);
}

/*
 *      Sound object playing GtkCList item destroy signal callback.
 */
void YPlayerSndObjPlayingListItemDestroyCB(gpointer data)
{
	YConnection *con;
	yplayer_struct *yp;
	yplayer_sndobj_playing_struct *sop = YPLAYER_SNDOBJ_PLAYING(data);
	if(sop == NULL)
	    return;

	/* Get YPlayer and Y server connection so that we can destroy
	 * the playing sound object.
	 */
	yp = sop->yp;
	con = YPLAYER_CONNECTION(yp);

	/* Connected andthe sound object has already been killed?
	 * Note that if we were not connected then the sound object
	 * would have already been killed so no need to worry about
	 * the yid.
	 */
	if((con != NULL) && (sop->yid != YIDNULL))
	{
	    /* Destroy sound playing object */
/* printf("Killing YID %ld\n", sop->yid); */
	    YDestroyPlaySoundObject(con, sop->yid);
	    sop->yid = YIDNULL;
	}

	g_free(sop->filename);
	g_free(sop);
}


/*
 *	Toplevel "delete_event" signal callback.
 */
gint YPlayerDeleteEventCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
)
{
	YPlayerCloseCB(widget, data);
	return(TRUE);
}

/*
 *	Toplevel "destroy" signal callback.
 */
void YPlayerDestroyCB(GtkObject *object, gpointer data)
{

}


/*
 *	DSPView "button_press_event" signal callback.
 */
gint YPlayerDSPViewButtonCB(
	GtkWidget *widget, GdkEventButton *button, gpointer data
)
{
	gint status = FALSE;
	yplayer_struct *yp = YPLAYER(data);
	if(yp == NULL)
	    return(status);

	switch(button->button)
	{
	  case 3:
	    switch((gint)button->type)
	    {
	      case GDK_BUTTON_PRESS:
		YCNewOptWinSelectPage(
		    YC_CORE(yp->core_ptr), YPLAYER_TOPLEVEL(yp),
		    1
		);
		status = TRUE;
		break;

	      case GDK_BUTTON_RELEASE:
		status = TRUE;
		break;
	    }
	}

	return(status);
}

/*
 *	YPlayer GtkCList "button_press_event" signal callback.
 */
gint YPlayerKeyEventCB(
	GtkWidget *widget, GdkEventKey *key, gpointer data
)
{
	gint status = FALSE;
	gboolean key_press;
	GtkCList *clist;
	yplayer_struct *yp = YPLAYER(data);
	if((widget == NULL) || (key == NULL) || (yp == NULL))
	    return(status);

	if(GTK_IS_CLIST(widget))
	    clist = GTK_CLIST(widget);
	else
	    return(status);

	key_press = (key->type == GDK_KEY_PRESS) ? TRUE : FALSE;

	switch(key->keyval)
	{
	  case GDK_Return:
	  case GDK_KP_Enter:
	  case GDK_ISO_Enter:
	  case GDK_3270_Enter:
	    if(key_press)
	    {
		if(widget == YPLAYER_SNDOBJ_CLIST(yp))
		{
		    YPlayerSndObjPlayCB(widget, data);
		}
		else if(widget == YPLAYER_SNDOBJ_PLAYING_CLIST(yp))
		{
		    YPlayerSndObjStopCB(widget, data);
		}
		else if(widget == YPLAYER_AUDIOCD_TRACKS_CLIST(yp))
		{
		    if(key->state & GDK_CONTROL_MASK)
			YPlayerAudioCDRandomCB(widget, data);
		    else
			YPlayerAudioCDPlayCB(widget, data);
		}
		status = TRUE;
	    }
	    break;

	  case GDK_Escape:
	    if(key_press)
	    {
		if(widget == YPLAYER_SNDOBJ_CLIST(yp))
		{
		    YPlayerSndObjStopCB(widget, data);
		}
		else if(widget == YPLAYER_SNDOBJ_PLAYING_CLIST(yp))
		{
		    YPlayerSndObjStopCB(widget, data);
		}
		else if(widget == YPLAYER_AUDIOCD_TRACKS_CLIST(yp))
		{
		    if(key->state & GDK_CONTROL_MASK)
			YPlayerAudioCDEjectCB(widget, data);
		    else
			YPlayerAudioCDStopCB(widget, data);
		}
		status = TRUE;
	    }
	    break;

	  case GDK_Insert:
	    if(key_press)
	    {
		if(widget == YPLAYER_SNDOBJ_CLIST(yp))
		{
		    if(key->state & GDK_SHIFT_MASK)
			YPlayerSndObjOpenPlayListCB(widget, data);
		    else
			YPlayerSndObjAddCB(widget, data);
		}
		else if(widget == YPLAYER_SNDOBJ_PLAYING_CLIST(yp))
		{
		    if(key->state & GDK_SHIFT_MASK)
			YPlayerSndObjOpenPlayListCB(widget, data);
		    else
			YPlayerSndObjAddCB(widget, data);
		}
		else if(widget == YPLAYER_AUDIOCD_TRACKS_CLIST(yp))
		{

		}
		status = TRUE;
	    }
	    break;

	  case GDK_Delete:
	    if(key_press)
	    {
		if(widget == YPLAYER_SNDOBJ_CLIST(yp))
		{
		    YPlayerSndObjRemoveCB(widget, data);
		}
		else if(widget == YPLAYER_SNDOBJ_PLAYING_CLIST(yp))
		{
		    YPlayerSndObjStopCB(widget, data);
		}
		else if(widget == YPLAYER_AUDIOCD_TRACKS_CLIST(yp))
		{

		}
		status = TRUE;
	    }
	    break;

	}

	return(status);
}

/*
 *	YPlayer GtkCList "button_press_event" signal callback.
 */
gint YPlayerButtonEventCB(
	GtkWidget *widget, GdkEventButton *button, gpointer data
)
{
	gint status = FALSE;
	GtkCList *clist;
	yplayer_struct *yp = YPLAYER(data);
	if((widget == NULL) || (button == NULL) || (yp == NULL))
	    return(status);

	if(GTK_IS_CLIST(widget))
	    clist = GTK_CLIST(widget);
	else
	    return(status);

	if(button->window != clist->clist_window)
	    return(status);

	switch(button->button)
	{
	  case 3:
	    switch((gint)button->type)
	    {
	      case GDK_BUTTON_PRESS:
		if(widget == YPLAYER_SNDOBJ_CLIST(yp))
		{
		    gint row, column;
		    GtkWidget *w;

		    gtk_clist_get_selection_info(
			clist, button->x, button->y, &row, &column
		    );
		    gtk_clist_freeze(clist);
		    if(!(button->state & GDK_CONTROL_MASK) &&
		       !(button->state & GDK_SHIFT_MASK)
		    )
			gtk_clist_unselect_all(clist);
		    gtk_clist_select_row(clist, row, column);
		    gtk_clist_thaw(clist);

		    w = YPLAYER_SNDOBJ_MENU(yp);
		    if(w != NULL)
			gtk_menu_popup(
			    GTK_MENU(w), NULL, NULL,
			    NULL, NULL,
			    button->button, button->time
			);
		}
		else if(widget == YPLAYER_SNDOBJ_PLAYING_CLIST(yp))
		{
		    gint row, column;
		    GtkWidget *w;

		    gtk_clist_get_selection_info(
			clist, button->x, button->y, &row, &column
		    );
		    gtk_clist_freeze(clist);
		    if(!(button->state & GDK_CONTROL_MASK) &&
		       !(button->state & GDK_SHIFT_MASK)
		    )
			gtk_clist_unselect_all(clist);
		    gtk_clist_select_row(clist, row, column);
		    gtk_clist_thaw(clist);

		    w = YPLAYER_SNDOBJ_PLAYING_MENU(yp);
		    if(w != NULL)
			gtk_menu_popup(
			    GTK_MENU(w), NULL, NULL,
			    NULL, NULL,
			    button->button, button->time
			);
		}
		else if(widget == YPLAYER_AUDIOCD_TRACKS_CLIST(yp))
		{
		    gint row, column;
		    GtkWidget *w;

		    gtk_clist_get_selection_info(
			clist, button->x, button->y, &row, &column
		    );
		    gtk_clist_freeze(clist);
		    gtk_clist_unselect_all(clist);
		    gtk_clist_select_row(clist, row, column);
		    gtk_clist_thaw(clist);

		    w = YPLAYER_AUDIOCD_TRACKS_MENU(yp);
		    if(w != NULL)
			gtk_menu_popup(
			    GTK_MENU(w), NULL, NULL,
			    NULL, NULL,
			    button->button, button->time
			);
		}
		status = TRUE;
		break;
	    }
	    break;

	  case 1:
	    switch((gint)button->type)
	    {
	      case GDK_2BUTTON_PRESS:
		if(widget == YPLAYER_SNDOBJ_CLIST(yp))
		{
		    YPlayerSndObjPlayCB(widget, data);
		}
		else if(widget == YPLAYER_SNDOBJ_PLAYING_CLIST(yp))
		{
		    YPlayerSndObjStopCB(widget, data);
		}
		else if(widget == YPLAYER_AUDIOCD_TRACKS_CLIST(yp))
		{
		    if(button->state & GDK_CONTROL_MASK)
			YPlayerAudioCDRandomCB(widget, data);
		    else
			YPlayerAudioCDPlayCB(widget, data);
		}
		status = TRUE;
		break;
	    }
	    break;
	}

	return(status);
}

/*
 *	YPlayer GtkCList "select_row" signal callback.
 */
void YPlayerCListSelectRowCB(
	GtkCList *clist, gint row, gint column, GdkEvent *event,
	gpointer data
)
{
	yplayer_struct *yp = YPLAYER(data);
	if((clist == NULL) || (yp == NULL))
	    return;

	if(gtk_clist_row_is_visible(clist, row) !=
	    GTK_VISIBILITY_FULL
	)
	    gtk_clist_moveto(
		clist,
		row, -1,	/* Row, column */
		0.5, 0.0        /* Row, column */
	    );

	YPlayerUpdateMenus(yp);
}

/*
 *      YPlayer GtkCList "unselect_row" signal callback.
 */
void YPlayerCListUnselectRowCB(
	GtkCList *clist, gint row, gint column, GdkEvent *event,
	gpointer data
)
{
	yplayer_struct *yp = YPLAYER(data);
	if((clist == NULL) || (yp == NULL))
	    return;

	YPlayerUpdateMenus(yp);
}


/*
 *	Connect callback.
 */
void YPlayerConnectCB(
	GtkWidget *widget, const gchar *s, gpointer data
)
{
	gchar *strptr;
	gchar *s2 = (s != NULL) ? g_strdup(s) : NULL;
	yplayer_struct *yp = YPLAYER(data);
	if(yp == NULL)
	    return;

	YPlayerSetBusy(yp, TRUE);

	/* Empty string? Then use default location */
	if(*s2 == '\0')
	{
	    g_free(s2);
	    s2 = g_strdup("127.0.0.1:9433");
	}

	/* No port number? Then append it */
	strptr = strchr(s2, ':');
	if(strptr == NULL)
	{
	    strptr = g_strdup_printf("%s:9433", s2);
	    g_free(s2);
	    s2 = strptr;
	}

	/* Do connect */
	YPlayerDoConnect(yp, s2, NULL);

	/* Delete coppied location */
	g_free(s2);

	YPlayerSetBusy(yp, FALSE);

}

/*
 *	Y server list callback.
 */
void YPlayerYServerListCB(GtkWidget *widget, gpointer data)
{
	yplayer_struct *yp = YPLAYER(data);
	if(yp == NULL)
	    return;

	YCNewYServerList(YC_CORE(yp->core_ptr), YPLAYER_TOPLEVEL(yp));
}

/*
 *	Close callback.
 */
void YPlayerCloseCB(GtkWidget *widget, gpointer data)
{
	yplayer_struct *yp = YPLAYER(data);
	if(yp == NULL)
	    return;

	/* Save default playlist */
	YPlayerSavePlayListFile(
	    yp,
	    YC_GET_CFG_S(
		YC_CORE(yp->core_ptr),
		YC_CFG_PARM_PLAYER_SNDOBJ_DEFAULT_PLAYLIST_FILE
	    ),
	    FALSE
	);

	/* Save positions and unmap */
	YPlayerSavePosition(yp);
	YPlayerUnmap(yp);
}

/*
 *	New YConsole window callback.
 */
void YPlayerNewYConsoleCB(GtkWidget *widget, gpointer data)
{
	yplayer_struct *yp = YPLAYER(data);
	if(yp != NULL)
	    YCNewYConsole(
		YC_CORE(yp->core_ptr),
		YPlayerGetLocation(yp)
	    );
}

/*
 *	New YMixer window callback.
 */
void YPlayerNewYMixerCB(GtkWidget *widget, gpointer data)
{
	yplayer_struct *yp = YPLAYER(data);
	if(yp != NULL)
	    YCNewYMixer(
		YC_CORE(yp->core_ptr),
		YPlayerGetLocation(yp)
	    );
}

/*
 *	New YAudio window callback.
 */
void YPlayerNewYAudioCB(GtkWidget *widget, gpointer data)
{
	yplayer_struct *yp = YPLAYER(data);
	if(yp != NULL)
	    YCNewYAudio(
		YC_CORE(yp->core_ptr),
		YPlayerGetLocation(yp)
	    );
}

/*
 *	New YPlayer window callback.
 */
void YPlayerNewYPlayerCB(GtkWidget *widget, gpointer data)
{
	yplayer_struct *yp = YPLAYER(data);
	if(yp != NULL)
	    YCNewYPlayer(
		YC_CORE(yp->core_ptr),
		YPlayerGetLocation(yp)
	    );
}

/*
 *	Map options window callback.
 */
void YPlayerUptionsCB(GtkWidget *widget, gpointer data)
{
	yplayer_struct *yp = YPLAYER(data);
	if(yp != NULL)
	    YCNewOptWin(
		YC_CORE(yp->core_ptr),
		YPLAYER_TOPLEVEL(yp)
	    );
}

/*
 *	Refresh callback.
 */
void YPlayerRefreshCB(GtkWidget *widget, gpointer data)
{
	gchar *s;
	yplayer_struct *yp = YPLAYER(data);
	if(yp == NULL)
	    return;

	YPlayerDoDisconnect(yp, FALSE);
	s = ConBarGetLocation(YPLAYER_CONBAR(yp));
	s = (s != NULL) ? g_strdup(s) : NULL;
	YPlayerDoConnect(yp, s, NULL);
	g_free(s);

}

/*
 *	Shellout callback.
 *
 *	Returns actual shellout state.
 */
gboolean YPlayerShelloutCB(
	GtkWidget *widget, gboolean shellout, gpointer data
)
{
	YConnection *con;
	yplayer_struct *yp = YPLAYER(data);
	if(yp == NULL)
	    return(yp->shellout);

	con = YPLAYER_CONNECTION(yp);
	if(con == NULL)
	    return(yp->shellout);

	if(shellout == yp->shellout)
	    return(yp->shellout);

	if(shellout)
	{
	    if(YSetAudioModeValues(
		con,
		8,      /* Sample size in bits */
		1,      /* Number of channels */
		0,      /* Sample rate set to 0 implies shellout */
		0,      /* Direction, 0 = play, 1 = record */
		True,
		2,
		1024
	    ))
		yp->shellout = FALSE;
	    else
		yp->shellout = TRUE;
	}
	else
	{
	    if(YChangeAudioModePreset(
		con,
		(yp->audio_mode_name != NULL) ?
		    yp->audio_mode_name : "Default"
	    ))
		yp->shellout = TRUE;
	    else
		yp->shellout = FALSE;
	}

	YPLAYER_SET_CONBAR_AUDIO(yp);

	return(yp->shellout);
}

/*
 *	Shutdown Y server callback.
 */
void YPlayerShutdownCB(GtkWidget *widget, gpointer data)
{
	yplayer_struct *yp = YPLAYER(data);
	if(yp == NULL)
	    return;

	YPlayerDoDisconnect(yp, TRUE);
}

/*
 *	YPlayer manage callback.
 */
void YPlayerManageCB(yplayer_struct *yp)
{
	YConnection *con;
	YID yid;
	YEvent yevent;
	dspview_struct *dv;
	connection_bar_struct *cb;


	if(yp == NULL)
	    return;

	cb = YPLAYER_CONBAR(yp);
	dv = ConBarGetDSPView(cb);
	if(dv == NULL)
	    return;

	/* Manage Y events */
	con = YPLAYER_CONNECTION(yp);
	while((con != NULL) && (YGetNextEvent(con, &yevent, False) > 0))
	{
	    switch(yevent.type)
	    {
	      case YAudioChange:
		/* Update Audio values on conbar and DSPView */
		if(yevent.audio.preset)
		{
		    YEventAudioStats stat_buf;

		    g_free(yp->audio_mode_name);
		    yp->audio_mode_name = g_strdup(yevent.audio.mode_name);

		    if(!YGetAudioStats(con, &stat_buf))
		    {
			yp->sample_rate = stat_buf.sample_rate;
			yp->sample_size = stat_buf.sample_size;
			yp->channels = stat_buf.channels;
			yp->cycle_us = stat_buf.cycle_us;
			yp->num_fragments = stat_buf.num_fragments;
			yp->fragment_size = stat_buf.fragment_size;
			yp->flip_stereo = stat_buf.flip_stereo;
			yp->shellout = FALSE;
		    }
		}
		else
		{
		    YEventAudioChange *audio = &yevent.audio;
		    yp->sample_rate = audio->sample_rate;
		    yp->sample_size = audio->sample_size;
		    yp->channels = audio->channels;
		    yp->num_fragments = audio->num_fragments;
		    yp->fragment_size = audio->fragment_size_bytes;
		    yp->flip_stereo = FALSE;
		    yp->shellout = FALSE;
		}
		YPLAYER_SET_CONBAR_AUDIO(yp);
		ConBarSetShellOut(cb, yp->shellout);
		DSPViewSetChannels(dv, yevent.audio.channels);
		break;

	      case YCycleChange:
		/* Update audio values */
		yp->cycle_us = yevent.cycle.cycle_us;
		YPLAYER_SET_CONBAR_AUDIO(yp);
		break;

	      case YDisconnect: case YShutdown:
		YPlayerDoDisconnect(yp, FALSE);
		con = NULL;	/* So we break out of this loop */
		break;

	      case YSoundObjectPlay:
	      case YSoundObjectPlayValues:
		yid = yevent.play.yid;
		if(yid == yp->sndobjadv_yid)
		{

		}
		break;

	      case YSoundObjectKill:
		yid = yevent.kill.yid;
		YPlayerSndObjPlayingKilledCB(yp, yid);
		YPlayerSndObjAdvKilledCB(yp, yid);
		break;

	      case YSHMSound:
		if(yevent.yshmsound.minor_op_code == YSHMSoundClose)
		{
		    /* Need to reopen YSHM sound buffer when Audio
		     * mode changes.
		     */
		    YSHMCloseSound(con, yp->yshm_buf);
		    yp->yshm_buf = (guint8 *)YSHMOpenSound(
			con, &yp->yshm_buf_len, &yp->yshm_id
		    );
		}
		break;
	    }
	}

	/* Redraw DSP view */
	ConBarDraw(cb, yp->yshm_buf, yp->yshm_buf_len);
}

/*
 *      YPlayer manage 2 callback.
 *
 *	This function is called in longer successions.
 */
void YPlayerManage2CB(yplayer_struct *yp)
{
	YConnection *con;
	YID yid;

	if(yp == NULL)
	    return;

	con = YPLAYER_CONNECTION(yp);
	if(con != NULL)
	{
	    YEventServerStats stat_buf;
/* Don't need this since avg volume not displayed
	    YCAnalyzeAverageChannelVolume(
		yp->yshm_buf, yp->yshm_buf_len,
		yp->channel_volume_avg,
		CLIP(yp->channels, 1, 4)
	    );
 */
	    if(!YGetServerStats(con, &stat_buf))
	    {
		yp->start_time = stat_buf.start_time;
		yp->cycle_load = stat_buf.cycle_load;
		yp->version_major = stat_buf.protocol_version_major;
		yp->version_minor = stat_buf.protocol_version_minor;
		g_free(yp->server_name);
		yp->server_name = g_strdup(stat_buf.server_name);
		g_free(yp->sound_name);
		yp->sound_name = g_strdup(stat_buf.sound_name);
		YPLAYER_SET_CONBAR_AUDIO(yp);
	    }
	}

	/* Get sound object advanced playing sound object statistics */
	yid = yp->sndobjadv_yid;
	if((con != NULL) && (yid != YIDNULL))
	{
	    YEventSoundPlay v;

	    if(!YGetPlaySoundObjectValues(con, yid, &v))
	    {
		gulong flags = v.flags;
		GtkWidget *w = yp->sndobjadv_position_scale;
		GtkAdjustment *adj = (w != NULL) ? 
		    GTK_RANGE(w)->adjustment : NULL;

		YPLAYER_SNDOBJADV_FREEZE(yp);

		/* Update position */
		if((flags & YPlayValuesFlagPosition) &&
		   (adj != NULL)
		)
		{
		    adj->value = (gfloat)v.position;
		    gtk_signal_emit_by_name(GTK_OBJECT(adj), "value_changed");
		}

		YPLAYER_SNDOBJADV_THAW(yp);

		YPlayerSndObjAdvDisplayDraw(yp);
	    }

	}
}


/*
 *	YPlayer sound object GtkCList "drag_motion" signal callback.
 */
gboolean YPlayerSndObjListDragMotionCB(
	GtkWidget *widget, GdkDragContext *dc, gint x, gint y, guint t,
	gpointer data
)
{
	if(dc == NULL)
	    return(FALSE);

	if(dc->actions & GDK_ACTION_COPY)
	    gdk_drag_status(dc, GDK_ACTION_COPY, t);
	else
	    gdk_drag_status(dc, 0, t);

	return(TRUE);
}

/*
 *	YPlayer sound object GtkCList "drag_data_received" signal
 *	callback.
 */
void YPlayerSndObjListDragDataReceivedCB(
	GtkWidget *widget, GdkDragContext *dc, gint x, gint y,
	GtkSelectionData *selection_data, guint info, guint t,
	gpointer data
)
{
	yplayer_struct *yp = YPLAYER(data);
	YConnection *con = YPLAYER_CONNECTION(yp);
	GtkCList *clist = (GtkCList *)YPLAYER_SNDOBJ_CLIST(yp);
	if((selection_data == NULL) || (clist == NULL))
	    return;

	if((selection_data->data == NULL) ||
	   (selection_data->length <= 0)
	)
	    return;

	if((info == YC_DND_TYPE_INFO_TEXT_PLAIN) ||
	   (info == YC_DND_TYPE_INFO_TEXT_URI_LIST) ||
	   (info == YC_DND_TYPE_INFO_STRING)
	)
	{
	    if(widget == YPLAYER_SNDOBJ_CLIST(yp))
	    {
		gint row, column;
		const gchar *filename = selection_data->data;
		if(!gtk_clist_get_selection_info(
		    clist,
		    x,
		    y - ((clist->flags & GTK_CLIST_SHOW_TITLES) ?
			clist->column_title_area.height +
			clist->column_title_area.y : 0),
		    &row, &column
		))
		{
		    row = -1;
		    column = 0;
		}
		if(row >= clist->rows)
		    row = -1;

		/* Seek filename past protocol prefix as needed */
		if(!strncmp(filename, "file://", 7))
		    filename += 7;

		YPlayerSetBusy(yp, TRUE);

		/* Seek past host name and other info until path portion
		 * is reached.
		 */
		filename = strchr(filename, '/');
		if((con != NULL) && (filename != NULL))
		{
		    YEventSoundObjectAttributes soa;
		    if(YGetSoundObjectAttributes(con, filename, &soa))
		    {
			gchar *buf = g_strdup_printf(
"Unsupported sound object format or unable to find sound object:\n\
\n\
    %s",
			    filename
			);
			CDialogSetTransientFor(YPLAYER_TOPLEVEL(yp));
			CDialogGetResponse(
			    "Sound Object Add Failed",
			    buf,
"Either the sound object you tried to add does not exist or\n\
its format is not supported by the Y server.",
			    CDIALOG_ICON_WARNING,
			    CDIALOG_BTNFLAG_OK | CDIALOG_BTNFLAG_HELP,
			    CDIALOG_BTNFLAG_OK
			);
			CDialogSetTransientFor(NULL);
			g_free(buf);
		    }
		    else
		    {
			gint new_row;

			/* Add row */
			new_row = YPlayerSndObjListAdd(
			    yp, row, filename,
			    soa.format,
			    soa.sample_rate, soa.sample_size, soa.channels,
			    soa.length
			);
/*			gtk_clist_select_row(clist, new_row, 0); */
		    }
		}
		YPlayerUpdateMenus(yp);
		YPlayerSetBusy(yp, FALSE);
	    }
	    else if(widget == YPLAYER_SNDOBJ_PLAYING_CLIST(yp))
	    {

	    }
	    else if(widget == yp->sndobjadv_display_da)
	    {
		const gchar *filename = selection_data->data;

		/* Seek filename past protocol prefix as needed */
		if(!strncmp(filename, "file://", 7))
		    filename += 7;

		YPlayerSetBusy(yp, TRUE);

		/* Seek past host name and other info until path portion
		 * is reached.
		 */
		filename = strchr(filename, '/');
		if((con != NULL) && (filename != NULL))
		    YPlayerDoOpenSoundObjectAdv(yp, filename);

		YPlayerUpdateMenus(yp);
		YPlayerSetBusy(yp, FALSE);
	    }
	    else if(widget == YPLAYER_AUDIOCD_TRACKS_CLIST(yp))
	    {

	    }
	}
}


/*
 *	YPlayer sound object play callback.
 */
void YPlayerSndObjPlayCB(GtkWidget *widget, gpointer data)
{
	gint repeats = 1;
	gboolean change_mode = FALSE;
	gboolean mode_changed = FALSE;
	gint row;
	const gchar *filename;
	GList *glist;
	GtkWidget *w;
	YEventSoundObjectAttributes soa;
	YAudioModeValuesStruct **mode = NULL;
	int total_modes = 0;
	yplayer_struct *yp = YPLAYER(data);
	GtkCList *clist = (GtkCList *)YPLAYER_SNDOBJ_CLIST(yp);
	YConnection *con = YPLAYER_CONNECTION(yp);
	if((con == NULL) || (clist == NULL))
	    return;

	/* Get play options */
	w = yp->sndobj_change_mode_tb;
	if((w != NULL) ? GTK_TOGGLE_BUTTON(w)->active : FALSE)
	    change_mode = TRUE;

	w = yp->sndobj_repeat_tb;
	if((w != NULL) ? GTK_TOGGLE_BUTTON(w)->active : FALSE)
	    repeats = -1;


	YPlayerSetBusy(yp, TRUE);

	/* If we plan to change Audio mode then we should get the
	 * list of Audio preset modes now.
	 */
	if(change_mode)
	    mode = YGetAudioModes(con, &total_modes);


	/* Iterate through all selected rows on sndobj list */
	glist = clist->selection;
	while(glist != NULL)
	{
	    row = (gint)glist->data;
	    filename = (const gchar *)gtk_clist_get_row_data(
		clist, row
	    );
	    /* Get sound object attributes */
	    if(!YGetSoundObjectAttributes(con, filename, &soa))
	    {
		YID yid;
		YEventSoundPlay v;

		/* Change audio mode to match this sound object */
		if(!mode_changed && change_mode &&
		   (soa.format == SndObjTypeDSP)
		)
		{
		    const YAudioModeValuesStruct *m = YCMatchAudioModeByValues(
			mode, total_modes,
			soa.sample_rate, soa.sample_size, soa.channels
		    );
		    if(m != NULL)
		    {
		      if((*m->name != '\0') &&
			 ((yp->sample_rate != m->sample_rate) ||
			  (yp->sample_size != m->sample_size) ||
			  (yp->channels != m->channels) ||
			  (yp->num_fragments != m->num_fragments) ||
			  (yp->fragment_size != m->fragment_size_bytes)
			 )
		      )
		      {
			if(!YChangeAudioModePreset(con, m->name))
			{
			    g_free(yp->audio_mode_name);
			    yp->audio_mode_name = g_strdup(m->name);
			    yp->sample_rate = m->sample_rate;
			    yp->sample_size = m->sample_size;
			    yp->channels = m->channels;
			    yp->num_fragments = m->num_fragments;
			    yp->fragment_size = m->fragment_size_bytes;
			    YPLAYER_SET_CONBAR_AUDIO(yp);

			    /* Mark that we have changed the Audio mode
			     * so that we do not change it again in this
			     * loop.
			     */
			    mode_changed = TRUE;
			}
		      }
		    }
		}

		/* Set up sound object playing values */
		v.flags =	YPlayValuesFlagPosition |
				YPlayValuesFlagTotalRepeats |
				YPlayValuesFlagVolume;
		v.position = 0;
		v.total_repeats = repeats;
		v.left_volume = 1.0;
		v.right_volume = 1.0;
		v.back_left_volume = 1.0;
		v.back_right_volume = 1.0;

		/* Start playing */
		yid = YStartPlaySoundObject(con, filename, &v);

		/* Succesfully playing? */
		if(yid != YIDNULL)
		{
		    gint new_row = YPlayerSndObjPlayingListAdd(
			yp,
			-1,		/* Always append */
			filename,
			soa.format,
			yid
		    );
		    if(new_row > -1)
		        gtk_clist_select_row(
			    (GtkCList *)YPLAYER_SNDOBJ_PLAYING_CLIST(yp),
			    new_row, 0
		        );
		}
	    }
	    glist = glist->next;
	}

	/* Delete modes list */
	YFreeAudioModesList(mode, total_modes);

	YPlayerSetBusy(yp, FALSE);

	YPlayerUpdateMenus(yp);
}

/*
 *	YPlayer sound object stop callback.
 */
void YPlayerSndObjStopCB(GtkWidget *widget, gpointer data)
{
	yplayer_struct *yp = YPLAYER(data);
	GtkCList *clist = (GtkCList *)YPLAYER_SNDOBJ_PLAYING_CLIST(yp);
	if(clist == NULL)
	    return;

	YPlayerSetBusy(yp, TRUE);

	/* Begin removing rows, when the row is removed the row data
	 * destroy callback will stop the sound object from playing
	 * and delete the row data.
	 */
	while(clist->selection_end != NULL)
	    gtk_clist_remove(clist, (gint)clist->selection_end->data);

	YPlayerSetBusy(yp, FALSE);

	YPlayerUpdateMenus(yp);
}

/*
 *	YPlayer sound object add signal callback.
 */
void YPlayerSndObjAddCB(GtkWidget *widget, gpointer data)
{
	gboolean status;
	gint row;
	const gchar *s;
	gchar *last_path, *filename = NULL;
	GList *glist;
	fb_type_struct **ftype = NULL, *ftype_rtn = NULL;
	gint total_ftypes = 0;
	gchar **path_rtn = NULL;
	gint total_path_rtns = 0;
	YEventSoundObjectAttributes soa;
	yplayer_struct *yp = YPLAYER(data);
	GtkCList *clist = (GtkCList *)YPLAYER_SNDOBJ_CLIST(yp);
	YConnection *con = YPLAYER_CONNECTION(yp);
	if((con == NULL) || (clist == NULL))
	    return;

	if(FileBrowserIsQuery())
	    return;

	YPlayerSetBusy(yp, TRUE);

	s = YC_GET_CFG_S(
	    YC_CORE(yp->core_ptr),
	    YC_CFG_PARM_PLAYER_SNDOBJ_LAST_PATH
	);
	last_path = (s != NULL) ? g_strdup(s) : NULL;

	FileBrowserTypeListNew(
	    &ftype, &total_ftypes,
	    ".wav .wave", "Microsoft Wave files"
	);
	FileBrowserTypeListNew(
	    &ftype, &total_ftypes,
	    ".voc", "Sound Blaster VOC files"
	);
	FileBrowserTypeListNew(
	    &ftype, &total_ftypes,
	    ".raw", "Raw sound files"
	);
	FileBrowserTypeListNew(
	    &ftype, &total_ftypes,
	    ".mid .midi", "Musical Instrument Data Interface files"
	);
	FileBrowserTypeListNew(
	    &ftype, &total_ftypes,
	    "*.*", "All files"
	);

	/* Query user for sound object */
	FileBrowserSetTransientFor(YPLAYER_TOPLEVEL(yp));
	status = FileBrowserGetResponse(
	    "Add Sound Object",
	    "Add", "Cancel",
	    last_path,
	    ftype, total_ftypes,
	    &path_rtn, &total_path_rtns,
	    &ftype_rtn
	);
	FileBrowserSetTransientFor(NULL);
	FileBrowserDeleteTypeList(ftype, total_ftypes);

	/* Get response as filename */
	if(status)
	{
	    gint i;

	    /* Get selected row */
	    glist = clist->selection_end;
	    row = (glist != NULL) ? (gint)glist->data : -1;

	    for(i = 0; i < total_path_rtns; i++)
	    {
		s = path_rtn[i];
		if(s == NULL)
		    continue;

		g_free(filename);
		filename = g_strdup(s);

		/* Get sound object attributes */
		if(YGetSoundObjectAttributes(con, filename, &soa))
		{
		    gchar *buf = g_strdup_printf(
"Unsupported sound object format or unable to find sound object:\n\
\n\
    %s",
			filename
		    );
		    CDialogSetTransientFor(YPLAYER_TOPLEVEL(yp));
		    CDialogGetResponse(
			"Sound Object Add Failed",
			buf,
"Either the sound object you tried to add does not exist or\n\
its format is not supported by the Y server.",
			CDIALOG_ICON_WARNING,
			CDIALOG_BTNFLAG_OK | CDIALOG_BTNFLAG_HELP,
			CDIALOG_BTNFLAG_OK
		    );
		    CDialogSetTransientFor(NULL);
		    g_free(buf);
		}
		else
		{
		    gint new_row;

		    /* Add row */
		    new_row = YPlayerSndObjListAdd(
			yp, row, filename,
			soa.format,
			soa.sample_rate, soa.sample_size, soa.channels,
			soa.length
		    );
		    if(new_row > -1)
		    {
/*			gtk_clist_select_row(clist, new_row, 0); */
			if(row > -1)
			    row++;	/* Increment row position for next add */
		    }
		}
	    }
	}

	/* Update last path */
	YC_SET_CFG_S(
	    YC_CORE(yp->core_ptr),
	    YC_CFG_PARM_PLAYER_SNDOBJ_LAST_PATH,
	    GetParentDir(filename)
	);

	YPlayerSetBusy(yp, FALSE);

	g_free(last_path);
	g_free(filename);
}

/*
 *	YPlayer sound object remove signal callback.
 */
void YPlayerSndObjRemoveCB(GtkWidget *widget, gpointer data)
{
	gint i;
	yplayer_struct *yp = YPLAYER(data);
	GtkCList *clist = (GtkCList *)YPLAYER_SNDOBJ_CLIST(yp);
	if(clist == NULL)
	    return;

	if(clist->selection == NULL)
	    return;

	YPlayerSetBusy(yp, TRUE);

	i = (gint)g_list_length(clist->selection);
	if(i > 0)
	{
	    gint response;
	    gchar *buf, *s2, *s = (gchar *)gtk_clist_get_row_data(
		clist, (gint)clist->selection->data
	    );
	    s2 = (s != NULL) ? strrchr(s, '/') : NULL;
	    s = g_strdup((s2 != NULL) ? (s2 + 1) : "");
	    s2 = strrchr(s, '.');
	    if(s2 != NULL)
		*s2 = '\0';

	    if(i == 1)
		buf = g_strdup_printf(
		    "Remove \"%s\" from the playlist?", s
		);
	    else
		buf = g_strdup_printf(
		    "Remove %i items from the playlist?", i
		);
	    CDialogSetTransientFor(YPLAYER_TOPLEVEL(yp));
	    response = CDialogGetResponse(
		"Confirm Remove",
		buf,
		NULL,
		CDIALOG_ICON_WARNING,
		CDIALOG_BTNFLAG_YES | CDIALOG_BTNFLAG_NO,
		CDIALOG_BTNFLAG_NO
	    );
	    CDialogSetTransientFor(NULL);
	    g_free(buf);
	    g_free(s);
	    if(response != CDIALOG_RESPONSE_YES)
	    {
		YPlayerSetBusy(yp, FALSE);
		return;
	    }
	}

	/* Begin removing rows */
	while(clist->selection_end != NULL)
	    gtk_clist_remove(clist, (gint)clist->selection_end->data);

	YPlayerSetBusy(yp, FALSE);

	YPlayerUpdateMenus(yp);
}

/*
 *      YPlayer sound object shift down callback.
 */
void YPlayerSndObjShiftUpCB(GtkWidget *widget, gpointer data)
{
	gint row;
	GList *glist;
	yplayer_struct *yp = YPLAYER(data);
	GtkCList *clist = (GtkCList *)YPLAYER_SNDOBJ_CLIST(yp);
	if(clist == NULL)
	    return;

	glist = clist->selection_end;
	row = (glist != NULL) ? (gint)glist->data : -1;
	if((row < 0) || (row < 1))
	    return;

	gtk_clist_freeze(clist);
	gtk_clist_swap_rows(clist, row, row - 1);
	gtk_clist_thaw(clist);

	YPlayerUpdateMenus(yp);
}

/*
 *	YPlayer sound object shift down callback.
 */
void YPlayerSndObjShiftDownCB(GtkWidget *widget, gpointer data)
{
	gint row;
	GList *glist;
	yplayer_struct *yp = YPLAYER(data);
	GtkCList *clist = (GtkCList *)YPLAYER_SNDOBJ_CLIST(yp);
	if(clist == NULL)
	    return;

	glist = clist->selection_end;
	row = (glist != NULL) ? (gint)glist->data : -1;
	if((row < 0) || (row >= (clist->rows - 1)))
	    return;

	gtk_clist_freeze(clist);
	gtk_clist_swap_rows(clist, row, row + 1);
	gtk_clist_thaw(clist);

	YPlayerUpdateMenus(yp);
}

/*
 *	YPlayer sound object change Audio mode on play "toggle"
 *	signal callback.
 */
void YPlayerSndObjChangeModeToggleCB(GtkWidget *widget, gpointer data)
{
	gboolean active = FALSE;
	GtkWidget *w;
	yplayer_struct *yp = YPLAYER(data);
	if((widget == NULL) || (yp == NULL))
	    return;

	/* Get active state */
	if(GTK_IS_CHECK_MENU_ITEM(widget))
	    active = GTK_CHECK_MENU_ITEM(widget)->active;
	else if(GTK_IS_TOGGLE_BUTTON(widget))
	    active = GTK_TOGGLE_BUTTON(widget)->active;

	/* Update other widgets as needed and if their active states
	 * differ.
	 */
	w = yp->sndobj_change_mode_tb;
	if((w != NULL) && (w != widget) &&
	   (GTK_TOGGLE_BUTTON(w)->active != active)
	)
	    gtk_toggle_button_set_active(
		GTK_TOGGLE_BUTTON(w), active
	    );

	w = yp->sndobj_change_mode_micheck;
	if((w != NULL) && (w != widget) &&
	   (GTK_CHECK_MENU_ITEM(w)->active != active)
	)
	    gtk_check_menu_item_set_active(
		GTK_CHECK_MENU_ITEM(w), active
	    );
}

/*
 *	YPlayer sound object repeat "toggle" signal callback.
 */
void YPlayerSndObjRepeatToggleCB(GtkWidget *widget, gpointer data)
{
	gboolean active = FALSE;
	GtkWidget *w;
	yplayer_struct *yp = YPLAYER(data);
	if((widget == NULL) || (yp == NULL))
	    return;

	/* Get active state */
	if(GTK_IS_CHECK_MENU_ITEM(widget))
	    active = GTK_CHECK_MENU_ITEM(widget)->active;
	else if(GTK_IS_TOGGLE_BUTTON(widget))
	    active = GTK_TOGGLE_BUTTON(widget)->active;

	/* Update other widgets as needed and if their active states
	 * differ.
	 */
	w = yp->sndobj_repeat_tb;
	if((w != NULL) && (w != widget) &&
	   (GTK_TOGGLE_BUTTON(w)->active != active)
	)
	    gtk_toggle_button_set_active(
		GTK_TOGGLE_BUTTON(w), active
	    );

	w = yp->sndobj_repeat_micheck;
	if((w != NULL) && (w != widget) &&
	   (GTK_CHECK_MENU_ITEM(w)->active != active)
	)
	    gtk_check_menu_item_set_active(
		GTK_CHECK_MENU_ITEM(w), active
	    );
}

/*
 *      YPlayer sound object select all selection callback.
 */
void YPlayerSndObjSelectAllCB(GtkWidget *widget, gpointer data)
{
	yplayer_struct *yp = YPLAYER(data);
	GtkCList *clist = (GtkCList *)YPLAYER_SNDOBJ_CLIST(yp);
	if(clist == NULL)
	    return;

	YPlayerSetBusy(yp, TRUE);

	gtk_clist_freeze(clist);
	gtk_clist_select_all(clist);
	gtk_clist_thaw(clist);

	YPlayerSetBusy(yp, FALSE);
}

/*
 *      YPlayer sound object unselect all callback.
 */
void YPlayerSndObjUnselectAllCB(GtkWidget *widget, gpointer data)
{
	yplayer_struct *yp = YPLAYER(data);
	GtkCList *clist = (GtkCList *)YPLAYER_SNDOBJ_CLIST(yp);
	if(clist == NULL)
	    return;

	YPlayerSetBusy(yp, TRUE);

	gtk_clist_freeze(clist);
	gtk_clist_unselect_all(clist);
	gtk_clist_thaw(clist);

	YPlayerSetBusy(yp, FALSE);
}

/*
 *	YPlayer sound object invert selection callback.
 */
void YPlayerSndObjInvertSelectionCB(GtkWidget *widget, gpointer data)
{
	gint i;
	GList *glist;
	yplayer_struct *yp = YPLAYER(data);
	GtkCList *clist = (GtkCList *)YPLAYER_SNDOBJ_CLIST(yp);
	if(clist == NULL)
	    return;

	YPlayerSetBusy(yp, TRUE);

	/* Get copy of selected rows (or NULL if none are selected) */
	glist = clist->selection;
	glist = (glist != NULL) ? g_list_copy(glist) : NULL;

	/* Iterate through all rows, inverting their selection */
	gtk_clist_freeze(clist);
	for(i = 0; i < clist->rows; i++)
	{
	    /* This row selected? */
	    if((glist != NULL) ?
		(g_list_find(glist, (gpointer)i) != NULL) : FALSE
	    )
		gtk_clist_unselect_row(clist, i, 0);
	    else
		gtk_clist_select_row(clist, i, 0);
	}
	gtk_clist_thaw(clist);

	/* Delete coppied selected rows list */
	if(glist != NULL)
	{
	    g_list_free(glist);
	    glist = NULL;
	}

	YPlayerSetBusy(yp, FALSE);
}

/*
 *	YPlayer sound object open play list callback.
 */
void YPlayerSndObjOpenPlayListCB(GtkWidget *widget, gpointer data)
{
	gboolean status;
	const gchar *s;
	gchar *last_path, *filename;
	fb_type_struct **ftype = NULL, *ftype_rtn = NULL;
	gint total_ftypes = 0;
	gchar **path_rtn = NULL;
	gint total_path_rtns = 0;
	yplayer_struct *yp = YPLAYER(data);
	if(yp == NULL)
	    return;

	if(FileBrowserIsQuery())
	    return;

	YPlayerSetBusy(yp, TRUE);

	s = YC_GET_CFG_S(
	    YC_CORE(yp->core_ptr),
	    YC_CFG_PARM_PLAYER_SNDOBJ_PLAYLIST_LAST_PATH
	);
	last_path = (s != NULL) ? g_strdup(s) : NULL;

	FileBrowserTypeListNew(
	    &ftype, &total_ftypes,
	    ".ini", "Play List files"
	);
	FileBrowserTypeListNew(
	    &ftype, &total_ftypes,
	    "*.*", "All files"
	);

	/* Query user for play list */
	FileBrowserSetTransientFor(YPLAYER_TOPLEVEL(yp));
	status = FileBrowserGetResponse(
	    "Open PlayList",
	    "Open", "Cancel",
	    last_path,
	    ftype, total_ftypes,
	    &path_rtn, &total_path_rtns,
	    &ftype_rtn
	);
	FileBrowserSetTransientFor(NULL);
	FileBrowserDeleteTypeList(ftype, total_ftypes);

	/* Get response as filename */
	s = (status && (total_path_rtns > 0)) ?
	    path_rtn[0] : NULL;
	filename = (s != NULL) ? g_strdup(s) : NULL;
	if(filename == NULL)
	{
	    YPlayerSetBusy(yp, FALSE);
	    g_free(last_path);
	    return;
	}

	/* Update last path */
	YC_SET_CFG_S(
	    YC_CORE(yp->core_ptr),
	    YC_CFG_PARM_PLAYER_SNDOBJ_PLAYLIST_LAST_PATH,
	    GetParentDir(filename)
	);

	/* Open play list */
	YPlayerLoadPlayListFile(yp, filename, FALSE);

	YPlayerSetBusy(yp, FALSE);

	g_free(last_path);
	g_free(filename);
}

/*
 *	YPlayer sound object insert playlist callback.
 */
void YPlayerSndObjInsertPlayListCB(GtkWidget *widget, gpointer data)
{
	gboolean status;
	const gchar *s;
	gchar *last_path, *filename;
	fb_type_struct **ftype = NULL, *ftype_rtn = NULL;
	gint total_ftypes = 0;
	gchar **path_rtn = NULL;
	gint total_path_rtns = 0;
	yplayer_struct *yp = YPLAYER(data);
	if(yp == NULL)
	    return;

	if(FileBrowserIsQuery())
	    return;

	YPlayerSetBusy(yp, TRUE);

	s = YC_GET_CFG_S(
	    YC_CORE(yp->core_ptr),
	    YC_CFG_PARM_PLAYER_SNDOBJ_PLAYLIST_LAST_PATH
	);
	last_path = (s != NULL) ? g_strdup(s) : NULL;

	FileBrowserTypeListNew(
	    &ftype, &total_ftypes,
	    ".ini", "Play List files"
	);
	FileBrowserTypeListNew(
	    &ftype, &total_ftypes,
	    "*.*", "All files"
	);

	/* Query user for play list */
	FileBrowserSetTransientFor(YPLAYER_TOPLEVEL(yp));
	status = FileBrowserGetResponse(
	    "Insert PlayList",
	    "Insert", "Cancel",
	    last_path,
	    ftype, total_ftypes,
	    &path_rtn, &total_path_rtns,
	    &ftype_rtn
	);
	FileBrowserSetTransientFor(NULL);
	FileBrowserDeleteTypeList(ftype, total_ftypes);

	/* Get response as filename */
	s = (status && (total_path_rtns > 0)) ?
	    path_rtn[0] : NULL;
	filename = (s != NULL) ? g_strdup(s) : NULL;
	if(filename == NULL)
	{
	    YPlayerSetBusy(yp, FALSE);
	    g_free(last_path);
	    return;
	}

	/* Update last path */
	YC_SET_CFG_S(
	    YC_CORE(yp->core_ptr),
	    YC_CFG_PARM_PLAYER_SNDOBJ_PLAYLIST_LAST_PATH,
	    GetParentDir(filename)
	);

	/* Insert play list */
	YPlayerLoadPlayListFile(yp, filename, TRUE);

	YPlayerSetBusy(yp, FALSE);

	g_free(last_path);
	g_free(filename);
}

/*
 *	YPlayer sound object save play list callback.
 */
void YPlayerSndObjSavePlayListCB(GtkWidget *widget, gpointer data)
{
	gboolean status;
	const gchar *s;
	gchar *last_path, *filename;
	fb_type_struct **ftype = NULL, *ftype_rtn = NULL;
	gint total_ftypes = 0;
	gchar **path_rtn = NULL;
	gint total_path_rtns = 0;
	struct stat stat_buf;
	yplayer_struct *yp = YPLAYER(data);
	if(yp == NULL)
	    return;

	if(FileBrowserIsQuery())
	    return;

	YPlayerSetBusy(yp, TRUE);

	s = YC_GET_CFG_S(
	    YC_CORE(yp->core_ptr),
	    YC_CFG_PARM_PLAYER_SNDOBJ_PLAYLIST_LAST_PATH
	);
	last_path = (s != NULL) ? g_strdup(s) : NULL;

	FileBrowserTypeListNew(
	    &ftype, &total_ftypes,
	    ".ini", "Play List files"
	);
	FileBrowserTypeListNew(
	    &ftype, &total_ftypes,
	    "*.*", "All files"
	);

	/* Query user for play list */
	FileBrowserSetTransientFor(YPLAYER_TOPLEVEL(yp));
	status = FileBrowserGetResponse(
	    "Save PlayList",
	    "Save", "Cancel",
	    last_path,
	    ftype, total_ftypes,
	    &path_rtn, &total_path_rtns,
	    &ftype_rtn
	);
	FileBrowserSetTransientFor(NULL);
	FileBrowserDeleteTypeList(ftype, total_ftypes);

	/* Get response as filename */
	s = (status && (total_path_rtns > 0)) ?
	    path_rtn[0] : NULL;
	filename = (s != NULL) ? g_strdup(s) : NULL;
	if(filename == NULL)
	{
	    YPlayerSetBusy(yp, FALSE);
	    g_free(last_path);
	    return;
	}

	/* Update last path */
	YC_SET_CFG_S(
	    YC_CORE(yp->core_ptr),
	    YC_CFG_PARM_PLAYER_SNDOBJ_PLAYLIST_LAST_PATH,
	    GetParentDir(filename)
	);

	/* Check if file exists and confirm overwrite */
	if(!lstat(filename, &stat_buf))
	{
	    gint response;
	    gchar *buf = g_strdup_printf(
"Overwrite existing file:\n\
\n\
    %s",
		filename
	    );
	    CDialogSetTransientFor(YPLAYER_TOPLEVEL(yp));
	    response = CDialogGetResponse(
		"Confirm Overwrite",
		buf,
		NULL,
		CDIALOG_ICON_WARNING,
		CDIALOG_BTNFLAG_YES | CDIALOG_BTNFLAG_NO,
		CDIALOG_BTNFLAG_NO
	    );
	    CDialogSetTransientFor(NULL);
	    g_free(buf);
	    if(response != CDIALOG_RESPONSE_YES)
	    {
		YPlayerSetBusy(yp, FALSE);
		g_free(last_path);
		g_free(filename);
		return;
	    }
	}

	/* Save play list */
	YPlayerSavePlayListFile(yp, filename, FALSE);

	YPlayerSetBusy(yp, FALSE);

	g_free(last_path);
	g_free(filename);
}

/*
 *      YPlayer sound object save selected callback.
 */
void YPlayerSndObjSaveSelectedCB(GtkWidget *widget, gpointer data)
{
	gboolean status;
	const gchar *s;
	gchar *last_path, *filename;
	fb_type_struct **ftype = NULL, *ftype_rtn = NULL;
	gint total_ftypes = 0;
	gchar **path_rtn = NULL;
	gint total_path_rtns = 0;
	struct stat stat_buf;
	yplayer_struct *yp = YPLAYER(data);
	if(yp == NULL)
	    return;

	YPlayerSetBusy(yp, TRUE);

	if(FileBrowserIsQuery())
	    return;

	s = YC_GET_CFG_S(
	    YC_CORE(yp->core_ptr),
	    YC_CFG_PARM_PLAYER_SNDOBJ_PLAYLIST_LAST_PATH
	);
	last_path = (s != NULL) ? g_strdup(s) : NULL;

	FileBrowserTypeListNew(
	    &ftype, &total_ftypes,
	    ".ini", "Play List files"
	);
	FileBrowserTypeListNew(
	    &ftype, &total_ftypes,
	    "*.*", "All files"
	);

	/* Query user for play list */
	FileBrowserSetTransientFor(YPLAYER_TOPLEVEL(yp));
	status = FileBrowserGetResponse(
	    "Save Selected To PlayList",
	    "Save", "Cancel",
	    last_path,
	    ftype, total_ftypes,
	    &path_rtn, &total_path_rtns,
	    &ftype_rtn
	);
	FileBrowserSetTransientFor(NULL);
	FileBrowserDeleteTypeList(ftype, total_ftypes);

	/* Get response as filename */
	s = (status && (total_path_rtns > 0)) ?
	    path_rtn[0] : NULL;
	filename = (s != NULL) ? g_strdup(s) : NULL;
	if(filename == NULL)
	{
	    YPlayerSetBusy(yp, FALSE);
	    g_free(last_path);
	    return;
	}

	/* Update last path */
	YC_SET_CFG_S(
	    YC_CORE(yp->core_ptr),
	    YC_CFG_PARM_PLAYER_SNDOBJ_PLAYLIST_LAST_PATH,
	    GetParentDir(filename)
	);

	/* Check if file exists and confirm overwrite */
	if(!lstat(filename, &stat_buf))
	{
	    gint response;
	    gchar *buf = g_strdup_printf(
"Overwrite existing file:\n\
\n\
    %s",
		filename
	    );
	    CDialogSetTransientFor(YPLAYER_TOPLEVEL(yp));
	    response = CDialogGetResponse(
		"Confirm Overwrite",
		buf,
		NULL,
		CDIALOG_ICON_WARNING,
		CDIALOG_BTNFLAG_YES | CDIALOG_BTNFLAG_NO,
		CDIALOG_BTNFLAG_NO
	    );
	    CDialogSetTransientFor(NULL);
	    g_free(buf);
	    if(response != CDIALOG_RESPONSE_YES)
	    {
		YPlayerSetBusy(yp, FALSE);
		g_free(last_path);
		g_free(filename);
		return;
	    }
	}

	/* Save selected to play list */
	YPlayerSavePlayListFile(yp, filename, TRUE);

	YPlayerSetBusy(yp, FALSE);

	g_free(last_path);
	g_free(filename);
}

/*
 *	YPlayer sound object killed callback.
 */
void YPlayerSndObjPlayingKilledCB(yplayer_struct *yp, YID yid)
{
	gint i;
	yplayer_sndobj_playing_struct *sop;
	GtkCList *clist = (GtkCList *)YPLAYER_SNDOBJ_PLAYING_CLIST(yp);
	if((yid == YIDNULL) || (clist == NULL))
	    return;

	for(i = 0; i < clist->rows; i++)
	{
	    sop = YPLAYER_SNDOBJ_PLAYING(
		gtk_clist_get_row_data(clist, i)
	    );
	    /* Given YID matches the sound object referenced by this
	     * sop structure?
	     */
	    if((sop != NULL) ? (sop->yid == yid) : FALSE)
	    {
		/* Mark that the YID on the sop structure that it has
		 * been killed.
		 */
		sop->yid = YIDNULL;

		/* Remove this row and the row data, this will delete
		 * the sop structure so it should not be referenced
		 * again.
		 */
		gtk_clist_remove(clist, i);
		sop = NULL;

		/* No need to check for duplicates */
		break;
	    }
	}
}


/*
 *	YPlayer sound object advanced event callback.
 *
 *	For the display and volume GtkDrawingArea widgets.
 */
gint YPlayerSndObjAdvEventCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
)
{
	gint status = FALSE;
	GdkWindow *window;
	GdkEventConfigure *configure;
	GdkEventFocus *focus;
	GdkEventKey *key;
	GdkEventButton *button;
	GdkEventMotion *motion;
	yplayer_struct *yp = YPLAYER(data);
	if((widget == NULL) || (event == NULL) || (yp == NULL))
	    return(status);

	window = widget->window;

	switch((gint)event->type)
	{
	  case GDK_CONFIGURE:
	    configure = (GdkEventConfigure *)event;
	    if(widget == yp->sndobjadv_display_da)
	    {
		GDK_PIXMAP_UNREF(yp->sndobjadv_display_pm)
		yp->sndobjadv_display_pm = GDK_PIXMAP_NEW(
		    configure->width, configure->height
		);
		gdk_window_set_back_pixmap(window, yp->sndobjadv_display_pm, FALSE);
		YPlayerSndObjAdvDisplayDraw(yp);
		status = TRUE;
	    }
	    else if(widget == yp->sndobjadv_volume_da)
	    {
		GDK_PIXMAP_UNREF(yp->sndobjadv_volume_pm)
		yp->sndobjadv_volume_pm = GDK_PIXMAP_NEW(
		    configure->width, configure->height
		);
		gdk_window_set_back_pixmap(window, yp->sndobjadv_volume_pm, FALSE);
		YPlayerSndObjAdvVolumeDraw(yp);
		status = TRUE;
	    }
	    break;

	  case GDK_EXPOSE:
	    gdk_window_clear(window);
	    status = TRUE;
	    break;

	  case GDK_FOCUS_CHANGE:
	    focus = (GdkEventFocus *)event;
	    if(focus->in)
		GTK_WIDGET_SET_FLAGS(widget, GTK_HAS_FOCUS);
	    else
		GTK_WIDGET_UNSET_FLAGS(widget, GTK_HAS_FOCUS);
	    if(widget == yp->sndobjadv_display_da)
		YPlayerSndObjAdvDisplayDraw(yp);
	    else if(widget == yp->sndobjadv_volume_da)
		YPlayerSndObjAdvVolumeDraw(yp);
	    status = TRUE;
	    break;

	  case GDK_KEY_PRESS:
	  case GDK_KEY_RELEASE:
	    key = (GdkEventKey *)event;
#define DO_STOP_KEY_SIGNAL      \
{ if(widget != NULL) \
 gtk_signal_emit_stop_by_name( \
  GTK_OBJECT(widget), \
  (keystate) ? "key_press_event" : "key_release_event" \
 ); }
	    if(widget == yp->sndobjadv_display_da)
	    {
		gboolean keystate = (event->type == GDK_KEY_PRESS) ? TRUE : FALSE;
		GtkRange *range = (GtkRange *)yp->sndobjadv_position_scale;
		GtkAdjustment *adj = (range != NULL) ? range->adjustment : NULL;

#define EMIT_SIGNAL_POSITION_CHANGE    \
{ if(adj != NULL) { \
 if(adj->value > (adj->upper - adj->page_size)) \
  adj->value = adj->upper - adj->page_size; \
 if(adj->value < adj->lower) \
  adj->value = adj->lower; \
 gtk_signal_emit_by_name(GTK_OBJECT(adj), "value_changed"); \
} }

		switch(key->keyval)
		{
		  case GDK_Return:
		  case GDK_KP_Enter:
		  case GDK_3270_Enter:
		  case GDK_ISO_Enter:
		  case GDK_space:
		    if(keystate)
		    {
			if(!YPLAYER_SNDOBJADV_IS_OPENED(yp))
			    YPlayerSndObjAdvOpenCB(
				yp->sndobjadv_open_btn, yp
			    );
			else if(YPLAYER_SNDOBJADV_IS_PLAYING(yp))
			    YPlayerSndObjAdvPauseCB(
				yp->sndobjadv_pause_btn, yp
			    );
			else
			    YPlayerSndObjAdvPlayCB(
				yp->sndobjadv_play_btn, yp
			    );
		    }
		    status = TRUE;
		    break;

		  case GDK_Escape:
		    if(keystate)
		    {
			if(YPLAYER_SNDOBJADV_IS_PLAYING(yp))
			    YPlayerSndObjAdvPauseCB(
				yp->sndobjadv_play_btn, yp
			    );
		    }
		    status = TRUE;
		    break;

		  case GDK_Insert:
		  case GDK_KP_Insert:
		    if(keystate)
		    {
			YPlayerSndObjAdvOpenCB(
			    yp->sndobjadv_open_btn, yp
			);
		    }
		    status = TRUE;
		    break;

		  case GDK_Delete:
		  case GDK_KP_Delete:
		    if(keystate)
		    {
			YPlayerSndObjAdvCloseCB(
			    yp->sndobjadv_close_btn, yp
			);
		    }
		    status = TRUE;
		    break;

		  case GDK_Left:
		  case GDK_KP_Left:
		  case '-': case '_': case ',': case '<':
		    if(keystate && (adj != NULL))
		    {
			if(key->state & GDK_SHIFT_MASK)
			    adj->value -= adj->page_increment;
			else
			    adj->value -= adj->step_increment;
			EMIT_SIGNAL_POSITION_CHANGE
		    }
		    DO_STOP_KEY_SIGNAL
		    status = TRUE;
		    break;

		  case GDK_Page_Up:
		  case GDK_KP_Page_Up:
		    if(keystate && (adj != NULL))
		    {
			adj->value -= adj->page_increment;
			EMIT_SIGNAL_POSITION_CHANGE
		    }
		    status = TRUE;
		    break;

		  case GDK_Right:
		  case GDK_KP_Right:
		    case '=': case '+': case '.': case '>':
		    if(keystate && (adj != NULL))
		    {
			if(key->state & GDK_SHIFT_MASK)
			    adj->value += adj->page_increment;
			else
			    adj->value += adj->step_increment;
			EMIT_SIGNAL_POSITION_CHANGE
		    }
		    DO_STOP_KEY_SIGNAL
		    status = TRUE;
		    break;

		  case GDK_Page_Down:
		  case GDK_KP_Page_Down:
		    if(keystate && (adj != NULL))
		    {
			adj->value += adj->page_increment;
			EMIT_SIGNAL_POSITION_CHANGE
		    }
		    status = TRUE;
		    break;

		  case GDK_Home:
		  case GDK_KP_Home:
		    if(keystate && (adj != NULL))
		    {
			adj->value = adj->lower;
			EMIT_SIGNAL_POSITION_CHANGE
		    }
		    status = TRUE;
		    break;

		  case GDK_End:
		  case GDK_KP_End:
		    if(keystate && (adj != NULL))
		    {
			adj->value = adj->upper - adj->page_size;
			EMIT_SIGNAL_POSITION_CHANGE
		    }
		    status = TRUE;
		    break;

		  case '1':
		    if(keystate && (adj != NULL))
		    {
			adj->value = (adj->upper - adj->lower) * 0.0;
			EMIT_SIGNAL_POSITION_CHANGE
		    }
		    status = TRUE;
		    break;

		  case '2':
		    if(keystate && (adj != NULL))
		    {
			adj->value = (adj->upper - adj->lower) * 0.1;
			EMIT_SIGNAL_POSITION_CHANGE
		    }
		    status = TRUE;
		    break;

		  case '3':
		    if(keystate && (adj != NULL))
		    {
			adj->value = (adj->upper - adj->lower) * 0.2;
			EMIT_SIGNAL_POSITION_CHANGE
		    }
		    status = TRUE;
		    break;

		  case '4':
		    if(keystate && (adj != NULL))
		    {
			adj->value = (adj->upper - adj->lower) * 0.3;
			EMIT_SIGNAL_POSITION_CHANGE
		    }
		    status = TRUE;
		    break;

		  case '5':
		    if(keystate && (adj != NULL))
		    {
			adj->value = (adj->upper - adj->lower) * 0.4;
			EMIT_SIGNAL_POSITION_CHANGE
		    }
		    status = TRUE;
		    break;

		  case '6':
		    if(keystate && (adj != NULL))
		    {
			adj->value = (adj->upper - adj->lower) * 0.5;
			EMIT_SIGNAL_POSITION_CHANGE
		    }
		    status = TRUE;
		    break;

		  case '7':
		    if(keystate && (adj != NULL))
		    {
			adj->value = (adj->upper - adj->lower) * 0.6;
			EMIT_SIGNAL_POSITION_CHANGE
		    }
		    status = TRUE;
		    break;

		  case '8':
		    if(keystate && (adj != NULL))
		    {
			adj->value = (adj->upper - adj->lower) * 0.7;
			EMIT_SIGNAL_POSITION_CHANGE
		    }
		    status = TRUE;
		    break;

		  case '9':
		    if(keystate && (adj != NULL))
		    {
			adj->value = (adj->upper - adj->lower) * 0.8;
			EMIT_SIGNAL_POSITION_CHANGE
		    }
		    status = TRUE;
		    break;

		  case '0':
		    if(keystate && (adj != NULL))
		    {
			adj->value = (adj->upper - adj->lower) * 0.9;
			EMIT_SIGNAL_POSITION_CHANGE
		    }
		    status = TRUE;
		    break;
		}
#undef EMIT_SIGNAL_POSITION_CHANGE
	    }
	    else if(widget == yp->sndobjadv_volume_da)
	    {
		gboolean keystate = (event->type == GDK_KEY_PRESS) ? TRUE : FALSE;
		GtkRange *range = (GtkRange *)yp->sndobjadv_volume_scale;
		GtkAdjustment *adj = (range != NULL) ? range->adjustment : NULL;

#define SIGNAL_VOLUME_CHANGE	\
 YPlayerSndObjAdvVolumeChangedCB(adj, yp);

		switch(key->keyval)
		{
		  case GDK_Up:
		  case GDK_KP_Up:
		    if(keystate)
		    {
			yp->sndobjadv_volume_y += 0.1f;
			if(yp->sndobjadv_volume_y > 1.0f)
			    yp->sndobjadv_volume_y = 1.0f;
			SIGNAL_VOLUME_CHANGE
		    }
		    DO_STOP_KEY_SIGNAL
		    status = TRUE;
		    break;

		  case GDK_Down:
		  case GDK_KP_Down:
		    if(keystate)
		    {
			yp->sndobjadv_volume_y -= 0.1f;
			if(yp->sndobjadv_volume_y < -1.0f)
			    yp->sndobjadv_volume_y = -1.0f;
			SIGNAL_VOLUME_CHANGE
		    }
		    DO_STOP_KEY_SIGNAL
		    status = TRUE;
		    break;

		  case GDK_Left:
		  case GDK_KP_Left:
		    if(keystate)
		    {
			yp->sndobjadv_volume_x -= 0.1f;
			if(yp->sndobjadv_volume_x < -1.0f)
			    yp->sndobjadv_volume_x = -1.0f;
			SIGNAL_VOLUME_CHANGE
		    }
		    DO_STOP_KEY_SIGNAL
		    status = TRUE;
		    break;

		  case GDK_Right:
		  case GDK_KP_Right:
		    if(keystate)
		    {
			yp->sndobjadv_volume_x += 0.1f;
			if(yp->sndobjadv_volume_x > 1.0f)
			    yp->sndobjadv_volume_x = 1.0f;
			SIGNAL_VOLUME_CHANGE
		    }
		    DO_STOP_KEY_SIGNAL
		    status = TRUE;
		    break;

		  case GDK_Home:
		  case GDK_KP_Home:
		  case GDK_KP_5:
		  case ' ':
		    if(keystate)
		    {
			yp->sndobjadv_volume_x = 0.0f;
			yp->sndobjadv_volume_y = 0.0f;
			if(adj != NULL)
			{
			    adj->value = 0.0f;
			    gtk_signal_emit_by_name(
				GTK_OBJECT(adj), "value_changed"
			    );
			}
		    }
		    status = TRUE;
		    break;

		}
#undef SIGNAL_VOLUME_CHANGE
	    }
#undef DO_STOP_KEY_SIGNAL
	    break;

	  case GDK_BUTTON_PRESS:
	    button = (GdkEventButton *)event;
	    yp->sndobjadv_button = button->button;
	    if(!GTK_WIDGET_HAS_FOCUS(widget) && GTK_WIDGET_SENSITIVE(widget))
		gtk_widget_grab_focus(widget);
	    if(widget == yp->sndobjadv_display_da)
	    {
		GtkWidget *w = yp->sndobjadv_display_menu;
		if((w != NULL) && (button->button == 3))
		    gtk_menu_popup(
			GTK_MENU(w), NULL, NULL,
			NULL, NULL,
			button->button, button->time
		    );
	    }
	    else if((widget == yp->sndobjadv_volume_da) &&
		    (yp->sndobjadv_button == 1)
	    )
	    {
		GtkRange *range = (GtkRange *)yp->sndobjadv_volume_scale;
		gint    x = (gint)button->x,
			y = (gint)button->y;
		gint width, height;

		gdk_window_get_size(window, &width, &height);
		if((width > 0) && (height > 0))
		{
		    yp->sndobjadv_volume_x = CLIP(
			(((gfloat)x / (gfloat)width) - 0.5f) * 2.0f,
			-1.0f, 1.0f
		    );
		    yp->sndobjadv_volume_y = CLIP(
			-(((gfloat)y / (gfloat)height) - 0.5f) * 2.0f,
			-1.0f, 1.0f
		    );
		}

		/* Send signal notifying change in volume and redraw
		 * volume drawing area.
		 */
		YPlayerSndObjAdvVolumeChangedCB(
		    (range != NULL) ? range->adjustment : NULL,
		    yp
		);

		gdk_pointer_grab(
		    window,
		    FALSE,
		    GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
			GDK_POINTER_MOTION_MASK,
		    NULL,
		    NULL,
		    button->time
		);
	    }
	    status = TRUE;
	    break;

	  case GDK_BUTTON_RELEASE:
	    button = (GdkEventButton *)event;
	    if(gdk_pointer_is_grabbed())
		gdk_pointer_ungrab(button->time);
	    yp->sndobjadv_button = 0;
	    status = TRUE;
	    break;

	  case GDK_MOTION_NOTIFY:
	    motion = (GdkEventMotion *)event;
	    if((widget == yp->sndobjadv_volume_da) &&
	       (yp->sndobjadv_button == 1)
	    )
	    {
		GtkRange *range = (GtkRange *)yp->sndobjadv_volume_scale;
		gint	x = (gint)motion->x,
			y = (gint)motion->y;
		gint width, height;
		GdkModifierType mask;

		if(motion->is_hint)
		    gdk_window_get_pointer(
			motion->window, &x, &y, &mask
		    );

		gdk_window_get_size(window, &width, &height);
		if((width > 0) && (height > 0))
		{
		    yp->sndobjadv_volume_x = CLIP(
			(((gfloat)x / (gfloat)width) - 0.5f) * 2.0f,
			-1.0f, 1.0f
		    );
		    yp->sndobjadv_volume_y = CLIP(
			-(((gfloat)y / (gfloat)height) - 0.5f) * 2.0f,
			-1.0f, 1.0f
		    );
		}

		/* Send signal notifying change in volume and redraw
		 * volume drawing area.
		 */
		YPlayerSndObjAdvVolumeChangedCB(
		    (range != NULL) ? range->adjustment : NULL,
		    yp
		);
		status = TRUE;
	    }
	    break;

	  case GDK_2BUTTON_PRESS:
	    button = (GdkEventButton *)event;
	    if(widget == yp->sndobjadv_display_da)
	    {
		switch(button->button)
		{
		  case 1:
		    if(!YPLAYER_SNDOBJADV_IS_OPENED(yp))
			YPlayerSndObjAdvOpenCB(
			    yp->sndobjadv_open_btn, yp
			);
		    else if(YPLAYER_SNDOBJADV_IS_PLAYING(yp))
			YPlayerSndObjAdvPauseCB(
			    yp->sndobjadv_play_btn, yp
			);
		    else
			YPlayerSndObjAdvPlayCB(
			    yp->sndobjadv_play_btn, yp
			);
		    break;
		}
		status = TRUE;
	    }
	    break;
	}

	return(status);
}

/*
 *	YPlayer sound object advanced sound object killed callback.
 */
void YPlayerSndObjAdvKilledCB(yplayer_struct *yp, YID yid)
{
	if((yp == NULL) || (yid == YIDNULL))
	    return;

	if(yid == yp->sndobjadv_yid)
	{
	    yp->sndobjadv_yid = YIDNULL;

	    YPlayerSndObjAdvRewindCB(yp->sndobjadv_rewind_btn, yp);
	    YPlayerSndObjAdvDisplayDraw(yp);
	    YPlayerUpdateMenus(yp);
	}
}

/*
 *      YPlayer sound object advanced volume "value_changed" signal
 *      callback.
 */
void YPlayerSndObjAdvVolumeChangedCB(
	GtkAdjustment *adj, gpointer data
)
{
	YID yid;
	gfloat a, c[4];
	yplayer_struct *yp = YPLAYER(data);
	YConnection *con = YPLAYER_CONNECTION(yp);
	if(yp == NULL)
	    return;

	if(YPLAYER_SNDOBJADV_FREEZE_COUNT(yp))
	    return;

	a = (adj != NULL) ?
	    CLIP(1.0f - adj->value, 0.0f, 1.0f) : 1.0f;
	YPlayerVolumeXYToQuad(
	    yp->sndobjadv_volume_x, yp->sndobjadv_volume_y, a,
	    &c[0], &c[1], &c[2], &c[3]
	);

	yid = yp->sndobjadv_yid;
	if((con != NULL) && (yid != YIDNULL))
	{
	    YEventSoundPlay v;

	    v.flags =   YPlayValuesFlagYID |
			YPlayValuesFlagVolume;
	    v.yid = yid;
	    v.left_volume = c[0];
	    v.right_volume = c[1];
	    v.back_left_volume = c[2];
	    v.back_right_volume = c[3];
	    YSetPlaySoundObjectValues(con, yid, &v);
	}

	YPlayerSndObjAdvVolumeDraw(yp);
}

/*
 *	YPlayer sound object advanced position "value_changed" signal
 *	callback.
 */
void YPlayerSndObjAdvPositionChangedCB(
	GtkAdjustment *adj, gpointer data
)
{
	YID yid;
	yplayer_struct *yp = YPLAYER(data);
	YConnection *con = YPLAYER_CONNECTION(yp);
	if((adj == NULL) || (yp == NULL))
	    return;

	if(YPLAYER_SNDOBJADV_FREEZE_COUNT(yp))
	    return;

	yid = yp->sndobjadv_yid;
	if((con != NULL) && (yid != YIDNULL))
	{
	    YEventSoundPlay v;

	    v.flags =	YPlayValuesFlagYID |
			YPlayValuesFlagPosition;
	    v.yid = yid;
	    v.position = (YDataPosition)adj->value;
	    YSetPlaySoundObjectValues(con, yid, &v);
	}

	YPlayerSndObjAdvDisplayDraw(yp);
}

/*
 *      YPlayer sound object advanced applied sample rate
 *	"value_changed" signal callback.
 */
void YPlayerSndObjAdvSampleRateChangedCB(
	GtkAdjustment *adj, gpointer data
)
{
	YID yid;
	yplayer_struct *yp = YPLAYER(data);
	YConnection *con = YPLAYER_CONNECTION(yp);
	if((adj == NULL) || (yp == NULL))
	    return;

	if(YPLAYER_SNDOBJADV_FREEZE_COUNT(yp))
	    return;

	yid = yp->sndobjadv_yid;
	if((con != NULL) && (yid != YIDNULL))
	{
	    YEventSoundPlay v;

	    v.flags =   YPlayValuesFlagYID |
			YPlayValuesFlagSampleRate;
	    v.yid = yid;
	    v.sample_rate = (YDataPosition)adj->value;
	    YSetPlaySoundObjectValues(con, yid, &v);
	}
}

/*
 *      YPlayer sound object advanced open signal callback.
 */
void YPlayerSndObjAdvOpenCB(GtkWidget *widget, gpointer data)
{
	gboolean status;
	const gchar *s;
	gchar *last_path, *filename;
	fb_type_struct **ftype = NULL, *ftype_rtn = NULL;
	gint total_ftypes = 0;
	gchar **path_rtn = NULL;
	gint total_path_rtns = 0;
	yplayer_struct *yp = YPLAYER(data);
	YConnection *con = YPLAYER_CONNECTION(yp);
	if((con == NULL) || (yp == NULL))
	    return;

	if(FileBrowserIsQuery())
	    return;

	YPlayerSetBusy(yp, TRUE);

	s = YC_GET_CFG_S(
	    YC_CORE(yp->core_ptr),
	    YC_CFG_PARM_PLAYER_SNDOBJ_LAST_PATH
	);
	last_path = (s != NULL) ? g_strdup(s) : NULL;

	FileBrowserTypeListNew(
	    &ftype, &total_ftypes,
	    ".wav .wave", "Microsoft Wave files"
	);
	FileBrowserTypeListNew(
	    &ftype, &total_ftypes,
	    ".voc", "Sound Blaster VOC files"
	);
	FileBrowserTypeListNew(
	    &ftype, &total_ftypes,
	    ".raw", "Raw sound files"
	);
	FileBrowserTypeListNew(
	    &ftype, &total_ftypes,
	    ".mid .midi", "Musical Instrument Data Interface files"
	);
	FileBrowserTypeListNew(
	    &ftype, &total_ftypes,
	    "*.*", "All files"
	);

	/* Query user for sound object */
	FileBrowserSetTransientFor(YPLAYER_TOPLEVEL(yp));
	status = FileBrowserGetResponse(
	    "Open Sound Object",
	    "Open", "Cancel",
	    last_path,
	    ftype, total_ftypes,
	    &path_rtn, &total_path_rtns,
	    &ftype_rtn
	);
	FileBrowserSetTransientFor(NULL);
	FileBrowserDeleteTypeList(ftype, total_ftypes);

	/* Get response as filename */
	s = (status && (total_path_rtns > 0)) ?
	    path_rtn[0] : NULL;
	filename = (s != NULL) ? g_strdup(s) : NULL;
	if(filename == NULL)
	{
	    YPlayerSetBusy(yp, FALSE);
	    g_free(last_path);
	    return;
	}

	/* Update last path */
	YC_SET_CFG_S(
	    YC_CORE(yp->core_ptr),
	    YC_CFG_PARM_PLAYER_SNDOBJ_LAST_PATH,
	    GetParentDir(filename)
	);

	/* Open sound object */
	YPlayerDoOpenSoundObjectAdv(yp, filename);

	YPlayerSetBusy(yp, FALSE);

	g_free(last_path);
	g_free(filename);
}

/*
 *      YPlayer sound object advanced play signal callback.
 */
void YPlayerSndObjAdvPlayCB(GtkWidget *widget, gpointer data)
{
	gfloat a, c[4];
	const gchar *path;
	YID yid;
	YEventSoundPlay v;
	GtkWidget *w;
	GtkAdjustment *adj;
	yplayer_struct *yp = YPLAYER(data);
	YConnection *con = YPLAYER_CONNECTION(yp);
	if((con == NULL) || (yp == NULL))
	    return;

	/* Already playing? */
	if(yp->sndobjadv_yid != YIDNULL)
	    return;

	/* No sound object opened */
	path = yp->sndobjadv_path;
	if(path == NULL)
	    return;


	YPlayerSetBusy(yp, TRUE);

	v.flags =	YPlayValuesFlagPosition |
			YPlayValuesFlagRepeats |
			YPlayValuesFlagTotalRepeats |
			YPlayValuesFlagVolume |
			YPlayValuesFlagSampleRate;

	v.yid = YIDNULL;

	w = yp->sndobjadv_position_scale;
	adj = (w != NULL) ? GTK_RANGE(w)->adjustment : NULL;
	v.position = (YDataPosition)((adj != NULL) ? adj->value : 0.0f);

	v.length = 0;

	v.repeats = 0;

	w = yp->sndobjadv_repeat_tb;
	v.total_repeats = (w != NULL) ?
	    ((GTK_TOGGLE_BUTTON(w)->active) ? -1 : 1) : 1;

	w = yp->sndobjadv_volume_scale;
	adj = (w != NULL) ? GTK_RANGE(w)->adjustment : NULL;
	a = (adj != NULL) ?
	    CLIP(1.0f - adj->value, 0.0f, 1.0f) : 1.0f;
	YPlayerVolumeXYToQuad(
	    yp->sndobjadv_volume_x, yp->sndobjadv_volume_y, a,
	    &c[0], &c[1], &c[2], &c[3]
	);
	v.left_volume = c[0];
	v.right_volume = c[1];
	v.back_left_volume = c[2];
	v.back_right_volume = c[3];

	w = yp->sndobjadv_sample_rate_spin;
	adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
	v.sample_rate = (gint)((adj != NULL) ? adj->value : 0.0f);

	/* Play sound object */
	yp->sndobjadv_yid = yid = YStartPlaySoundObject(
	    con, path, &v
	);


	YPlayerSetBusy(yp, FALSE);
	YPlayerSndObjAdvDisplayDraw(yp);
	YPlayerUpdateMenus(yp);
}

/*
 *      YPlayer sound object advanced pause signal callback.
 */
void YPlayerSndObjAdvPauseCB(GtkWidget *widget, gpointer data)
{
	YID yid;
	yplayer_struct *yp = YPLAYER(data);
	YConnection *con = YPLAYER_CONNECTION(yp);
	if((con == NULL) || (yp == NULL))
	    return;

	YPlayerSetBusy(yp, TRUE);

	yid = yp->sndobjadv_yid;
	if(yid != YIDNULL)
	{
	    YDestroyPlaySoundObject(con, yid);
	    yp->sndobjadv_yid = yid = YIDNULL;
	}

	YPlayerSetBusy(yp, FALSE);
	YPlayerSndObjAdvDisplayDraw(yp);
	YPlayerUpdateMenus(yp);
}

/*
 *      YPlayer sound object advanced rewind signal callback.
 */
void YPlayerSndObjAdvRewindCB(GtkWidget *widget, gpointer data)
{
	GtkAdjustment *adj;
	GtkWidget *w;
	yplayer_struct *yp = YPLAYER(data);
	YConnection *con = YPLAYER_CONNECTION(yp);
	if((con == NULL) || (yp == NULL))
	    return;

	YPlayerSetBusy(yp, TRUE);

	w = yp->sndobjadv_position_scale;
	adj = (w != NULL) ? GTK_RANGE(w)->adjustment : NULL;
	if(adj != NULL)
	{
	    /* Move position to beginning and emit signal so that
	     * the sound object (if playing) will be rewinded.
	     */
	    adj->value = adj->lower;
	    gtk_signal_emit_by_name(GTK_OBJECT(adj), "value_changed");
	}

	YPlayerSetBusy(yp, FALSE);
}

/*
 *      YPlayer sound object advanced change Audio mode signal callback.
 */
void YPlayerSndObjAdvChangeModeCB(GtkWidget *widget, gpointer data)
{
	YAudioModeValuesStruct **mode;
	const YAudioModeValuesStruct *m;
	int total_modes;
	YEventSoundObjectAttributes *soa;
	yplayer_struct *yp = YPLAYER(data);
	YConnection *con = YPLAYER_CONNECTION(yp);
	if(con == NULL)
	    return;

	/* No sound object opened? */
	if(yp->sndobjadv_path == NULL)
	    return;

	YPlayerSetBusy(yp, TRUE);

	/* Get attributes of the opened sound object and a list of
	 * preset Audio modes.  Then find the preset Audio mode that
	 * best matches this sound object's attributes (if any).
	 */
	soa = &yp->sndobjadv_soa;
	mode = YGetAudioModes(con, &total_modes);
	m = YCMatchAudioModeByValues(
	    mode, total_modes,
	    soa->sample_rate, soa->sample_size, soa->channels
	);
	if(m != NULL)
	{
	    /* Found suitable preset Audio mode */
	    if(!YChangeAudioModePreset(con, m->name))
	    {
		g_free(yp->audio_mode_name);
		yp->audio_mode_name = g_strdup(m->name);
		yp->sample_rate = m->sample_rate;
		yp->sample_size = m->sample_size;
		yp->channels = m->channels;
		yp->num_fragments = m->num_fragments;
		yp->fragment_size = m->fragment_size_bytes;
		YPLAYER_SET_CONBAR_AUDIO(yp);
	    }
	}
	else
	{
	    /* No suitable preset Audio mode found, so instead set
	     * explicit Audio values.
	     */
	    if(!YSetAudioModeValues(
		con,
		soa->sample_size, soa->channels, soa->sample_rate,
		1,
		True,
		2, 2048
	    ))
	    {
		yp->sample_rate = soa->sample_rate;
		yp->sample_size = soa->sample_size;
		yp->channels = soa->channels;
		yp->fragment_size = 2048;
		yp->cycle_us = YCalculateCycle(
		    con,
		    yp->sample_rate, yp->channels, yp->sample_size,
		    yp->fragment_size
		);
		YSetCycle(con, yp->cycle_us);
		YPLAYER_SET_CONBAR_AUDIO(yp);
	    }
	}


	/* Delete preset Audio modes list */
	YFreeAudioModesList(mode, total_modes);

	YPlayerSetBusy(yp, FALSE);
	YPlayerUpdateMenus(yp);
}

/*
 *      YPlayer sound object advanced close signal callback.
 */
void YPlayerSndObjAdvCloseCB(GtkWidget *widget, gpointer data)
{
	yplayer_struct *yp = YPLAYER(data);
	if(yp == NULL)
	    return;

	YPlayerSetBusy(yp, TRUE);
	YPlayerDoCloseSoundObjectAdv(yp);
	YPlayerSetBusy(yp, FALSE);
	YPlayerUpdateMenus(yp);
}


/*
 *	YPlayer audio CD play signal callback.
 */
void YPlayerAudioCDPlayCB(GtkWidget *widget, gpointer data)
{
	gint row, track_number;
	GList *glist;
	yplayer_struct *yp = YPLAYER(data);
	YConnection *con = YPLAYER_CONNECTION(yp);
	GtkCList *clist = (GtkCList *)YPLAYER_AUDIOCD_TRACKS_CLIST(yp);
	if((con == NULL) || (clist == NULL))
	    return;

	glist = clist->selection_end;
	row = (glist != NULL) ? (gint)glist->data : -1;
	if((row < 0) || (row >= clist->rows))
	    return;

	track_number = row + 1;

	YPlayerSetBusy(yp, TRUE);
	YPlayAudioCDTrack(con, track_number);
	YPlayerSetBusy(yp, FALSE);
	YPlayerUpdateMenus(yp);
}

/*
 *	YPlayer audio CD play random signal callback.
 */
void YPlayerAudioCDRandomCB(GtkWidget *widget, gpointer data)
{
	gint row;
	yplayer_struct *yp = YPLAYER(data);
	YConnection *con = YPLAYER_CONNECTION(yp);
	GtkCList *clist = (GtkCList *)YPLAYER_AUDIOCD_TRACKS_CLIST(yp);
	if((con == NULL) || (clist == NULL))
	    return;

	if(clist->rows <= 0)
	    return;

	srand(time(NULL));
	row = (gint)rand() % (gint)clist->rows;
	gtk_clist_select_row(clist, row, 0);

	YPlayerAudioCDPlayCB(widget, data);
}

/*
 *	YPlayer audio CD stop signal callback.
 */
void YPlayerAudioCDStopCB(GtkWidget *widget, gpointer data)
{
	yplayer_struct *yp = YPLAYER(data);
	YConnection *con = YPLAYER_CONNECTION(yp);
	if(con == NULL)
	    return;

	YPlayerSetBusy(yp, TRUE);
	YStopAudioCD(con);
	YPlayerSetBusy(yp, FALSE);
	YPlayerUpdateMenus(yp);
}

/*
 *	YPlayer audio CD stop signal callback.
 */
void YPlayerAudioCDEjectCB(GtkWidget *widget, gpointer data)
{
	yplayer_struct *yp = YPLAYER(data);
	YConnection *con = YPLAYER_CONNECTION(yp);
	if(con == NULL)
	    return;

	YPlayerSetBusy(yp, TRUE);
	YEjectAudioCD(con);
	YPlayerSetBusy(yp, FALSE);
	YPlayerUpdateMenus(yp);
}

/*
 *	YPlayer audio CD refresh signal callback.
 */
void YPlayerAudioCDRefreshCB(GtkWidget *widget, gpointer data)
{
	yplayer_struct *yp = YPLAYER(data);
	if(yp == NULL)
	    return;

	YPlayerSetBusy(yp, TRUE);
	YPlayerAudioCDTracksListUpdate(yp);
	YPlayerSetBusy(yp, FALSE);
	YPlayerUpdateMenus(yp);
}


/*
 *	YPlayer reconfigured signal callback.
 */
void YPlayerReconfiguredCB(
	yplayer_struct *yp, gpointer core_ptr
)
{
	GtkRcStyle *rcstyle;
	GtkWidget *w;
	dspview_struct *dv;
	yc_core_struct *cp = YC_CORE(core_ptr);
	if((yp == NULL) || (cp == NULL))
	    return;

	/* Update connection bar's audio display */
	if(TRUE)
	{
	    GdkColor *c, cs[3];
	    const cfg_color_struct *color;

#define DO_CONVERT_COLOR	\
{ if((c != NULL) && (color != NULL)) { \
 c->red		= (gushort)-1 * color->r; \
 c->green	= (gushort)-1 * color->g; \
 c->blue	= (gushort)-1 * color->b; \
} }
	    color = YC_GET_CFG_COLOR(
		cp, YC_CFG_PARM_AUDIO_DISPLAY_COLOR_BG
	    );
	    c = &cs[0];
	    DO_CONVERT_COLOR

	    color = YC_GET_CFG_COLOR(
		cp, YC_CFG_PARM_AUDIO_DISPLAY_COLOR_FG
	    );
	    c = &cs[1];
	    DO_CONVERT_COLOR

	    color = YC_GET_CFG_COLOR(
		cp, YC_CFG_PARM_AUDIO_DISPLAY_COLOR_FG_DIM
	    );
	    c = &cs[2];
	    DO_CONVERT_COLOR

	    ConBarSetAudioColors(
		YPLAYER_CONBAR(yp), &cs[0], &cs[1], &cs[2]
	    );
#undef DO_CONVERT_COLOR

	    ConBarSetAudioFonts(
		YPLAYER_CONBAR(yp),
		YC_GET_CFG_S(cp, YC_CFG_PARM_AUDIO_DISPLAY_FONT_LARGE),
		YC_GET_CFG_S(cp, YC_CFG_PARM_AUDIO_DISPLAY_FONT_MEDIUM),
		YC_GET_CFG_S(cp, YC_CFG_PARM_AUDIO_DISPLAY_FONT_SMALL)
	    );
	}

	/* Update connection bar's DSPView */
	dv = ConBarGetDSPView(YPLAYER_CONBAR(yp));
	if(dv != NULL)
	{
	    const cfg_color_struct *color;
	    gfloat alpha_coeff;

	    DSPViewSetDisplayStyle(
		dv,
		YC_GET_CFG_I(cp, YC_CFG_PARM_DSPVIEW_DISPLAY_STYLE)
	    );
	    DSPViewSetMergeMode(
		dv,
		YC_GET_CFG_I(cp, YC_CFG_PARM_DSPVIEW_MERGE_MODE)
	    );
	    color = YC_GET_CFG_COLOR(cp, YC_CFG_PARM_DSPVIEW_COLOR);
	    alpha_coeff = YC_GET_CFG_F(
		cp, YC_CFG_PARM_DSPVIEW_TRANSPARENCY
	    );
	    if(color != NULL)
		DSPViewSetDSPColor(
		    dv,
		    (guint32)(
			((guint32)(color->r * 0xff) << 0) |
			((guint32)(color->g * 0xff) << 8) |
			((guint32)(color->b * 0xff) << 16) |
			((guint32)(alpha_coeff * 0xff) << 24)
		    )
		);
	    DSPViewSetFadeCoeff(
		dv,
		YC_GET_CFG_F(cp, YC_CFG_PARM_DSPVIEW_FADE_COEFF)
	    );
	}	/* Update connection bar's DSPView */


	/* Get style for GtkCLists */
	rcstyle = YPlayerAudioDisplayRCStyleNew(yp);

	w = YPLAYER_SNDOBJ_CLIST(yp);
	if((w != NULL) && (rcstyle != NULL))
	    gtk_widget_modify_style(w, rcstyle);

	w = YPLAYER_SNDOBJ_PLAYING_CLIST(yp);
	if((w != NULL) && (rcstyle != NULL))
	    gtk_widget_modify_style(w, rcstyle);

	w = yp->sndobjadv_display_da;
	if((w != NULL) && (rcstyle != NULL))
	    gtk_widget_modify_style(w, rcstyle);

	w = yp->sndobjadv_volume_da;
	if((w != NULL) && (rcstyle != NULL))
	    gtk_widget_modify_style(w, rcstyle);

	w = yp->sndobjadv_sample_rate_spin;
	if((w != NULL) && (rcstyle != NULL))
	    gtk_widget_modify_style(w, rcstyle);

	w = YPLAYER_AUDIOCD_TRACKS_CLIST(yp);
	if((w != NULL) && (rcstyle != NULL))
	    gtk_widget_modify_style(w, rcstyle);

	if(rcstyle != NULL)
	    gtk_rc_style_unref(rcstyle);
}
