#include <stdlib.h>
#include <string.h>

#include <Y2/Y.h>
#include <Y2/Ylib.h>

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include "guiutils.h"
#include "cdialog.h"
#include "fb.h"
#include "pdialog.h"

#include "cfg.h"
#include "yconsolewin.h"
#include "yaudio.h"
#include "ymixer.h"
#include "yplayer.h"
#include "yserverlist.h"
#include "yserverlistcb.h"
#include "yc.h"
#include "ycop.h"
#include "yccfglist.h"
#include "config.h"


void YServerListItemDestroyCB(gpointer data);
gint YServerListDeleteEventCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
);
void YServerListDestroyCB(GtkObject *object, gpointer data);
gint YServerListKeyEventCB(
	GtkWidget *widget, GdkEventKey *key, gpointer data
);
gint YServerListButtonEventCB(
	GtkWidget *widget, GdkEventButton *button, gpointer data
);

void YServerListSelectRowCB(
	GtkCList *clist, gint row, gint column, GdkEvent *event,
	gpointer data
);
void YServerListUnselectRowCB(
	GtkCList *clist, gint row, gint column, GdkEvent *event,
	gpointer data
);

static gchar *YServerListPromptBrowseCB(
	gpointer p, gpointer data, gint prompt_num
);
void YServerListAddCB(GtkWidget *widget, gpointer data);
void YServerListEditCB(GtkWidget *widget, gpointer data);
void YServerListRemoveCB(GtkWidget *widget, gpointer data);
void YServerListShiftUpCB(GtkWidget *widget, gpointer data);
void YServerListShiftDownCB(GtkWidget *widget, gpointer data);
void YServerListConnectCB(GtkWidget *widget, gpointer data);
void YServerListStartCB(GtkWidget *widget, gpointer data);
void YServerListCloseCB(GtkWidget *widget, gpointer data);

void YServerListReconfiguredCB(
	yserver_list_struct *sl, gpointer core_ptr
);


/*
 *	YServer list item "destroy" signal callback.
 */
void YServerListItemDestroyCB(gpointer data)
{
	yserver_list_item_struct *item = YSERVER_LIST_ITEM(data);
	if(item != NULL)
	{
	    g_free(item->name);
	    g_free(item->address);
	    g_free(item->start_command);
	    g_free(item->comments);
	    g_free(item);
	}
}

/*
 *	Toplevel "delete_event" signal callback.
 */
gint YServerListDeleteEventCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
)
{
	YServerListCloseCB(widget, data);
	return(TRUE);
}

/*
 *	Toplevel "destroy" signal callback.
 */
void YServerListDestroyCB(GtkObject *object, gpointer data)
{

}

/*
 *      Servers list "key_press_event" signal callback.
 */
gint YServerListKeyEventCB(
	GtkWidget *widget, GdkEventKey *key, gpointer data
)
{
	gint status = FALSE;
	gboolean key_press;
	GtkCList *clist;
	yserver_list_struct *sl = YSERVER_LIST(data);
	if((key == NULL) || (sl == NULL))
	    return(status);

	clist = (GtkCList *)YSERVER_LIST_SERVERS_CLIST(sl);
	if(clist == NULL)
	    return(status);

	key_press = (key->type == GDK_KEY_PRESS) ? TRUE : FALSE;

	switch(key->keyval)
	{
	  case GDK_Return:
	  case GDK_KP_Enter:
	  case GDK_ISO_Enter:
	  case GDK_3270_Enter:
	    if(key_press)
	    {
		if(key->state & GDK_CONTROL_MASK)
		    YServerListEditCB(widget, data);
		else
		    YServerListConnectCB(widget, data);
		status = TRUE;
	    }
	    break;

	  case GDK_Escape:
	    if(key_press)
	    {
		YServerListCloseCB(widget, data);
		status = TRUE;
	    }
	    break;

	  case GDK_Insert:
	    if(key_press)
	    {
		YServerListAddCB(widget, data);
		status = TRUE;
	    }
	    break;

	  case GDK_Delete:
	    if(key_press)
	    {
		YServerListRemoveCB(widget, data);
		status = TRUE;
	    }
	    break;

	}

	return(status);
}

/*
 *	Servers list "button_press_event" signal callback.
 */
gint YServerListButtonEventCB(
	GtkWidget *widget, GdkEventButton *button, gpointer data
)
{
	gint status = FALSE;
	gint row, column;
	GtkWidget *w;
	GtkCList *clist;
	yserver_list_struct *sl = YSERVER_LIST(data);
	if((button == NULL) || (sl == NULL))
	    return(status);

	clist = (GtkCList *)YSERVER_LIST_SERVERS_CLIST(sl);
	if(clist == NULL)
	    return(status);

	if(button->window != clist->clist_window)
	    return(status);

	switch(button->button)
	{
	  case 3:
	    switch((gint)button->type)
	    {
	      case GDK_BUTTON_PRESS:
		gtk_clist_get_selection_info(
		    clist, button->x, button->y, &row, &column
		);
		gtk_clist_freeze(clist);
		gtk_clist_unselect_all(clist);
		gtk_clist_select_row(clist, row, column);
		gtk_clist_thaw(clist);

		w = sl->menu;
		if(w != NULL)
		    gtk_menu_popup(
			GTK_MENU(w), NULL, NULL,
			NULL, NULL,
			button->button, button->time
		    );
		status = TRUE;
		break;
	    }
	    break;

	  case 1:
	    switch((gint)button->type)
	    {
	      case GDK_2BUTTON_PRESS:
		if(button->state & GDK_CONTROL_MASK)
		    YServerListEditCB(widget, data);
		else
		    YServerListConnectCB(widget, data);
		status = TRUE;
		break;
	    }
	    break;
	}

	return(status);
}


/*
 *	Servers list "select_row" signal callback.
 */
void YServerListSelectRowCB(
	GtkCList *clist, gint row, gint column, GdkEvent *event,
	gpointer data
)
{
	yserver_list_struct *sl = YSERVER_LIST(data);
	if((clist == NULL) || (sl == NULL))
	    return;

	if(gtk_clist_row_is_visible(clist, row) !=
	    GTK_VISIBILITY_FULL
	)
	    gtk_clist_moveto(
		clist,
		row, -1,	/* Row, column. */
		0.5, 0.0	/* Row, column. */
	    );

	YServerListUpdateMenus(sl);
}

/*
 *	Servers list "unselect_row" signal callback.
 */
void YServerListUnselectRowCB(
	GtkCList *clist, gint row, gint column, GdkEvent *event,
	gpointer data
)
{
	yserver_list_struct *sl = YSERVER_LIST(data);
	if(sl == NULL)
	    return;

	YServerListUpdateMenus(sl);
}

/*
 *	Prompt bar browse path callback.
 */
static gchar *YServerListPromptBrowseCB(
	gpointer p, gpointer data, gint prompt_num
)
{
	gboolean status;
	gchar **strv = NULL;
	gint strc = 0;
	fb_type_struct **type = NULL;
	gint typec = 0;
	fb_type_struct *type_rtn;
	const gchar *s = PDialogGetPromptValue(prompt_num);
	yserver_list_struct *sl = YSERVER_LIST(data);
	if(sl == NULL)
	    return(NULL);

	if(FileBrowserIsQuery())
	    return(NULL);

	FileBrowserTypeListNew(
	    &type, &typec,
	    "*.*", "All files"
	);

	FileBrowserSetTransientFor(YSERVER_LIST_TOPLEVEL(sl));
	status = FileBrowserGetResponse(
	    "Select Object",
	    "Select", "Cancel",
	    s,
	    type, typec, &strv, &strc, &type_rtn
	);
	FileBrowserSetTransientFor(NULL);
	FileBrowserDeleteTypeList(type, typec);

	if(!status || (strc < 1))
	    return(NULL);
	else
	    return(strv[0]);
}

/*
 *      Add callback.
 */
void YServerListAddCB(GtkWidget *widget, gpointer data)
{
	gint row;
	GList *glist;
	GtkCList *clist;
	yserver_list_struct *sl = YSERVER_LIST(data);
	if(sl == NULL)
	    return;

	clist = (GtkCList *)YSERVER_LIST_SERVERS_CLIST(sl);
	glist = (clist != NULL) ? clist->selection_end : NULL;

	/* Get selected row or -1 if no row is selected. */
	row = (glist != NULL) ? (gint)glist->data : -1;

	/* Add a new row with default values. */
	row = YServerListAddItem(
	    sl, row,
	    "Untitled", "127.0.0.1:9433", TRUE, NULL, NULL
	);
	/* Failed to add new row? */
	if(row < 0)
	    return;

	/* Select new row. */
	gtk_clist_freeze(clist);
	gtk_clist_unselect_all(clist);
	gtk_clist_select_row(clist, row, 0);
	gtk_clist_thaw(clist);

	/* Edit the new row. */
	YServerListEditCB(NULL, data);
}

/*
 *	Edit callback.
 */
void YServerListEditCB(GtkWidget *widget, gpointer data)
{
	gint row;
	GList *glist;
	GtkCList *clist;
	yserver_list_item_struct *item;
	yserver_list_struct *sl = YSERVER_LIST(data);
	if(sl == NULL)
	    return;

	clist = (GtkCList *)YSERVER_LIST_SERVERS_CLIST(sl);
	glist = (clist != NULL) ? clist->selection_end : NULL;
	if(glist == NULL)
	    return;
	row = (gint)glist->data;
	if((row < 0) || (row >= clist->rows))
	    return;

	/* Get list item data stucture */
	item = YSERVER_LIST_ITEM(gtk_clist_get_row_data(clist, row));
	if(item != NULL)
	{
	    gchar **strv;
	    gint strc;

	    PDialogDeleteAllPrompts();
	    PDialogAddPrompt(
		NULL, "Name:", item->name
	    );
	    PDialogAddPrompt(
		NULL, "Address:", item->address
	    );
	    PDialogAddPromptToggle(
		NULL, "Is Local", item->is_local
	    );
	    PDialogAddPromptWithBrowse(
		NULL, "Start Command:", item->start_command,
		sl, YServerListPromptBrowseCB
	    );
	    PDialogAddPrompt(
		NULL, "Comments:", item->comments
	    );
	    PDialogSetSize(450, -1);
	    PDialogSetTransientFor(YSERVER_LIST_TOPLEVEL(sl));
	    strv = PDialogGetResponse(
		"Edit Item", NULL, NULL,
		PDIALOG_ICON_EDIT,
		"OK", "Cancel",
		PDIALOG_BTNFLAG_SUBMIT | PDIALOG_BTNFLAG_CANCEL,
		PDIALOG_BTNFLAG_SUBMIT,
		&strc
	    );
	    PDialogSetTransientFor(NULL);

	    if((strv != NULL) && (strc >= 4))
	    {
		const gchar *s;
		int n;

		s = strv[0];
		g_free(item->name);
		item->name = (s != NULL) ? g_strdup(s) : NULL;

		s = strv[1];
		g_free(item->address);
		item->address = (s != NULL) ? g_strdup(s) : NULL;

		s = strv[2];
		n = (s != NULL) ? atoi(s) : 0;
		item->is_local = (n) ? TRUE : FALSE;

		s = strv[3];
		g_free(item->start_command);
		item->start_command = (s != NULL) ? g_strdup(s) : NULL;

		s = strv[4];
		g_free(item->comments);
		item->comments = (s != NULL) ? g_strdup(s) : NULL;

		YServerListUpdateItem(sl, row);
	    }
	}

	YServerListUpdateMenus(sl);
}

/*
 *	Remove callback.
 */
void YServerListRemoveCB(GtkWidget *widget, gpointer data)
{
	gint row;
	GList *glist;
	GtkCList *clist;
	yserver_list_struct *sl = YSERVER_LIST(data);
	if(sl == NULL)
	    return;

	clist = (GtkCList *)YSERVER_LIST_SERVERS_CLIST(sl);
	glist = (clist != NULL) ? clist->selection_end : NULL;
	if(glist == NULL)
	    return;
	row = (gint)glist->data;
	if((row < 0) || (row >= clist->rows))
	    return;

	/* Confirm. */
	if(TRUE)
	{
	    yserver_list_item_struct *item = YSERVER_LIST_ITEM(
		gtk_clist_get_row_data(clist, row)
	    );
	    gchar *buf = g_strdup_printf(
		"Remove item \"%s\"?",
		(item != NULL) ? item->name : "Untitled"
	    );
	    gint response;

	    CDialogSetTransientFor(YSERVER_LIST_TOPLEVEL(sl));
	    response = CDialogGetResponse(
		"Confirm Remove", buf, NULL,
		CDIALOG_ICON_WARNING,
		CDIALOG_BTNFLAG_YES | CDIALOG_BTNFLAG_NO,
		CDIALOG_BTNFLAG_NO
	    );
	    CDialogSetTransientFor(NULL);
	    g_free(buf);

	    if(response != CDIALOG_RESPONSE_YES)
		return;
	}

	/* Remove item. */
	gtk_clist_freeze(clist);
	gtk_clist_remove(clist, row);
	gtk_clist_thaw(clist);

	YServerListUpdateMenus(sl);
}

/*
 *	Shift up callback.
 */
void YServerListShiftUpCB(GtkWidget *widget, gpointer data)
{
	gint row;
	GList *glist;
	GtkCList *clist;
	yserver_list_struct *sl = YSERVER_LIST(data);
	if(sl == NULL)
	    return;

	clist = (GtkCList *)YSERVER_LIST_SERVERS_CLIST(sl);
	glist = (clist != NULL) ? clist->selection_end : NULL;
	if(glist == NULL)
	    return;
	row = (gint)glist->data;

	/* Can shift up? */
	if((row - 1) < 0)
	    return;

	/* Shift up. */
	gtk_clist_freeze(clist);
	gtk_clist_swap_rows(clist, row, row - 1);
	gtk_clist_thaw(clist);

	YServerListUpdateMenus(sl);
}

/*
 *	Shift down callback.
 */
void YServerListShiftDownCB(GtkWidget *widget, gpointer data)
{
	gint row;
	GList *glist;
	GtkCList *clist;
	yserver_list_struct *sl = YSERVER_LIST(data);
	if(sl == NULL)
	    return;

	clist = (GtkCList *)YSERVER_LIST_SERVERS_CLIST(sl);
	glist = (clist != NULL) ? clist->selection_end : NULL;
	if(glist == NULL)
	    return;
	row = (gint)glist->data;

	/* Can shift down? */
	if((row + 1) > (clist->rows - 1))
	    return;

	/* Shift down. */
	gtk_clist_freeze(clist);
	gtk_clist_swap_rows(clist, row, row + 1);
	gtk_clist_thaw(clist);

	YServerListUpdateMenus(sl);
}

/*
 *	Connect callback.
 */
void YServerListConnectCB(GtkWidget *widget, gpointer data)
{
	gboolean got_match = FALSE;
	const gchar *address, *start_command;
	gint row;
	GList *glist;
	GtkWidget *w;
	GtkCList *clist;
	yc_core_struct *core_ptr;
	yserver_list_item_struct *item;
	yserver_list_struct *sl = YSERVER_LIST(data);
	if(sl == NULL)
	    return;

	core_ptr = YC_CORE(sl->core_ptr);
	if(core_ptr == NULL)
	    return;

	clist = (GtkCList *)YSERVER_LIST_SERVERS_CLIST(sl);
	if(clist == NULL)
	    return;

	glist = clist->selection_end;
	row = (glist != NULL) ? (gint)glist->data : -1;
	if((row < 0) || (row >= clist->rows))
	    return;

	item = YSERVER_LIST_ITEM(
	    gtk_clist_get_row_data(clist, row)
	);
	if(item == NULL)
	    return;

	address = item->address;
	if((address != NULL) ? (*address == '\0') : TRUE)
	{
	    gchar *buf = g_strdup_printf(
"There is no address defined for item\n\
\"%s\".  To define the address\n\
please click on \"Edit\".",
		(item->name != NULL) ? item->name : "*Untitled*"
	    );
	    CDialogSetTransientFor(YSERVER_LIST_TOPLEVEL(sl));
	    CDialogGetResponse(
		"Connect Failed", buf, NULL,
		CDIALOG_ICON_WARNING,
		CDIALOG_BTNFLAG_OK,
		CDIALOG_BTNFLAG_OK
	    );
	    CDialogSetTransientFor(NULL);
	    g_free(buf);
	    return;
	}

	/* Get start command (may be NULL). */
	start_command = item->start_command;


	/* Begin searching through all windows to see which window's
	 * toplevel widget matches the reference toplevel on this
	 * YServer list window.
	 */
	w = sl->rel_toplevel;
	if(w != NULL)
	{
	    gint i;


	    if(!got_match)
	    {
		/* Iterate through all YConsole windows. */
		yconsole_struct *yc;
		for(i = 0; i < core_ptr->total_yconsoles; i++)
		{
		    yc = core_ptr->yconsole[i];
		    if(w == YCONSOLE_TOPLEVEL(yc))
		    {
			YConsoleDoConnect(yc, address, start_command);
			got_match = TRUE;
		    }
		}
	    }
	    if(!got_match)
	    {
		/* Iterate through all YAudio windows. */
		yaudio_struct *ya;
		for(i = 0; i < core_ptr->total_yaudios; i++)
		{
		    ya = core_ptr->yaudio[i];
		    if(w == YAUDIO_TOPLEVEL(ya))
		    {
			YAudioDoConnect(ya, address, start_command);
			got_match = TRUE;
		    }
		}
	    }
	    if(!got_match)
	    {
		/* Iterate through all YMixer windows. */
		ymixer_struct *ym;
		for(i = 0; i < core_ptr->total_ymixers; i++)
		{
		    ym = core_ptr->ymixer[i];
		    if(w == YMIXER_TOPLEVEL(ym))
		    {
			YMixerDoConnect(ym, address, start_command);
			got_match = TRUE;
		    }
		}
	    }
	    if(!got_match)
	    {
		/* Iterate through all YPlayer windows. */
		yplayer_struct *yp;
		for(i = 0; i < core_ptr->total_yplayers; i++)
		{
		    yp = core_ptr->yplayer[i];
		    if(w == YPLAYER_TOPLEVEL(yp))
		    {
			YPlayerDoConnect(yp, address, start_command);
			got_match = TRUE;
		    }
		}
	    }

/* Add more windows to check here. */

	}


	/* If we still did not find a window, then create a new YConsole
	 * window and have it connect.
	 */
	if(!got_match)
	{
	    YCNewYConsole(core_ptr, address);
	    got_match = TRUE;
	}

	YServerListUpdateMenus(sl);
}

/*
 *	Start server callback.
 */
void YServerListStartCB(GtkWidget *widget, gpointer data)
{
	const gchar *cmd;
	gint row;
	GList *glist;
	GtkCList *clist;
	yc_core_struct *core_ptr;
	yserver_list_item_struct *item;
	yserver_list_struct *sl = YSERVER_LIST(data);
	if(sl == NULL)
	    return;

	core_ptr = YC_CORE(sl->core_ptr);
	if(core_ptr == NULL)
	    return;

	clist = (GtkCList *)YSERVER_LIST_SERVERS_CLIST(sl);
	if(clist == NULL)
	    return;

	glist = clist->selection_end;
	row = (glist != NULL) ? (gint)glist->data : -1;
	if((row < 0) || (row >= clist->rows))
	    return;

	item = YSERVER_LIST_ITEM(
	    gtk_clist_get_row_data(clist, row)
	);
	if(item == NULL)
	    return;

	cmd = item->start_command;
	if((cmd != NULL) ? (*cmd == '\0') : TRUE)
	{
	    gchar *buf = g_strdup_printf(
"There is no start server command defined for item\n\
\"%s\".  To define the start server command\n\
please click on \"Edit\".",
		(item->name != NULL) ? item->name : "*Untitled*"
	    );
	    CDialogSetTransientFor(YSERVER_LIST_TOPLEVEL(sl));
	    CDialogGetResponse(
		"Start Server Failed", buf, NULL,
		CDIALOG_ICON_WARNING,
		CDIALOG_BTNFLAG_OK,
		CDIALOG_BTNFLAG_OK
	    );
	    CDialogSetTransientFor(NULL);
	    g_free(buf);
	    return;
	}

	YCNewStartupDialog(
	    core_ptr, YSERVER_LIST_TOPLEVEL(sl),
	    cmd, FALSE
	);
}

/*
 *	Close callback.
 */
void YServerListCloseCB(GtkWidget *widget, gpointer data)
{
	yserver_list_struct *sl = YSERVER_LIST(data);
	YServerListSavePosition(sl);
	YServerListUnmap(sl);
}


/*
 *	YServer list reconfigured signal callback.
 */
void YServerListReconfiguredCB(
	yserver_list_struct *sl, gpointer core_ptr
)
{
	const cfg_item_struct *cfg_list;
	yc_core_struct *cp = YC_CORE(core_ptr);
	if((sl == NULL) || (cp == NULL))
	    return;

	cfg_list = cp->cfg_list;
	if(cfg_list == NULL)
	    return;






}
